sha1_signatures = {
"application/x.atari5200.cart": {
  b'\xe1\x8a`;\x18\xb8^$\xf9\xc98\xb3;`w\xa7g\xc1\xfa\x9c': 'A.E. (USA) (Proto)',
  b'\xab10%m\x16\xaf\t\xca\xc8\xb1\x17I\xdf\xaf\xf6\xaa\xaa>\xa3': 'Blaster (USA) (Proto)',
  b'\xcf\xdd/\x8d\xab\xb0=\x87\x84\xf36\xbb\x0c\x80Bp\xdc:\x8al': 'Bounty Bob Strikes Back! (USA)',
  b'\xc1\xa5j\x1f\xc5\x1e\t\xb1\xd8\xe9:\xf4\r\xddu\x96\x84~ad': 'Castle Blast (USA) (Unl)',
  b'\x9f\x1e\xe1\x911\x9e\xf0&\x1c\xc7\x07J\x81\xc0\xe6\xd6\xb7FL\xd2': 'Choplifter! (USA)',
  b'\r\xf0\xd7\xf1\x1d\x89\xe2\xc5\xb5\x9c\xc9\xf9\x8dg\xd9\x8e 81\xaa': 'Galaxian (USA)',
  b'\x95G\xc5n)Q\xe1\xb3\xa7\x13\xb0\xb9\x9d>\xdbC\xc9\xd3\xcd\xd8': 'Koffi - Yellow Kopter (USA) (Unl)',
  b'W\xdf\xc2,\x98\x05U_\x95~\xd5\x1a-\xc7\xbf5\xad\x00\xc9\xd9': 'Star Raiders (USA)',
  b'j\xd7\xa1\xe8\xc9\xfa\xd4\x86\xfb\xec\x94\x98\xcbH\xbf[\xc3\xad\xc50': '[BIOS] Atari 5200 (USA)',
},
"application/x.atari5200.cart.type16": {
  b'\xdb 9s\xdf\x13\xb0q\xdaF\xd5/\xdb\x07\x8av\xf4\xdbB\x8d': 'Activision Decathlon, The (USA)',
  b'\xb7]U]\x05\x9a\xd6\x19\x92\xf9\xf5\xa24\x12\x9e\xcb\xf9PL5': 'Beamrider (USA)',
  b"\xe5l\xcck\xf4'\xa82\xce\x00T\x0b\x02\xd0e\xa2h\x82`\xdc": 'BerZerk (USA)',
  b'\x1e\x1a \xc1\x04\xbfMU\xec\xa0\xca\x1er\xdc]\xc6O\xa6\xb1\xb5': 'Blue Print (USA)',
  b'\xb0Q\xf0\xfb.\x81g\x87\xabn\xed\x0c\xb0\x08\xb9\xf8\xcc\x99c\xdc': "David Crane's Pitfall II - Lost Caverns (USA)",
  b'HbZ\xf4\xf3\xe9\xfeS\x8d\xab*\x0bF8\x87\xfc\x8ck\x13\x8d': 'Final Legacy (USA) (Proto)',
  b'\xd2\x86\xec\xb6y:\r\xda\x9c\x9ch\xf65]\xbe:\xdei"\xf9': 'H.E.R.O. (USA)',
  b'@F\xf6\x06\x02%J}\x86\xd2\xa6\xef\xd1\xa30E^\xae\xc1\x19': 'Last Starfighter, The (USA) (Proto)',
  b"2\xd8j8\x1f\xef\xf41I\xb6c6\xdc\xc4\xba%#'2(": 'Meteorites (USA)',
  b"\x12C\xa8\xec\xfc\x02\xa7J\xe4\xe1\xf4\xfc\xda8P'v\xf4\\\xb0": 'Millipede (USA) (Proto)',
  b'\x05d\xb1\x86z\x0bW\rf\xdf\xcb\xc1\x1a\xdc>Q\xa2\xc6\xf2\x8c': 'Miner 2049er Starring Bounty Bob (USA)',
  b'\x02\xdd[\x92x\xc0j\xe8\xcb\xb6\xda\xcc\tN#\x8f\xba\xcf\x08\xd2': 'Moon Patrol (USA)',
  b'\x17\xa4\x1c_\x9a2\xf8\xa7\x199\xcf\xaa\x0e\x96\x80{\x95\x1a\xed\xad': 'Quest for Quintana Roo (USA)',
  b'\x91u\xe6\xda\xcaX\x10*\xc8\xb9r\x9d{\x036\xcf\t\x97}\x12': 'Robotron 2084 (USA)',
  b'Z\xe2\x0ev}\x98\x19\xfe`\x1f;\x1e\xa6\xd4Pj\xe2\xf2\xce\xef': 'Space Shuttle - A Journey Into Space (USA)',
  b'#\x8c\xb4HU]\x8b\xb5\xc9j\xa05G\x18A\x1e\x87w\xc6\x13': 'Super Pac-Man (USA) (Proto)',
  b'\x85\xfc\xea\nMM\x1f"Cx\x01\xc9b)\xba\x96\xe0\xf6\xb7\x11': 'Tempest (USA) (Proto)',
  b'\x9d\xfb\xb5\xd8\xb7u\x04\xc6l\xe3\xacC\x82\xda\xf7\x98 9=\xd8': 'Track and Field (USA) (Proto)',
  b'I\xde\xbd\xf3\xc13\xfd\xe9\x90]m\x917\xda]\x8d\xff\x8db\x83': 'Wizard of Wor (USA)',
  b'\x181[\xec\x1eb\x10"\x90\xfe\xebw\xfb)\x0e\x12f\xcf[\xfc': 'Zone Ranger (USA)',
},
"application/x.atari5200.cart.type19": {
  b'\xe6O\xb1\xa3S\xc7+\x87\xc5H;\xb8\x8c\xb3R>\xa5\xd3{Y': '5200 Menu (USA) (Proto)',
  b'\x18i}\xd8\xe6@\xfa\x15x\xb3\xb4\xae\t\xafHL\x14\xfb\x84\x10': 'Asteroids (USA) (Proto)',
  b'\xd3\xff\x99\xe0\x80\x98\\\x9c\x8e%8\x9d\xfat\x83^\x8c\x9b\xe5\x08': "Atari PAM - Pete's Test (USA)",
  b'\xaf\x81\xc4\x00\xf1\xaf\x0e\xe1:$\xda^\x8bd$B\xa0\xfc\x94y': 'Atari PAM System Test (USA) (v1.02)',
  b'\xce\x9b\xd3\xb0\x84|^\x18|>\x8ff|\x8f^\xa7q\xc4\x99e': "Carol Shaw's River Raid (USA)",
  b'\xea{@\xce\x11\x067\x8e\x1d@\x9c\xabf4\xfb.H?\xf28': 'Dreadnaught Factor, The (USA)',
  b'\xd9\xa6\x95u\x7f\x04\x94\xae\xe9*P\xab\x8aIU\x87E\xe0\xce\xdf': 'Frogger (USA)',
  b'\x80\x92k\xb2p\xeaf\xcb\\\xac\xc3y\xde\x1f\xc1m\x8a=\x81K': 'Gorf (USA)',
  b'\xc7\xeb+JF\xa1\x97\xad\xfe\xdf3\xc2\xcc\xc1}\x84$\xd6\x81\xc4': 'K-Razy Shoot-Out (USA)',
  b'\x9f1\xc8Z\x13\xe9\x19\x13T8Y\x06\x01\xe3;\xef\x11\xa8\xdf\x8a': 'Keystone Kapers (USA)',
  b'\xf5Q$`\xc6\xb6\xd3\xa1\xba\xa0\xb2:\xdfj:7\xfe0\x8e\xb3': 'MegaMania (USA)',
  b'\\\x8e\xc2.\x13\x18\x10\xcc\xac\x08\x06l\xa9n\x99D\xcc\xfaN\xcb': 'Missile Command (USA)',
  b'.z\xe5r`bM\x1fW\x10\xd4E\xa0\x93o\x0c(\nfU': 'Mountain King (USA)',
  b'\x00.o\x8cu3\xb0?\x80A\xe7\xf1\x89e\x89\xc0\xab\x0e\x17\xa6': "Mr. Do!'s Castle (USA)",
  b'\xd1\xc7\xdc=\xe0\x97\xa4\xbf\xab\x82#Z\xce\xc6\x9a\x97\xe6\xe0_\x18': 'Pitfall! (USA)',
  b'\xe6\x96k\xde\\\xd4\x16p \xc9\xd2\x1b\xaba:\xf1d\x8bMu': 'Q-bert (USA)',
  b'\xcfA\x1d\xb7p!{Z\xf5\xfc\xfa2r\\\xc5\x87\xa8\xb9\xe7\x10': 'Space Invaders (USA)',
  b'\xbfw\x04\x8a\xbc6\xba,\xba^\xee\xac\x81(\x02\xd6\xa0\xeb\xa8\x91': 'Star Wars - Return of the Jedi - Death Star Battle (USA)',
  b'\x82\x05\t?_\xb1\xfd\xf0\xcc\xb7\x1df~\xa0\xa3\x84\x9f\xb0\xa0\xf2': 'Super Cobra (USA)',
  b'U\xdaV;C\x94P\x8c#\xd8f\xfb\x1eT2b\xfbESm': 'Zenji (USA)',
},
"application/x.atari5200.cart.type20": {
  b'R6\xbc\x13\xe4\xb0m.\x16\xd7\x0b\n\x0c\x98X\x8e\x17&\xda\x0c': 'Boogie (USA) (Demo)',
  b'\xa8\x0f\x87j\x89\xa8\xce\xac\xc7\x1d\xa5\xd9\xc0+Tip\xb1w0': 'Kaboom! (USA)',
  b'\xf5\xf9\xc7\xf9{\xf0\x9eVG\xe8\x08\xd8\x92\xb6V\xf3\xee\x1e\x19\xe0': 'Super Breakout (USA)',
  b'\xe5\x8e\x1f%y#iM\xbe\xb5\x86\xc8`\xf3\x85\xe7NF\x0f\x90': 'Yellow Submarine (USA) (Demo)',
},
"application/x.atari5200.cart.type4": {
  b'\x1d\xfd\x9c\xb5\xeb`\x85\x90\xc8N\x147b\x91\xf8\xc0/\xb1\xab\x93': 'Ballblazer (USA)',
  b'Q\x0c\xcb \xdf.\xcd\xbe\x7f\x83s\xdej\x9f\xc1\x14\x93\xe8\xc3\xf2': 'Barroom Baseball (USA) (Proto)',
  b'\x9b~\xcdDKb9D.\xf1\xf5L\xd8\xe1L)L\x06M\xa2': 'Black Belt (USA) (Proto)',
  b"N\x02.\x10\x1dsF\xa0\xa8a\x80\x187'@\xe73;J6": 'Castle Crisis (USA) (Unl)',
  b'\xf3\xc3+9X\rX\x97\x99\xdc\xea\xd9\x8b\xb2\x06\xfb\xf3\xdf\x05\x04': 'Gremlins (USA)',
  b'\xca}]\x08q\xbe\x19\xca"\xd0\x10\x19\xfdQ+X6\xe3\x13\n': 'Joust (USA) 32k',
  b'\xa2\x93\xb6\xb0`P\xe1\x10)\xd9q7\xf2\xb1\xb3`YVrs': 'Mario Bros. (USA)',
  b'\xff\x0f\x1d\x9e\xd0\xaeE\xe3\xc8\x1b%/=\n3^\x03\xb0\xaaK': 'Meebzork (USA) (Proto)',
  b'\x19\xffv"\xdb\xbc)\x15\xe3\xe9\x8d\x8d\x96c@\xe7\xe8\x93\xff4': 'Pengo (USA)',
  b'\xa9M\x06s\x9f\xa1\x94S\xfaXj\x83n\xaa\xca\xe1\x1f\xe99\x89': 'RealSports Baseball (USA)',
  b'\x9b!\xac\x7f\xec\x8dUe\x85\xc6\x97\x0211\x8e\xf1\xf5\x01\xb2\xb5': 'RealSports Basketball (USA)',
  b'\x9e\xdb\x06j7\xcavW\x97\x99R\xc9F\xab\x89\xbey&\xf4\xf8': 'RealSports Basketball (USA) (83-10-31) (Proto)',
  b'\x92S\xd7(l\xf8y\x99GK7\xdf$+\xa2;CX\xefJ': 'Rescue on Fractalus! (USA)',
  b'h\x813v\x1cF-he\xbfg>\xd2\x98\xb4\xd8(\x12\t@': 'Spitfire (USA) (Proto)',
  b'lc=\xe5\x0el\xc0\xc0\xc4\x00a\xcd$\xe6\xf7\x8fU*C\x8f': 'Vanguard (USA)',
  b'\xc3\xcf3!m\xfaHC\t\xb9#\xf2\x18)\xd0@Hk\xc4\x81': 'Xevious (USA) (Proto)',
  b'JLJ%\xc8\xed\x1b\n\xe7\x9c\xa2&\x84\xd7\xd3\x11\xa5d\xa7\xb5': 'Zaxxon (USA)',
},
"application/x.atari5200.cart.type6": {
  b'g\xe4\x05\xc3~U\xde\xc4\x07\x10\x18\xe8\x16j\xb4\xfa\xf9\xf2\x1d\xf5': 'Astro Chase (USA)',
  b'\\\xc3\xb4$_\xaf\x01\xc3\x0fw\xac \xb0\xb4\x94\x0fO\x8f\x83\xd8': 'Atari PAM Diagnostics (USA) (v2.0)',
  b'\xe0\x86\xe9\x86\x8a\xa9\x13\xe8\x7f\x94go\xd6\x00\xf0\x03\x89.s\xa1': 'Atari PAM Diagnostics (USA) (v2.3)',
  b'\xa39\xcd\xdf\xc2\x84w\xc2\x1cXx=-;\x0c\xf0\xf4bm\xdf': 'Battlezone (USA) (Proto)',
  b'\xbc|\x81\xf7X\xdd>"NB\xac\x8a\x16\xc1\xa1-i\xae\xd7"': 'Buck Rogers - Planet of Zoom (USA)',
  b'\xeb4;\x90\xd3\xdeini\xb1\xf09\xd4\xfb,j\xfb\r\x80\xc0': 'Centipede (USA)',
  b'\xe0 <!$\x12~\xba-g\x9b\xe5H\xfa%~\xcc[&\x07': 'Congo Bongo (USA)',
  b'\xfamw\xe6\xe9\xaf\xd0\xeaB\xfc\x9f\x98\x07\x93+\xf9\xe8\xfb\x17\x0e': 'Countermeasure (USA)',
  b'D\xf6\xdb\xf5\x01:p\xe6F\xd5\xc5#)\x81\xd5\x17\x06\x9f\xc4I': 'Defender (USA)',
  b'8zv\xbaY\x12%\x1e"\x85\xb7r \x04\xadu\xe9\xed\xe5#': 'Dig Dug (USA)',
  b'i\xdc\xc8\xf5\x9c\xa4I\xf1=\xc8\x96\xb4\xb2\x9bSGb\xd9pd': 'Frisky Tom (USA) (Proto)',
  b'=\x81U\x94\xcfgF\x94\xb0K\xb0\xff\x90\xbfc]}@/\xc2': 'Frogger II - Threeedeep! (USA)',
  b'\xd0\xd8e\x0e\xb0}\x16y\x00MI\x13`\xc2\xc0A\xe3|a\xcd': 'Gyruss (USA)',
  b'\xaevYC\xae\x8d\x90\x9bl\xa6\xff\x15`\xba\x02z\xde\xea\xa5\x15': 'James Bond 007 (USA)',
  b'\x0ft\x12^\xda\xc9\x0f\xfd\x12t\xa1(\xdf)p\xc294\x8f\xb5': 'Joust (USA)',
  b'\xa3\xe7`\xd0\x9eM\xc6\x052\x86\x96uv\xb5\xb7\xd4.\xe2\xfd8': 'Jr. Pac-Man (USA) (Proto)',
  b'\xfa\x02\xee\x04\xcf\xad\x90\xc6\xcc\xf9\xd9\xb9\x84\xe3\xb7\xb32\x7f\xec\x12': 'Jungle Hunt (USA)',
  b'\xc3\xdb\xee\xd7\xad\xc0\xe2\x82\xbfv3*\xb5\xa3 \x95P\xe2m\x02': 'Kangaroo (USA)',
  b'\xbcJSf\x16\xacEh\x9b:\xd7\x90\xf6\xb19N\xbb\xd8\xcc~': 'Looney Tunes Hotel (USA) (Proto)',
  b'/\x15*\x85z\x8aX\xd6\x02\xb3\xc7\x17a-\x97DB>\xe4\x95': 'Microgammon SB (USA) (Proto)',
  b'\xefM>\xe7\xb4\xda.\xf7\x9coM\r\x183\xf1\x0c\x1e\xb7r^': 'Miniature Golf (USA) (Proto)',
  b'~*\x05bU\x98\xc7\xe6a_\x96\x9a\xa6\x95\xe6C\x15\xcd\x98F': "Montezuma's Revenge featuring Panama Joe (USA)",
  b'\xf9=rDMv\x07Rs\xe5\x08UT#\x93e0\x87QH': 'Ms. Pac-Man (USA)',
  b'\xeeK<o\x08\xaa\x9f\xdf\xabS.\xec\x98\x13c\x06\xb6\x92?u': 'Pac-Man (USA)',
  b'\xd5\x04vLX\x94\xee\x99\xc0\xb5\xe2\x02L\xbe\x83>\xbb\xa3\xab\x17': 'Pole Position (USA)',
  b'\xe6\x15\xaa\xa0\xc1\xd5j\xc8;\x91\xf6\x08yp\xfap\x8dd3\xbf': 'Popeye (USA)',
  b'xl\xdfT\xd1b^\r>\xa8\x81E\x06\xfd\xee\xba^N\xceH': 'QIX (USA)',
  b'+C\xf7M\xb4\x832r\xd2\x04\x8d\xe8\xfa2\xb8W\xaf\x8e\t\x0f': 'RealSports Basketball (USA) (82-11-05) (Proto)',
  b'\xa0\xdb\xd6T\xed;\xa2X<\xa1u\xa15i$\xd9R\x98\x84\xdf': 'RealSports Football (USA)',
  b'\xc9\x11\xcc\x91O9#\xd5\xae\xb1\xa8\x97\xe6\xdcJU\x1a\x9aS\xda': 'RealSports Soccer (USA)',
  b'.ltr\xe3y&\xcd\xfd=\x85ZK\xd3z\x18\xfa\x83\xd8\xcf': 'RealSports Tennis (USA)',
  b"\nz\xae\xf5\xb0\xed\xe9'\xce\xa1>rI\x06 ~\x1d\xb4\x80h": 'Road Runner (USA) (Proto)',
  b'\x12~\xd1\xc3\x1c>\x85\xd1\x02\x12\xee\xe7\xb9\xe6\xd7\x1a\xef}\xd8\x91': 'Space Dungeon (USA)',
  b'_\x03T\xd8y-e#\x13\xe5\xa0\xfa]9^\t\x1c\x80\x1d\x97': 'Sport Goofy (USA) (Proto)',
  b'\x19\x80\x14T\xab\x86\xf2\xa5\xcfi\x1f\xcdWzq\xad+\x8d\xc7\xd5': 'Star Trek - Strategic Operations Simulator (USA)',
  b'\xabgo\xfe\x05K#\xbd\xd4<\xee$\x98\xaf\xef\xaa\xbb\xf95[': 'Star Wars - The Arcade Game (USA)',
  b'\t\xd1\x92\xa3Y\x8c\xda\xdf\xb6\x1e\xce\xb40\xb3\xd5\x89.@\x96/': 'Stargate (USA) (Proto)',
  b'\x8bU\x12\xd3\xd2\xa9R\xc9\x9b\x93w\x91\x9f\x0e\x06H\x11a\xd8\x1e': 'Xari Arena (USA) (Proto)',
},
}  # end sha1_signatures
