# Warning! This is generated code.
# 
# Any edits will be overwritten with the next call to parse_gen.py


cpu_id_to_name = {10: '6502', 11: '6502 (including undocumented instructions)', 12: '65816', 13: '65C02', 14: '6800', 15: '6809', 16: '6811', 17: '8051', 18: '8080', 32: 'Jumpman Level', 0: 'Data', 30: 'Antic Display List', 31: 'Jumpman Harvest Table'}

cpu_name_to_id = {'6502': 10, '6502 (including undocumented instructions)': 11, '65816': 12, '65C02': 13, '6800': 14, '6809': 15, '6811': 16, '8051': 17, '8080': 18, 'Jumpman Level': 32, 'Data': 0, 'Antic Display List': 30, 'Jumpman Harvest Table': 31}

valid_cpu_ids = [10, 11, 12, 13, 14, 15, 16, 17, 18]
