sha1_signatures = {
"application/x.vectrex": {
  b'J\x83>Hi"\x1f\x83-\xac\xb7\xef\x0f\xb2\x99|\x87\x0b}\x7f': '3-D Crazy Coaster (1983)',
  b'1\x99t`\x9eB\x92\xa0\xe6\xb6\xa7\xc3Bd\xcb\xa0\x82\x85U\t': '3-D Crazy Coaster (1983) [b1]',
  b'\x171\xe8\x92\xda\x89E\xfeIE\xb48G\xd29`\xc5\x81h*': '3-D Mine Storm (1983)',
  b"\x9e]#\xcb~\xf1\x119N\x18\x04\xe1\x05\xb6\xdd'|\x06-\xf6": '3-D Narrow Escape (1983)',
  b'O\xf2\xca\xa1\x1b\xa1\x11\xb2Va\r4\xa1\x9d\xe9{(\x050&': '3-D Narrow Escape (1983) [b1]',
  b'v2\\\xe2\xd6\x96\xdd\xbf\xde\xf7\xc9\xff2\xe0\xb4\xe3\xd3Q\xe1\xe6': '4-D Rotating Cube Demo (19xx) (PD)',
  b'\x85X\x8c\xa1\x16\xfe6\x95\xf1\x9e\x1d\x06\xdd\x08S"\n[/9': 'All Good Things by John Dondzila (1996)',
  b'q\x94k\xe4}\xeb\xae\x89\xad\x19\x1a\xc3\x82\xc2\x01]j\x86 r': 'Animaction (1983) (light pen)',
  b'\xebv\xf8`\x13\x0f\x8c\xc8x&\xfa\x93\x13PFw`Q\x12\x13': 'Armor Attack (1982)',
  b'\x8f5\xba\xc5\xe8\xe0\xfcJ\xd7)\x1b\xbf\xe5\xaeni2r\xde\x06': 'Art Master (1983) (light pen)',
  b'?\xf3\xfaO\xed\xa6\xd5b\xde\xdf\xf2\xa1\x0c\xfa\x86y\xd81\x80D': 'Bedlam (1983)',
  b'\xe4\xba\xddIN\xd5Hr\x13(\xef\x86\x9d3\xadp\xe3\xe5"\xc9': 'Berzerk (1982)',
  b'\x9e"!\xce_u\xb4\x9a\x08L(L\xc9\xdb\r\xf4\xcc\x8ab\xea': 'Berzerk (1982) (Bugfixed Prototype)',
  b'\x88\x03\x18#\xed*\xf9^\x8azb9\xcd\x83\xb1X\xf9\xb7\xc1j': 'Berzerk (1982) (Bugfixed Prototype) [o1]',
  b'\xaa\x1a\xa8\x11\xc9\xce^\x19\xf4;x\xa6S\x1c\xdd\x0eps7\xbd': 'Blitz! - Action Football (1982)',
  b'`\x89=\xfa\xdaB\xcfP\xc9@\x83\xd4Z\xd5\xbd\x14\xd0\xc01\xc8': 'Blitz! - Action Football (1982) [b1]',
  b'g\xf8Q9X\xc0N\x93k\x13W@\xedN\xc6\xe6\xfa\x17c\xd5': 'Clean Sweep (1982)',
  b')\x98A\xa1\x16\xe5\x93\x99\xafNt\xcc\xae#\xdd\xfc\x07\x8e\x00{': 'Clean Sweep (Mr Boston Version) (1982)',
  b'V\xa7\xf2w\xbf\t\xf9Da\xdd\xa5B\xc0L6&m\x83\x8b\x86': 'Cosmic Chasm (1982)',
  b'\xdb\xb0\xb3x\xaa\xcc*\x18\xdf\xb4v\xaa\x99\xab\x00\xd6\xab9R\x05': 'Dark Tower (1983) (Prototype)',
  b'\xfb\xfdXFMB\x98m\x10\xf1\xb3\x7f2\x10q\xectRTi': 'Dark Tower (1983) (Prototype) (Fred Taft Hack) [h1]',
  b"\xeb\x15\xf5\xc9\x0c|1*y\x98\x16\xaa\x9fi\xc6\xfc'\x06>\xce": 'Dark Tower (1983) (Prototype) [b2]',
  b'u\xf9\x1f\xd2\xc1\xefC\xdc\xc8\xf2\x15\xda;\xb9\x92\xfe\x8c!\xf4\x8b': 'Dark Tower (1983) (Prototype) [b3]',
  b'\xbf\x12BY<\x1f\xa0\x8e\x8c6p\xcbnK\x8dQxc\xd0.': 'Dual Vectrex Test #1 (1983) (PD)',
  b'\xc7\xbc\xa2>\x90\xe6\xf8\xad\xe5\xc6h\xf7\xffU\xc6\xc7,*\x08q': 'Dual Vectrex Test #2 (1983) (PD)',
  b'\xc2&\xd7\x83\x01\xdb\xe4\x0f!\x94\xc1e\xf2\x12MZ\xfa\x88[\xe4': 'Engine Analyzer (1983) (light pen)',
  b'(n\x06\x131(\x9b\x03\x1d\xe37%\xa4\x96l7\x9c\x1d\xf4\x7f': 'Fortress of Narzod (1982)',
  b'A \xa6W\xce*p#q\xd0\x9f\xe7\xfdX\xdc\x9c\xc4\xb7\xe5\xcb': 'Gravitrex by John Dondzila (2002)',
  b'\x04\x06\xc19R\xfb\x0b\xd6\xad\xc6\x82\x98\xae\x16>\xbca`HL': 'Heads Up - Action Soccer (1983)',
  b'\xa2F\xf1\xd9J[\x06+`xG\x87\xa1\xd6\xd4\x99\xb2\x1c\x10\xcd': 'Hyperchase (1982)',
  b'X\xae\xcc\x7f\xb1\xd2j"\xb0\x9e\x9a\x13\xd8E\xe4\xed\x12\xe2D\x9d': 'Melody Master (1983) (light pen)',
  b'e\xd0t&\xb5 \xdd\xd3\x11]@\xf2UQ\x1e\x0f\xd2\xe2\n\xe7': 'Mine Storm (1982)',
  b'\xe0\x90\x0b\xe6\xd6\x85\x8b\x98_\xd7\xf0\x99\x9d\x86K/\xce\xaf\x01\xa1': 'Mine Storm (1982) (Karrsoft Hack) [h1]',
  b'\xa9aD\xbb6\xa9N\xb0\xa2\xf9\x81p\x12\x13\xceYy\x11\x85\xd1': 'Mine Storm (1982) (RLB Hack) [h1]',
  b'\xc6\xc4\x0bOV\xd6\xf8L\xb6GX\xd0\ntT\x96\xe3\xe8\x1dh': 'Mine Storm (1982) (RLB Hack) [o1]',
  b'\x11B\xf7W\x0e\xfc\xcd}\x03\xb4\xf5\xa6DrSC\x03\xbf\xcf\xdf': 'Mine Storm (1982) (RLB Hack) [o2]',
  b'\x1f\xdc\xc6\xe5J\xe5\x17{\xc9\xcd\xc7\x9c\xe6\x16\xae4\x01\xe5\xc2)': 'Mine Storm II (1983)',
  b'|\x07R\xc2K6E\xb7\x163{W\xfeoK\xc4\x06N\x9e\xb0': 'Moon Lander Demo by Clay Cowgill (1997) (PD) [a1]',
  b'>\x8f\xfd?V\x9ci\xddJ\xa6G\xaa\x8f9.\xa38\xfe\xfeB': 'Patriots III - BALListics Busters by John Dondzila (2002)',
  b'H\x1a\xa3\xb7a\xf3GG\xfev\x9c\x84 VNO7R\x9c&': 'Patriots by John Dondzila (1996) [a1]',
  b'\x87$\x8d\xcd=\xb5\x1c\x1e\xbf?\x8e\x07\xe2\x1e\xc0!\xee\xc0\xb9\xbe': 'Polar Rescue (1983)',
  b'\xb7\xbd\x8c\x80\x03\xc6\x055\xe1\xe7vu\x8f\\(\x9a\xa9\xa7~\x81': 'Polar Rescue (1983) [a1]',
  b'\xfe\xc1\x13\xdf\xd3\x93\x1f\x14\xa6\xe7\xca\xd2%W\x18`\xf9\xa7"\x1f': 'Polar Rescue (1983) [a1][o1]',
  b'\x84T\xd4rC^zjL\xa5\xfe\x12\x91bH\x0c\xa6\xe8K\x87': 'Pole Position (1982)',
  b'\x9f[\x0b^l\xafm\x88}67\xac\xe7\n\xd6\xa6%\x8b\x99\x9d': 'Pole Position (1982) [f1]',
  b'"\xf7\xa4\x90\xf6\xf7\xaa\xa3\x1e\x89+\x03\xa8Qx\xaf!\x1fb\xe3': 'Rip-Off (1982)',
  b'\xa1\x0f\xfbn\xfd\xf3*c\xf7\xc4\xda\x8b\xb0\xb7\xf8\xd5\xfdl\xdb\x98': 'Rockaroids Remix - 3rd Rock by John Dondzila (1996) [a1]',
  b'8\xe3\x8b\\`FaF\xd4d\x8f\x89)\xb5\xce:\x08\xdc\xbe\r': 'Scramble (1982)',
  b'8\x07-\xc4\x91\xd5\xc7)\xc0oLP\xbc1\r6\x1e\x0f\xf6*': 'Solar Quest (1982)',
  b'\x0c"\x9d\xd1*U\x04F\x81X_\x7f\xda\xbd7\xdb\xc1$;Y': 'Space Wars (1982)',
  b'|\xe8]\xb8\xdd2hz\xd7b\x961\xae\x118 7\x1f\xaf|': 'Spectrum I+ Demo (1982) (PD)',
  b'\x8b\xbag\x05\xe6\xf6\x9fU\xabm\x10^\xac/\xd5\xb0\xc0\xd1m\x9f': 'Spike (1983)',
  b"w]n_\xd48$j%o'\xa2\xa5\x14Sh\xec},\xbe": "Spike Hoppin' by John Dondzila (1998)",
  b'.\x82\x85\x06\x150\xcc\xad\xfa-\x8d\x11\xc3t\xae$DI\xeb\xfa': 'Spinball (1983)',
  b'\x16\x1d\xb6\x8a\\\x87e?\x87\xbb4\xe0\xbd\xd3\x95\xfc\xbf\x0c\xe0y': 'Spinball (1983) [b2]',
  b'\xc9\xb7\x9f\x1b\xa3\r\xcb\xceZ\x06\x8b`j\x97w\\\xf8c\xfe\n': 'Star Castle (1983)',
  b'\xd6\xb3\x88\xb3\x95\x1e4\x97\x1c\x1b\x00\xc7\xfd\xf8\xdd\x00\xadR\xdf\t': 'Star Hawk (1982)',
  b'\xd1\x82\xea\xc5\xa5O\xfcd/\xcb\xe4\xa4\x95\xc7\xc0\x9f\x87\xb45\x85': 'Star Ship (1982)',
  b':.\xe6\x13\x16\xfb1m\x9e\x9e\xfd\xf5\xee@^\x814u\xb4f': 'Star Trek - The Motion Picture (1982)',
  b'!px\x11\xe5D\xed\xaef\xca\x07cQ\xbf\x07C\x84+F\xe5': 'Star Trek - The Motion Picture (1982) [a1]',
  b'\xf3\xd2\xfa,\x1ex\xda\x1faV\x92c\x1e\t\xee\xe1\xe4\xbe\x18\x84': 'Star Trek - The Motion Picture (1982) [b1]',
  b'\xe3A\x80U\xf1\xe7\x9a#\x00W,U~q\xbf,\xe8\xba\xdb\xfa': 'Test Rev. 4 (1982) (PD)',
  b'OF\x1b\x0bQ\xc8$3\xe7\x010\xea\xc5IO&#\x9bDS': 'Tour De France (1983) (Prototype)',
  b'\x7f\x01F\x970\x9e\xfc\xa2/1\x04H\xac\xb0,\x8e]\xa8\x9cj': 'V-Frogger by Chris Salomon (1998) (PD)',
  b'{\xd5\xc2\x96\x86j:q6\xa9\x90\xe1\x1f1&\xd1\x90A\xb3\xce': 'V-Frogger by Chris Salomon (1998) (PD) [a1]',
  b'\xee\xc5t\x1a\xe8Z\xe1\x0eJ\x0b\xa7\x8e\xf7\xee)F\x8b"1@': 'V-Frogger by Chris Salomon, Sound by Kurt Woloch (2001) (PD)',
  b'\x9f`\xae\xf3\xdd\x8a\x8b\x11\x06\xf8Z\x8e\xdfMe:\x0bQW2': 'Vecmania Part 1 (1999) (PD)',
  b'\x93f\xbb\xf1\x7f\xb4J(\xfe\x0e\x1d\xb2\xf4\xc0\xb8\xc3\x1c\xe0\xcf\xe8': 'Vecmania Part 1 (1999) (PD) [a1]',
  b'\x0c\x8b\x89\x9b"\xd3W\x84\xb7>\x1c;=\x91V\xc19\xaa\xc1L': 'Vecmania Part 2 (1999) (PD)',
  b'xYp)3H\x99j\x18\x03\xa8\xd5\x08?\x81\xa0\xf9\xbe\x14\xb9': 'Vectopia by John Dondzila (2001) (PD)',
  b'\x83\xa6&\x9a\x92u\xd8\xceH\xc5\x8bU`\xfb-s$\xea\xee\xa0': 'Vector Vaders by John Dondzila (1996)',
  b'\xb9\xbb\xf5\xbb\x0e\xacR\xd09\xa4\xa9\x93\xa2\xd8\x06K\x86,\x9e(': 'Vectrex BIOS (1982)',
  b'\x96\x81\xdc]8\xa3\x1c\xb0\xd0\xc6\xc3\xbd\x0f\xc7k\xfd\xdeS\x1c\xce': 'Web Warp (1983)',
  b'\x08\xe2\xdf\x9fwi\xc1jMs\n\xdf\x8de\x9f\x8e\xa4\xbe\x7f\x9d': 'Web Wars (1983)',
  b't]\xdb\xdc(_\x1d\xa4\xf7\xc3\xeb^D\xfe%-P<z\x15': 'Wormhole by John Dondzila (2001) (PD)',
},
}  # end sha1_signatures
