sha1_signatures = {
"application/x.atari2600.starpath": {
  b'L\xda\x82\x17l\xab\xc3\xe52\xeb\xca\xd0\xf63\x17\x9al@;\xae': 'Escape from the Mindmaster (Labyrinth) (1982) (Starpath Corporation, Dennis Caswell) (5) (AR-4200) (PAL)',
  b'\xad\xb2B\x98E~\x15\xfa\xac\xdaS-P\xb7%;\xef\x0e\xf3\xb2': 'Escape from the Mindmaster (Labyrinth) (1982) (Starpath Corporation, Dennis Caswell) (5) (AR-4200) ~',
},
}  # end sha1_signatures
