/***************************************************************
Warning! This is generated code.

Any edits will be overwritten with the next call to parse_gen.py
***************************************************************/


#ifndef _HOME_ROB_SRC_OMNIVORE_LIBUDIS_STRINGIFY_UDIS_CPU_H
#define _HOME_ROB_SRC_OMNIVORE_LIBUDIS_STRINGIFY_UDIS_CPU_H
#include "libudis.h"
#include <stdio.h>
#include <string.h>

extern int stringify_entry_6502(history_entry_t *entry, char *t, char *hexdigits, int lc, jmp_targets_t *jmp_targets);
extern int stringify_entry_6502undoc(history_entry_t *entry, char *t, char *hexdigits, int lc, jmp_targets_t *jmp_targets);
extern int stringify_entry_65816(history_entry_t *entry, char *t, char *hexdigits, int lc, jmp_targets_t *jmp_targets);
extern int stringify_entry_65c02(history_entry_t *entry, char *t, char *hexdigits, int lc, jmp_targets_t *jmp_targets);
extern int stringify_entry_6800(history_entry_t *entry, char *t, char *hexdigits, int lc, jmp_targets_t *jmp_targets);
extern int stringify_entry_6809(history_entry_t *entry, char *t, char *hexdigits, int lc, jmp_targets_t *jmp_targets);
extern int stringify_entry_6811(history_entry_t *entry, char *t, char *hexdigits, int lc, jmp_targets_t *jmp_targets);
extern int stringify_entry_8051(history_entry_t *entry, char *t, char *hexdigits, int lc, jmp_targets_t *jmp_targets);
extern int stringify_entry_8080(history_entry_t *entry, char *t, char *hexdigits, int lc, jmp_targets_t *jmp_targets);
extern int stringify_entry_jumpman_level(history_entry_t *entry, char *t, char *hexdigits, int lc, jmp_targets_t *jmp_targets);
#endif /* _HOME_ROB_SRC_OMNIVORE_LIBUDIS_STRINGIFY_UDIS_CPU_H */
