python gen-sha.py application/x.vectrex ~/retro/vectrex/*
python gen-sha.py application/x.atari2600.cart ~/retro/atari2600/cart/*
python gen-sha.py application/x.atari2600.starpath ~/retro/atari2600/starpath/*
python gen-sha.py application/x.atari5200.cart ~/retro/atari5200/uncat/*
python gen-sha.py application/x.atari5200.cart.type4 ~/retro/atari5200/type4/*
python gen-sha.py application/x.atari5200.cart.type6 ~/retro/atari5200/type6/*
python gen-sha.py application/x.atari5200.cart.type16 ~/retro/atari5200/type16/*
python gen-sha.py application/x.atari5200.cart.type19 ~/retro/atari5200/type19/*
python gen-sha.py application/x.atari5200.cart.type20 ~/retro/atari5200/type20/*
python gen-sha.py application/x.atari800.cart.type2 ~/retro/atari800/cart/type2/*
