sha1_signatures = {
"application/x.atari2600.cart": {
  b'\xfa$\xb0\xf7\x99>\xa9V\xec\x8d\xf9\xa5Q\x9c(1\xb8w\xfcQ': '128 in 1 Game Select ROM (128 in 1) (Unknown) ~',
  b'\xbf\x06\x9aM\xd5\xb6T\x03\xbe\xeal\xc6\r\x91 ew\xea\x1c\x0e': '2 Pak Special - Cavern Blaster, City War (1992) (HES) (773-867) (PAL)',
  b'\x05fM\xccE\xd0,\xe5\x85\xcdy\xd5Q\xc5\xc2\x0f\xc2\xc7\xa83': '2 Pak Special - Challenge, Surfing (1990) (HES) (771-333) (PAL)',
  b'\x0c^\xd7\x16\xe4G\x7f\x8e"\xf6\xdd^bul\xd6\x18\xac\x15K': '2 Pak Special - Dolphin, Oink (1990) (HES) (771-341) (PAL)',
  b'\xe4\xa2j\x82\x85r\xac\x87~P\x86/\xd9,!1\\\x7f.\xd2': "2 Pak Special - Dolphin, Pigs 'n' Wolf (1990) (HES) (771-341) (PAL)",
  b'\x96\xe4(ff\x92%\xc1\xa8\xc0x\xc2c\xe1g\x03\xd0\x80\xff\xdf': '2 Pak Special - Dungeon Master, Creature Strike (1992) (HES) (773-891) (PAL)',
  b'\xb0\x8cn\x94\xbe\xca\xddv\xa7\xbe\rG6"\xa6\x98\xf0\x02U\xef': '2 Pak Special - Hoppy, Alien Force (1992) (HES) (773-875) (PAL)',
  b'\xbf\xcc5\xfe\xee7\x13\xf5\x02\x81"\x84M\x02|\x88\xa8Gr\xb6': '2 Pak Special - Motocross, Boom Bang (1990) (HES) (701-157) (PAL)',
  b'j\xd5\xc4\x120N\xf9*\xf3~\xde\x1d\x8f\xe4.\xc4\x84\xda\xc0\x04': '2 Pak Special - Space Voyage, Fire Alert (1992) (HES) (773-883) (PAL)',
  b'\xf4\x88V\x10P;\xff,L\xa8\x16\xf4\xf2\x8d\x1f\xe5\x17\xb9/5': '2 Pak Special - Star Warrior, Frogger (1990) (HES) (771-422) (PAL)',
  b'@\x05\x8f\x86H\x8c\x90Zg\xcf\xb3\xf6Y]\xbbMb\x15\x00\x11': '2 Pak Special - Wall Defender, Planet Patrol (1990) (HES) (771-406) (PAL)',
  b' \xd4\x1e|\x06\x00Md\xb9\x92-\xd1\xc3\xb4\x9c\xba\xff\xa7\tU': '208 in 1 Game Select ROM (208 in 1) (Unknown) (PAL) ~',
  b'\x0e\x14n^\xb1\xa6\x8c\xbaO\x8f\xa5^\xc6\xf1%C\x8e\xfc\xfc\xb4': '3-D Genesis (1983) (Amiga - Video Soft, Jerry Lawson, Dan McElroy) (Prototype) ~',
  b')\xe41\xc2}\xdb\x86I\xec?\x15\x96`\x11PX"\x90q9': '3-D Ghost Attack (1983) (Amiga - Video Soft, Michael K. Glass, Jerry Lawson) (Prototype) ~',
  b'\xceamf\xd4\xed\x90\xdd\xc9r\xea"U]\xf7?kA\xaf\x8c': '3-D Havoc (1983) (Amiga - Video Soft, Frank Ellis, Jerry Lawson) (2110) (Prototype) ~',
  b'!\xd9\x83\xf2\xf5+\x84\xc2.\xca\xe8K\tCg\x8a\xe2\xc3\x1c\x10': '3-D Tic-Tac-Toe (1980) (Atari, Carol Shaw - Sears) (CX2618 - 49-75123) ~',
  b'k\x16:\xa9g\xe4 J[\xd9\x8aY\xbd\x8e\x80\xf1Y\x00N4': '3-D Tic-Tac-Toe (1980) (Atari, Carol Shaw) (CX2618, CX2618P) (PAL)',
  b'[\xef)=\x89\x17G\xe7\x90\xcc\x08?\x1df>\xfe7,\xec;': '3-D Zapper (12-15-82) (U.S. Games Corporation - JWDA, Todd Marshall) (Prototype) ~',
  b'\xafu\xd0\xe4\xb9\xc3\xf4mDA!\xc9\xcf\x96\xce\xb3\xa3I\xfc\x1d': '3-D Zapper (U.S. Games Corporation - JWDA, Todd Marshall) (Prototype)',
  b'\x8d\xcc\xc9\x0b\xe6\x02\x90:\x01\x96\xe8Yb\xed\xa6/\x99\x18\x1a\x17': '32 in 1 Console ROM (02-10-1989) (Atari) (Prototype) (PAL)',
  b'\x97\xff\xc2RC\x8a\\\x93a\tj\x11Q\xbd\xbfy\xbdq|\xb3': '32 in 1 Game ROM (1988) (Atari) (Prototype) (PAL)',
  b"\xfd\x8f*n\xb9$\x82'\xed\xcaj\xe6\r\xe9<\xa3\x8c\x0e\xd2\xea": '3D Tic-Tac-Toe (32 in 1) (1988) (Atari, Carol Shaw) (CX26163P) (PAL)',
  b'rL\x1c\xe3R\xd0!\x96\x99\x89/|x\x08?\x82Zq\xac\x1a': '4 Game in One - Ice Hockey, Phantom UFO, Spy Vs. Spy, Cosmic Avenger (1983) (Bit Corporation) (PAL)',
  b'\xb6\x198&\xdfQ\x1f9\x07p\x136\x9eb^VW\x7f/6': '4 Game in One - Rodeo Champ, Open Sesame, Bobby Is Going Home, Festival (1983) (Bit Corporation) (P460) (PAL)',
  b'\xbc`\x98{f\x8do\xa2lI\xf5\xab\xa5\x81V"q\x01@\xd9': '4 in 1 - Canyon Bomber, Home Run, Night Driver, Sky Diver (02-19-1987) (Atari) (CX26137) (Prototype)',
  b'\xc0f\x94$F\xc3\xfcpu\xee\x840\xa3\x81\xd5\xdaJ\xda\xecA': '8 in 1 (Supergames 8 in 1) (01-16-92) (Atari) (CX26193) (Prototype)',
  b'SA5w\xaf\xe7\xde\xf1\xd3\x90\xe3\x89,E\x82$\x05Q<\x07': 'A-Team, The (AKA Saboteur) (03-30-1984) (Atari, Jerome Domurat, Howard Scott Warshaw) (CX26133) (Prototype)',
  b'\xbc\xbd68\x9cV\xe9\xb4\x08\xe8\xfcK\xc5\x95\xdcs\x1a\xaf\x0e\x02': 'A-Team, The (AKA Saboteur) (05-08-1984) (Atari, Jerome Domurat, Howard Scott Warshaw) (CX26133) (Prototype)',
  b'\xd7\xc6-\xf80\nh\xb2\x1c\xe6r\xcf\xaaM\x0f/K=\x0c\xe1': 'Acid Drop (1992) (Salu, Dennis M. Kiss) (460758) (PAL) ~',
  b'\x9eo\xb0G\xee\x9f\xa0\xa4T\xca#g>\xd9i40\x03-\xc6': 'Action Man - Action Force (AKA G.I. Joe - Cobra Strike) (Paddle) (1983) (Parker Brothers, John Emerson) (931511) (PAL)',
  b'\x18\x04\xad\x16\x9e*\xf9\xe1Hv\xff\xb1&\x18\x17\xb7\xeb\x84o\x1b': 'Actionauts (Microbots) (1984-2008) (Imagic - Advanced Program Technology, Rob Fulop) ~',
  b"\xe0\xd4ue\xdf\x93\\\x06K\xc4\x05V6\xf3z\x042\xab7'": 'Activision Decathlon, The (1983) (Activision, David Crane) (AG-930-04, AZ-030) [fixed] ~',
  b'qvV\xf5a\x82>\xda\xa6\x92@G\x1c1\x06\x96?\\0~': 'Activision Decathlon, The (1983) (Activision, David Crane) (AG-930-04, AZ-030) ~',
  b'}\xa9\xcc#\x83\xd0\x985\xd3E\xec}nv\xd2#Y\x1d**': 'Activision Decathlon, The (1983) (Activision, David Crane) (EAZ-030) (SECAM)',
  b'\x08/\xc9\x14\xd6\xeb\xf4\x15\x92o\xbbl\xf7\x81\xde \x9aM\x05-': 'Activision Decathlon, The - Zehnkampf (1983) (Activision, David Crane - Ariola) (EAZ-030, EAZ-030-04B, EAZ-030-04I - 711 030-725) (PAL)',
  b'\xe0~H\xd4c\xd3\x03!#\x9a\x8a\xcc\x00\xc4\x90\xf2\x7f\x1ft"': 'Adventure (1980) (Atari, Warren Robinett - Sears) (CX2613 - 49-75154) ~',
  b"\xec#\xbb\xea\x83\xcfG't\x82\xe7u\xadX\xac\x8em\x14\x83\xfe": 'Adventure (1980) (Atari, Warren Robinett) (CX2613, CX2613P) (PAL)',
  b'\x03\xa4\x95\xc7\xbf\xa0g\x1e$\xaaM\x94`\xd22s\x1fh\xcbC': 'Adventures of TRON (TRON Joystick) (1983) (M Network - APh Technological Consulting, Hal Finney, Glenn Hightower, Peter Kaminski - INTV) (MT4317) ~',
  b'nB\x05D\xbf\x91\xf6\x03c\x91\x88\x82J+W\x078\xbb~\x02': 'Adventures on GX-12 (AKA Adventures of TRON) (1988) (Telegames) (4317 A009) (PAL)',
  b';\x02\xe7\xda\xcbA\x8cD\xd0\xd3\xdcw\xd6\n\x96c\xb9\x0b\x0f\xbc': 'Air Raid (Men-A-Vision) (PAL) ~',
  b")\xf5\xc7=\x1f\xe8\x06\xa4(EG'M\xd7?\x99r\xa7\xedp": 'Air Raiders (Air Battle) (1983) (M Network - APh Technological Consulting, Larry Zwick - INTV) (MT5861) ~',
  b'\xe6Z\x0clZ\x1f\x9f\x05\xeb\xcf\xaa\xa7\xb2\xc9\xeef%\xbf-\x83': 'Air Raiders (Unknown) (PAL)',
  b'\xea\xc7\xd1\xfb"}%z j`\xdaK\xf7n\xe5\xe5q\xb3\x94': 'Air-Sea Battle (32 in 1) (1988) (Atari, Larry Kaplan) (CX26163P) (PAL)',
  b'\xaf[\x9f3\xcc\xb7w\x8bB\x95}\xa4\xf2\x0f+\xc0\x00\x99#f': 'Air-Sea Battle (Hack) (Unknown) (4K)',
  b'\xb9lzP\x9b\xf6\x10\xf6\x1f\x827{\xfdPm\xb3\xdb\xa2\xb4#': 'Air-Sea Battle - Air Sea Battle (Anti-Aircraft) (1977) (Atari, Larry Kaplan) (CX2602, CX2602P) (PAL)',
  b'\xa7F\xfd\xc8+3j\x9dI\x9b\xf1\x7fP\xb4\x1e\x01\x93\xbaY^': 'Air-Sea Battle - Air Sea Battle - Target Fun (Anti-Aircraft) (1977) (Atari, Larry Kaplan - Sears) (CX2602 - 99802, 6-99802, 49-75102) ~',
  b'\t\xae7\x15\x87:\x888k;\xc1\x1eB\x9f\x82\x91\x9b\x91\xfa\x1a': 'Airlock (1982) (Data Age) (DA1004) (Prototype)',
  b'\x03v\xc2B\x81\x9bxS\x10\xb8\xafC\xc0;\x1d\x11V\xbd_\x02': 'Airlock (1982) (Data Age) (DA1004) ~',
  b'\xa9K\x108\x06\xd35F\xaa\x8f\xac\x00V\xcdW\xb8\x18\xbf\xed\x8c': 'Airlock (1983) (Gameworld) (133-004) (PAL)',
  b'{\xe2\x89\x9dwl\xc2\x8f\x1c\xa8\x15\x84\xa0.\\"\x8aO`>': 'Alices Abenteuer - Lilly Adventure (1983) (Quelle - Otto Versand) (732.273 8 - 600273, 781644) (PAL)',
  b'\x01\xd9\x9b\xf3\x07&(%\xdbXc\x1e\x80\x02\xdd\x00\x8aB\xcb\x1e': "Alien (1982) (20th Century Fox Video Games, Douglas 'Dallas North' Neubauer) (11006) ~",
  b'\xab\xf1\x86\xc0\x10\x87\x91\xd9\x98\xf7s\x83$\xd0\xe4\x9a\x97{\x01[': 'Alien (CCE)',
  b'\xa1y]cq\xdd\xb0\xbc\x93\xbam\xf7\x02\x99\xccT\x8d\xc8\x8b\x99': 'Alligator People (1983) (20th Century Fox Video Games, John Russell) (Prototype) ~',
  b'\xea\x11$\xa3\xea\xc3\x913\xf5\x9dfEE\xc3<\xc5\x14\x02 g': "Alpha Beam with Ernie (Alpha Beam) (Kid's Controller) (06-03-1983) (Atari - CCW, Michael Callahan, Preston Stuart) (CX26103) (Prototype) (PAL)",
  b'q\x18\xf2@\x1a\r4\x99gb\xd9\x83\x8d/K\xac\xf7E\xfd\xe3': "Alpha Beam with Ernie (Alpha Beam) (Kid's Controller) (12-22-1983) (Atari - CCW, Michael Callahan, Preston Stuart) (CX26103) (Prototype)",
  b'K\xe3\xffI\x04\xfcu\xe0G\xc7.\xc7\xe1\xb1\xb3a\xc7\xaemP': "Alpha Beam with Ernie (Alpha Beam) (Kid's Controller) (1983) (Atari - CCW, Michael Callahan, Preston Stuart) (CX26103) (PAL)",
  b"\xa1\xf6`\x82|\xe2\x91\xf1\x97\x19\xa5g/,]'}\x90;\x03": "Alpha Beam with Ernie (Alpha Beam) (Kid's Controller) (1983) (Atari - CCW, Michael Callahan, Preston Stuart) (CX26103) ~",
  b'\xfb\x9b\xbaeV\xfcE\x90M\xac\x87P\xfa\x18\x15^a\x96\xb2\xc0': 'Amidar (1982) (Parker Brothers, Ed Temple) (931504) (PAL)',
  b'\xb8\x9aZ\xc6Y>\x83\xfb\xeb\xee\x1f\xe7\xd4\xce\xc8\x1ap2\xc5D': 'Amidar (1982) (Parker Brothers, Ed Temple) (PB5310) ~',
  b'=6#\xfa\r\x87\xea\x02\xf2\xa0\x80\xd9Y\xa1\'">u{?': 'AndroMan on the Moon (1984) (Androbot - Western Technologies, Michael Case, Lenny Carlson) (Prototype) ~',
  b'\xa7c\x8a.yRK\x06\x808\xf0\xbe\x03[F\xd0\xcee\xf7\xea': 'Angeln I (AKA Fishing Derby) (Videospielkassette - Ariola) (PGP237) (PAL)',
  b'\xacX\xac\x94\xce\xabxrZ\x11\x82\xcc{\x90sv\xc0\x11\xb0\xc8': 'Angriff der Luftflotten - Paris Attack (AKA M.A.D.) (1983) (Quelle - Otto Versand) (495.463 2 - 746381) (PAL)',
  b'TZ\xbc\x0c\xdaZ\xb3\xf7k_\x1f\xb6]\x95M9\xda\xf3\xe1\xc3': 'Ant Party (AKA Cosmic Swarm) (32 in 1) (1988) (Atari) (CX26163P) (PAL)',
  b'b\xe9%\x08\xcfC\xac\xc2\x8b\xff\x8b\x03\xa1\x97\x13}V\xa5\xdfy': 'Apples and Dolls (AKA I Want My Mommy) (CCE)',
  b'\xcb@\r\xe2e>\x12^pJ\xbd\x8b\x0f\xe5\xdd\xdbC\xe3C\x8b': 'Aquaventure (CCE)',
  b'}\x13*\xb7v\xffu[\x86\xbfO Ae\xaaT\xe9\xe1\xf1\xcf': 'Aquaventure (Sea Sentinel) (08-12-1983) (Atari, Tod Frye, Gary Shannon) (Prototype) ~',
  b'a""\x0c\x8f\xbfB\xfd\xf0Cps\xeb\xe2@}o\x87Z@': 'Armor Ambush (1988) (Telegames) (PAL)',
  b'\x9bjT\x96\x92@\xba\xf6I(\x11\x87A\xc3\xaf\xfe\xe1H\xd7!': 'Armor Ambush (Tank Battle) (1982) (M Network - APh Technological Consulting, Hal Finney - INTV) (MT5661) ~',
  b'\x011B\xaa\x138\xd12rHv\xc1\xd5Z2*\x1d\x04\x93\x80': 'Artillery Duel (1983) (Xonox - K-Tel Software - Action Graphics, John Perkins, David Thiel) (6230, 7210, 06004, 99004) (PAL)',
  b'\x8c$\x9e\x9e\xaa\x83\xfck\xe1`9\xf0^\xc3\x04\xef\xdf\x98{\xeb': 'Artillery Duel (1983) (Xonox - K-Tel Software - Action Graphics, John Perkins, David Thiel) (6230, 7210, 06004, 99004) ~',
  b'M\x06\x89\xa9\x80\x08\xbf\x89\xca\xab\x94\x92I\xff\x0fT:\xbf^\x0c': 'Ases do Ar (AKA Sky Jinks) (Dismac)',
  b'\x0c\x03\xeb\xa9}\xf5\x17\x8e\xec]M\n\xeaJo\xe2\xf9a\xc8\x8f': 'Assault (AKA Sky Alien) (1983) (Bomb - Onbase) (CA281)',
  b"\x1a\tO\x92\xe4j\x81'\xd9\xc2\x98\x89\xb58\x98eV\x1c\no": 'Asterix (AKA Taz) (1984) (Atari, Jerome Domurat, Steve Woita) (CX2696)',
  b'\xff1\xd8\x85\xea=\x96\x85\x05G\xfd\xb8\x97\x8d\x12\xd0\x96Z,\x1f': 'Asterix (AKA Taz) (1984) (Atari, Jerome Domurat, Steve Woita) (CX2696) (PAL)',
  b'\x8fJ\x00\xcbJ\xb6\xa6\xf8\t\xbe\x0e\x05]\x97\xe8\xfe\x17\xf1\x9e}': 'Asteroid Fire (1983) (Home Vision - Gem International Corporation - VDI) (VCS83111) (PAL) ~',
  b'\x84#\xf9\x90\x92\xb4T\xae\xd8\x9f\x89\xf5\xd7\xdae\x8c\xafz\xf0\x16': 'Asteroids (1981) (Atari, Brad Stewart - Sears) (CX2649 - 49-75163) [no copyright] ~',
  b'\xd6\x897\xe5z6~a\xea\xa4\xb4EP\xae\x8b\x9diEfa': 'Asteroids (1981) (Atari, Brad Stewart - Sears) (CX2649 - 49-75163) ~',
  b'\x1c\xb8\xf0W\xac\xadm\xc6_\xef\x07\xd3  \x88\xffJ\xe3U\xcd': 'Asteroids (1981) (Atari, Brad Stewart) (CX2649, CX2649P) (PAL)',
  b"|\x0fR-W\x96\xc4\xf3\x07\xdc\xcf\x9c\xe5\x15\xab\x02'x\xb3\xc7": 'Asteroids (1981) (Atari, Brad Stewart) (CX2649, CX2649P) (PAL) [no copyright]',
  b'\x8e\x83\xfdw\xe1\x98\xb2~\xb5\xb0*m\xc0*Z\xb3{Zv\xbd': 'Astroblast (Paddle) (1982) (M Network - APh Technological Consulting, Hal Finney - INTV) (MT5666) [fixed] ~',
  b'\xb8P\xbdr\xd1\x89\x06\xd9hN\x1crQ\xcbi\x95\x88\xcb\xcfd': 'Astroblast (Paddle) (1982) (M Network - APh Technological Consulting, Hal Finney - INTV) (MT5666) ~',
  b'\xab\x8a\xf3\xc4\xe0\xd8\x85\x84\x81\xf3\xe5+Y\xb5\x04\x0b\x18\xce\xb4\xd7': 'Astroblast (Paddle) (1988) (Telegames) (PAL)',
  b'\xb6\xf0\x17\x97\xe3\xe8\xd8\x0fN3\xdas\xe6J\xa2\x93Z\xf9`8': 'Astrowar (Unknown)',
  b'\xb8L4\xaa\xed\xd8O\xca0\xb6\xe8">\xe0bC\x9a\xcfO\xe0': 'Atari VCS Point-of-Purchase ROM (1982) (Atari) ~',
  b'\xd1V<$ \x87f\xcf\x8d(\xdez\xf9\x95\x02\x1a\x9f\x89\xd7\xe1': 'Atari Video Cube (Atari Cube, Video Cube) (1983) (Atari - GCC) (CX2670) ~',
  b'\xe9\xd6\xac?\xd8\x07\xccq\xdem\x91D\x17|\x18\xf3\xa7\x88\x1b[': 'Atlantis (1983) (CCE) (C-832)',
  b"\\\x9b\xe3\xbd\x18\x0eF'\x11\x7f\xba\xb3\xfaJ\x9a\x80\xf3%R\xd5": 'Atlantis (1983) (CCE) (C-832) [a]',
  b'\x08l\xa9\xc0\xba\xb6\x92\x10E\x903\x8c\xe8\xaa\xa0a\xdf\xa2\xdd\xe4': 'Atlantis (1983) (Digitel)',
  b'\xcb\xda\xcef\t#\xe0\xb2\xc6\xd9|$\x87{\r\xc1e!\xd9\x13': 'Atlantis (1983) (Dynacom)',
  b'\xc2/\xde1\xc8\x85\xb3\xb6\xf51\xbb\xee\x96\xf9\x16\x90~^=\xaf': 'Atlantis (Dactari - Milmar)',
  b'\x08\xc5Cq\xfd|r\xdf\x1f6\xaa\xe41\x88pUa\xa3\xf8\xbd': 'Atlantis (Hack) (208 in 1) (Unknown) (PAL)',
  b'\x8e\xeb\x85\xd9\x8eC\x19\xa2;v\xcc\xca\xb9%sKx-\xca\x8b': 'Atlantis (Hack) (Unknown) (PAL)',
  b'\xc6\xb1\xdc\xdb/\x02J\xb6\x821m\xb4Wc\xba\xcciI\xc3<': 'Atlantis (Lost City of Atlantis) (1982) (Imagic, Dennis Koble) (720103-1A, 720103-1B, IA3203, IX-010-04) ~',
  b'{\x99u\xdc\x1a\xab\xc3M\xef-\x19\x85K9\xf27\xe4\xaf\xe0\x88': 'Atlantis (Lost City of Atlantis) (1982) (Imagic, Dennis Koble) (720103-2A, IA3203P, EIX-010-04I) (PAL)',
  b'\xf4\xe88\xde\x91Y\xc1I\xac\x08\n\xb8^O\x83\r[)\x99c': 'Atlantis II (1982) (Imagic, Dennis Koble) (720103-1A, IA3203)',
  b'\xff*L\xec\x80\x05\xa1\xb0L g\x1f\x9d\x06\xae\xfe\xbb\xb0\xfd\xa1': 'Atom Smasher (1984) (Video Soft) (Prototype) ~',
  b'\x045\x90y\xd6\x96\r\n\xbc/I3\x08s\xfa\xc1\x0c\x9d\x9e!': 'Autorennen (AKA Grand Prix) (Videospielkassette - Ariola) (PGP232) (PAL)',
  b'lV\xfa\xd6\x88\xb2\xe9\xbbx?\x8aZ#`\xc8\n\xd23\x8eG': 'BASIC Programming (Keyboard Controller) (1979) (Atari, Warren Robinett) (CX2620) ~',
  b'\xffM\xa0\x13\x19\x15:+\xde-,\xb1\xbb\x1d\xd9\x00\xae\x10^\x9a': 'BASIC Programming (Keyboard Controller) (1979) (Atari, Warren Robinett) (CX2620, CX2620P) (PAL)',
  b"0\xd5 A\xc5\x9a$\n\x8e\xbe'I\xe4\xde2\xbb\xc7I\x88\xee": 'BMX Air Master (1989) (Atari) (CX26190) (PAL)',
  b'\xb5J\xb7\x00\xec\x94A\xae\xb7&e\xd0\x95GP5\xa1U\xaa"': 'BMX Air Master (1989) (TNT Games - Sculptured Software, Adam Clayton) (26192) (PAL)',
  b'\xff%\xed\x06-\xccC\x04H\xb3X\xd2\xactW\x87A\x0e\x11i': 'BMX Air Master (1989) (TNT Games - Sculptured Software, Adam Clayton) (26192) ~',
  b'u\xe7\xef\xa8a\xf7\xe7\xd8\xe3g\xc0\x9b\xf7\xc0\xcc5\x1bG/\x03': 'Bachelor Party (Paddle) (1982) (PlayAround - J.H.M.) (205)',
  b'\xb8\x8c\xa8#\xaa\xa1\nzJ=2P#\x88\x1b-\xe9i\xc1V': 'Bachelorette Party (AKA Bachelor Party) (Paddle) (1982) (PlayAround - J.H.M.) (202)',
  b'\x9b\x1d\xa7\xfb\xd0\xbfo\xca\xdf\x1b`\xc1\x1e\xeb1\xb6\xa6\x1a\x03\xc3': 'Backgammon (Paddle) (1979) (Atari, Craig Nelson - Sears) (CX2617 - 49-75183) ~',
  b'\xbd\xa9\xfb\xd4\xc7\\,p\xab\njc\x9a\n\x83\x9d\x03\x97\xb56': 'Backgammon (Paddle) (1979) (Atari, Craig Nelson) (CX2617, CX2617P) (PAL)',
  b'\xc4\xce\t\xbeK\x99{7\xb0\x98\x03\xd0@\x1eJu\xb6\xf2\x95\xdb': 'Bank Heist (208 in 1) (Unknown) (PAL)',
  b'\x07\xa6\xad\xc7\x82\x99\xd9\x81\xcf\xdf-\x96\xe7\x08\xb1\xf8\xe4\xf7\xfbK': 'Bank Heist (Action Hi-Tech - Hi-Score) (PAL)',
  b'\x80\xd4\x02\x05u\xb1N\x13\x0f(\x14k\xf4Y!\xe0\x01\xf9\xf6I': "Bank Heist (Bonnie and Clyde, Cops 'n' Robbers, Holdup, Roaring 20's) (1983) (20th Century Fox Video Games, Bill Aspromonte) (11012) ~",
  b'7&c\t{A\x9c\xedd\xf4N\xf7C\xfe\x8d\n\xf41\x7fF': 'Barnstorming (1982) (Activision, Steve Cartwright) (AX-013) ~',
  b'r\xf0\xbd5\xb4\x9eJ5\xf6\xe5\x1d\xf6\xec\x8e\xa5y\xb5\xa6s\xeb': 'Barnstorming (CCE)',
  b'V\x8dy\r\x1d\xc8\xc2\x81g\xb6\xc0a\xaeL\x00\xbe\xa785<': 'Barnstorming (Unknown) (PAL)',
  b"\x0e\x14k@\x81\xae\x08\xd7\xa5'\xbd\x0ez\xbd#\xbb\x1dB\xdf\xfb": 'Barnstorming - Die tollkeuhnen Flieger (1982) (Activision, Steve Cartwright - Ariola) (EAX-013, PAX-013 - 711 013-720) (PAL)',
  b'\xd0\xbd\xd6\t\xeb\xc6\xe6\x9f\xb3Q\xbaF\x9f\xf3"@k\xcb\xabP': 'Base Attack (Hack) (Unknown)',
  b'*\xe3\xbf"\x95\x02\x01\x92\xac\xa5\x87\xaaZO$&T?\xbf\x80': 'Baseball (AKA Super Challenge Baseball) (1988) (Telegames) (5665 A016) (PAL)',
  b'\xdd\xa6\x85\xce\xa7N .\x1bG;\x0f\x89\x0c9\x10\xc8\x9c\x7fy': 'Basic Math (208 in 1) (Unknown) (PAL)',
  b'\xbf\xfe\x99EL\xb0UU.]a/\r\xba%G\x0172\x8d': 'Basketball (1978) (Atari, Alan Miller - Sears) (CX2624 - 6-99826, 49-75113) ~',
  b"\x93u\xc5\xc5)\x8e\x81\xb3r$\xdb\xde\x9b\xc5\xaf\x15\x11\x81\xca'": 'Basketball (1978) (Atari, Alan Miller) (CX2624, CX2624P) (PAL)',
  b'\xa2}\xe2?\x80 y-\xec\xdf\x94v\xcc\x04zp\x91Ob\x00': 'Basketball (208 in 1) (Unknown) (PAL)',
  b':W6\xb2\xfb\x19E\x14c\x80\x1b\xb7j\x81\x8a\x93\xf6\xc5\xdc)': 'Basketball (32 in 1) (1988) (Atari, Alan Miller) (CX26163P) (PAL)',
  b'\xb3\x10\xd3\x82i\x16\x889\xc9S\xeeD[z\xd2W\x0b\xc2\xf8\xeb': 'Basketball (Hack) (208 in 1) (Unknown) (PAL)',
  b'\xfa\xfdR.6\xa4\x88\x8d\xfc*@\xb1\xd5\x87\x9e\xe8\xa3=y1': 'Battlezone (05-02-1983) (Atari - GCC, Mike Feinstein, Brad Rice) (CX2681) (Prototype)',
  b'>"m\x8f\xfa\x97[\x93}\x80\x8fs\xf7\x9b?\x05,;\x06~': 'Battlezone (05-12-1983) (Atari - GCC, Mike Feinstein, Brad Rice) (CX2681) (Prototype)',
  b'\xe4\x13J;J\x06\\\x85h\x02\xbc\x93\\\x12\xfa~\x98h\x11\n': 'Battlezone (1983) (Atari - GCC, Mike Feinstein, Brad Rice) (CX2681) ~',
  b'\xd3:6\x7f\xc4\x89\x0b\xea\x8c-\x9ar\x83\xd8G \xb2\x0c\xab\xf8': 'Battlezone (1983) (Atari - GCC, Mike Feinstein, Brad Rice) (CX2681, CX2681P) (PAL)',
  b'Ga\x9e\xdb5/\x7f\x95_\x81\x1c\xbb\x03\xa0\x07F\xc8\xe0\x99\xb1': 'Beamrider (1984) (Activision - Cheshire Engineering, David Rolfe, Larry Zwick) (AZ-037-04) ~',
  b'\n\xe5\xd0\xf0\x1ec\xe0S\xba\xbaK\xda\xae\xd8\xa8.Y3J\xaf': 'Beamrider (1984) (Activision - Cheshire Engineering, David Rolfe, Larry Zwick) (EAZ-037-04, EAZ-037-04I) (PAL)',
  b'\xfa\xd0\xc9s1\xa5%\xa4\xae\xbag\x98uR\xba2F)\xa7\xa0': 'Beany Bopper (1982) (20th Century Fox Video Games - Sirius Software, Grady Ward) (11002) ~',
  b's\xd5\\\xf6<\xca\xbf\x88\x11H\xe7\x1d\xfd\xbd}DE\x88\x9f\x80': 'Beany Bopper (1983) (CCE) (C-835)',
  b'\x8b}j\x0c^\x14\xc5-\xdb\x18\xe7\xbe|\x84\xf4Z\xbe\xf9K\xc0': "Beat 'Em & Eat 'Em (Paddle) (1982) (Mystique - American Multiple Industries, Joel H. Martin) (PAL)",
  b'\xe2\xc2\x9d\ns\xa4WP(\xb6-\xcatTv\xa1\x7f\x07\xc8\xf0': "Beat 'Em & Eat 'Em (Paddle) (1982) (PlayAround - J.H.M.) (204)",
  b'\x8e\x8e3O\xa4\x81i\x8d\n\xdb\x7f\x9c\xb7\x85.I\xacF.\xb9': "Beat 'Em & Eat 'Em (Paddle) (1983) (Dynacom)",
  b'\xc3\xaf\xd7\x90\x9br\xb4\x9c\xa7\xd4HTe\xb6"\xd5\xe5_\x89\x13': 'Berenstain Bears (Kid Vid Voice Module) (1983) (Coleco) (2658) ~',
  b'?V\xcf\xd8\xfe\xa7\x1a.\x96i\xe0\xf79oDr\x92\xdc\xa2~': 'Bermuda (AKA River Raid) (Unknown) (PAL)',
  b'\xfc\xad\x0eQ0\xde$\xf0k\x98\xfb\x86\xa7\xc3!HA\xcaB\xe2': 'Bermuda Triangle (1983) (Data Age, J. Ray Dettling) (112-007) ~',
  b'\x9aVj\xe95O\xc9M\xa9\xfezS\xdbI\xdb\xa5\x7fS\xc2`': 'Bermuda Triangle (1983) (Gameworld, J. Ray Dettling) (133-007) (PAL)',
  b"'\\\xc1&H\x1b\xdct~\x1ea\xe4\x8fR%\x00b\xda\x97R": 'Bermuda, The (AKA River Raid) (Rainbow Vision - Suntek) (SS-009) (PAL)',
  b'\xbd\t\x89\xa2\xbat\xd2\xb0T>\\|\x88\xc3\x1fRJ\xa9\xef\x11': 'Bermuda, The (AKA River Raid) (Rainbow Vision - Suntek) (SS-009) (PAL) [a]',
  b'\x08\xbc\xbc\x89TG>\x8f\x02B\xb8\x811[\n\xf4Fi\x98\xae': 'Berzerk (1982) (Atari, Dan Hitchens - Sears) (CX2650 - 49-75168) ~',
  b"\xaf\xb9\x83\xa2\xf8\x81\xa0\x17\xb2\xcf\xb5\x93\xb0\xd3'B\t6\x12\xdd": 'Berzerk (1982) (Atari, Dan Hitchens) (CX2650) (PAL)',
  b'k\xff\xbb\xe1 vb\xde\x89\xc7\xea\x02\xc1\xfc\xd0\x04\x07\x1f\xb9\xe4': 'Berzerk (CCE)',
  b'T\x9a\xbaro\x9a\x11\xc4\x14\xdd\xb1a\x93\x10N.Y\xb3\xde\x96': 'Berzerk (Unknown) (PAL)',
  b'\x0bM\xee\xf0\xd1\xf6\xd9^&\x92>\x08\x89\x1eI\xdb\xc3\xc5-[': 'Berzerk (Unknown) (PAL) [a]',
  b'\xd3p\xe0 V~H\x8e\x9aiM\xff\xcf6\xbe\xf8\xe4TH\xbc': 'Bi! Bi! (AKA Skindiver) (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co) (PAL)',
  b'\xaa\xfb\x88\xef\x8f\x95\xb1\x05\x1c\xb8\xb6\xa1S\xb1e!\\\x95\xc7\xbd': "Big Bird's Egg Catch (Grover's Egg Catch) (Kid's Controller) (05-02-1983) (Atari - CCW, Christopher H. Omarzu) (CX26104) (Prototype)",
  b'\xe2\xb00M\xe6\x95\xce\xa7\xc9\x83Dd"T\xb7\xf1\x81\xe9(\xd2': "Big Bird's Egg Catch (Grover's Egg Catch) (Kid's Controller) (05-17-1983) (Atari - CCW, Christopher H. Omarzu) (CX26104) (Prototype)",
  b'\xa3\xab6\x99\xd9E\xac5j\xef\xf67\x04o\x19\x7f\x9dopc': "Big Bird's Egg Catch (Grover's Egg Catch) (Kid's Controller) (12-08-1982) (Atari - CCW, Christopher H. Omarzu) (CX26104) (Prototype)",
  b'\xfd\x98"\xecT\xad\xd0Ei\xb6)\xd61\x18-\x9d\xea\xc7\x18M': "Big Bird's Egg Catch (Grover's Egg Catch) (Kid's Controller) (1983) (Atari - CCW, Christopher H. Omarzu) (CX26104) (PAL)",
  b'^E\x17\xdb\x83\xc0a\x92a0\xabe\x97^;\x83\xd9@\x1c\xc9': "Big Bird's Egg Catch (Grover's Egg Catch) (Kid's Controller) (1983) (Atari - CCW, Christopher H. Omarzu) (CX26104) ~",
  b"\x92\xc4\x99n\xf6X\xbd\xcc\x9a\xe2'\xba.?i\xedzk\xa0\x82": 'Billard (AKA Trick Shot) (1983) (Quelle) (626.610 0) (PAL)',
  b"\xcd\xd8\x90'm\xf3AJ\xf0\xfcf\xec\xf2Q\xfbHR:\xd4Q": 'Bingo (AKA Dice Puzzle) (1983) (CCE) (C-868) (PAL)',
  b'\xee\xedS\x12\xb6/\x9a\\+\xcexYV\xbda\xc4!\xddD1': 'Bionic Breakthrough (Headband - Super Breakout) (Mindlink Controller) (06-22-1984) (Atari, Paul Donaldson) (Prototype)',
  b'\xf6\xa4\x15\x07\xb8\xcf\x89\n\xb7\xc5\x9b\xb1BO\x05\x00SC\x85\xce': 'Bionic Breakthrough (Headband - Super Breakout) (Mindlink Controller) (1984) (Atari, Paul Donaldson) (Prototype) ~',
  b'\xfaj\x18oU=}\x14\x1eK/mK\xe0E|H\x96\xf8W': 'Black Hole (AKA Cosmic Ark) (Double-Game Package) (1983) (Otto Versand) (311388) (PAL)',
  b'b\x8a\xd1\xbc;0:x"\xac\xa2\xeb\t \x0f-\x11\x81\x85\xb5': 'Blackjack (32 in 1) (Paddle) (1988) (Atari, Bob Whitehead) (CX26163P) (PAL)',
  b'Y\xb7K\xb1l\xb0F\xda\x0f\xb3\x8c(u\r(\xc3\xf0q\xf1\xc0': 'Blackjack (Gambling) (Paddle) (1977) (Atari, Bob Whitehead) (CX2651) (PAL)',
  b'\x9b\x15\x90o@\x11\x92zG\xef5=\xc2\xf2\x99\xe8!\xe3\x83A': 'Blackjack (Paddle) (Unknown) (PAL)',
  b'\xed\xfd\x90Z4\x87\x01\x96\xf8\xac\xb2\xa9\xcdA\xf7\x9fC&\xf8\x8d': 'Blackjack - Black Jack (Gambling) (Paddle) (1977) (Atari, Bob Whitehead - Sears) (CX2651 - 99805, 49-75602) ~',
  b'\xfe\xb0\xd0_O\x92\x03\xd6\xf4\xdf\x01\t\\\xeb\xac\xae\x9a\xd3\xf4/': 'Bloody Human Freeway (Human Freeway) (Freeway Beta) (Activision, David Crane) (AG-009) (Prototype)',
  b'fu:\x7f\xff\xaaC+\xda\xcd\xb1\x8e}\xcf\tv\xe2}\x8a\xbe': 'Blueprint (1983) (CBS Electronics, Tom DiDomenico) (4L 2486 5000) (Prototype)',
  b'\x0f\xad\xef\x01\xce(\x19(\x80\xf7E\xb2:_\xbbd\xc5\xa9n\xfe': 'Blueprint (1983) (CBS Electronics, Tom DiDomenico) (4L 2486 5000) ~',
  b'\x1b{\xdcG{[\x84\x88\xfc\x1cUf\xca\x91\x02\x97\xdc\x81\xf4\x88': 'Blueprint (1983) (CBS Electronics, Tom DiDomenico) (4L2477, 4L2482, 4L2485, 4L4171) (PAL)',
  b'\x08\x88X,!\xb5\x04\xf8\xf7\xeb\xe1{=Z:\xcd{o\xfa\x93': 'Bob Is Going Home (AKA Bobby Is Going Home) (JVP)',
  b'h\x95s\xac\xea+g\xf6S\xf4\x1e!\xc6\xb1\r\xe1\xa3\xb5\x18\xd6': 'Bobby Is Going Home (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'}^\xd06\xa5\x7f\xdb`\x92q!\xb9\xa35>\x8a<\xda\xae\xfa': 'Bobby Is Going Home (Rentacom)',
  b'\xe0\x02\x02\xdaB\x8d\xb2\xd0^0\xcd\xd5\x94\x88\xe5\xe0\xd4\x05\xc6x': 'Bobby Is Going Home - Bobby Vai Para Casa (1983) (CCE) (C-803)',
  b'\x87l;\x01\xe6\x85h\x10\r\xb7\x16\xbci\xbfX\x18\x9a\xcbr\xc2': 'Bobby Is Going Home - Bobby Vai Para Casa (1983) (CCE) (C-803) (PAL)',
  b'lJ\x96\xae\xc2\x1fl\xed\xe3\xf4<\xd1(B\xbcr8d\x87p': 'Bobby Is Going Home - Bobby geht Heim (1983) (Bit Corporation) (PG206)',
  b'J\xef0\xb42Ib\x1e\t\xa6\x90\xd5N\xbe#7r\x89\xa9\x97': 'Bobby Is Going Home - Bobby geht Heim (1983) (Bit Corporation) (PG206) (PAL) [demonstration cartridge]',
  b'P\xe2f\x88\xfd\xd3\xea\xdc\xfa\x83$\x06\x16&z\x8f`!l%': 'Bobby Is Going Home - Bobby geht Heim (1983) (Bit Corporation) (PG206) (PAL) ~',
  b'o\x89\xec\x06\xbc6\x00\x13\xefL\x00\xd5\x1cT\xdfP*\x80\xe3\xbf': 'Bobby geht nach Hause (AKA Bobby Is Going Home) (1983) (Quelle) (476.774 5) (PAL)',
  b'(,\xad\x17H/_\x87\x80Pe\xd1\xa6.I\xe6b\xd5\xb4\xbb': 'Bogey Blaster (AKA Air Raiders) (1988) (Telegames) (5861 A030)',
  b'n:\xf3a\xbb\xf7\nJ\x85\xd6\xc8\xd4\xed@p\xe8\xfe\xd9c\xcb': 'Bogey Blaster (AKA Air Raiders) (1988) (Telegames) (5861 A030) (PAL)',
  b'J\xcc\xd4\xccO\x8f:\x93\\\xa1xD\x9aJ\xc6"\x83c\x9cS': 'Boggle (08-07-1978) (Atari, David Crane) (Prototype) ~',
  b'\xd7\xa0y\xa8\x87M\xcf\x05\x9d\x10\xdcI\x83\x1dm\x9e-\xb6\x89\x17': 'Boing! (Bubbles, Soap Suds, The Emphysema Game) (1983) (First Star Software, Alex Leavens, Shirley Ann Russell) (PAL)',
  b'\xd1\x06\xbbA\xa3\x8e\xd2"\xde\xad`\x8d\x83\x9e\x8a?\r\x0e\xcc\x18': 'Boing! (Bubbles, Soap Suds, The Emphysema Game) (1983) (First Star Software, Alex Leavens, Shirley Ann Russell) ~',
  b'W\xe0\xc1\xdbG\xcd\x8cLg\x1fp\\t\xfaS\xbe\xdf\x13O\xd1': 'Boom Bang (AKA Crackpots) (1983) (CCE) (C-849)',
  b")Tz[c `\xfb\xbc\x93Rg\xb7\t\x81\x1aw\xff\xf1'": 'Boom Bang (AKA Crackpots) (HES) (PAL)',
  b'\xc3g9b\x81\xba\xd3\xfe^\xf8\xcd\x9d\x99\x0c=)\x94\xd3+\xf1': 'Boom Bang (AKA Crackpots) (Unknown)',
  b'\xba\xe5\n\x199\xf1\x1e\xbc\xb8\x08\xf1\xf8\xb9x\xc4\xe2Xme\xee': 'Boom Bang (AKA Crackpots) (Unknown) (PAL)',
  b'\xafJ\x8a1\xf8\xdd$\xfe\x99\x04\xb5h]\xa8x\x0b\xea\x07\x826': "Bouncin' Baby Bunnies (Bouncing Baby Monkeys) (1983) (Telesys, Alex Leavens) (Prototype) ~",
  b'\xcfl\xe2D\xb3\xed\xaa\xd7\xad^\x9c\xa5\xf0\x16h\x13\\/\x93\xd0': 'Bowling (1979) (Atari, Larry Kaplan - Sears) (CX2628 - 6-99842, 49-75117) ~',
  b'\xda^\xd7\xa4\xa1\x00\xfaH0\xb65\x98\x17t\xa1\x1b\xa4\xb0\xa6\r': 'Bowling (1979) (Atari, Larry Kaplan) (CX2628, CX2628P) (PAL)',
  b'\xe7\xff\x9d\xedg\xc5\x99XjS:<o\xb9hJ\xfd\x9e&\xbd': 'Bowling (208 in 1) (Unknown) (PAL)',
  b'h\xeaC\xea\x0eg\x8c\xb5A\x9aCz\x9b\xda\xb5\xc8O^\xe1\xb9': 'Bowling (32 in 1) (1988) (Atari, Larry Kaplan) (CX26163P) (PAL)',
  b'La\x89\xfcF\\\xc05;\xec\x82\xde|\xdd\xdez\xad&\x9e\x93': 'Bowling (Dactar - Milmar)',
  b'\xdeq7Yb#\xf0\xe8\xd0\xf4\xe6\xcdg^\xa54H\x92\xc6\xf4': 'Boxen (AKA Boxing) (Videospielkassette - Ariola) (PGP234) (PAL)',
  b'\xa7qj\xf5D\xf0\x81<\xe7e{\xb8\xa1\xa5_\xf6\x08\x8f\t\xc3': 'Boxing (1983) (CCE) (C-861)',
  b'U\x9dF\xd1\xf9\xdf\xe6\xba.\xd8\xf2\x00\x1d\xb8\xdb\xe7\xbc\xd2\x07\xef': 'Boxing (32 in 1) (1988) (Atari) (CX26163P) (PAL)',
  b"\xb6z\x81F\xb6O'\xf9\xbb\x89u\xdaje\x9f;\xe8=(K": 'Boxing (Dactari - Milmar)',
  b'\xa7J\xb3\xe9U\xd8\xf7en\x97r\x82cf\x86c+\xa7\x96H': 'Boxing (Unknown) (PAL)',
  b'Oux\x06\x08\x14\x8fQ\xed\xdb0\xce-\xa0\xc1\xa4\x0b\xe7\xd3\x7f': 'Boxing - Box-Champion (1980) (Activision, Bob Whitehead - Ariola) (EAG-002, EAG-002-04I, PAG-002 - 711 002-715) (PAL)',
  b'\x14\xb9\xcd\x91\x18\x8c\x7f\xb0\xd4VdB\xd69\x87\x0f\x8do\x17M': 'Boxing - La Boxe (1980) (Activision, Bob Whitehead) (AG-002, CAG-002, AG-002-04) ~',
  b'#\x89\x15\xca\xfd&\xf6\x9b\xc8\xa3\xb9\xaa}\x88\r\xdeY\xf6\xf1-': 'Brain Games (Keyboard Controller) (1978) (Atari, Larry Kaplan - Sears) (CX2664 - 6-99818) ~',
  b'{\x9d\x9a\xc2\x88\xea\x1f\xda\x8e\x8c\xbb\xdeu\xe8KN.\xd3z^': 'Brain Games (Keyboard Controller) (1978) (Atari, Larry Kaplan) (CX2664, CX2664P) (PAL)',
  b'M7\xcc\xf1\xd9\x83\x91\x7f\xf5\x96Y\xb7\x85s\xd1C\n\xd4\xf2s': 'Breakout (Paddle) (1978) (Atari, Brad Stewart) (CX2622, CX2622P) (PAL)',
  b'\x8dG;\x87\xb7\x0e&\x89\x02h\xe6\xc4\x17\xc0\xbb\x7f\x01\xe4\x02\xeb': 'Breakout - Breakaway IV (Paddle) (1978) (Atari, Brad Stewart - Sears) (CX2622 - 6-99813, 49-75107) ~',
  b'\x06a\x93\xa1:\xcb\xc1\xab\xfao\xde\xc2\x9dA\xf7\xad\x14\x86\x9a[': 'Bridge (1980) (Activision, Larry Kaplan - Ariola) (EAX-006, PAX-006 - 771 006-720) (PAL)',
  b'm\xf4;u\xae\x08\x02@\xd9wT\xabi\xc3\x9c27\x08\x00\xa7': 'Bridge (1980) (Activision, Larry Kaplan) (AX-006) [fixed] ~',
  b'(s\xebn\xff\xd3P\x03\xd1>/\x8f\x99{v\xdb\xc8]\x0fd': 'Bridge (1980) (Activision, Larry Kaplan) (AX-006) ~',
  b'.\xc3\x18\x13\x90mO\xcb\x11`\xbaum\x0f\x93\xdb\x8f\x93\x08\x9a': 'Bridge (208 in 1) (Unknown) (PAL)',
  b'p\xf61\nR-\xa7l\x18\xf3\xde\xdc\x05\x0e\xf4B\x17\xef\xb7q': 'Buck Rogers - Planet of Zoom (1983) (Sega - Teldec) (005-10) (PAL)',
  b'\xa6]\xea-\x97\x90\xf3\xeb0\x8c\x04\x8a\x01Vn5\xe8\xc2EI': 'Buck Rogers - Planet of Zoom (1983) (Sega) (005-01) ~',
  b'g8}\r=H\xa4H\x00\xc4H`\xbf\x153\x9a\x81\xf4\x1a\xa9': 'Bugs (Paddle) (1982) (Data Age) (DA1005) ~',
  b"\xa8\xae\xde\xa6'\xc6|8\x03,\xdd\xe4A\xfb\x98\xda\xd2&\x91j": 'Bugs (Paddle) (1983) (Gameworld) (133-005) (PAL)',
  b'\xb3\xff\x12H\x91\xde\x0f\xb3\xd4L5\x11]\x83\x8f\xd7\xe15\xca\x04': 'Bugs Bunny (08-04-1983) (Atari, Alan J. Murphy, Robert C. Polaro) (CX26100) (Prototype) ~',
  b"\x9c\x0e\x13\xaf3j\x98l'\x1f\xe8(\xfa\xfd\xca%\n\xfb\xa6G": 'Bugs Bunny (1983) (Atari, Alan J. Murphy, Robert C. Polaro) (CX26100) (Prototype)',
  b'\xf3\xa6\x1b\xe1\xe2\x9bB\xc0g]\xb5\x1c\x01\xf8\x08R\xad1\xd9\x15': "Bump 'n' Jump (1983) (M Network, David Akers, Joe King, Patricia Lewis Du Long, Jeff Ratcliff - INTV) (MT7045) ~",
  b'5\xbc@H\xf5\x8b\xb1p18r\xa0\xbc\xe4O\xb1\xca2\x17\xcc': "Bump 'n' Jump (1988) (Telegames) (7045 A015) (PAL)",
  b"l\x19\x97\x82\xc7\x96\x86\xdc\x0c\xbc\xe6\xd5\xfe\x80_'j\x86\xa3\xf5": 'Bumper Bash (Paddle) (1983) (Spectravideo, David Lubar) (SA-218) ~',
  b'y\xc0\x08V\x81g\xbe\xc1Bb\x80i\x0cE\xe7\xba[\xd7\x1b>': 'Bumper Bash (Paddle) (1983) (Spectravideo, David Lubar) (SA-218, SA-218C) (PAL)',
  b'\xadH\xf4\x95.\x86z+\x97\x90\r\x96[i\xf5\x0f\xdd\xf8\xba-': 'Bumper Bash (Unknown) (PAL)',
  b'I\xe0\x1b\x80H\xae4L\xb6X8\xf6\xb1\xc1\xde\x0e\x1fAo)': 'BurgerTime (1983) (M Network, Patricia Lewis Du Long, Ron Surratt - INTV) (MT4518) ~',
  b'\xfc\xf8\xee\xac#\xe1F\xd2\x0f\x93\x81*\xb5\xa9\x87 E\xf9[R': 'Burning Desire (1982) (Mystique - American Multiple Industries) (1003) (PAL) ~',
  b'\xb23\xc3z\xa5\x16JT\xe2\xe7\xcc=\xc6!\xb31\xdd\xc6\xe5[': 'Burning Desire (1982) (PlayAround - J.H.M.) (202)',
  b'\xa6c\xf1<\xe7\xf3\x1e4Q\xe9D\xe9\xf0<\x9a\xac,>\x8f!': 'Busy Police (AKA Keystone Kapers) (Zellers)',
  b'\x9b\x92+Y\x0f\x1a\xe0A\xf1G36\x06\x8c\xdc\xa7e\xa6.n': 'Cabbage Patch Kids - Adventures in the Park (05-24-1984) (Coleco - Individeo, Ed Temple) (Prototype)',
  b'3\xee\x9f6`\x8e\x9e\x8b\xb8y\xa2r5\xd0B&\x12\x1c_k': 'Cabbage Patch Kids - Adventures in the Park (06-14-1984) (Coleco - Individeo, Ed Temple) (Prototype)',
  b'\xa4\xec\xf3u{x(-\xac\x1d\x16f\x8f8 \xca\x9d\xa4\x1a\xd9': 'Cabbage Patch Kids - Adventures in the Park (06-XX-1984) (Coleco - Individeo, Ed Temple) (Prototype)',
  b'\xcd\x7f\x18\xbf\x86U\x01\xe7\xb9\x18R.\x00\xdd\xbeCO\xd5~!': 'Cabbage Patch Kids - Adventures in the Park (07-03-1984) (Coleco - Individeo, Ed Temple) (Prototype)',
  b'\x15\x9b.\xad.TU\xc8rdV\\b\x83\x96\xcc\xa9\x97\xa7\xf3': 'Cabbage Patch Kids - Adventures in the Park (07-27-1984) (Coleco - Individeo, Ed Temple) (Prototype)',
  b'c\xce7\x95\xf4\x8a+\x1fY\x88\xf0\x1d\x8b\xf7\xcc\x10\xf3\x11">': 'Cabbage Patch Kids - Adventures in the Park (08-21-1984) (Coleco - Individeo, Ed Temple) (Prototype)',
  b'\x8c~\x04\xc6\xed\x08M\x8d\xed\xc1\xe8\xea\xe1+r\x16\x1c\x04\xf56': 'Cabbage Patch Kids - Adventures in the Park (09-04-1984) (Coleco - Individeo, Ed Temple) (Prototype)',
  b'*\x122O\x04\xe2,\x86\xfdO\x993\xa2#\x93l\x11\x17\xbb\xa2': 'Cabbage Patch Kids - Adventures in the Park (09-07-1984) (Coleco - Individeo, Ed Temple) (Prototype)',
  b"R\x105\xaf\xc4'\xe4\xbe\\\x85\xa5\xc1\xa4\xb8A&\xc4\x9b\xd1\x0f": 'Cabbage Patch Kids - Adventures in the Park (09-13-1984) (Coleco - Individeo, Ed Temple) (Prototype) [a]',
  b'\xff\xdb\xa8\xae"xL\xcc\x81\xa5\xa2\xe8\x1a#j\xce\t\xe5\xb7\xf4': 'Cabbage Patch Kids - Adventures in the Park (09-13-1984) (Coleco - Individeo, Ed Temple) (Prototype) ~',
  b'?\x1f\x17\xcfb\x0fF#U\x00\x9fS\x02\xcd\xdf\xfas\x0f\xa2\xfa': 'Cakewalk (Bakery) (1983) (CommaVid, Irwin Gaines) (CM-008) ~',
  b'9*2\x16\xff\xbbak\x8f\x12>\xaa\xb3\xcd98\xa6>\x11\x91': 'Cakewalk - Alarm in der Backstube (Bakery) (1983) (CommaVid, Irwin Gaines - Ariola) (CM-008 - 712 008-720) (PAL)',
  b'\xde\xf9P,Z7p\n\xe04a\xb2\xd7\xcf/s\xe9\x1bL\xec': 'California Games (1988) (Epyx, Steven A. Baker, Peter Engelbrite) (80561-00286) (PAL)',
  b"`\x9c 6\\:q\xceE\xcb'|f\xec<\xe6\xb2\xc5\t\x80": 'California Games (1988) (Epyx, Steven A. Baker, Peter Engelbrite) (80561-00286) ~',
  b"]'\xb7\xe9\xc4\x92\x11'4\x16\n\xbb\xb4Ep\xf0.*\xecg": 'Canyon Bomber (208 in 1) (Unknown) (PAL)',
  b"\xb8\x94C\xa0\x02\x9ev\\'\x16wO\xe2X+\xe3vP\x11\\": 'Canyon Bomber (Paddle) (1979) (Atari, David Crane - Sears) (CX2607 - 6-99828, 49-75115) ~',
  b'HWj$\xa4\x04\x82\x15\x1fk<\x96\x87\xbeD-\xb2q\x15\xa8': 'Care Bears (1983) (Parker Brothers, Laura Nikolich) (Prototype) ~',
  b'\xe1\xac\xf7\xa8E\xb5nK=\x18\x19*u\xa8\x1cz\xfao4\x1a': "Carnival (1982) (Coleco - Woodside Design Associates, Steve 'Jessica Stevens' Kitchen) (2468) ~",
  b"\x0c{\xa0'*\xc8hA\x05\xda\xa2\x88C\x87P\x17dM\xcd\xf9": "Carnival (1983) (CBS Electronics - Woodside Design Associates, Steve 'Jessica Stevens' Kitchen) (4L1717, 4L1718, 4L1719, 4L2277) (PAL)",
  b'\xec\xa4i8\xb9\x8c\xe7\xd9\x08\xe1LJ\x12e`\xa9\xb4S\xf6\x9c': 'Carnival (Hack) (208 in 1) (Unknown) (PAL)',
  b'\n|A\xb4!_y.I\x9d \xff\x86\xfdm\xa9\xc2d\x8d\x17': 'Casino (Paddle) (1979) (Atari, Bob Whitehead) (CX2652, CX2652P) (PAL)',
  b'\x08Y\x81\x01\xe3\x87V\x91f\x13\xf3u\x81\xef\x1ba\xc7\x19\x01o': 'Casino - Poker Plus (Paddle) (1979) (Atari, Bob Whitehead - Sears) (CX2652 - 6-99816, 49-75151) ~',
  b'\xcc\xd3H\'\xa8\xd9`\xd4\xd9\x06" \xbf\xca\x84\x8dv)\x12\xf7': 'Cat Trax (1983) (UA Limited) (1) [a]',
  b'\xe9y\xdeq\x9c\xec\xab!\x15\xaf\xfd\x9c\x05R\xc6\xc5\x96\xb1\x99\x9a': 'Cat Trax (1983) (UA Limited) (1) ~',
  b'7r&\x88\xd2{S8\xd9A\xa7\xa7\xf04\xa5\xef\xa7\xcfG\xc4': 'Catch Time (AKA Plaque Attack) (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'j\xdfp\xe0\xb7\xb5\xda\xb7L\xf4w\x8fV\x00\r\xe7`^\x87\x13': 'Cathouse Blues (1982) (PlayAround - J.H.M.) (201)',
  b'\xc5\xffd\xd7Y?\r\x01\x03\xd6\xbbP\xa1P\xe8\xa8}".\xf3': 'Centipede (1982) (Atari - GCC) (CX2676) (Prototype) (PAL)',
  b"\x18A\xce'\xfc\x14\xcc*\xa7A\xd6~\x91\xf6Oy$\xe3\xbf>": 'Centipede (1983) (Atari - GCC) (CX2676) (PAL)',
  b'\x0bY\x14\xbc\x15&\xa9\xbe\xafT\xd7\xfd\x11@\x81u\xcd\x8f\xccr': 'Centipede (1983) (Atari - GCC) (CX2676) ~',
  b'\xbcx3\x03\xf5\xb0_\xd7\x1cA\x8e,\xb4\xcf\x07\xe0\xba/\xd0\x0f': 'Challenge (208 in 1) (Unknown) (PAL)',
  b'\x11\x91\xed<\xacI\x1b\xabHz\xb3\xc3\xb5.\xcfV\xbd\x81{X': 'Challenge (HES) (PAL)',
  b'\xe8\x1b^I\xcf\xbb(>\xdb\xa2\xc8\xf2\x1f1\xa8\x14\x8d\x86E\xa1': 'Challenge (Unknown)',
  b'\xb2\xb1\xbd\x16[<\x10\xcd\xe51n\xd0\xf9\xf0ZP\x9a\xac\x82\x8d': 'Challenge (Zellers)',
  b'\x15\xed_\x153%h\x16\x98\x0e\x0f1]\x1d\x1aZC\xb3\xd5\xaa': 'Challenge of.... Nexar, The (1982) (Spectravision - Spectravideo - Sirius Software, David Lubar) (SA-206) (PAL)',
  b'\xac\x9b\x0cb\xba\x0c\xa7\xa9u\xd0\x8f\xab\xbb\xc7\xc7D\x8e\xcd\xf1\x8d': 'Challenge of.... Nexar, The (1982) (Spectravision - Spectravideo - Sirius Software, David Lubar) (SA-206) ~',
  b'Jk\xbd\x96\xf2\xc8\x05\xcd\xa128Hl\x82\x83y%|ZD': 'Challenge of.... Nexar, The (Hack) (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\xc7\xc9\xca\x98nv\xb5".\xe9v\xa7`=L\xcf\x87\x00\x14\xfc': 'Challenge of.... Nexar, The (Unknown) (PAL)',
  b'\x87+/\x9a\xa7\xed\xbc\xbb#h\xde\r\xb3il\x90\x99\x8f\xf0\x16': 'Chase the Chuckwagon (1983) (Spectravideo - TMQ Software - Video Games Industries Corporation, Mike Schwartz - Ralston Purina) ~',
  b"9\xb5\xbb'\xa6\xc4\xcbe2\xbd\x9dL\xc5 A\\Y\xda\xc6S": 'Checkers (1980) (Activision, Alan Miller) (AG-003) ~',
  b'\xd0\x99\xcb\xbcr\xa9\xcc\xc4\xf0\x84\xec\xe0\x9f\xab\xf2\x07\x18\x04\xfe\x0f': 'Checkers (208 in 1) (Unknown) (PAL)',
  b'\x9b\x9a\x9bkm4aH\\1\xd8\n\x83\tH\xcep\xea\x7f\xf1': 'Checkers (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\x0c\x0c\xd0\xf0\xf7h\xaf\xce\xf0\xe7\xce2\xa8d\x0ci\x14\xd9\xff\x99': 'Checkers (32 in 1) (1988) (Atari) (CX26163P) (PAL)',
  b'\x1b\xdd-\x9c\xd6s\xde\xf7Z\xfd\xebZ\xee(8\xc45\x1b\x8e\x8b': 'Cheese (Dragonfire Beta) (05-21-1982) (Imagic, Bob Smith) (720020-1A, IA3611) (Prototype)',
  b'\xfa\xac\x85\xc0\xc6\x12if\x027\x97\x89d\xa4\xcfW\x88|\xcc\xfc': 'China Syndrome (1982) (Spectravision - Spectravideo) (SA-205) (PAL)',
  b'\x0b\x1b\xb7gi\xae?\x8bI6\xf0\xf9_IA\xd2vy\x1b\xde': 'China Syndrome (1982) (Spectravision - Spectravideo) (SA-205) ~',
  b'Q\xa5;\xbf\xdb\xcc"\x92U\x15\xae\n\xf7\x9d\xf44\xdfn\xe6\x8a': 'Chopper Command (1982) (Activision, Bob Whitehead) (AX-015, AX-015-04) ~',
  b'\x84\x07\xa8\xbe\x1f\x95\xc5S\x99Ev\x04\xca\xe2\xfc\x11\xa5%\xd9G': 'Chopper Command (1983) (CCE) (C-827)',
  b"\x80N:\xca\xec\x15e\x0bl\xa8\xd2\x96l[\xfdzeZ'.": 'Chopper Command (1983) (CCE) (C-827) [a]',
  b'\xfc\x97d:\x90Dz\xdc\x90M\xa5\xe7\x94\xc3aF\x83\xf9\xfcR': 'Chopper Command (1983) (Digitel)',
  b'R\x964\xd1\xea\xf0\xe1<>\x80\x7f\x05*\x0f\x11~m\x83\x06\xdc': 'Chopper Command (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b's\x8f\xb2\x8b\xc0C\xaa\x1bG\x86\xde\x02/0\xcc\xd7\xf8\xbc\xaf\xc6': 'Chopper Command (Hack) (208 in 1) (Unknown) (PAL)',
  b'gk\x8e\x99\x91\xa0\x99\x8c\xba\x85l\xdb\xd8\xb3\xca\xf3`\xa0R,': 'Chopper Command (Hack) (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b"\n\x1c\xa0\x82'|HI\xf5\xf3\xde\xe3\x92T\xc0\xd7\x19s!?": 'Chopper Command (SuperVision) (405, 427, 806, 808, 813, 816) (PAL)',
  b'\xa4{\xf3\xc3[Gr\x930\xd3\xe1\xb0\xa1\xe0\xb5\xde\xeck\xa2\x13': 'Chopper Command - Captain Helicopter (1982) (Activision, Bob Whitehead - Ariola) (EAX-015, EAX-015-04I - 711 015-725) (PAL)',
  b'h!\xc34\xd9\xcb\x85\xda\x17\xfa*\x96\x066b\tf\xb9\x1f\x96': 'Circus (AKA Circus Atari) (Zellers)',
  b'4,\xab\xe3\x05\x92\xdb\x8d\xcfi)\x94fn@\xfb\x17l\x8d\x86': 'Circus Atari (Paddle) (1980) (Atari, Mike Lorenzen) (CX2630, CX2630P) (PAL)',
  b'\x04a\xf1\x1a\xec\x13\xc0a\x06\x12\xbe\x937\xfc--L\xbel\xab': 'Circus Atari (Unknown)',
  b'(\x83\xdfS\x0b\x082\xe9\xd8J\xdet\n\xb9_\xbc~:{\xac': 'Circus Atari (Unknown) (PAL)',
  b'\x8a\x91\xec\xdb\xd8\xbf\x9dA-\xa0Q\xc3B*\xbb\x00N\xab\x86\x03': 'Circus Atari - Circus (Paddle) (1980) (Atari, Mike Lorenzen - Sears) (CX2630 - 49-75122) ~',
  b'?V\xd1\xa3vp+d\xb3\x99+-VR\xa3\x84,V\xff\xad': 'Coco Nuts (1982) (Telesys, Jim Rupp, Jack Woodman) (1001) ~',
  b'\x02R\xbba\xf8\xca\x04b\xccW\xa2s\xd5u\xc4\xbc\xdbZ\x95\xf2': 'Codebreaker (Numbers) (Keyboard Controller) (1978) (Atari) (CX2643) (PAL)',
  b'\x13{\xd3\xd3\xf3n%I\xc6\xe1\xcc:`\xf2\xa7WOvwu': 'Codebreaker - Code Breaker (Numbers) (Keyboard Controller) (1978) (Atari - Sears) (CX2643 - 6-99815) ~',
  b'S\xc3$\xaesj\xfa\x92\xa8=a\x9b\x04\xe4\xfer\x18"\x81\xa6': 'Color Bar Generator (1984) (Video Soft, Jerry Lawson, Dan McElroy) (VS1008) ~',
  b'3\xdb1Q\xc8\xadyN\xeb{\xdf#\xc4\x15\xb9#\xf9\x83\x9a\x93': 'Colors (1980) (Atari) (Prototype) (PAL) ~',
  b'\xf4_\xd0\x0e\x9d\xdeY.7)\xae\xf2\x88\xabr\xbc\xa0\x95\x8c$': 'Comando Suicida (AKA Chopper Command) (Dismac)',
  b'Z\x07\x8c\xe2\x08z\x07\xbc<5\xf8G\x9f\x05p\xbf\xb4\xed\x9c\x8d': 'Combat (32 in 1) (1988) (Atari, Joe Decuir, Larry Kaplan, Steve Mayer, Larry Wagner) (CX26163P) (PAL)',
  b'a\xc8\xa4\xb0\x11\xb2\xcc1\xd1\xfa\xa3\x04\xa6\xb7j\xcb\xd4\xca\xd2\\': 'Combat (Tank) (1977) (Atari, Joe Decuir, Larry Kaplan, Steve Mayer, Larry Wagner) (CX2601, CX2601P) (PAL)',
  b'f\x01M\xe1\xf8\xe9\xf3\x94\x83\xee?\x97\xca\r\x97\xd0&\xff\xc3\xbb': 'Combat Two (Super Combat) (1982) (Atari - GCC) (CX2663) (Prototype) ~',
  b'\xe4}\xe0Sj\x87JP\x0b\xd6Xn\xb6\xec\x81\x9e\x14\xb4\x7f\x12': 'Commando (1988) (Activision - Imagineering, Mike Reidel) (AK-043-04) [different logo] ~',
  b'h\xa7\xcb?\xf8G\xcd\x98zU\x1f=\xd9\xcd\xa5\xf9\x0c\xe0\xa3\xbf': 'Commando (1988) (Activision - Imagineering, Mike Reidel) (AK-043-04) ~',
  b'\xdeJ\x12\x03fPW{TL/\xc0\x91\xb9\x81\xf1\x8fw\x1e4': 'Commando (1988) (Activision - Imagineering, Mike Reidel) (EAK-043-04I) (PAL)',
  b'\x8d\xad\x05\x08VW\xe9^V\x7fG\x83e\x02\xbeQ[B\xf6k': 'Commando Raid (1982) (U.S. Games Corporation - JWDA, Todd Marshall, Wes Trager, Henry Will IV) (VC1004) ~',
  b'\x8fc\xc5\x01\x00\x08U\x1f\x8e_\xe5\xad\xa4\x14\xf0\xc2\xbcD\xe5\xe5': 'Commando Raid (208 in 1) (Unknown) (PAL)',
  b'\xaf\x7f\xa7t\x11\x83\xa4?j\xeb\xb03\xc3[\x8b\xe5\x1f!D\t': 'Commando Raid (Unknown) (PAL)',
  b'\xd8]\xae\xd1\xb59zw\x9e\x882b\x89y\xf59fP\t\r': 'CompuMate (1983) (Spectravideo - Universum) (SV-010) (PAL)',
  b'\x19G\xb2d\x19\xfa\xdf\xc6\xb2\xc9\xb5\xd5R\xfd=+\xd2 \xbca': 'CompuMate (1983) (Spectravideo) (SV-010) ~',
  b'\xdb\xc0\xc0E\x1d\xeeDBX\x10\xe0M\xf8\xf1\xd2m\x1c-9\x93': 'Computer Chess (07-07-1978) (Atari, Larry Wagner, Bob Whitehead) (Prototype)',
  b'\xe1\x9d\xe1+\xde\x02\x95\x8d\xc5~%K\x0f\xe9\x17\xb3\x90\xf5\xf9D': 'Condor Attack (Unknown) (PAL)',
  b'U\x12\xa0\xedC\x06\xed\xc0\x07\xa7\x8b\xb5-\xbc\xf4\x92\xad\xf7\x98\xec': 'Confrontation (1983) (Answer Software Corporation - TY Associates) (ASC2001) (Prototype) ~',
  b'_\xecX\x0e\x99\xc9\x93\x80\x88\x9c\x05\x8e\xf4\x93\x18\xe4Q\x8eo\x7f': 'Congo Bongo (1983) (Sega - Beck-Tech, Steve Beck, Phat Ho - Teldec) (006-01 - 3.60105 VG) (PAL)',
  b'>\xc2\x1f\xcd\xc1K\xdbn\x9c\xae\xe1\xf7\xd9\xb8\x11\x07p=\xa5[': 'Congo Bongo (1983) (Sega - Beck-Tech, Steve Beck, Phat Ho) (006-01) [a]',
  b':w\xdbC\xb6X>\x86\x89C_\x0f\x14\xaa\x04\xb9\xe5{\xdd\xed': 'Congo Bongo (1983) (Sega - Beck-Tech, Steve Beck, Phat Ho) (006-01) ~',
  b'\xfdG\xdfp\x08 )^\xad4\xa9\xd9\xcf\\s\xec\xcf\x11\x01\xbd': "Cookie Monster Munch (Cokie Monster's Maze, Cookie Monster's Garden) (Kid's Controller) (1983) (Atari - CCW, Gary Stark) (CX26102) (PAL)",
  b'>\x031\x9a\xc9\xfd\xb0\x95\xac\xe6|\xa1Xk\xe2\xcd\x1d\xbb\xbfd': "Cookie Monster Munch (Cokie Monster's Maze, Cookie Monster's Garden) (Kid's Controller) (1983) (Atari - CCW, Gary Stark) (CX26102) (PAL) [a]",
  b'\xf4\xa6+\xa0\xffY\x80<_@\xd5\x9e\xee\xd1\xe1&\xfe7\x97\x9b': "Cookie Monster Munch (Cokie Monster's Maze, Cookie Monster's Garden) (Kid's Controller) (1983) (Atari - CCW, Gary Stark) (CX26102) ~",
  b'9V\xdd\xe2k&%\xfa7t\xe2\xbf\xc9>^,\x95\xbf\xa8\x86': 'Corrida da Matemática (AKA Math Gran Prix) (CCE)',
  b'\xa0\xed\xdf4\xfaJ\xb3~\xd9\\-\x1a\x1a0d\x00\xb7\x06l\xc7': 'Cosmic Ark (1983) (CCE) (C-831)',
  b';]\x82\xff\xcc~\x9d\xd7\x96m\xf9\xae\xb5:\x8b\xb1\xf4\x99\x88!': 'Cosmic Ark (1983) (CCE) (C-831) [a]',
  b'\x16|\xba^\xad\x06q\xb8\x0fB\xf4\xe9\x08\x1cg\xa9h\xea=o': 'Cosmic Ark (Canal 3 - Intellivision)',
  b"C@\xfd\xd7h\xd6k\xca\xe7\xe5FL1I\xd4'\xa3*\x84\x8a": 'Cosmic Ark (Reaction) (1982) (Imagic, Rob Fulop) (720104-1A, 720104-1B, IA3204) [selectable starfield] ~',
  b'\x18y\x83\xfd\x14\xd3t\x98C}\x0e\xf8\xf3\xfb\xd0Vu\xfe\xb6\xae': 'Cosmic Ark (Reaction) (1982) (Imagic, Rob Fulop) (720104-1A, 720104-1B, IA3204) ~',
  b'\x99\xe3]\x9b\xeb\xd3u\x0e\x9e\xd5\xdaDq\x91\x08`1\x1c"\x81': 'Cosmic Ark (Reaction) (1982) (Imagic, Rob Fulop) (720104-2A, IA3204P, EIX-008-04I) (PAL)',
  b'\x02\xae\x14\x1c~"\xcc4\xce\xbd\x9bl\xb3n9\xc7\xf9\xce!E': 'Cosmic Ark (Reaction) (1982) (Imagic, Rob Fulop) (720104-2A, IA3204P, EIX-008-04I) (PAL) [selectable starfield]',
  b'B\x12\x8e\x80\xe7;\xaa(rY\xbf\xe8@\xa5\x9f\x14\x99:\x19\x80': 'Cosmic Ark (Unknown) (PAL)',
  b'\x82v\xa9\x81\x0f\x05e\x8e\x1f\xb2\xc0\xc8\x16Ez\xd86a-\x85': 'Cosmic Avenger (AKA StarMaster) (4 Game in One) (1983) (Bit Corporation) (PGP214) (PAL)',
  b'7\x17\xc9{\xbb\x0fT~C\x89\xdb\x8f\xc9T\xd1\xba\xd9\x92DL': 'Cosmic Commuter (1984) (Activision, John Van Ryzin) (AG-038-04) ~',
  b'L\xbe\x08/"H\xce\x8c\xc62\xa7\x866\xce\x8dHnvR,': 'Cosmic Commuter (CCE)',
  b'\x8b\x9d\xfe\xf6\xc6uzjY\xe0\x1dx;u\x1eJ\xb9T\x1d\x9e': 'Cosmic Corridor (1983) (ZiMAG - Emag - Vidco) (708-111 - GN-040) ~',
  b'\xcdqn\xa26\xc6\xa74P\xf4\x15\x13\x1a\xc7\xac\x8e\xe9\x1aB\x8d': "Cosmic Creeps (Space Maze, Spaze Maze) (1982) (Telesys, Don 'Donyo' Ruffcorn) (1002) (PAL)",
  b'"\xff(\x1b\x1ei\x8e\x8a]zoas\xc8lF\xd3\xcd\x85a': "Cosmic Creeps (Space Maze, Spaze Maze) (1982) (Telesys, Don 'Donyo' Ruffcorn) (1002) ~",
  b'\xeb\x99.A\xc6&\xc2#\xb7\xd4"\x11q\x96\'\x0e\xbd\npp': 'Cosmic Creeps (Unknown) (PAL)',
  b'L\xe1\x0f\x7fx\xfe\xbd)Tr9\xf3dd\xbaM\xef\x80\xc0n': 'Cosmic Swarm (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\xc0\x13Tv\x0f,\xa8\xd6\xe4\xd0\x1b#\x0f1a\x19s\xc6\xae-': 'Cosmic Swarm (Termite) (1982) (CommaVid, John Bronstein) (CM-003) ~',
  b'\xd0\xb5\xe8N\xa3\xc8+7+\xd0\xb9\xd1[\x98\xf3\x1e\xe5R\xa4]': 'Cosmic Swarm - Angriff der Termiten (Termite) (1982) (CommaVid, John Bronstein - Ariola) (CM-003 - 712 003-720) (PAL)',
  b'\xb1\xe0g-\xc5\xcf\xd4\xd7U\xc9Z\xf9\xab\xdaK\xb3\x08\x89H\x8a': 'Cosmic Town (AKA Base Attack) (Perry Rhodan-Serie) (1983) (ITT Family Games) (554-33 375) (PAL)',
  b'\x89\xb38\x9b\x814\xfb\x996\xb5\xe3V\xe0\xbc\xef\x12\xe2\xbbFD': 'Cosmic War (AKA Cosmic Corridor) (1983) (Home Vision - Gem International Corporation - VDI) (VCS83136) (PAL)',
  b'\xd2&\xe8\xafN8\xd1\xd4\xeb\x8b\xb6\x9c\xdfk\xcc\xda\xd5a\xc8\x04': "Crack'ed (11-28-1988) (Atari - Zip Technology, Randy Bowker, Bruce Williams) (CX26142) (Prototype) ~",
  b'n\xe0\xa2j\xf4d?\xf2P\x19\x8d\xfc\x1c+|eh\xb4\xf2\x07': 'Crackpots (1983) (Activision, Dan Kitchen) (AX-029) ~',
  b'b\xb6|\xc0\xb7`\x1b\x96b\x88\n\xa3\x9e\xad2:\xec\xd4\x1cD': 'Crackpots (1983) (CCE) (C-862)',
  b'\xb1\x86\xa2\xeb\x0fW\xc3\x03\xc3\x19\x1b\xe0&2+N9\xce\xa8\x01': 'Crackpots (1983) (CCE) (C-862) [a]',
  b' \x86\n\xc7\x87\xeaXt\xc7C\x18\xe6\xf8t\x18:\xb8\x01\x15(': 'Crackpots (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'<3p\xa2\xf8B\xf2\x19G~\x89\x84(p\xe7\xea#\t\x7f\x8b': 'Crackpots (Digivision)',
  b'<\xe4\x11[\xfe\xeb\x00\xdaQ\xb9]\x11a\xf81\x94{\x0e\x97\xf6': 'Crackpots (Unknown) (PAL)',
  b's\xd6\x8f2\xd1\xfbs\x88<\xeb\x18=QP\xbf\xf5\xf1\x06]\xe4': 'Crash Dive (Voyage to the Bottom of the Sea) (1983) (20th Century Fox Video Games, Bill Aspromonte) (11031) ~',
  b'\x87\x86G\xd7\x8f\x94?\x95\xe0\xae\xc5`{\xa6@\xc0\x11"o\x91': 'Crash Dive (Voyage to the Bottom of the Sea) (1983) (20th Century Fox Video Games, Bill Aspromonte) (11131) (PAL)',
  b'p\xe7#\xaag\xd6\x8f\x85I\xd9\xbd\x8f\x96\xd8\xb1&,\xbd\xac<': 'Crazy Climber (1983) (Atari - Roklan, Joe Gaucher, Alex Leavens) (CX2683) ~',
  b'\xb1\xb3\xd8\xd6\xaf\xe9Ks\xa4<6f\x8c\xc7V\xc5\xb6\xfd\xc1\xc3': 'Cross Force (Cross Fire) (1982) (Spectravision - Spectravideo) (SA-203) ~',
  b'\xe1iN\xc2RR\x9d\x92;\x08y{{\xaap0\xfb\x9a\xe7\x9d': 'Cross Force (Unknown) (PAL)',
  b'3\x039G\xa4\xf45\x86\xfa,\xfa8\x8c\x9b\x9eB@h`X': 'Cross Force (Unknown) (PAL) [a]',
  b'\x8a|g\xd8=\x8e]\x0f\x9f\xc0\xa7\x8c\x17\x157V\xaa\x049r': 'Cross Force - Kreuzfeuer (Cross Fire) (1982) (Spectravision - Spectravideo - Quelle) (SA-203 - 413.223 9) (PAL)',
  b']\xa3\xd0\x89\xcc\xda\x96\x0c\xe2D\xad\xb8U\x97Xw\xc6p\xe6\x15': 'Crossbow (1988) (Atari - Imagineering, Dan Kitchen) (CX26139) ~',
  b'\x1e3&\x144A\xf2\xe6\xf9p\x9f\x81x\x19x"\x05U(\x8f': 'Crossbow (1988) (Atari - Imagineering, Dan Kitchen) (CX26139P) (PAL)',
  b'\xa5pb\xf4Nz\xc7\x93\xd4\xc3\x9d\x13PR\x1d\xc5\xbc*f_': 'Crypts of Chaos (1983) (20th Century Fox Video Games, John W.S. Marvin) (11009) ~',
  b'\x9d\x0bnp\xde[\xa3\x83\x0f!\xfaGt8\xbe\x15\xea\xc4\x84\xed': 'Crystal Castles (01-04-1984) (Atari, Michael Kosaka, Peter C. Niday, Robert Vieira) (CX26110) (Prototype)',
  b'3\x11\xeaS\xaf\xc8\x90\x1df^Hp6\xe7\xda\n\xe1\xcd\xb2v': 'Crystal Castles (1984) (Atari, Michael Kosaka, Peter C. Niday, Robert Vieira) (CX26110) (PAL)',
  b'.N\xe5\xee\x04\x0b\x08\xbe\x1f\xe5h`-\x18YfN`~\xfb': 'Crystal Castles (1984) (Atari, Michael Kosaka, Peter C. Niday, Robert Vieira) (CX26110) ~',
  b'\r\xd7*4a\xb4\x16\x7f-h\xc95\x11\xedI\x85\xd9~j\xdc': 'Cubicolor (1982) (Imagic, Rob Fulop) (Prototype) ~',
  b'\xf2\xaf/\xd5\xacgo\xd6\xf3\xce%\x97n5u\x9d\xe8e5\xc4': 'Cubo Mágico (AKA Cubicolor) (CCE)',
  b'K=\x02\xb5\x9e\x17R\x0bM`#eh\xd5\xcbP\xa4\xe6\xae\xb3': "Custer's Revenge (1982) (Mystique - American Multiple Industries, Joel H. Martin) (1001) ~",
  b'0\xa0v\xf1\xdbS\xf8W\x99\xf0\xad\x9c\xfc\xe9g$\xc3\x98\x87\xf7': "Custer's Revenge (1982) (Mystique - American Multiple Industries, Joel H. Martin) (PAL)",
  b'\x85P\r\xc5\xf7\xb0\x83\xce-t<"<V\x16?\xf9[9\xe3': 'Dancing Plate (AKA Dishaster) (1982) (Puzzy - Bit Corporation) (PG205) (PAL)',
  b'\x07\xe9J}5~\x85\x9d\xcf\xf7q\x80\x98\x17\x13\xce\x81\x192N': 'Dancing Plate - Dancing Plates - Tanzende Teller (AKA Dishaster) (1982) (Bit Corporation) (PG205) (PAL)',
  b'\n\xe2\xfc\x87\xf8z\\\xc1\x99\xc3\xb9\xa1tD\xbf</j\x82\x9b': 'Dark Cavern (1982) (M Network - APh Technological Consulting, Hal Finney) (MT5667) ~',
  b'rV?\x87.\xbf\xd2\x85\xefQ\xb9\xf2\xfe}:\x8fU\xf1\xb0\xfa': 'Dark Chambers (Dungeon, Dungeon Masters) (1988) (Atari - Sculptured Software, Adam Clayton) (CX26151, CX26151P) (Prototype) (PAL)',
  b'\xfb\xb4\x81Is\xfc\xb4\xe1\x01R\x15\x15\xe0M\xaad$\xc4_\\': 'Dark Chambers (Dungeon, Dungeon Masters) (1989) (Atari - Sculptured Software, Adam Clayton) (CX26151) ~',
  b'\x0e\xa0\xde\xa5p\xac\xed\xc8\x06\xaa\xe60\xa9/h\xb3\xe1\xd68h': 'Dark Chambers (Dungeon, Dungeon Masters) (1989) (Atari - Sculptured Software, Adam Clayton) (CX26151, CX26151P) (PAL)',
  b'&\xdeS\x1a\xb1P*\x8d\xfb{p\x0c\x1d\xb1\x02E%\xdcA\r': 'Deadly Discs (AKA TRON - Deadly Discs) (1988) (Telegames) (PAL)',
  b'h\xde)\x1d^\x9c\xbe\xbf\xedr\xd2\xf9\x03\x9e`X\x1bm\xbd\xc5': 'Deadly Duck (1982) (20th Century Fox Video Games - Sirius Software, Ed Hodapp) (11004) ~',
  b'_q\n\x11Ht\x07`\xb4\xeb\xccB\x86\x1a\x1f\x9c3\x84y\x9e': 'Death Trap (1983) (Avalon Hill, Jean Baer, Jim Jacob) (5001002) ~',
  b'z\x89\x169\x82\xd0\xb3\xc8\x9e\xd8sq3Kj\x10\xc5n\x1d*': 'Decathlon (AKA Activision Decathlon, The) (HES) (PAL) (16K)',
  b'\xce\xe6\xf8\xe8\x14\xb8\xd6I\x9c\x1a*\x89.3\xeb{\x17E\x14\x8b': 'Defender (10-30-1981) (Atari, Robert C. Polaro, Alan J. Murphy - Sears) (CX2609 - 49-75186) (Prototype)',
  b'y\xfa\xcc\x1b\xf7\x0ed&\x85\x05y\x99\xf5\xc2\xb8\xe9K\x10$9': 'Defender (1982) (Atari, Robert C. Polaro, Alan J. Murphy - Sears) (CX2609 - 49-75186) ~',
  b'#\x1f\nTA\xc4\xb9\x10GOE\xd7\xdc\xf9\x80,\x9096I': 'Defender (1982) (Atari, Robert C. Polaro, Alan J. Murphy) (CX2609, CX2609P) (PAL)',
  b'\xec\xa3V\xc4 o\xbf\xe6&\x7f,B\x7f\xf5s@\xd1Z\xc77': 'Defender (CCE)',
  b'{\x10\xa7\x1e\xbe\x07\x9ff\xa6\x17\x02\xd7.-\xf55\xec\x07\xaa\xb2': 'Defender (Dactari - Milmar)',
  b':\x0b@F\x08\x81\x15\xff\xd5\xb6\xdf\x1a\x1e:\xbc\xc1\x00=E\xe1': 'Defender (Digivision)',
  b's\x05n6X\xf2\xeeF\xe6\x12\x0eAP2\x9c\xcc\xd7Gq\xb0': 'Defender (Hack) (Unknown)',
  b'8y\x0b\x82Bc\xbe\xc2\x92)fD\xfev\x03l\xd6\x1a\xa7H': 'Defender (Unknown) (PAL)',
  b'\xd7\xb5\x06\xb8O(\xe1\xb9\x17\xa2\x97\x87S\xd5\xa4\x0e\xb1\x97Sz': 'Defender II (AKA Stargate) (1987) (Atari, Bill Aspromonte, Andrew Fuchs) (CX26120)',
  b'\xab\xa4\xe9M\x01\xb1\x0c\xdf\x9b\x88/%\xe6 \xa1&\xdd\x80\xdf\x19': 'Defender II (AKA Stargate) (1987) (Atari, Bill Aspromonte, Andrew Fuchs) (CX26120) (PAL)',
  b'\xc9#$\xfc*5\xe3H\xce;\x00\x8cC"w\x1a\xb0*0\xc7': 'Demolition Herby (1983) (Telesys, Don Ruffcorn) (1006) (PAL)',
  b'Z\xaea\x82\x92\xa7(\xb5Z\xd7\xf0\x02B\xd8pskSV\xd3': 'Demolition Herby (1983) (Telesys, Don Ruffcorn) (1006) ~',
  b'CN\x87b&\xbe\xa0\n\xb1\x93$\x80K\x18\xde\x14\xc9=\tk': 'Demolition Herby (Unknown) (PAL)',
  b'\x82\xb2\xccwJ-\x15l*u\xd7@\x9c\xc9\xa8\xbb\x86\x1bY\xf8': 'Demon Attack (1983) (CCE) (C-823)',
  b'r\xd41$k\xbc\xcc\\\x1amg\x01\xa2\xa0)J\xf6\xe8\x18h': 'Demon Attack (Canal 3 - Intellivision) (C 3016)',
  b'\xe5\x01\x1b\x85\xa0\xe3\xc1H\x08xe\xbe\xbd~\x9bf\x08\xc3"\x92': 'Demon Attack (Death from Above) (1982) (Imagic, Rob Fulop) (720000-200, 720101-1B, 720101-1C, IA3200, IA3200C, IX-006-04) [fixed] ~',
  b'\xa5\x80\xd8\x86\xc1\x91\xa0i\xf6\xb9\x03l?\x87\x9e\x83\xe0\x95\x00\xc2': 'Demon Attack (Death from Above) (1982) (Imagic, Rob Fulop) (720000-200, 720101-1B, 720101-1C, IA3200, IA3200C, IX-006-04) ~',
  b'\x91\xa7\x8b2\xfeb)P\x90\x08\xbe?/\xa0\x87\x04\x9f\x16\xda<': 'Demon Attack (Death from Above) (1982) (Imagic, Rob Fulop) (720101-2B, IA3200P, EIX-006-04I) (PAL)',
  b'\x12p\xf2\xda\x04\xf4o\x86\xc1\xd8\xc4\xa4\x12\x02H\n\xa3\x18\xc3e': 'Demon Attack (Robby)',
  b'\x86\\&C\x10\x81\x11\x119\x1b\xc8\x1e2\xa1|W\xfd\xc6\xe3\xbf': 'Demon Attack (Supergame)',
  b'\x03\xa9\xeaA\x915\x8e\xa6Wx\xb7K/&P\x89pj\x9a\x9a': 'Demon Attack (Unknown) (PAL)',
  b'\xb4U\x82\xde\x81\xc4\x8b\x04\xc2\xbbu\x8di\x02\x1e\x80\x88\xc7\x0c\xe7': "Demons to Diamonds (Hot Rox) (Paddle) (1982) (Atari, Alan J. Murphy, Nick 'Sandy Maiwald' Turner - Sears) (CX2615 - 49-75140) ~",
  b'\x18\xc890\x84\x1b\x05+C9\xcc\xf0T\xcd\x0f\xceid\xfa\xcf': "Demons to Diamonds (Hot Rox) (Paddle) (1982) (Atari, Alan J. Murphy, Nick 'Sandy Maiwald' Turner) (CX2615) (PAL)",
  b"8\xe2Hq\xd3\\l%\xb8\xfa\x91W$J\xaa\xff++n'": 'Demons to Diamonds (Paddle) (CCE)',
  b'\xd7/\xd7x1=\xca\rp -/\xf2\xc2\x02\x13\xb0\xd3\xed0': 'Depth Charge (1983) (Amiga - Video Soft) (Prototype) ~',
  b'\x18\xa1\x05\xbf)z\xe1\xe7\xe947G\xa0\x02\xff\xa9~\xc0\x8e\xa9': 'Der Geheimkurier (AKA Mr. Postman) (1983) (Quelle) (802.744 3) (PAL)',
  b'\x03\xbcT\xea\xa88r\xd4b%\x13\xb20\x94\xc0++8\xa6\x19': 'Der Vielfrass (AKA Fast Food) (1983) (Quelle) (176.543 7) (PAL)',
  b']_-\x98\xc3s\x1fx\xa7D$\x15\xe0\xa3\x91\x15\xaa\xae\xca,': 'Der kleine Baer (AKA Frostbite) (1983) (Quelle) (685.640 5) (PAL)',
  b'\n$\x05\x99\x14\xac\x07\x8c\xae\xe8W!\x95U\xae%\x90TG\xd4': 'Der moderne Ritter - Mr. T (AKA Fast Eddie) (1983) (Quelle - Otto Versand) (700.223 1 - 781627) (PAL)',
  b']\x96\x19A\x0f\xd0X\xab\x83\x15\xf2Y\x10\xa3\x8fH\x05\xc4\xb1\xc2': 'Desert Falcon (Nile Flyer, Sphinx) (05-27-1987) (Atari - Bobco, Robert C. Polaro) (CX26140, CX26140P) (Prototype) (PAL)',
  b'\xcc\xea-P\x95D\x1d~\x1b\x14h\xe3\x87\x9aj\xb5V\xdc\x8bz': 'Desert Falcon (Nile Flyer, Sphinx) (1987) (Atari - Bobco, Robert C. Polaro) (CX26140) ~',
  b'\xb6\xed\x92W\xbdT*\xcf\xa8\xc4\xa3B`\x94\xf4j\xdb\xc6\xa9;': 'Desert Falcon (Nile Flyer, Sphinx) (1987) (Atari - Bobco, Robert C. Polaro) (CX26140, CX26140P) (PAL)',
  b'wP#\x86q\xb8\xee\x1eA7\x8a\xdfFD\x1a\xdb\xa3$\xea,': 'Diagnostic Test Cartridge 2.0 (1980) (Atari) (50008) (Prototype) ~',
  b'6\xd0\r\xfcJ/\xe1\x98\x9d\xa4\xc8\xac\xb8\x19\xba\xdf\xc2\x81@\xd9': 'Diagnostic Test Cartridge 2.6 (1982) (Atari) (MA017600) ~',
  b'\xd9\xa6\x19\xe5\x85\xcd\xf0\xd3\xf0\xed\xec\x1f\xd7J\xbc\x14(I\x84Q': 'Diagnostic Test Cartridge 2.6P (1982) (Atari) (TE016643) (PAL) (4K)',
  b'\x81\x02.\xf3\x0eh!\x83\xd5\x1b\x18\xbf\xf1E\xceB\\o\x92N': 'Dice Puzzle (1983) (Panda) (106)',
  b'P\x9e\xeb{\x11;\t\xe72f\x18\xb7O\x90\xfa.\xb1\xdd\xfc\x95': "Dice Puzzle (1983) (Sancho - Tang's Electronic Co.) (TEC005) (PAL) ~",
  b'a\xad\xf5\x89\xc1\x9f\x98Y(\xb5\x93\xc6\x98d\xceA\x07\xa1\xf6\x18': 'Die Ente und der Wolf (AKA Pooyan) (1983) (Quelle) (688.383 9) (PAL)',
  b'>h\xcd)\xa1\xf8P\xcb\x88}o\x06]%\xf1\x19\xb9\xa8\x19\xad': 'Die Ratte und die Karotten (AKA Gopher) (1983) (Quelle) (687.463 0) (PAL)',
  b'r\xbe[\x87\x1d\x9b\xd4\x12\x0c\x9b\xbfR\x03A-\xcaV\x15\x01\x91': 'Die hungrigen Froesche (AKA Frogs and Flies) (1983) (Quelle) (043.151 0, 874.382 5) (PAL)',
  b'y\xe7FRE \xdaTbI\x14\x9c3aO\xc2:O*Q': 'Dig Dug (1983) (Atari - GCC, Douglas B. Macrae) (CX2677) ~',
  b'\x81j\xbc\xef\x9c\xc8\xfc\xa5\x00|\xc4\x00SN\x84\xd0\x02\x85\xf1\x99': 'Dig Dug (1983) (Atari - GCC, Douglas B. Macrae) (CX2677, CX2677P) (PAL)',
  b't\x85\xcfU \x1e\xf9\x8d\xed \x1a\xecs\xc4\x14\x1f\x9ft\xf8c': 'Dishaster (1983) (ZiMAG - Emag - Vidco) (711-111 - GN-020) ~',
  b'\x94\xa4;J\xa6\x86\x03\x15\xd9\x99\x1b\xba\xa1D\xeb[\t\xa00\x81': 'Dishaster (Unknown) (PAL)',
  b'\xcd\x9129\x91\xb4\xb6NS8\xa3\xf7\x1eo@R"`\x84R': "Dodge 'Em (Head On) (1980) (Atari, Carla Meninsky) (CX2637, CX2637P) (PAL)",
  b'\xb2ft\xd6\xe3\r\x1a\x0b\xb2q\x9b\x9b\xb1\xb3\xcc\xfa4b`\xcf': "Dodge 'Em (Head On) (1980) (Atari, Carla Meninsky) (CX2637, CX2637P) (PAL) [fixed]",
  b'\xe2A\x8fr\xc1D\xb0\xe6\xce\xd5I\xcb\xb9c\x84\x9dS\x02\x0e\xf4': "Dodge 'Em (Unknown) (PAL)",
  b'\x15q\x17\xdf#\xcbR)8m\x06\xbb\xdb:\xf2\n \x87"\xe0': "Dodge 'Em - Dodger Cars (Head On) (1980) (Atari, Carla Meninsky - Sears) (CX2637 - 49-75158) [fixed] ~",
  b'\x0f\xfc\x02\xc5A\x90\xe9\xddQ\xac\x1eR\xc9\x11\xd3\xd7\xd70\xa4\n': "Dodge 'Em - Dodger Cars (Head On) (1980) (Atari, Carla Meninsky - Sears) (CX2637 - 49-75158) ~",
  b'\xe3\x98]u\x9f\x8a\x8fG\x05\xf5C\xce~\xb5\xe9;\xf67"\xb5': 'Dolphin (1983) (Activision, Matthew L. Hubbard, Bob Whitehead) (AX-024) ~',
  b'{4\xf3\x19\x92\xdaF\x0e,t\x12E\xad\xd5\xf7\xf1\x9d\xb6\n\x04': 'Dolphin (CCE)',
  b'F\x06\xc0u\x1fV\x02\x00\xae\xdee\x98\xec\x9c\x8ebI\xa1\x05\xf5': "Donald Duck's Speedboat (Donald Duck's Regatta) (04-12-1983) (Atari, Peter C. Niday) (CX26108) (Prototype) ~",
  b'\xb4\x96\xe5@\xdb\xc8K\x19\x18^.Y}U\x98h\xcb\x1b\x10\xbc': "Donald Duck's Speedboat (Donald Duck's Regatta) (04-18-1983) (Atari, Peter C. Niday) (CX26108) (Prototype) (PAL)",
  b'HR;^3\xaa\xa5\xec\xda\x88\xfa\npWQ,e\x8f\xcc\x03': 'Donkey Kong (1983) (CBS Electronics - Woodside Design Associates - Imaginative Systems Software, Dan Kitchen, Garry Kitchen) (4L1700, 4L1701, 4L1702, 4L1802, 4L2274) (PAL)',
  b"D\x91NkA\x087\xc4-\\EX\x95'6\n\xb0\xd3g\xc2": 'Donkey Kong (1983) (Pet Boat) (PAL)',
  b'nn7\xec\x8df\xae\xa1\xc1>\xd4D\x86>=\xb9\x14\x97\xaa5': 'Donkey Kong (1987) (Atari) (CX26143)',
  b'X\xc4\xdd\xd4\xaaTf\xb6\x88\\T\x10q~VMq\xff)\xd8': 'Donkey Kong (208 in 1) (Unknown) (PAL)',
  b'\x01\xab\xbc-c\x0b\x92sU\\z\xaf\x1e\x9b\xeb\xa2*\xa0\xe4\x8c': 'Donkey Kong (AKA Inca Gold) (Star Game) (001)',
  b';\xc3Z\xe02\xf1\x9fxW\x063\x0c\x19\xac\x11vtxx\x15': 'Donkey Kong Jr (AKA Donkey Kong Junior) (1983) (CCE) (C-1003)',
  b'a\\t\x88\x0cH\xc9\x9b\x89ad,NZ\x96\xf6\xc6<j1': 'Donkey Kong Jr. (AKA Donkey Kong Junior) (Tron)',
  b'\xf6Mh\x96K\xc2+jB\x17\xb7y\xc1\xd5\x8b\xc2\x04\xa5C\x07': 'Donkey Kong Junior (1983) (CBS Electronics - Woodside Design Associates, Harley H. Puthuff Jr.) (4L1802) (PAL)',
  b'\x98\xf9\x8a\xc0r\x8ch\xdef\xaf\xdae\x00\xca\xfb\xdf\xfe\x8a\xb5\n': 'Donkey Kong Junior (1987) (Atari) (CX26144)',
  b'\xcc\x99\xdb\xa0\xa7\x8f\xed\xd1q8\x7fI.\x98\x10\xf3\x03z_\x05': 'Double Dragon (1989) (Activision - Imagineering, Dan Kitchen) (AK-050-04) ~',
  b'?H*\xcc\x19\x14\xc6\xbc\x92\xfd\x1f\xd8#\xd6\xed\x10\xc7\xcae\xa8': 'Double Dragon (1989) (Activision - Imagineering, Dan Kitchen) (EAK-050-04) (PAL)',
  b'\xfc3O\xed\xe2\xebPjl[U~\x01\t\xcda\xf7\xa3\t\x19': 'Double Dragon (CCE) (PAL)',
  b'x\xbb\xa6\xd2\x8b\x86\xf5\xec\x08\nn\xae\x17\xd2\x0e\xe5\xb2]c\xd7': 'Double Dunk (Super Basketball) (1989) (Atari, Matthew L. Hubbard) (CX26159) (PAL)',
  b'\x8e.\xa3 \xb29\x94\xdc\x87\xab\xe6\x9da$\x94\x89\xf3\xa0\xfc\xcc': 'Double Dunk (Super Basketball) (1989) (Atari, Matthew L. Hubbard) (CX26159) ~',
  b'\xe4t\xad\x18d\xdfVu\r\xedq\xe9\xe4W\x05\x1e\xab\xfb\x03\xee': 'Dragon Defender (Video Game Cartridge - Ariola) (TP-605) (PAL)',
  b'\xb4F8\x1f\xe4\x80\x15`w\xb0\xb3\xc5\x17G\xd1V\xe5\xdd\xe8\x9f': 'Dragon Treasure (AKA Dragonfire) (Zellers)',
  b'}\xff\xde\x1b\x8c\xcf\xd7\xafhd\\Mt\xf5 K<\x13\xc9s': 'Dragonfire (Cheese) (1982) (Imagic, Bob Smith) (720020-2A, IA3611P) (PAL)',
  b'\xde3\xee\xcf0\xe3\x9cZh\xdb\x9a\xfc<\x86j5\xccs\x9c\xa1': 'Dragonfire (Unknown) (PAL)',
  b'Z\x11\x13\xb3\x95G\x95J\x8b,\xf1RX\xa5\xa5&\x12\xfe\xfbi': 'Dragster (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\x94LR\xde\x85F@p\xa9F\x81;\x05\x05\x18\x97wP\xe99': 'Dragster (Drag Strip) (1980) (Activision, David Crane) (AG-001) ~',
  b'/\xda\xd1j`\xcd:I\xd3\xbd\xfb0\x1b5\x90\xf2\xaa\xa7\x1d\xa3': 'Dragster - Dragster Rennen (Drag Strip) (1980) (Activision, David Crane - Ariola) (EAG-001, PAG-001, EAG-001-04B, EAG-001-04I - 711 001-715) (PAL)',
  b"\xfb\x1a\x08\xa3\xa8\xd3\xd4\xcdf'oR9\xe0'\xdbC\xbb\x06\xbc": 'Dream Flight (AKA Nightmare) (1983) (Goliath - Hot Shot) (83-215) (PAL)',
  b'\xc0a\xd7SC]\xcbru\xa8vOJ\xd0\x03\xb0_\xa1\x00\xed': 'Dukes of Hazzard (1983) (Atari, Mark R. Hahn) (CX2678) ~',
  b']\xa9\x8e9\xd2\xce-\xbf\x88\x9ew\x0e\x96>S\x1a\xc9\t\x8d\xf0': 'Dukes of Hazzard (AKA Stunt Cycle) (Paddle) (1980) (Atari, Robert C. Polaro) (Prototype)',
  b'\xfa\x8b25\x905\xc5\x1d\xf9\xba\xca(\x81X+\xb0\x9a\xb4\xa3\xd4': "Dumbo's Flying Circus (Dumbo Flies Home) (05-05-1983) (Atari, Jerome Domurat, Peter C. Niday) (CX26115) (Prototype) ~",
  b'\xd1n\xba\x13\xab\x13\x13\xf3u\xe8kH\x81\x81V\x7f\x84o\x1d\xc4': "Dumbo's Flying Circus (Dumbo Flies Home) (07-11-1983) (Atari, Jerome Domurat, Peter C. Niday) (CX26115) (Prototype) (PAL)",
  b'\x97\xbe#8\xaf\xd4\x0e\x90\x92\xde7L\xa8ju@F\xda\xd5V': "Dumbo's Flying Circus (Dumbo Flies Home) (1983) (Atari, Jerome Domurat, Peter C. Niday) (CX26115) (Prototype) (PAL)",
  b'^+-\x07\xdb\xa3i-\xb0\xec\x05\x82\xb0\xa2\xcbL+j\xd3\x1f': 'Dune (07-10-1984) (Atari, Bruce Poehlman, Gary Stark) (Prototype) ~',
  b'\x9a\x8b\xf8\xd1@\x019cl\xfc\x14T7>\xee\xa0\x89\xe7BQ': 'Dungeon (Dark Chambers Beta) (11-22-1985) (Atari - Sculptured Software, Adam Clayton) (CX26151, CX26151P) (Prototype)',
  b'T\x82\x85&\xfc{\xb7\x96\xbbB\xebT\x13\xe0I\xfc\xb3N!\xd7': 'E.T. - The Extra-Terrestrial (1982) (Atari, Jerome Domurat, Howard Scott Warshaw) (CX2674) (PAL)',
  b'\x9e4\xf9\xcaQW<\x92\x91\x87 \xf8\xa2Y\xb9D\x9a\x0c\xd6^': 'E.T. - The Extra-Terrestrial (1982) (Atari, Jerome Domurat, Howard Scott Warshaw) (CX2674) ~',
  b'm\x17\xf1:(\x9d\xcc\x93j\x91\x9a\x1d\xbd}A^\x84<0\xeb': 'E.T. - The Extra-Terrestrial (CCE)',
  b'g\xb6`\xfcze6&\xf5\x95\x8c\xff\xb1\xf8\xa9\xc3\x17\xd9\xaa\xa3': 'Earth Attack (AKA Defender) (Zellers)',
  b'\x1f\x83I#\xea\xc2q\xbf\x04\xc1\x86!\xac*\xad\xa6\x8dBi\x17': 'Earth Dies Screaming, The (The Day the Earth Stood Still) (1983) (20th Century Fox Video Games - Sirius Software, Dan Thompson) (11020) ~',
  b'\xe7\x10q\xd9\x9fXw\xa4\x98\xd5=\xf4\x8c@e\xe7\xfd\x1e\x07\xaa': 'Earth Dies Screaming, The (Unknown) (PAL)',
  b'N\x1a\xc80\xa8\xcd,\xa2\x7f\x92\xcb]~x\xc8\x9cV>L\x10': 'Eddy Langfinger, der Museumsdieb (AKA A Mysterious Thief) (1983) (Quelle) (732.052 6) (PAL)',
  b'q\xb9\xdf\xc46\xc4\xb6\xa2\x18\xabp\x81\x15\xf0\x95\x8c\xddY\x81!': 'Eggomania (Canal 3 - Intellivision)',
  b'h\xcb\xfa\xdf\tz\xe2\xd1\xe88\xf3\x15\xc7\xcc{p\xbb\xf2\xcc\xc8': 'Eggomania (Weird Bird) (Paddle) (1982) (U.S. Games Corporation - JWDA, Todd Marshall, Robin McDaniel, Wes Trager, Henry Will IV) (VC2003) ~',
  b'[\x1a}\xc9\xc5\xa6\xb7R?&\xa9\x1a\xd9\xed\xb3\x04g\xaf\xf5s': 'Eggomania - Eierregen (Paddle) (1983) (Carrere Video - JWDA, Todd Marshall, Robin McDaniel, Wes Trager, Henry Will IV - Teldec - Prism) (USC2003) (PAL)',
  b'\xc0\x8e\xda"=\x9b\xbdt\xd2e\xf8=\xab%\xe1\xc5\xae\xd2\x1d)': 'Eishockey-Fieber (AKA Ice Hockey) (1983) (Quelle) (873.790 0) (PAL)',
  b'\xba\xb8r\xeeAi\\\xef\xe4\x1d\x88\xe4\x93!2\xec\xa6\xc4\xe6\x9c': 'Elevator Action (1983) (Atari, Dan Hitchens) (CX26126) (Prototype) ~',
  b"\xf1\xb3p\x0f\xec\x94\xef\x84\x195\xf0'1\xdc\xc4\xb9\x91i\xe6z": 'Elf Adventure (04-22-83) (Atari, Warren Robinett) (Prototype)',
  b'\xc5\xd0\xf9\x15\xfflg\xd3`\x86\xbc\xb5@\xb2\x86\xcc}\x07\x88\xab': 'Elf Adventure (05-02-83) (Atari, Warren Robinett) (Prototype)',
  b'\x93\xcf\x13\xc1m\x99\\i\xc7\x80\x85\xef\xfbU_\x17\x10_\x8b\x01': 'Elf Adventure (05-25-83) (Atari, Warren Robinett) (Prototype) ~',
  b'G_\xc2\xb2<\x0e\xe2s8\x859\xa4\xee\xaf\xa3O\x8f\x8d?\xd8': "Eli's Ladder (1984) (Simage) ~",
  b'9\x83\xe1\t\xfc\x0b8\xc0\xb5Y\xa0\x9a\x00\x1f>_+\xb1\xdc*': 'Elk Attack (1987) (Atari, Mark R. Hahn) (Prototype) ~',
  b' Z\xf4\x05\x1e\xa3\x9f\xb5\xa08\xa8T\\x\xbf\xf9\x1d\xf3!\xb7': 'Encounter at L-5 (Megalon Invasion) (Paddle) (1982) (Data Age) (DA1001) ~',
  b'k\x1e\x07\x87\xefYZ\xf5\xf8_\xef;\xf7\x80U;?\xbaL}': 'Encounter at L-5 (Megalon Invasion) (Paddle) (1983) (Gameworld) (133-001) (PAL)',
  b"\xfc\xf3DP\xbf\xd4\xa3jJ*'~\xff\r\xbe_\x13\x18\xe9\x7f": 'End of the World, The (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\x82\xe9\xb2\xddm\x99\xf1S\x81Pjv\xef\x95\x8a\x17s\xa7\xba!': 'Enduro (1983) (Activision, Larry Miller) (AX-026, AX-026-04) ~',
  b'\xdd\x89gj8\xee\xac\x90?2m+\xc4\x02/|\xea\xc3ZJ': 'Enduro (1983) (CCE) (C-810)',
  b'\x8e\xb0&\xdf.\xd0\xe0\x7f\xe6\xafM5\x8a\xe9\xcf\xdd\xff\rkS': 'Enduro (1983) (Digitel)',
  b'\x1e\xd9\xee2\xa5\xbe\x89U\xbe{H}E\x83\x8e3\x1c_s\xa5': 'Enduro (1983) (Dynacom)',
  b'\xea\x9d\xfb\xda$=\xd7\xbc\xcb\x14\x8e3Sy}ZvM\xcab': 'Enduro (1984) (Supergame)',
  b'\xc4\xba\x92\xa7\xf2\x91\x8f?\x8e" \x91\xd6\xc4}\x03\xcc\xb7@;': 'Enduro (Canal 3 - Intellivision)',
  b'\x9eya\xd2\r\xba\x83i\xcd\x1d\xb6\\\x1a\x07\xe9\x13\xb0\xde\xd9\x0f': 'Enduro (Dactari - Milmar)',
  b'|\x93\x1a\x1a\xb9T\x8e\x02\xf5XS\xea6\x9bh\x84|-\x8d\x01': 'Enduro (Digivision)',
  b'\x15\x94h\x9e\xdd\xcf\xcd\x17@\xe8\x08O)\xb3\xe3}\xf47_V': 'Enduro (Robby)',
  b'\\2|N\x87\x89\xf4\x9d8\x1fl\x8e\xd3u\x1eM3N\xfd\x91': 'Enduro (Tron)',
  b"\x9dQ\xd6\xb8\xc4]\x04\xb6\x84\xa2\x15\xbd?9'b\xf7\xf2\xaf\x8f": 'Enduro (Unknown)',
  b'\x9b\x19 .;\xa1\xf1m2a\xc2\xce\x7fq\x06\xab9\xc0\xce\x19': 'Enduro (Unknown) (PAL)',
  b'\n/\xc2\xf0H\xa9\x90G\x99\xdf&\xdb\x1e\x17\xff\xed\xad\x8f\xf3\x02': 'Enduro - Transamerika-Rennen (1983) (Activision, Larry Miller - Ariola) (EAX-026, EAX-026-04B, EAX-026-04I - 711 026-725) (PAL)',
  b'\x18\rj\xadO_\xe5\xde\xa5S\xef./\xf23\xbdcY-\x00': 'Entity, The (1983) (20th Century Fox Video Games, Mark Klein) (11036) (Prototype) ~',
  b'y\x05\xae\xe9\nm\xd6M\x958\xe0\xb8\xe7r\xf83\xba\x9f\xeb\x83': "Entombed (Maze Chase, Pharaoh's Tomb, Zombie) (1983) (U.S. Games Corporation - Western Technologies, Jeff Corsiglia, Paul Allen Newell, Steven B. Sidley, Tom Sloper) (VC2007) ~",
  b"'\xd9%\xd4\x82U=\xef\xf2?\x08\x89\xb3\x05\x10\x91\x97}i ": 'Espial (1984) (Tigervision - Teldec) (7-012 - 3.60016 VC) (PAL)',
  b']\xb1h\xbbE\r\xc8/a\x8d\xfa`\xb9\xf2q\xad\xe3\xa0W\xc7': 'Espial (1984) (Tigervision) (7-012) ~',
  b'\xd8\xcf\xd0PY\x15\xe5\x9c?$Q\r6E\xbdxx\xea\x87*': 'Euro Gen (02-01-83) (JWDA, Todd Marshall) (PAL) ~',
  b"\xc1x\x01\xc0\x19\n\xde'\xf48\xe2\xaa\x98\xdd\xe8\x1b:\xe6bg": 'Exocet (1983) (Panda) (109)',
  b'\xc7\xbaY\x12\xd4\x0e\xb2\x99/\xda\x0f\x99\xebp(\xd9\x9e\xda\xd7,': 'Extra Terrestrials (1984) (Skill Screen Games, Herman Quast) ~',
  b'b\x97\xdd3jcC\xf9\x8c\xd1B\xd1\xd3\xd7l\xe8Gp\xa4\x88': 'Fantastic Voyage (1982) (20th Century Fox Video Games - Sirius Software, David Lubar) (11008) ~',
  b']\x84\xb5Og+\x92\xf8p\x98Rw\xd6\xea\n\xc9\xc1\xf6\xe9\xf5': 'Fantastic Voyage (208 in 1) (Unknown) (PAL)',
  b'c\xa2MO\x86R\x99t\xee8\xf0\xe3\xbdf\x02G\x085\xc9\x93': 'Farmyard Fun (AKA Play Farm) (Video Game Cartridge - Ariola) (TP-617)',
  b'\xa5aLu\x1f)\x11\x8d\xdb=\xec\x97\x94a+\x98\xa0\xf0\x0b\x98': 'Fast Eddie (1982) (20th Century Fox Video Games - Sirius Software, Mark Turmell) (11003) ~',
  b'\xb3\xb3\r\xe8 \xb8^EJ\x15\x10\xc6\x0c\xad\xb3\x05\x11\xb6U\xf7': 'Fast Eddie (1983) (CCE) (C-834)',
  b'\x81\x91QJ\xf6.(\x0c\xd2&\x1f\x9f$\x03\xea\xb6\xaa\xe1\xc0\x98': 'Fast Food (1982) (Telesys, Don Ruffcorn, Jack Woodman) (1003) (PAL)',
  b'\xc6*pdY9H\x0b\x18N;.7\x8e\xc4\xbc\xbdHK\xc7': 'Fast Food (1982) (Telesys, Don Ruffcorn, Jack Woodman) (1003) ~',
  b'0\x174/Y*\xcc\x87\x8f\x16\xedA\xbd\xb2\xa9\xdb]UA\x12': 'Fast Food (Unknown) (PAL)',
  b'\xae&\xb1\x81\xc6\x8d\x17\x9bl\x191gwk\x95\xa6c\xc3\xa5O': 'Fast Food (Zirok)',
  b'Y\xd0\x04T~i:\xa0[\xb33?\x05Qc\xad\xe2\xc9\xea\x95': 'Fatal Run (Ultimate Driving) (1989) (Atari - Sculptured Software, Steve Aguirre) (CX26162) (Prototype)',
  b'\xd0\xbbX\xea\x1f\xc3~\x92\x9e_|\xde\xad\x03{\xb1J\x16dQ': 'Fatal Run (Ultimate Driving) (1991) (Atari - Sculptured Software, Steve Aguirre) (CX26162) (PAL) ~',
  b"hd'\xccG\xb6\x99\x80\xd2\x92\xd0E\x97'\x03G\x94's\xff": 'Fathom (Scuba) (1983) (Imagic, Wilfredo Aguilar, Michael Becker, Rob Fulop) (720111-1A, 03205) ~',
  b'\x80n>\xb9w\xdf\xfe\x86\xfb\t\xad-\x8dN\x17f\xfbz\xa59': 'Fathom (Scuba) (1983) (Imagic, Wilfredo Aguilar, Michael Becker, Rob Fulop) (720111-2A, 13205) (PAL)',
  b'e\x93\xf8c%j1t{x\xeb\x8d\x02\xc5l\xb1\x81\x1a\x84\xb4': 'Festival (AKA Carnival) (4 Game in One) (1983) (Bit Corporation) (PGP230) (PAL)',
  b'"\xf6\xb8\xfc\x84gB[\xd1\x99\xf3\xcb\xc9\xd6\x9c\xe7\x91\x82\x0e%': 'Feuerwehr im Einsatz (AKA Fire Fighter) (1983) (Quelle) (343.173 1) (PAL)',
  b'hBu\xb2/+\xac}W|\xf4\x8c\xf4/\xa1O\xa6\xf6\x96x': 'Fighter Pilot (AKA Tomcat - The F-14 Fighter Simulator) (1988) (Activision, Dan Kitchen) (EAK-046-04B) (PAL)',
  b'\xba\x9a\x8c\xcf\xebU-\xd7V\xc6`\xea\x84:9a\x9d<w\xe9': 'Final Approach (1982) (Apollo) (AP-2009) ~',
  b'\xe2\x84;\x86\xda\x00\rnJo\xe6\x83$:{>>\x16\x0cw': 'Fire Bird (AKA Phoenix) (Video Game Program) (PAL)',
  b'R\xe5\x06q\t\xb7\xe3r\x95*}\x00\x94\xce\xf8\xd1m2\xa99': 'Fire Birds (AKA Sky Alien) (Perry Rhodan-Serie) (1983) (ITT Family Games) (554-33 383) (PAL)',
  b'\xf7l\xc1J\xfdz\xef6|Z]\xef\xbd\x84\xf3\xbb\xb2\xf9\x8b\xa3': 'Fire Fighter (Fire Fighters) (1982) (Imagic, Brad Stewart) (720105-1A, IA3400) ~',
  b'\xcf\x15&\x00\xad\xf9E\xd7A\xc4!\xf2\x1fQ\x0c\x8ew\xae\x1e\xcb': 'Fire Fighter (Fire Fighters) (1982) (Imagic, Brad Stewart) (720105-2A, IA3400P, EIX-005-04I) (PAL)',
  b'\xdfT \xeb\x0fq\xe6\x81\xe7".\xde\x8e!\x1av\x01\xe7\xa3\'': 'Fire Fly (1983) (Mythicon, Bill Bryner, Bruce de Graaf) (MA1002) ~',
  b'\xb10\xa3\x9c7\x0b\xe1\x1c)\xbdwF\x005"$\x95fH\xe1': 'Firebug (AKA Spinning Fireball) (Suntek) (SS-028) (PAL)',
  b'\xc11\xafNE\x9bJh\xd4w<\xfd\\\xe9\xee\xecp\xc5&\x14': 'Fisher Price (AKA Skindiver) (1983) (CCE) (C-863)',
  b'\x16(R\xd5\xd3}F\xda\xea\xf8$\xb7\x10\xdd\xf2\x03\xf9`\x1a\xc9': 'Fishing (AKA Fishing Derby) (32 in 1) (1988) (Atari) (CX26163P) (PAL)',
  b'S\x1e\x99Z\xefl\xd4{\x0e\xfe\xa7*\xe3\xe5j\xee\xe4I\xd7\x98': 'Fishing Derby (1980) (Activision, David Crane) (AG-004) ~',
  b'\xe7|\x83}\x07\xe5\xbb\x1ea\xfb"\xe2\xff\x8d\x89\x8d\xf74jY': 'Fishing Derby (32 in 1) (1988) (Atari) (CX26163P) (PAL)',
  b'\xf1\xcfF\xf3\x83p\xb0.\x12\xef\xaf\x92_\xa5\x87\x96\x8drgN': 'Fishing Derby (CCE)',
  b'mw\xd4[\x0ctC\xd2\xd0\xeb\x07\x18%\xb2 \xb3\xa2\xbe\xfe\x0b': 'Fishing Derby (Canal 3 - Intellivision)',
  b'\x16w\xcc\x90\xce\xd42\x05\xe09~\t\xc3\x0c\xae\xdcJq\xa3E': 'Fishing Derby (Hack) (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\x1fr5\xf81\xfeS\xf6?Z\x03Bb\x07q\x89\x1cL\x80\x05': 'Fishing Derby (Hack) (Unknown) (PAL)',
  b"\xc1\x0f\xb7\x96'\x94\xf0\xe2'\x0e\x81\xed\xad\xc1\xe1\xe0sa_O": 'Fishing Derby - Schneller als der Hai (1980) (Activision, David Crane - Ariola) (EAG-004, PAG-004 - 711 004-715) (PAL)',
  b'_O\xa8\xc7\xa4\x0f\x1e\xc6+\x03\xb8\xe8h\x07\xc9\x8c\x86\xa0\x81\x9e': 'Flag Capture (208 in 1) (Unknown) (PAL)',
  b'\xa5\x19\xd2]`sv\x05\xb9\xb8?\xc4\xd1\xcbt\x02\xbf\xed\x08-': 'Flag Capture (32 in 1) (1988) (Atari, Jim Huether) (CX26163P) (PAL)',
  b'\xac\x05\xf0_3e\xf5\xe3H\xe1\xe6\x18A\x00e\xa1\xc2\xa8\x8e\xe4': 'Flag Capture - Capture (Capture the Flag) (1978) (Atari, Jim Huether - Sears) (CX2644 - 6-99824) ~',
  b'\x9d\xf3\xf1>\x14:\x8f6\t][u\x18g\xba\xf8"\xd0M\xca': 'Flash Gordon (Unknown) (PAL)',
  b'\x86\xd4\xfe\xde\x19z\xfc\xd3f\xa5A\x98\x1a\x17=\x9et\xfa\xadn': 'Flippern (AKA Video Pinball) (Double-Game Package) (1983) (Otto Versand) (781698) (PAL)',
  b'\xc6\xfeL\xe2K\xc1\xeb\xd58%\x8d\x98\xcf\xe8)\x963#\xac\xca': 'Football (1979) (Atari, Bob Whitehead - Sears) (CX2625 - 6-99827, 49-75114) ~',
  b"\x91W]'\xf4\xd8q\x1cg{\xd5(\xdc\xae\xff\xee\r\xf7\xe5\xb0": 'Football (208 in 1) (Unknown) (PAL)',
  b'\x9a&\xfa\xdd\xd1\x02\\\x0f\xec\x88\xab\x0e\x03\xbe\xf91\xf3ra\xff': 'Football (AKA Super Challenge Football) (1988) (Telegames) (5658 A088) (PAL)',
  b'y\x0e\nYrp\xa5\t\x01\x10\xe3\xbf\xbd\xc3\xa0v\xa4|\xed\x14': "Forest (1983) (Sancho - Tang's Electronic Co.) (TEC006) (PAL) ~",
  b'l\xedlv;\x98C\xc1\x1a\xf6\xc4_\xfaz\xb9\xa2\x95\xccx\xbb': 'Fox & Goat (AKA Nuts) (Double-Game Package) (1983) (Otto Versand) (311377) (PAL)',
  b'\xc6\x02;\xf78\x18\xc7\x8b.Gz\x9cm\xacA\x1c\xdb\xf9\xc0\xaa': "Frankenstein's Monster (1983) (Data Age, J. Ray Dettling) (112-008) ~",
  b'\xd3\xdf\xa6\x185\xb9j\xaehU\xf7\xca\xba\x0cEj*\x98\x16\xa4': "Frankenstein's Monster (1983) (Gameworld, J. Ray Dettling) (133-008) (PAL)",
  b'\x91\xcc~\\\xd6\xc0\xd4\xa6\xf4.\xd6cS\xb7\xee{\xb9r\xfa?': 'Freeway (1981) (Activision, David Crane) (AG-009, AG-009-04) ~',
  b'\n\x02\xe9\xb7P\xb5\xda\x8e\xe2\x89X\xa3\x83\x16u\xdb\x9bp\xe7.': 'Freeway (CCE)',
  b'\xd9\xca\x1a/K9\xc0E\xa6\xc4\xf1\x84\xd8\x87f\x8e\xd7!\x93o': 'Freeway (Canal 3 - Intellivision)',
  b'0\x81<i\x1b\xac\xc2{\x18\xbd\xee\xeb\x9f\xd5|\xd8X\xa2g\xb6': 'Freeway (Dactari - Milmar)',
  b"R\xee\xe6\x9b\x08\xe7\xdc2\xe1s'\x02\xa5\xfb\xf3\xac\x89\x93\xf0\x99": 'Freeway (Zellers)',
  b'e\x19/Q\xa2q\xec\xbd\xd0\x16\x8c\x9d$\xac;\x8dd\xd6?I': 'Freeway - Das verrueckte Huhn (1981) (Activision, David Crane - Ariola) (EAG-009, PAG-009 - 711 009-720) (PAL)',
  b'7S[y\x1ac\xc3\xe3]\x04\xea\r\x14\xc0!V\x80\xd7\x1e\xda': 'Freeway Chicken (AKA Freeway) (32 in 1) (1988) (Atari) (CX26163P) (PAL)',
  b')!V\xb1\xad\x05\xbb\xcb\x97I\xde\xd7\x87\x80\x18A\xd6\xbe\xed\r': 'Freeway Rabbit (AKA Freeway) (32 in 1) (1988) (Atari) (CX26163P) (PAL)',
  b'nzZ\x17\x1a+\xb7\x0c\x8d\xee\x1d\xf8H\x04\x98\xa6\x8b\xa3l\xe6': 'Frisco (Hack) (Unknown) (PAL)',
  b'\x15\xd9\x89\xe4\x9c_\x1a\xf2\xa0\xfe\x11L\xdb\xa8\xf6\xda9f\x0eu': 'Frisco (Unknown)',
  b'\xde\xd4\xc6M`\xff\xceS\x04\xd7>\x01k\xe7\xde\xb5\xf2\xdbE\xd4': 'Frog Demo (1983) (CommaVid, Joseph Biel) (PAL) ~',
  b'\rj\x96\xf8W\xae\x0e\x81;MI8f\xe2B\x0c\xc5\xc4\xba\xd5': "Frog Pond (08-27-1982) (Atari, Nick 'Sandy Maiwald' Turner) (CX2665) (Prototype) ~",
  b'\xdeo\xc1\xb5\x1dA\xb3M\xcd\xa9/W\x9b*\xa4\xdf\x8e\xcc\xf5\x86': "Frog Pond (1982) (Atari, Nick 'Sandy Maiwald' Turner) (CX2665) (Prototype)",
  b'\xb2z\xa8T2\xf8pLK>\x14\xf6\x04\x9eb\\\xf9r\x9a\xa8': 'Frogger (1982) (Parker Brothers, Ed English, David Lamkins) (931502) (PAL)',
  b'\xe8Y\xb95\xa3d\x94\xf3\xc4\xb4\xbfUG9&\x00\xfb\x9c\x96\xf0': 'Frogger (1982) (Parker Brothers, Ed English, David Lamkins) (PB5300) ~',
  b'\xbc\xe9-\xe2/\xc8\x95\x0f\x0e\xeb\xb6C\xd9o\xe9\xfa]\xba+r': 'Frogger II - Threeedeep! (1984) (Parker Brothers, Mark Lesser) (PB5590) (PAL)',
  b'k\x9eY\x1c\xc58DyW%\xfcf\xc5d\xf06M\x1f\xbe@': 'Frogger II - Threeedeep! (1984) (Parker Brothers, Mark Lesser) (PB5590) ~',
  b'v\x82\xb4\xb6\xd4Xe\xb7\xee\xc1$N]v\xf0\xf2\xef\xdb\x17\xe7': 'Frogs and Flies (1988) (Telegames) (PAL)',
  b'\xb5\xa1C\xcb\x04iS\xa6\x9fv\xcb/\x90>a5\x0f\r\xe2e': 'Frogs and Flies (Hack) (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\xf3D\xd5\xa8\xdc\x89\\Z*\xe0(\x8f<l\xb6fP\xe4\x91g': "Frogs and Flies - Frogs 'n' Flies (1982) (M Network - APh Technological Consulting, David Rolfe - INTV) (MT5664) ~",
  b'\xcf2\xbf\xcd\x7f,;}*j\xd2\xf2\x98\xae\xa2\xdf\xad\x82B\xe7': 'Front Line (1984) (Coleco - Individeo, Ed Temple) (2665) ~',
  b'\xceu\x80\x05\x9e\x8bA\xcbJ\x1esL\x9b5\xce7t\xbfwz': 'Frontline (AKA Combat) (Zellers)',
  b'\xb9\xe6\x047\xe7i\x1d^\xf9\x00,\xfc}\x15\xae\x95\xf1\xc0:\x12': 'Frostbite (1983) (Activision, Steve Cartwright) (AX-031) ~',
  b':]\xe0s2:&\xfd\xa9\xa6\xad\x15\x90\xd2lt\xd7\xa4!\x92': 'Frostbite (1983) (Digitel)',
  b'{\xab|y\xd8g\xc4\xfd\xbd\xe8g\x8f\xd2\xb5\x94.&\xd3\xce\x02': 'Frostbite (1983) (Video Mania)',
  b'/\x0e\xd5\xe0\x15h\x9a\x0c\xea3\xa9]\x99\r\xed\xca*\xb7aB': 'Frostbite (CCE)',
  b'\x7f\xbf\r\x1f\x8a\xa9\xf9\xff\x85\xb5\x13\x14<\x9f\xcee>\x8bei': 'Frostbite (Digivision)',
  b'\x8a\xd06g\xbb\xf7=<w`\xcb\x82\xf2\xc4D/\x87EH<': 'Frostbite (Hack) (Unknown) (PAL)',
  b'\x1cR\x9b\x13\xc1\xf7z\xf1\xc9\xf4\xd2\xa7Nz\xd6\xab+D\xa1\xfd': 'Frostbite (Unknown) (PAL)',
  b'X\xa6\xf8$4\xcc\xf4\x9c\xa4 Yh\t\xce\x95E78E\xa2': 'Frostbite - Expedition ins Eis (1983) (Activision, Steve Cartwright - Ariola) (EAX-031, EAX-031-04B - 711 031-717) (PAL)',
  b'M\xe5\xd9\x15\xb5\x84\x02\xf5\x01\x81s\x12\x0fB\xbez}\xfa\xbb\x8e': 'Fuchs & Schweinchen Schlau (AKA Oink!) (1983) (Quelle) (806.174 9) (PAL)',
  b'Q\xf4w\xf3rR\xb5\xc3$i\xfc\x1falV\xed|\xad$\xbc': 'Fun with Numbers (32 in 1) (1988) (Atari, Gary Palmer) (CX26163P) (PAL)',
  b'\\\xc4\x01\x0e\xb2\x85\x8a\xfe\x8c\xe7\x7fS\xa8\x9d7\xc7XN\x15\xb4': 'Fun with Numbers (AKA Basic Math) (1980) (Atari, Gary Palmer) (CX2661)',
  b'\xeb\xdc\x9f\xc9)\xa6\xefG\xda\xea\x8a\xb59\xb7i\xa2D$:\x1a': 'Fun with Numbers (AKA Basic Math) (1980) (Atari, Gary Palmer) (CX2661P) (PAL)',
  b'\xfb\xa4a\xd2\xa2\xd19YE\x80l\x88?M\xca\x92W\x12\x88^': 'Funky Fish (1983) (UA Limited) (Prototype) ~',
  b'&\x1b\x1c\xc4\x97\xa2\xdeZ\xa5\x14\x83\xdbO\xd7\xd3\xc0\xe6\xfd\x99\xef': 'Fussball (AKA International Soccer) (Videospielkassette - Ariola) (PGP235) (PAL)',
  b'\x9c\xfbb\x88\xa5\xc2\xda\xe6>\xe6\xf5\xe92R\x00\xcc\xd2\x1a0U': 'G.I. Joe - Cobra Strike (Paddle) (1983) (Parker Brothers, John Emerson) (PB5920) ~',
  b'>\xeeD9\x08-\xc9?\x9eMy\xc4\x19\xfa\xa42\xf2\x80a\x7f': 'Galactic (AKA Condor Attack) (Funvision - Fund. International Co.) (PAL)',
  b"Q\x17\xb0\x8e'\t&\x1b\x0e\xedA\xbd\xaceq:\x07;\x90H": 'Galactic (AKA The Challenge of.... Nexar) (Rainbow Vision - Suntek) (SS-002) (PAL)',
  b'gyzU\xbes\xb3s\xad\n\xfcg\xe4(\xa6\xe8\xdfV\xceB': 'Galatic (AKA Challenge of.... Nexar, The) (208 in 1) (Unknown) (PAL)',
  b'o\xa5&\xdaqs\xaew\xf1\xd3_\x19{\xde_!\xc9\x15\x1f\xbf': 'Galaxian (01-05-1983) (Atari - GCC, Mark Ackerman, Tom Calderwood, Glenn Parker) (CX2684) (Prototype)',
  b'\xa7\xc9\xc5%c*\x8e\\?z\xcc\\\x9b\tQ8\xa8D\x1fW': 'Galaxian (02-04-1983) (Atari - GCC, Mark Ackerman, Tom Calderwood, Glenn Parker) (CX2684) (Prototype)',
  b"\xb0\x81\xb3'\xac2\xd9Q\xc3l\xb4\xb3\xff\x81+\xe9V\x85\xd5/": 'Galaxian (1983) (Atari - GCC, Mark Ackerman, Tom Calderwood, Glenn Parker) (CX2684) ~',
  b'\xfb;2H\xd9pZ\x88?\xcdy\xc2=\xc6\x07Yv\xc2\x06F': 'Galaxian (1983) (Atari - GCC, Mark Ackerman, Tom Calderwood, Glenn Parker) (CX2684, CX2684P) (PAL)',
  b'I\xe3\xfd\xc8\xe4\x86\x05\xaa\x1f\xab7\xce>\xc0\xe2cq\x82/\xa5': 'Galaxian (CCE)',
  b'\xee\x0b\x91\xee\xc3+\xd77\xac\t\xe3\xea\x1f\xea\x844\xf0%1\x14': 'Galaxian (Digivision)',
  b'x\xf7\xbd"\x15+jf\xa0\x86\xec\xa90\n\xe3\xcd\xbb}\x97\xbf': 'GameLine Master Module ROM (1983) (Control Video Corporation) ~',
  b' \xab\xf6\xbaj\x96\x85br \x12\x8b\xab\xf7L0;|\x8d0': 'Gamma-Attack (1983) (Gammation, Robert L. Esken Jr.) ~',
  b'\xae@1r\xbe\xa0\xac\xa1\xf1.\x11]C\x1c\x93\xfav\x97e`': 'Gangster (AKA Outlaw) (Videospielkassette - Ariola) (PGP238) (PAL)',
  b'\x8c\xf4\x9dC\xbdb0\x8d\xf7\x88\xcf\xac\xbf\xcd\x80\xe9"lu\x90': 'Gangster Alley (1982) (Spectravision - Spectravideo) (SA-201) ~',
  b'\t\xc5\x01|\xff\t\xeb\xb4\x00W7B\xcc\x99\xab\xc6\xe1Wsq': 'Gangster Alley - Gangster Ruine (1982) (Spectravision - Spectravideo - Quelle) (SA-201 - 412.783 3) (PAL)',
  b'\xbc\r\x1e\xdc%\x1d\x8dM\xb3\xd5#N\xc8=\xee\x17\x16B\xa5G': 'Garfield (Garfield on the Run) (06-21-1984) (Atari, Mimi Nyden, Steve Woita) (CX26132) (Prototype) ~',
  b'\xe9\x19\xee\x03\xe8\xa1\x7f\xd8[\xe1\xe5;M_\xf7m\xaeT\xd0\x99': 'Gas Hog (1983) (Spectravideo, Mark Turmell) (SA-217) [fixed] ~',
  b'\r\xa1\xf2\xdeZ\x9bZf\x04\xcc\xdb\x0f0\xb9\xdaN_y\x9b@': 'Gas Hog (1983) (Spectravideo, Mark Turmell) (SA-217) ~',
  b's\xad\xae8\xd8mP6\x0b\x1a$rD\xdf\x05\x89.3\xdaF': 'Gauntlet (1983) (Answer Software Corporation - TY Associates) (ASC1002) ~',
  b'i\xa8\xb8\xa5Cw\xf5vg"\x14,2\x1d\xcf\xb0J\xb9I ': 'Gefecht im All (AKA Space Jockey) (1983) (Quelle) (147.443 6) (PAL)',
  b';\x1f\xb93B\xc7\xf0\x14\xa2\x8d\xdd\xf6\xf1h\x95\xd1\x1a\xc7\xd6\xf0': "General Re-Treat (AKA Custer's Revenge) (1982) (PlayAround - J.H.M.) (206) (PAL)",
  b'\x12\xd2\xf7\xe7^\x91\x07\t1\x82\xdf\xcf?\xb7Cp\xb0\x16|B': 'Ghost Manor (1983) (Xonox - K-Tel Software - Beck-Tech, Steve Beck) (6210, 06002, 06004, 99002) (PAL)',
  b'KS7v\xdc\xd9\xd58\xf9 j\xd1\xe2\x8b0\x11m\x08\xdf\x1e': 'Ghost Manor (1983) (Xonox - K-Tel Software - Beck-Tech, Steve Beck) (6210, 06002, 06004, 99002) ~',
  b'^\xd0\xb2\xcb4m r\x0e<Rm\xa31U\x1a\xa1j#\xa4': 'Ghostbusters (1985) (Activision, David Crane, Dan Kitchen) (AZ-108-04) ~',
  b'\x9aO\x19\xb0\xeb\x7f<v\xea\x06F\xc6\xb8\xe0\x98Xcia\x8c': 'Ghostbusters (1985) (Activision, David Crane, Dan Kitchen) (EAG-108-04, EAZ-108-04B) (PAL)',
  b'\x1b\xcf\x03\xe1\x12\x90\x15\xa4j\xd7\x02\x8e\x0et%6S\x94N\x86': 'Ghostbusters II (1992) (Salu - Avantgarde Software, Michael Buetepage) (460741) (PAL) [different tune] ~',
  b'\xe02\x87c\x05dz\x95\xb6"\xe5\xc4\x97\x1fp\x96\xefr\xac\xdb': 'Ghostbusters II (1992) (Salu - Avantgarde Software, Michael Buetepage) (460741) (PAL) ~',
  b'\xb6N\xd2\xd5\xa2\xf8\xfd\xacO\xf0\xceV\x93\x9b\xa7.4?\xec3': 'Gigolo (1982) (PlayAround - J.H.M.) (205)',
  b'=\x86}I\xf3)S$\xf2\x14\x13\x86@\xf0\xb2\xfds\xae\x89\xfb': 'Glacier Patrol (1988) (Telegames - VSS, Ed Salvo) (5665 A016) (PAL)',
  b':=r\x06\xaf\xee6x`&\xd6(\x7f\xe9V\xc2\xeb\xc8\x0e\xa7': 'Glacier Patrol (1988) (Telegames - VSS, Ed Salvo) (5667 A106) ~',
  b"{\xca\x0fz\x0f\x99'\x82\xe4\xe4\xc9\x07r\xba\xc9v\xca\x96:m": 'Glib - Video Word Game (1983) (QDI, Mike Montana, Rich Montana - Selchow & Righter) (87) ~',
  b'\xf7\x8cG\x8a\xac\xf6Se"\xe8\xd3z8\x88\xa9u\xe1\xa3\x83\xcd': 'Go Go Home (Unknown)',
  b'<9\xd2\xfd\xec\xec\x8b?\xcf6\x03\xd9\xc6c\xb8Lh/c|': 'Golf (208 in 1) (Unknown) (PAL)',
  b'\xe7(\xc4ol\x16\x08\xb4:Y"\x9f\xb4c\xf3\xf8\x7f\x00\xe9\x93': 'Golf (32 in 1) (1988) (Atari, Tom Rudadahl) (CX26163P) (PAL)',
  b'\xa2]Rw\x04\x081M\xecoA\xaa\xf5\xf9\xf0\xa2\xa3\xe2\xc1\x8f': 'Golf (Championship Golf) (1980) (Atari, Tom Rudadahl - Sears) (CX2634 - 49-75121) ~',
  b'\xb8\x7f\xbe\xb3i\xb2\x1fs>\x0e\xeb(\xb9\xa1\x9a\x8e$\xf3\x0c\x9e': 'Golf (Championship Golf) (1980) (Atari, Tom Rudadahl) (CX2634, CX2634P) (PAL)',
  b'\xb0\xc2\xef\xef\xc8\xd8\x06\x93VZd\xee\x02[\xe4\x80\xaa?\x17F': 'Golf Diagnostic (1983) (Video Soft, Jerry Lawson, Dan McElroy) (Prototype) ~',
  b'9T\x84%\xa9{\x18Z\xec\x93d\xa3C\xe9-Q\x8b\xb7\x02\xc0': 'Good Luck, Charlie Brown (04-18-1984) (Atari, Christopher H. Omarzu, Courtney Granner) (CX26112) (Prototype) ~',
  b'Y\x99\x10\xea\xebe\xde\x07\x882\x80\xba\x8d\x86\xe1\x0b\xd3\x82\xe9\xaf': 'Gopher (208 in 1) (Unknown) (PAL)',
  b'\x97\xfbH\x9b\xa4\xce\x0f\x8a0ec\x065cas!5,\xfb': 'Gopher (Gopher Attack) (1982) (U.S. Games Corporation - JWDA, Sylvia Day, Todd Marshall, Robin McDaniel, Henry Will IV) (VC2001) ~',
  b'\x0f\x15\xeeH\xcchc\x0e\xbb\xf3\xe4=7R\x8cP\x88PbE': 'Gopher (Unknown) (PAL)',
  b'\xafc\\\xf2~{zo@\x0fY\xd2;K\x01)\xac>\x87L': 'Gopher - Vorsicht Wühlmaus! (1983) (Carrere Video - JWDA, Sylvia Day, Todd Marshall, Robin McDaniel, Henry Will IV - Teldec - Prism) (USC2001) (PAL)',
  b'\xd50\xe7>:D\xb9\x1f\x0f,\xf7LR\x0e\xdf\x1c%\x8d\x07\xce': 'Gorf (1982) (CBS Electronics - Roklan, Joe Gaucher, Alex Leavens) (4L1751, 4L1752, 4L1753, 4L2275) (PAL)',
  b'5\xf84\x1cs\xc7\xe6\xe8\x96\xcb\x06YwB{?\x98\xae\x9f\x08': 'Gorf (1982) (CBS Electronics - Roklan, Joe Gaucher, Alex Leavens) (M8776, M8793) ~',
  b'2\xc9\xf5~y\xf0~V\x80\x8c\xec\xe2NP?\xbc2Z`\xdf': 'Grand Prix (1982) (Activision, David Crane - Ariola) (EAX-014, PAX-014, EAX-014-04B, EAX-014-04I - 711 014-720) (PAL)',
  b'$\xfa\xb8\x17r\x82\x16X+m\x95U\x8c6\x1a\xcef\xab\xf9o': 'Grand Prix (1982) (Activision, David Crane) (AX-014, AX-014-04) ~',
  b'\\W\xd8\x95CWe\x1e\x11\xbf\xbaI$D\xf6\x9a\x9ctC\xeb': 'Grand Prix (1983) (CCE) (C-826)',
  b'\x87\x01up\x11\x89\x1eI\xb6\xdf{\r\xcb\x90\x16\xa2[\x1b\x00!': 'Grand Prix (1983) (CCE) (C-826) [a]',
  b'\xc2\xda5\x8a^\xef\xb4\xa16\xfdg\x8d\x03\xe8\xfe\x88y\xa3\xa6\xf1': 'Grand Prix (Robby)',
  b'\xe3\xe4\xf4\x9e\xb6\xf8\xe0\xe7!\xfd\x01W\xed\xe8\x88e\xb7]\xe1\xb2': 'Grand Prix (Unknown) (PAL)',
  b'P\xfc\xec\xb6x\x84\xc6\x83\x07\x10\x91\x93kJ\xe3;P\xf6\t\xeb': 'Grand Prize (AKA Enduro) (Funvision - Fund. International Co.)',
  b'\xac\xf8?\xcdd\xb3&\xd7\xe2\xf0\xcd9ts\xe4Kah\xc1\x0c': 'Gravitar (04-12-1983) (Atari, Dan Hitchens, Mimi Nyden) (CX2685) (Prototype)',
  b'\xa3r\xd4\xdd=\x95\xb3\x86eS\xca\xe23nEe\xe0\x0c\xc2[': 'Gravitar (1983) (Atari, Dan Hitchens, Mimi Nyden) (CX2685) ~',
  b'\xaf\xdd\xee\xf0\x96\xb7p\xf1\x9c\x82\x9c\x01S\xdcv~\xc6\xe2t\xbe': 'Gravitar (CCE)',
  b'\xef\x08C\xc0\xda\xa1q\xbdlz*\x150\xba\xfb\x18\x9bY\xda\x86': 'Great Escape (AKA Asteroid Fire) (1983) (Bomb - Onbase) (CA282)',
  b"\xdfPjPn\xb33)^V\xcb\x9eB\x04\xe3'\xefc\x12}": 'Gremlins (Gargoyle) (03-12-1984) (Atari, Mimi Nyden, Scott Smith, Robert Vieira) (CX26127) (Prototype)',
  b'z\x02s)0\x9e\x01\x8b\rQ\xad\xcbj\xe1<\x9d\x13\xe5OJ': 'Gremlins (Gargoyle) (1984) (Atari, Mimi Nyden, Scott Smith, Robert Vieira) (CX26127) ~',
  b'\x010\xf9\xd13N\x10o/\xbd\x9b\xbd\\\x01\xef\xaeO9\xca\xe1': 'Ground Zero (AKA River Raid) (1983) (Goliath - Hot Shot) (83-113) (PAL)',
  b'\xb9v\x0f\xfb\xa0Q9\xbc\xa0\xfa\xc3\xf7\xd3\xdc\x1e]W`\x0e\xda': "Grover's Music Maker (Monkey Music) (Kid's Controller) (01-18-1983) (Atari - CCW, Stephan R. Keith, Laura Scholl, Preston Stuart) (CX26106) (Prototype) ~",
  b'\xc9\n\xca\xee\x06o\x97\xef\xc6\xa5 \xde\xb7\xfa>W`\xa4q\xfa': "Grover's Music Maker (Monkey Music) (Kid's Controller) (12-29-1982) (Atari - CCW, Stephan R. Keith, Laura Scholl, Preston Stuart) (CX26106) (Prototype)",
  b'\xa04\t\x0eJr\xac@\x06\x95*\x7f\x02\x1e\xa0\xd0\x869\xf5w': 'Guardian (Cosmic Combat) (Paddle) (1982) (Apollo, Larry Martin) (AP-2008) (Prototype)',
  b'}0\xffVZ\xd7\xb2\xa3\x14=\x04\x9c[9\xe4\xa6\xac?\x9c\xd5': 'Guardian (Cosmic Combat) (Paddle) (1982) (Apollo, Larry Martin) (AP-2008) ~',
  b'Y\x8fq\x0btt\xednN|\x13\xcc\x02\x94\xa4\xde\xba\x8f\xaa\x87': 'Gyruss (1984) (Parker Brothers - Roklan) (PB5080) (PAL)',
  b"K\xd8{\xa8\xb3\xb6\xd7\x85\x0e>\xa4\x1bMIL;\x12e\x9f'": 'Gyruss (1984) (Parker Brothers - Roklan) (PB5080) ~',
  b'(/\x94\x81t\x01\xe3r\\b+s\xa0\xc0V\x85\xcev\x17\x83': 'H.E.R.O. (1984) (Activision, John Van Ryzin) (AZ-036-04) ~',
  b'\x04\xe9F"T\xae\x9d\xb4\x80K=3\x1f*\xd2,a\x90\x13\x11': 'H.E.R.O. (1984) (Activision, John Van Ryzin) (EAZ-036-04) (SECAM)',
  b'\x9eY\x92\x84\xcd\xc8\xadJ\xa8\x03!\n\x19\xc0H_\x14\x12\xb3_': 'H.E.R.O. (CCE)',
  b'\x9d\x1fH\xfd0\x02o\x84\xf6\xc1\xc0\x8f\xe6\x9dB?\xa47\xf4\x0f': 'H.E.R.O. (Tron)',
  b'^O\x98!\x0e\xf5\xa6h%\x81\xb6\n\x9f\xc99\xfc\xe6\xb2-\x97': 'H.E.R.O. - Helicopter-Held (1984) (Activision, John Van Ryzin - Ariola) (EAZ-036-04, EAZ-036-04B, EAZ-036-04I - 711 036-720) (PAL)',
  b'#"\xe3\xac\xfd\x02\xb1 \x98\xffP\xeetV\x13\xf3\xbd\xc3F\xad': 'Halloween (1983) (Wizard Video Games - MicroGraphic Image, Robert Barber, Tim Martin) (007) (Prototype)',
  b'\xec\xc6\t <.&\xab \xfb\xd9-\x13\x94?\xc1\xfa\xed^\xcb': 'Halloween (1983) (Wizard Video Games - MicroGraphic Image, Robert Barber, Tim Martin) (007) (Prototype) [a]',
  b'Lr\xce\xc1Q\xf2\x19\x86k\xf8p\xfaz\xc7I\xa1\x9c\xa5\x01\xc9': 'Halloween (1983) (Wizard Video Games - MicroGraphic Image, Robert Barber, Tim Martin) (007) ~',
  b'tG|\xfay\x0ed\xb4\xc4\xfd\x16t\xc6\x82a\xc0\xe1\x875+': 'Hangman (1978) (Atari, Alan Miller) (CX2662P) (PAL)',
  b'V\x1b\xcc\xf5\x08\xe1b\xbcp\xc4-\x85\xc1p\xcf\r\x1dF\x91\xa3': 'Hangman - Spelling (1978) (Atari, Alan Miller - Sears) (CX2662 - 6-99811) ~',
  b'\xd6\xdbq\xda\x02\xae\x19A@\xbf\x81+\xe3Mn\x8ag\x85\xd18': 'Harbor Escape (AKA River Raid) (1983) (Panda) (110)',
  b'\xa2\\K3\xe1\xc8h\x95\xcbcc[\xd6\xca[\x83y\xa2\x06\xb3': 'Hard-Head (1983) (Activision) (Prototype) ~',
  b"\x90I[\xea\x80\x15\x840\xffc\xbd=\xb9w$\xcf<\xe2'6": 'Harem (1982) (Multivision, Michael Case) ~',
  b'\xb8AA\x1b,q\x9aJ\x89I\x8e\x8c\xac\x06\xc1\xc1\x98lyM': "Haunted House (Mystery Mansion, Graves' Manor, Nightmare Manor) (09-28-81) (Atari, James Andreasen - Sears) (CX2654 - 49-75141) (Prototype)",
  b'\x14v\xc8ia\x90u\xb5Q\xb2\x0f,\x7f\x95\xb1\x1e\r\x16\xae\xc1': "Haunted House (Mystery Mansion, Graves' Manor, Nightmare Manor) (1982) (Atari, James Andreasen - Sears) (CX2654 - 49-75141) ~",
  b'x\xf0\x89J\xb98o\xf1\xdf\xb6\xf3\xe96\xe2G<_\xa1\xa7 ': "Haunted House (Mystery Mansion, Graves' Manor, Nightmare Manor) (1982) (Atari, James Andreasen) (CX2654) (PAL)",
  b' x&e{\x94\x8a\xb1^\xa3yY,(\xd7J\xa6\xef\x85R': 'Hey! Stop! (AKA Keystone Kapers) (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'M*\xb4\x1bx\x18\xbfqV\xcf\xda\x93\x98oT\xd0\xbcP\\\xbf': 'Hili Ball (AKA Racquetball) (1983) (Quelle) (689.302 8) (PAL)',
  b'\xf2u\xc7\xca\x9av\xfbu\x8fRT\x8b\x85\x97\xbc\x13$Rc\xb6': 'Hole Hunter (AKA Topy) (Video Game Cartridge - Ariola) (TP-606)',
  b'\x81\x96 \x9e\xf7\x04\x8cT\x94\xdb\xdc\x93*\xdb\xf1\xc7\xab\xf7\x9fN': "Holey Moley (Kid's Controller) (02-29-1984) (Atari, Robert C. Polaro) (CX26130) (Prototype) ~",
  b'{\xda\xb24v-I\x8b|e\x17\xee\xd8\xa2\xf2\x92\xf5\xfd\xd1)': 'Home Run (Unknown) (PAL)',
  b'\xf3b\xd2\xb3\xa5\x0eZ\xe3\xc2\xb4\x12\xb6\xc0\x8e\xcd\xcf\xeeG\xa6\x88': 'Home Run - Baseball (1978) (Atari, Bob Whitehead - Sears) (CX2623 - 6-99819, 49-75108, 49-75125) ~',
  b'N\xf3(d\xa5\x9c\xefn\x1d\xaa\xfb\x80\x9aD\xa5\x88\xfb\xeb5L': 'Homerun - Horrorrun (32 in 1) (1988) (Atari, Bob Whitehead) (CX26163P) (PAL)',
  b'\x1f6\x9cvv\xe3:\x11Yo\xac\x07b\t\x80\xb9tq"\xd2': 'Hot Action Pak - Ghostbusters, Tennis, Plaque Attack (1990) (HES - Activision) (542) (PAL)',
  b'\xf60\xb9TB\x05K\xdaRY\xaf\\h(z\xf5t\xdd\x18\\': 'Hot Wave (AKA Ram It) (Double-Game Package) (1983) (Otto Versand) (746422) (PAL)',
  b'Y\xe8\xef\xa3\xea\xad\x08)\x91[\xcf\xfe\x9b*\xe2H~f6\xbe': 'Human Cannonball (128-in-1 Junior Console) (PAL)',
  b'\xad\xc8\xc6\xec,[\r\x0fI]\x08\x8b^\x01]J5\x1e\x1c$': 'Human Cannonball (1979) (Atari) (CX2627, CX2627P) (PAL)',
  b'\xc5\xdf\x16\xd4\xea\x93_\xc8\x84`\xc4 (\xcf\xa1\xeeo\xd5A\x82': 'Human Cannonball (32 in 1) (1988) (Atari) (CX26163P) (PAL)',
  b'\xd4\xb0\xb2\xaa7\x98\x935lrAN\xe0\x06Z:\x91\xcf\x9f\x97': 'Human Cannonball - Cannon Man (1979) (Atari - Sears) (CX2627 - 6-99841) ~',
  b'VK\xa5\xe3\xa7\x0b\xd7m\xd7\x90/\xd9\xeeC\xb6t)\xde\xa2\x9c': 'Hunt & Score (Keyboard Controller) (1978) (Atari, Alan Miller) (CX2642) (PAL)',
  b"\xa6\xe4,c\x13\x8a/\xd5'\xcd\xbe\x9b~`\xf5\xfe\xab\xddU\xc8": 'Hunt & Score - Memory Match (Keyboard Controller) (1978) (Atari, Alan Miller - Sears) (CX2642 - 6-99814) ~',
  b"\xf7\xe7\x82!K_\x92'\xe3L\x00\xf5\x90\xbePSO\x1f\xda\x91": 'I Want My Mommy (KidStuff) (1983) (ZiMAG - Emag - Vidco) (710-111 - GN-010) ~',
  b'\xfa\x05\xa3]\xe4\xc4\x01pE8\xcc%oFD\x0fB\xde=\xdb': 'I.Q. Memory Teaser (AKA IQ 180) (Suntek) (SS-033) (PAL)',
  b'\xf1W\xdez\xda\xc8\xd1B1\xd3\xf8\xe6\xf3]\xf8\xc3\x9b\xb1\xb7_': 'IQ 180 (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\x84\x83\xfd\x18\xdd\xda\xfb\xdf\xbc\xbd\x1cC\xbc}\xe2\xcc\x05\xa1\x89M': 'IQ 180 (Unknown)',
  b'L\x02\xf0\xd3\xa80\xd9tW3\xd3\x84\x00)\x15\xa9Ry\xc9\xd5': 'Ice Hockey (CCE)',
  b'^\xbc\xc5\x93\x8b\xb8\xb2;\xb48\xef\xa2\x1d8\x8a\x85\xfc\x9b\x7f\xe9': 'Ice Hockey (Canal 3 - Intellivision)',
  b'[\xd3K\xa7\xe2`\xbdi\xf1un\xf7\xcb\xd9[\x91\xbe\x8d\x91\x1b': 'Ice Hockey - Eishockey (1981) (Activision, Alan Miller - Ariola) (EAG-012-04I, EAX-012, EAX-012-04B - 711 012-720) (PAL)',
  b'\xb3c_\x88\xc7\x1e!\xe0*\x80e_*v\xa2xe+\xb1\xd3': 'Ice Hockey - Hockey, Hockey (4 Game in One) (1983) (Bit Corporation) (PGP210) (PAL)',
  b'!\xde\x0f\x03N]\xad\x03\xfa\x91\xebz\xe6\xcc\x08\x1c\x14+\xe3\\': 'Ice Hockey - Le Hockey Sur Glace (1981) (Activision, Alan Miller) (AX-012, CAX-012, AX-012-04) ~',
  b'\\\xed\x92\xd1?\x8b\xec\xd9m\x17[8\xd1y\xe3S\xf8\xc36`': 'Ikari Warriors (1991) (Atari - Imagineering, Dan Kitchen) (CX26177) (PAL)',
  b'\xd8\xf7\xb9\x08\xf6\x0f\xe4\x96g\xc7\xc5]H\xce\x15\xa0Z\xd9Z(': 'Ikari Warriors (1991) (Atari - Imagineering, Dan Kitchen) (CX26177) ~',
  b'yF%GXc\xc5"YG\x8e\xbd\x92\x8e\xad\x82\x9c\x1a\xae\xe3': 'Im Schutz der Drachen (AKA Dragon Defender) (1983) (Quelle) (719.252 9) (PAL)',
  b'Z\x9b\x9c\xc7?\xd3g\xd0)\x86\xe8i/B\xca\xaa`-Gb': 'Imagic Selector ROM (1982) (Imagic) (PAL)',
  b'\x0f\x184~\xde\x938\xf4\xaa\x85E\x06g\xd1d\xa7\xde\xa4\x11\x01': 'Imagic Selector ROM (1982) (Imagic) [a]',
  b'\xce\xb3\xc7"#F3\xb3\x9aW\t>b\xcd\x11\xaeGc\x9f\xa1': 'Imagic Selector ROM (1982) (Imagic) ~',
  b'^\t\xa8\xac0?p\x1c\xca\xc4\x83u]\x1a)\xdf\xcd!h\x1f': 'Immies & Aggies (1983) (CCE) (C-838)',
  b'\xc4:\xb1\x96\xec\xbb2\xb7\x1d+\xd0\xac\\\x86\xed\x86b\x86A]': 'Immies & Aggies (Unknown)',
  b'`;N\xccyyr\xd3//;\x8ef\xeb\xdd\xe5\xaa\xf5\x1a\x07': 'Immies & Aggies - Immies and Aggies (1983) (ZiMAG - Emag - Vidco) (715-111 - GN-060) (Prototype) ~',
  b'\x03\xa9\xf78\x14\xe0f\xbe^\xdbT\x1d\xdb\x7f\x17\xa9\xd1{V\xf2': 'Inca Gold (Unknown) (PAL)',
  b'\x1f\xe7\x9a\xa4\x82\x1d~\x17.\xc7\xd6\x9e\xa3K\xd1$\x03D"\x97': 'Inca Gold (Zellers)',
  b'\xf6\xca<\xa4\xafm\x9a\xc5\xa59\xf1\xbcLz:\x04\x83|.D': 'Indy 500 (Race Car) (Driving Controller) (1977) (Atari, Carla Meninsky, Ed Riddle) (CX2611, CX2611P) (PAL)',
  b'b\n\xb8\x8dc\xcd\xd3\xf8\xceg\xde\xac\x00\xa2"W\xc7 \\\x8b': 'Indy 500 - Race (Race Car) (Driving Controller) (1977) (Atari, Carla Meninsky, Ed Riddle - Sears) (CX2611 - 99821, 49-75149) ~',
  b'\x93>\xa6\xfaa/\xbb\xb3\xd5$\x96A\xdf\x16\xb8\x00\xd3w\x7fq': 'Infernal Tower (AKA Towering Inferno) (1983) (Carrere Video - Western Technologies, Jeff Corsiglia, Paul Allen Newell, Tom Sloper - Teldec - Prism) (USC1009) (PAL)',
  b'\x92,\xd1q\xef\x13+\xf6\xc5\xbe\xd0\n\xd0\x14\x10\xad\xa4\xb2\x07)': 'Infiltrate (1982) (Apollo - Games by Apollo) (AP-2006) ~',
  b'\x8aH\xee\xeb\xab\xd8\xd7\x86\xaex\xcd\xe6\xb2\xcd}\xc9\xcf\xf8\x95\x19': "Infiltrate - Nid d'espions (1982) (Apollo - Games by Apollo - RCA Video Jeux) (AP-2006) (PAL)",
  b"\x94<{\xb9&\xc7\x1e\x94\xf3\xa51\x85\xac;\xec\x1c'\xff\xbe\xe0": 'Innerspace (1983) (VentureVision, Dan Oliver) (Prototype)',
  b'\xfa[\xf1\r\x90X\xee\xb2\xe3*\xb6Z%\xa9\x06\xc1{DY\x03': 'International Soccer (1982) (M Network - APh Technological Consulting, Kevin Miller) (MT5687) ~',
  b'\x05\x9c\x0bZ5h\xf1:\xcec\xe5\xc6\x12\x08\xbe\xea\x9e\x92O\xfd': 'International Soccer (1988) (Telegames) (5687 A279) (PAL)',
  b'0F\xf6\xe7\xad\x1d[\xb9\xffqx\x92\x9c\xb3\x12\x7f\x94\xe0\x14?': 'International Soccer (Unknown) (PAL)',
  b'\xa1\x158\x15u)\xb4*(@\xf5\x18\xb9Z\xf5\xc5\x91C\xcc\xed': 'Ixion (1984) (Sega, Jeff Lorenz) (Prototype) ~',
  b'^H\xb5\xe9\x8b\xe4\xe8\xd2\xd8`+\x009\xc3\xea\x95\x9dj\x071': 'Jacky Jump (AKA Bobby Is Going Home) (1983) (Home Vision - Gem International Corporation - VDI) (PAL)',
  b'+\xbc\x12L\xea\xd9\xaaI\xb3d&\x875\xda\xd8\xcb\x1e\xb6YO': 'James Bond 007 (James Bond Agent 007) (1984) (Parker Brothers - On-Time Software, Joe Gaucher, Louis Marbel) (PB5110) ~',
  b'\xafMhg\xa8\xbcH\x18\xfck\xb7\x01\xa7e\xa3\xc9\x07\xfe\xb6(': 'Jaw Breaker (1983) (CCE) (C-824)',
  b'\x98\xac\x19\xf7\xb4Eh{r\xafL\x1d\xab\xbc\xe8\xe2i\xf4\x1e\x91': 'Jawbreaker (1982) (Tigervision, John Harris - Teldec) (7-002) (PAL)',
  b'\x8b\xd8b\x03F\xdd\x08\xb4\xbe\x17\xda\xd7Z\xe2J\xab\xef\xb4\xf6\xe8': 'Jawbreaker (Unknown) (PAL)',
  b'\xa6\xe5\xc5\x19\xca\x8a\xe3l\xa2Z f4u\xab\xe7Ao\xa2N': 'Jawbreaker (Unknown) (PAL) [a]',
  b'\x92\x8e\xaaBK6\xd9\x80x\xf9%\x1dg\xfb\x13\xa8\xfd\xdf\xaf\xbd': 'Journey Escape (1983) (Data Age, J. Ray Dettling) (112-006) ~',
  b'#g\x83\n\x136\x1d\xc0p\xac\x96\x10\t\xe4E\ra\x91\xdc\xcd': "Journey Escape - Rock 'n' Roll Escape (1983) (Gameworld, J. Ray Dettling) (133-006) (PAL)",
  b'\x07\x94\x91]\xa9\xbd)\xe7w\x174vJ\xc8g3\x0e\xcf.7': 'Joust (07-05-1983) (Atari - GCC, Mike Feinstein, Kevin Osborn) (CX2691) (Prototype)',
  b'Q\x88\x1d"l3\xca\xd7\xd4)\xd3\xa5k~\xe6D\xad\x05>W': 'Joust (08-09-1983) (Atari - GCC, Mike Feinstein, Kevin Osborn) (CX2691) (Prototype)',
  b'\xcb\x94\xdc1l\xba(*\x006\x87\x1d\xb2ArW\xe9`xk': 'Joust (1983) (Atari - GCC, Mike Feinstein, Kevin Osborn) (CX2691) ~',
  b'\x03?\xa9\xec\xad\x9e\xb4\xd5\xcf\xeb\xf6P\\\xb7j!&\xdeeg': 'Joust (1983) (Atari - GCC, Mike Feinstein, Kevin Osborn) (CX2691, CX2691P) (PAL)',
  b'\xcd,\xf2E\xd6\xe9$\xff!\x00\xcc\x93\xd2\x02#\xc4\xa21\xe1`': 'Jr. Pac-Man (1986) (Atari - GCC, Ava-Robin Cohen) (CX26123) ~',
  b'\xa3C\xd6U?\xa06\x7fIu/\xf88\x91\x16\xce\xe3`TQ': 'Jr. Pac-Man (1986) (Atari - GCC, Ava-Robin Cohen) (CX26123, CX26123P) (PAL)',
  b'p\xf6\xa5/MK)\xa5\x0b\x7f\x99\x07O\x8f\x81`.\xbe0\xcd': 'Jumping Jack (AKA Bobby Is Going Home) (1983) (Dynamics) (DY-293005) (PAL)',
  b'\x9a\x0e\xe8E\xd9\x92\x8dM\xb0\x03\xb0{\x92{\xb2\xc1\xf6(\xe7%': 'Jungle Fever (1982) (PlayAround - J.H.M.) (203)',
  b'0K"\x91\x9c\xe4\xa2x\xfaU`\xa7\t\x01\xf0\xaf\x08\x11N\x8d': 'Jungle Hunt (02-03-1983) (Atari - GCC, John Allred, Mike Feinstein) (CX2688) (Prototype)',
  b'\x1a\x91\x1b\x08\xc4\x87!\xe9(\x0e\r\xf0}}\xf4\xf7\xbf\xec\xb7$': 'Jungle Hunt (02-25-1983) (Atari - GCC, John Allred, Mike Feinstein) (CX2688) (Prototype)',
  b'\x83\xa3*-hcUC\x8c\x91U@\xcf\xe0\xbb\x13\xb7l\x11\x13': 'Jungle Hunt (1983) (Atari - GCC, John Allred, Mike Feinstein) (CX2688) ~',
  b'\xb3\xab[\x0c\x1cI)\xdc\xcaw\x7f\xe2D\x1f\x0c\x8a\xba\x03\x94O': 'Jungle Hunt (1983) (Atari - GCC, John Allred, Mike Feinstein) (CX2688, CX2688P) (PAL)',
  b'\xfeR\xaf\x1bT\xe1F\xbe\x18~B_\xef"\x97Q!.\xee\xe1': 'Jungle Hunt (1983) (CCE) (C-1007)',
  b"\x89\x03\xaa\xdaV;OW5'@\x92\xd4\xe1\xb2F\xf3\xd0\xf7m": 'Jungle Hunt (1983) (CCE) (C-1007) [a]',
  b'\xce\x8a\xc8\x8by\x9c(%gI\\\xe5\t@*ZL,M\x82': 'Kabobber (07-25-1983) (Activision, Rex Bradford) (Prototype) ~',
  b'@\xd4\xdfO\x8eJi\xa2\x99\xaevx\xc1~r\xbe\xde\xb7\x01\x05': 'Kaboom! (Paddle) (1981) (Activision, Larry Kaplan, David Crane) (AG-010, AG-010-04) ~',
  b'\xaeZ\xf5\x9b5\x03\x85\x8b\xd8G\xc0\x7f\xea\x803\x13\x98\x90\xff4': 'Kaboom! (Paddle) (CCE)',
  b'\xa2WTbq\xb3=~;\x9e~\xa0\xecF\x06\xecG\xcf\xbc\x94': 'Kaboom! - Schneller als der Knall (Paddle) (1981) (Activision, Larry Kaplan, David Crane - Ariola) (EAG-010, PAG-010 - 711 010-720) (PAL)',
  b'\xa8*\xae\xefD\xad\x88\xde`\\P\xd2?\xb4\xf6\xce\xc7?:\xb4': 'Kamikaze Saucers (1983) (Syncro, Daniel Wolf) (Prototype) ~',
  b'\xdd4\x94\xa1\xe3\xff\xa0c\x80;\x1f\x02qn\x87\xe8\x9d\xc3\xc5t': 'Kampf dem Steinfresser - Boom Bang (AKA Crackpots) (1983) (Quelle - Otto Versand) (463.574 4 - 781393, 986153) (PAL)',
  b'\x01\xfd01\x1e\x02\x89D\xea\xfbm\x14\xbb\x00\x105\xf8\x16\xce\xd7': 'Kangaroo (1983) (Atari - GCC, Kevin Osborn) (CX2689) ~',
  b'\xda\xc5;\x95bkh\xd3hd0\xcc\xfd\xa8\xc1\xa1\tF\xba\xd2': 'Kangaroo (1983) (Atari - GCC, Kevin Osborn) (CX2689, CX2689P) (PAL)',
  b'\x16i\x02\xbf\x07\x01:\xd2\x98<\x11Tr&\xd3=rt\x026': 'Kanguru (AKA Kangaroo) (1983) (Zirok)',
  b'\xc7\x8f\xea\xecE`L\xda\x85\xebX\t\xe8\x81L>I\xe8\xc7<': 'Karate (1982) (Funvision - Fund. International Co.) (PAL)',
  b'\xc0\xdb})^,\xe5\xe0\x0e\x00\xb8\xa80u\xb1\x106\x88\xea\x15': 'Karate (1987) (Froggo) (FG1001)',
  b'\xc8\xcd\n\x04\xed\xeaD\x91Q\xe64\x9d\xb2E]H\xf2=\xc2\x13': 'Karate (Unknown) (PAL)',
  b'\xd5\xcem\xaer\x87\x00\xef$\xa7\xcd\xdc\x95\xe1\xe4-\x0e- \xf9': 'Katastrophen-Einsatz (AKA M.A.S.H.) (1983) (Quelle) (876.013 4) (PAL)',
  b'T\xcd8HpEc9\x97\x84Bn}\x00\xec\xcd\x7f\x96\x83k': 'Kaystone Kapers (AKA Keystone Kapers) (1983) (Digitel)',
  b"(6\xa6yT\xf3\xd9L'\xe0\x96B\xe1#\xa2\xb0\xc1\xd4\xab9": 'Keystone (AKA Keystone Kapers) (Tiger Vision - Eram)',
  b'>\xef\xc1\x93\xde\xc3\xb2B\xbc\xfdC\xf5\xa4\xd9\xf0#\xe5Sx\xa4': 'Keystone Kapers (1983) (Activision, Garry Kitchen) (AX-025, AX-025-04) ~',
  b'\xb3\xdfP\xf1\xdfx\n\x8f|\xbc\xdf\x15\xd8\x03\xd2\xb8[\xb5\x86\xff': 'Keystone Kapers (1983) (Activision, Garry Kitchen) (EAX-025) (SECAM)',
  b"~r/\x11\x84\xd5)\x1c\xa5;\n\xcaa\x99\x15\xd3\xec')&": 'Keystone Kapers (Canal 3 - Intellivision) (C 3014)',
  b'\xc0]"Wf\xdf\xde\x9a\xae\xab%\xc9u\xaa\xda\xf4\xe8\xb9\xc5V': 'Keystone Kapers (Digivision)',
  b't\xfc\xce\xdb5T\xa4\x11\xf1^\r\x7f\x17<\xc1\x8a\x15\x05\xf6\x0f': 'Keystone Kapers (Robby)',
  b'J\xe6\xb6O\x9a\x93\xcfN|\xb3\xdd\xd4\x95\xaa\x0c\xa0\x88\xbaei': 'Keystone Kapers (Shock Vision)',
  b'K9}qe\x82^\xbf\xe8\xf8\xddm\xdd\xbapQ\xb0q\x85\xeb': 'Keystone Kapers (Unknown)',
  b'\x14nR\x0f\xeb\xeb\x9aM]\xf3\xab\\7M\xa6\x06\x84\xeb5H': 'Keystone Kapers (Unknown) (PAL)',
  b"R^\x90T49\x16\xe2\xa7'\x12x\xcf6\x10\xe0\xed\x1e\xfbq": 'Keystone Kapers - Raueber und Gendarm (1983) (Activision, Garry Kitchen - Ariola) (EAX-025, EAX-025-04I - 711 025-725) (PAL)',
  b'\xde*Gq\x02\x87\xf5\x05+\xfe\x12;J5o[k\xc1"\xdc': 'Keystone Kappers (AKA Keystone Kapers) (1983) (CCE) (C-816)',
  b'\x13z\xd1\x03\xf7L\x0f\xa4\xf5\xec\r\xf3\xf5lrkc\xc13a': 'Keystone Kappers (AKA Keystone Kapers) (1983) (CCE) (C-816) [a]',
  b'@\x1b\x81\x1c\xc0\x1c3p\xf5\xe0\xd1\x8d\xc1j\x18\xb0\xbb\xf3\xbf\x87': 'Keystone Keypers (AKA Keystone Kapers) (1983) (Zirok)',
  b':j\x04[T&\xa6\t\x0f\xf2\x1d\xb1:\xff\x1b}\xdegY\x90': 'King Arthur (AKA Dragonfire) (Double-Game Package) (1983) (Otto Versand) (600273) (PAL)',
  b'\xafg\x9e\xb1\x9d~J\xc0\xeab\x032\xed\xbbp\x81VzU\xdb': 'King Kong (1982) (Tigervision - Software Electronics Corporation, Karl T. Olinger - Teldec) (7-001 - 3.60001 VE) (PAL)',
  b'\x83\x9e\x13\xff\xed\xcb\xed"\xd5\x1a$\xc0\x01\x90\x0c4t\xa0x\xf2': 'King Kong (1982) (Tigervision - Software Electronics Corporation, Karl T. Olinger - Teldec) (7-001 - 3.60001 VE) (PAL) [a]',
  b'Y\xf4\x85\xea4\\\xf1\xb9\xe3f;E\xb2F\xf196\xa3)r': 'King Kong (1982) (Tigervision - Software Electronics Corporation, Karl T. Olinger) (7-001) ~',
  b'\x99\x10\x89\xa674\xa8Y\xc8\x9e\xb4e\xf3\xfc\x07NW\xc3\xeeM': 'Klax (06-14-1990) (Atari - Axlon, Steve DeFrisco) (CX26192) (Prototype)',
  b'1b%\x9cm\xbf\xbbW\xa2\xeaA\xd8I\x15W\x02\x15\x1e\xe3\x9b': 'Klax (08-18-1990) (Atari - Axlon, Steve DeFrisco) (CX26192) (Prototype)',
  b'Eb:\x1c\x8f\xb5\x07M\xe9\x8c7\xf0\x05\xed\xd5\xb1\xd0\x93}\xae': 'Klax (1991) (Atari - Axlon, Steve DeFrisco) (CX26192) (PAL) ~',
  b'u\x95\x97\xd1\xd7y\xcf\xdf\xd7\xaa0\xfd(\xa5\x9a\xccX\xca%3': 'Knight on the Town (1982) (PlayAround - J.H.M.) (203)',
  b'\xd8\x03>\x1b\xa6\xef\xca\xc9\xd1\x98u{AB;ic\xb6\xff\xd7': 'Knight on the Town (1982) (PlayAround - J.H.M.) (203) (PAL)',
  b'/U\x07C\xe27\xf6\xdc\x8cu\xc3\x89\xa0\x1b\x02\xe9\xa3\x96\xfd\xad': 'Kool-Aid Man (Kool Aid Pitcher Man) (1983) (M Network, Stephen Tatsumi, Jane Terjung - Kool Aid) (MT4648) ~',
  b'\x82\xe6Cfy[\x01\x1c*/\x17U\xbf\x89\x9c\xc2\xc0a\x7f\xe8': 'Krieg der Sterne (AKA Atlantis) (Videospielkassette - Ariola) (PGP233) (PAL)',
  b'\xfbt\x02\xffUM2D\x1e\xa90f(\x8aFZ\x1c\xc6\x94\xb9': 'Krull (05-27-1983) (Atari, Jerome Domurat, Dave Staugas) (CX2682) (Prototype)',
  b'K\xdf\x1c\xf73\x16\xbd\xb0\x00&\x06\xfa\xcf\x11\xb6\xdd\xcb(r\x07': 'Krull (1983) (Atari, Jerome Domurat, Dave Staugas) (CX2682) ~',
  b'\x07\xa1\xc1\xb1\xa2)|N\xdd\xe0\xe1oa\x0b^\xc2<wR\x17': 'Krull (CCE)',
  b'\xb5\xfa\xd0\xb6\x08\t\\~\xa3\xc1\x1bw\xad\xe4`\xae\xc5\x06\x14\xa6': 'Kung Fu (AKA Karate) (4 Game in One) (1983) (Bit Corporation) (PGP229) (PAL)',
  b'\x13$\r\r\xec\xd9D\xf7\xd5\xd7\xa8z}\x1bGW\x8fH\x97\x11': 'Kung Fu Master (CCE)',
  b'\x167\xb6\xb9\xcd\x1a\x91\x839\xec\x05L\xf9[\x92N|\xe4x\x9a': 'Kung Fu Superkicks - Pursuit of the Ninja (AKA Chuck Norris Superkicks) (1988) (Telegames) (6082 A145)',
  b'\x06\x1eQ\x96\xaf;U\x97f\xd4\x15\x8a\x13\xfa\xef\x04\xf76\\\xd5': 'Kung Fu Superkicks - Pursuit of the Ninja (AKA Chuck Norris Superkicks) (1988) (Telegames) (6082 A145) (PAL)',
  b';\x93\xa3K\xa2\xa6\xb7\xdb8~\xa5\x88\xc4\x8d\x93\x9e\xee]q\xa1': 'Kung-Fu Master (1987) (Activision, Dan Kitchen) (AG-039-04) ~',
  b'\x81\xab\xede\x06\x00\xed\xe0\n\x00\xe24\x91\xb7\x9dF\xa9\xb4\xf1s': 'Kung-Fu Master (1987) (Activision, Dan Kitchen) (EAX-039-04B, EAX-039-04I) (PAL)',
  b'il\xc5\xc9M\xe1\x07\x94\x0e\x0b"\xfe\xb3\x08\x9a\xafb\xe9i\x96': 'Kwibble (Quick Step! Beta) (1983) (Imagic, Dave Johnson) (720119-1A, 03211) (Prototype)',
  b'\xb4\xb5^\x0bk]t\x90V[\x00\xc03a\x19\x9f\x05\xfbp\xb8': 'Kyphus (1982) (Apollo, Tim Martin) (AP-2007) (Prototype) ~',
  b'\xad\xbc\x88F\x01\x01n\xbf\x1b;\xca\x03\xdc\xe5m\x95\xb7\xa9\n"': 'Labyrinth (AKA Maze Craze) (1983) (Quelle) (805.784 6) (PAL)',
  b'mY\xdf\xea&\xb7\xa0eE\xa8\x17\xf0?b\xa5\x9b\xe8\x995\x87': 'Lady in Wading (1982) (PlayAround - J.H.M.) (204)',
  b'\xe8\xba\xae\x84\xff\x05\xb2\x80\xb7\x12>\x94\n\xc2)\x14.\x0b\xdb.': 'Landung in der Normandie (AKA Commando Raid) (1983) (Quelle) (876.482 1) (PAL)',
  b'\xf0\x8f\x89\xb5DU\xa6\x8a[\xc1a\xecx\x1a\\\xa0\xa8\xdd\xbd\xd3': 'Landungskommando (AKA Strategy X) (1983) (Quelle) (176.433 1) (PAL)',
  b'\xe1\xf3z\x85\xfe\x11\x8b\xce\xe6V\xa8.\x9c\xc0\xe3y\x8f?\x9a1': 'Laser Base (AKA The End of the World) (Perry Rhodan-Serie) (1983) (ITT Family Games) (554-33 367) (PAL) [a]',
  b'\xc7\xdfY\x04\xcd\xc6\xa1e\x8c\xad\x89\xb8>>R;B\xa2\xb4]': 'Laser Blast (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'|\x01\xdc\x0b\xc9\xf8|\x01\xdc\x93\xf6\x8b<\xc9\xbf\xab\xed\xa7\x7f\x83': 'Laser Blast (32 in 1) (1988) (Atari) (CX26163P) (PAL)',
  b'LhI{5\xcc\x02Ts\x8b\xf9\xbf\xfe\xe6\xee\xd6r\xc7\xce\xa3': 'Laser Blast (CCE)',
  b'\xccE\xaf\nE\xa6\xe5\xd0X\x16\x83\xc3X\x94v\xc0\xdaKV\x00': 'Laser Blast (Hack) (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\xabZ\x84\x1c\xaa\xf0\xd5\xabPn\xeb\xd2\xb1=b\xa5o\x81/\x98': 'Laser Blast (Lazer) (08-08-1980) (Activision, David Crane) (AG-008) (Prototype)',
  b'\xea\x8e\xcc/h\x18\xe1\xc9G\x9fU\xc0\xa35n\xdc\xf7\xa4\xd6W': 'Laser Blast (Lazer) (1981) (Activision, David Crane) (AG-008) ~',
  b'\xf40F\xe4A\x00\xf1#\xe0D2\xdf\xeda\xa4\x87\xc1\x0e6\xd9': 'Laser Blast - Duell im Weltall (Lazer) (1981) (Activision, David Crane - Ariola) (EAG-008, PAG-008, EAG-008-04I - 711 008-720) (PAL)',
  b'\xc7h\xf4B5vk\xde\xbd\xb7\xc2\xc6\xb6\x1c\xfa\x84\xb4\x0b\x04\xbb': 'Laser Gate (AKA Innerspace) (1983) (CCE) (C-860)',
  b'V\xd3\xd4\xabu\xa6F\xa4s\xda\x04:\xe7\xe3\xdd\x13\x91l\x8a\x91': 'Laser Gate (AKA Innerspace) (Future Video Games) (PAL)',
  b'\xcd\xf5[s\xb42$(\xa0\x01\xe5E\x01\x9e\xaaY\x1d4y\xcf': 'Laser Gates (AKA Innerspace) (1983) (Imagic, Dan Oliver) (720118-1A, 03208) ~',
  b'u\x9e\xcc\xf0\xcc\x814\xa2\xffx{\xa3\x90\x87\xda\x04]\xdf\xac\xe1': 'Laser Gates (AKA Innerspace) (1983) (Imagic, Dan Oliver) (720118-2A, 13208, EIX-007-04I) (PAL)',
  b'\xaf\xabyW\x198jwk_\xb2\x16_\xc8OHX\xe1n\x05': 'Laser Volley (AKA Innerspace) (Zellers)',
  b'\xb6\xd4\x0f\x1b\xdb\x1e&`%M\x88\xff0\x89\xf2\xaf\xc5x\xcf\xfe': 'Lasercade (1983) (20th Century Fox Video Games - Videa, Lee Actor) (Prototype) ~',
  b'\x1a\xe0^\xe1\x95M\xb2\xa1\xa5\x86\xa5\x92+!\x06"\xd8\x06=\x12': "Last Starfighter, The (Universe) (Solaris Beta) (1984) (Atari, Douglas 'Solaris' Neubauer, Mimi Nyden) (CX26134) (Prototype)",
  b'\x8f\xf6\xb50\xc0 \x8d\xef\xac\x18\xc9jQ\xde\x95\xfd\xa7"\x8b\xe8': 'Lilly Adventure (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\xfe \x8a\xd7u\xcb\xf9R>z\x99c+\x9f\x10\xf2\xc9\xc7\xaa\x87': 'Lochjaw (1982) (Apollo - Games by Apollo, Steve Stringfellow) (AP-2005) ~',
  b'\xfc=u\xd4m\x91tW\xaa\x17\x01\xbfG\x84H\x17\xd0\xba\x96\xc3': "Lock 'n' Chase (1982) (M Network - APh Technological Consulting, Bruce Pedersen - INTV) (MT5663) ~",
  b'\x07^\xc3g\x8c\xea\xbb\xeeF\xa7\xd5v\xfd\xd1\x99\xa4\xab\x90\x8c\xc0': "Lock 'n' Chase (1988) (Telegames) (PAL)",
  b'\x11\x91q\x93^\xd9\xea\x87}\xe4\xf2\xa6\x8a\xee1\xcdU\xbci}': "Lock 'n' Chase (Unknown) (PAL)",
  b'\xf9+\x0b\x83\xdb<\xd8@\xd1n\xe2r`\x11\xf5\xf0\x14A\x03\xd5': "London Blitz (1983) (Avalon Hill, Jean Baer, Bill 'Rebecca Ann' Heineman, William O. Sheppard) (5002002) ~",
  b'^\x04\xfa\x03 \x16t4\xda\xb92\xf6\xb71\x83\xda\xf1\xa5\x0e\xc7': 'Looping (1983) (Coleco - Individeo, Ed Temple) (2654) (Prototype) ~',
  b'\x90\x10S\t\x04\xf7\x8a\xd1\nad\x17L\xb6W\xd1\xf4qs7': 'Lord of the Rings, The - Journey to Rivendell (The Lord of the Rings I) (1983) (Parker Brothers, Mark Lesser) (PB5950) (Prototype) [a]',
  b'\xef\x02\xfd\xb9J\xc0\x92${\xfc\xd5\xf5V\xe0\x1ah\xc0jH2': 'Lord of the Rings, The - Journey to Rivendell (The Lord of the Rings I) (1983) (Parker Brothers, Mark Lesser) (PB5950) (Prototype) ~',
  b'2\x00a\r\xab\xfck\xbe9)\xe1y\xd5+\x0e\x93y\xb0Du': 'Los Angeles 1984 Games (AKA Track and Field) (Track & Field Controller) (1984) (Atari - GCC, Jaques Hugon, Seth Lipkin) (CX26125) (Prototype) (PAL)',
  b'\x7f\x03\x16/\xa0\xa3\xb0\x81E\xe1}\x1d\xe0\x12\x081N\xbfD\x1f': 'Lost Luggage (Airport Mayhem) (1982) (Apollo - Games by Apollo, Larry Minor, Ernie Runyon, Ed Salvo) (AP-2004) [no opening scene] ~',
  b"\xe8I/\xe9\xd6'P\xdfh#X\xfeY\xa4\xd4'&U\xeb\x96": 'Lost Luggage (Airport Mayhem) (1982) (Apollo - Games by Apollo, Larry Minor, Ernie Runyon, Ed Salvo) (AP-2004) ~',
  b'\xe80$\xe0\x07\xe1\x03\xb8O\x83\xd3\x123b}\xaft\x1d\xf6\t': 'Lost Luggage - La valise piégée (Airport Mayhem) (1982) (Apollo - Games by Apollo, Larry Minor, Ernie Runyon, Ed Salvo - RCA Video Jeux) (AP-2004) (PAL)',
  b'\x86\xdfW\x8e\x914\x85\x08\x9d\x9b;>s\x84\x07kS\xa9\x84P': 'M.A.D. (1983) (Carrere Video - Western Technologies - Teldec - Prism) (USC1012) (PAL)',
  b'\xd6\xe2\xb7vZ\x9d0\xf9\x1c\x9b+\x8d\n\xdfa\xec]\xc2\xb3\n': 'M.A.D. (Missile Intercept) (1983) (U.S. Games Corporation - Western Technologies) (VC1012) ~',
  b'\xdc\xd9i\x13\xa1\xc8@\xc8\xb5xH\x98bB\xee\xb9(\xbf\xd2\xff': "M.A.S.H (1983) (20th Century Fox Video Games, Frank Cohen, Douglas 'Dallas North' Neubauer) (11011) ~",
  b'Xs\xae\x07h\xdf\xdc\xc8\\\x0b\x03\x19\x92\x92\x1a\xe4\xba\xdc:\xc8': "M.A.S.H (1983) (20th Century Fox Video Games, Frank Cohen, Douglas 'Dallas North' Neubauer) (11111) (PAL)",
  b'\xa2\xef\x131w\x12\xf5f\x855[d\x8c\xa95\x8e\xfa\xda\xcb<': 'M.A.S.H (208 in 1) (Unknown) (PAL)',
  b'\x98\x94\xf8\xcb\xabj\xcf\xdc\xcf"12\x9f\x9f\xbe*\x94\x8f\x92\xc5': 'M.A.S.H (Hack) (Unknown) (PAL)',
  b'\xfb\xc0uvc\x13\xf0J\x80\x826\x18\x9d\xaa+\xf4\x9fXOD': 'MASH (AKA M.A.S.H) (1983) (CCE) (C-859)',
  b'U\xae1\x9f\x11XO\x99{{R\x00\n\x1d\x98\xcd\xfa\xa0}]': 'MASH (AKA M.A.S.H) (1983) (CCE) (C-859) [a]',
  b'\x9e\x8bAz\t\x0f\x84\x18\xf6=\xec\xaa\xcd\xd4\xd8\x18\xa2;K\xea': 'Mafia (AKA Gangster Alley) (Rainbow Vision - Suntek) (SS-010) (PAL)',
  b'Lf\xb8J\xb0\xd2^Fr\x9b\xbc\xf2?\x98]Y\xca\x85 \xad': 'MagiCard (Keyboard Controller) (1981) (Computer Magic - CommaVid, John Bronstein) (CM-001) ~',
  b'\xcd\xc7\xe6]\x96Zz\x00\xad\xda\x1e\x8b\xed\xfb\xe6 \x0e4\x94\x97': 'Malagai (1983) (Answer Software Corporation - TY Associates) (ASC1001) ~',
  b'1\x1b\xf4:G\x9e\x99\x81\xc7\x0f8W=\xbfaD)\xd0\xa5\xb5': "Mangia' (1983) (Spectravideo) (SA-212) (PAL)",
  b'\xee\x8f\x9b\xf7\xcd\xb5_%\xf4\xd9\x9e\x1a#\xf4\xc9\x01\x06\xfa\xdc9': "Mangia' (1983) (Spectravideo) (SA-212) ~",
  b'\xdf\x02\xbbT\x18r/\xa5\xb8\xc1\xf1\x12\xf7\x17e*\x7f\x08}\x8c': 'Marauder (1982) (Tigervision, Rorke Weigandt - Teldec) (7-005) (PAL)',
  b'$\x9a\x11\xbbHr\xa2O"\xdf\xf1\x02\x7f\xf2V\xc1@\x81@\xc2': 'Marauder (1982) (Tigervision, Rorke Weigandt) (7-005) ~',
  b'\x01\n\xdd|6\x00R\x19\x8aO\x81\xb2\xbe\x965\x0fAx\xda.': 'Marauder (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\xfaI\xff(*\xb7\xc3\xc7\xb1%\xd2\xd2\xea}\x1d\xa1\x13N\x90\xd9': 'Mariana (AKA Seaquest) (Rainbow Vision - Suntek) (SS-017) (PAL)',
  b"\xads?{\xb6,\xe9'_4\x8aZ\xea\x07m\x9d\xf6\x10p\x85": 'Marine Wars (1983) (Gakken) (011) (PAL)',
  b'\xdd\x9e\x94\xca\x96\xc7Z!/\x14\x14\xaaQ\x1f\xd9\x9e\xcd\xad\xafD': 'Marine Wars (1983) (Konami) (RC 102-X 02) ~',
  b'\x91\x9f\xb7\xd1S\x1a\xa6\xb1\x81\xe9o\x86H\x16\x84\x91\xbez/ ': 'Marineflieger - River Raid II (AKA Seahawk) (1983) (Quelle - Otto Versand) (176.764 9 - 781644) (PAL)',
  b'IB_\xf1T\xb9,\xa0H\xab\xb4\xce^\x8dH\\$\x93P5': 'Mario Bros. (1983) (Atari, Dan Hitchens, Mimi Nyden) (CX2697) ~',
  b'\xa8?_\x9a\x07\\\xcf\xee\x1e{t\xc8\x08.\xc2\xff=~\x1fx': 'Mario Bros. (1983) (Atari, Dan Hitchens, Mimi Nyden) (CX2697, CX2697P) (PAL)',
  b'b\x82\xda\x81G,,\x95\xf8\xef\xc1\xe2\x0e\x0e\xeb\xad\xe7\xbe\x8c\xc1': 'Mario Bros. (Zirok)',
  b'j\xe4\x1e\x84SM\xbe\x0c\x8c\xff\xdb\xf0\x13\xd16uO#\x9e\x9f': "Mario's Bros. (AKA Mario Bros.) (1983) (CCE) (C-1004)",
  b"6\x96WY\x88\xbeh\x14\xb6\xc0\xee\xba\xb2(\\G\x1f\x94\xa1'": 'Marspatrouille (AKA Gas Hog) (1983) (Quelle) (292.542 8) (PAL)',
  b'\xfb\xe7\xa7\x87d@wC\xb4:\x91\x13i\x03\xed\xe6S\x06\xf4\xe7': 'Master Builder (1983) (Spectravideo) (SA-210) ~',
  b'\xa5?M`\x1c\x17\xaa\x14o\xdc\xbd\xd5\xd6!\x07\xec\xa9\x82\x12X': 'Master Builder - Super Baumeister (1983) (Spectravideo - Quelle) (SA-210 - 413.582 8) (PAL)',
  b'm\xb8\xfaeu]\xb8d8\xad\xa3\xd9\x0fL9\xcc(\x8d\xcf\x84': 'Masters of the Universe - The Power of He-Man (1983) (M Network, Connie Goldman, Joe King, Patricia Lewis Du Long, Gerald Moore, Mike Sanders, Jossef Wagner - INTV) (MT4319) ~',
  b'12Y\x03T\xba\x0e\xc6\x92.\xf0d\xbb\xec\xd4\x1a3#\xabX': 'Math Gran Prix (208 in 1) (Unknown) (PAL)',
  b'\x13\xc4\xfdK\xd1\xb6 \\N\xe8h\xd2\x8e\xe4\x16\x93\xa1h?\t': 'Math Gran Prix (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\x18\xfa\xc6\x06@\x0c\x08\xa0F\x9a\xeb\xd9\xb0q\xae:\xec*<\xf2': 'Math Gran Prix (Math Game) (1982) (Atari, Suki Lee - Sears) (CX2658 - 49-75128) ~',
  b':\xc9%)\xe4\xc1F\xeb\x156T\x86w\x80R\xe6\x89+\xab2': 'Math Gran Prix (Math Game) (1982) (Atari, Suki Lee) (CX2658) (PAL)',
  b'\xb7\xb9\xe6QY\xaf#0}\xa5L\xf3q\xaa\xc8Sl]"|': 'Maze Craze (Unknown)',
  b"J%\\\x1d'\x7f\x9c\xeb\x98\xdf\xe6\x1e\xc2\xc6\xf7S|\xa4\xe7\xbf": "Maze Craze - A Game of Cops 'n Robbers (The Maze) (1980) (Atari, Richard Maurer) (CX2635, CX2635P) (PAL)",
  b'\xab\xa2P\x89\xd8|\xd6\xfe\xe8\xd2\x06\xb8\x80\xba\xa5\xd98\xaa\xe2U': "Maze Craze - A Game of Cops 'n Robbers - Maze Mania - A Game of Cops 'n Robbers (The Maze) (1980) (Atari, Richard Maurer - Sears) (CX2635 - 49-75157) ~",
  b'\x01\x03\xb3[\x1a\xefk\x10\xc1\xc0\xa4K!>\xbf0\xafp\x8d\xf6': "McDonald's - Golden Arches Adventure (Big Mac) (06-06-1983) (Parker Brothers, Dave Engman, Isabel Garret) (Prototype) ~",
  b'\n\xe1\x187<{\xda\x97\xda/\x8d\x9c\x11>\x1e\t\xea~I\xe1': "Mega Force (1982) (20th Century Fox Video Games, Douglas 'Dallas North' Neubauer) (11005) ~",
  b'\xda`\xccE\x97h\x1b\xf9U\xee\xd5\xd0?s#\xad?\xa5\xc5%': "Mega Force (1982) (20th Century Fox Video Games, Douglas 'Dallas North' Neubauer) (11105) (PAL)",
  b'T-E\xc4\xb4\x87>X\xb4\xdac\xfa/\xa9\xdf\xc3\x89\xaaK\xf6': "Mega Force (1982) (20th Century Fox Video Games, Douglas 'Dallas North' Neubauer) (11105) (PAL) [a]",
  b'\x86O\\\x82\xf0U\xc3\xd2\x1c?\x10\xb6\xee\x89$\x93\xb3k\x83Z': 'Mega Funpak - Gorf, P. Patrol, Pacman, Skeet Shoot (HES) (PAL)',
  b'F\x97{\xaf\x0e\x1e\xe6\x12KRBX\x87\x9cF\xf8\rbO\xae': 'MegaBoy (Dynacom) ~',
  b'\xe21<2\x8e\xf8[EVf\xe0a\xbdGJ xR\xe7\x86': 'MegaMania (208 in 1) (Unknown) (PAL)',
  b'+j\x82`d)R P.\x1c\xee\x1d\xf8\xc3+\xf1\xcdq\x89': 'MegaMania (Unknown) (PAL)',
  b'\x9cWH\xb3\x86a\xdb\xad\xcb\xc9\xcd\x1e\xc6\xa6\xb0\xc5P\xb0\xe3\xda': 'MegaMania - A Space Nightmare (1982) (Activision, Steve Cartwright) (AX-017, AX-017-04) ~',
  b'uc\x96\xb5\xd8\x981\x90\x9d=U\xcb\xa1C\xf5\xd1\xa1\x01\xe3f': 'MegaMania - A Space Nightmare - Ein Alptraum im Weltall (1982) (Activision, Steve Cartwright - Ariola) (EAX-017, EAX-017-04I - 711 017-720) (PAL)',
  b'\xd0\xbfv\x8b^\x944\xd9\xae\xfa\xddC\xdbB\xb7Z(\xc36?': 'Megamania (1982) (Dynacom)',
  b'\x98~$^\x83\xec\xd4U\x19gP\xf5Xs\xd8\x0eu\xfe9A': 'Megamania (1983) (CCE) (C-829)',
  b'\x02\x9f\xac5L\xeb-:\x8d\x8c\xc4\xd4\x00\xb3w\xc0\xcb\x0e\xdd\xea': 'Megamania (1983) (Digitel)',
  b'\x9f\xadb\xbd\x9aG\xb4\xb4Z\xf8\x14\xd8\x0ew\\\xe7O7\x80\x0c': 'Megamania (Dactari - Milmar)',
  b'h?\xa0\x95x\xbc\xdd\x12\x95\x00\xc5\x1f\xdc]\xd0Y\xff/|h': 'Mein Weg - My Way (AKA Challenge) (1983) (Quelle - Otto Versand) (686.561 2 - 781627) (PAL)',
  b'\xde\xbb\x15r\xea\xdb \xbe\xb0\xe4\xcd-\xf89m\xef\x8e\xb0 \x98': 'Meltdown (Atom Smasher) (1983) (20th Century Fox Video Games - Videa, David Ross) (11029) (Prototype) ~',
  b'\xdf\xdc\xbb\xa1f9o\xae\xab\x1f\xfc@\xf1g\x18\x94\x98\xdf\xc1\x13': 'Midnight Magic (Pinball Wizard) (1986) (Atari, Glenn Axworthy) (CX26129) (PAL)',
  b'\x7f\xcf\x95E\x9e\xa5\x97\xa32\xbf[oV\xc8\xf8\x910{E\xb4': 'Midnight Magic (Pinball Wizard) (1986) (Atari, Glenn Axworthy) (CX26129) ~',
  b'\xa1\x9e\x8bS\xa0\x9e\xb0[\x07?\xa7Y\xff\x02X\xa6\r\xa7\x14M': 'Mighty Mouse (AKA Gopher) (Funvision - Fund. International Co.)',
  b'k\xe6\xc8\xc8\x19>C\xe9\xfd\x0f\xd3~SD\x84\x8dZB\xbbO': 'Millipede (01-04-1984) (Atari, Jerome Domurat, Andrew Fuchs, Dave Staugas, Robert Vieira) (CX26118) (Prototype)',
  b"\x06\t)\x98\xa3\x0e\x81j\x97\xfe\x9eM?'\x08[\x91>\xc9\xf9": 'Millipede (1984) (Atari - GCC) (CX26118) (Prototype) (PAL)',
  b'B\xe9}\x1b\xa1\x97\x81\x18~`T0\xfdp\x04\x9f\x04P\xefu': 'Millipede (1984) (Atari, Jerome Domurat, Andrew Fuchs, Dave Staugas, Robert Vieira) (CX26118) (Prototype)',
  b'\x06\x16\xf0\xdd\xe6\xd6\x97\x81m\xda\x92\xed\x9eZL=w\xa3\x94\x08': 'Millipede (1984) (Atari, Jerome Domurat, Andrew Fuchs, Dave Staugas, Robert Vieira) (CX26118) ~',
  b"{\xa2\xe2\xc0\xb3\xb6jPi/M\xec\xe8\x176\x85'\xe2\x91\x80": 'Millipede (1984) (Atari, Jerome Domurat, Andrew Fuchs, Dave Staugas, Robert Vieira) (CX26118, CX26118P) (PAL)',
  b'8D\xb7\x9d\xbe\xc5\xff\xc9\x9e\xaa,\x9f_\xa4\xf0\xa2l\x08\xc0m': 'Mind Maze (Mind Race) (Mindlink Controller) (10-10-1984) (Atari, Gary Shannon, Howard Scott Warshaw) (Prototype) ~',
  b'x`l/=\xd4\x18\xe9\x03\xa0v\x07F\xd1\x17\x14\xe7JU\xa6': 'Miner 2049er - Starring Bounty Bob (1982) (Tigervision) (7-008) (Prototype)',
  b'\xc4K/\x87iT@\x82\x8dhD\xaec\r?\x97[X#\x85': 'Miner 2049er - Starring Bounty Bob (1983) (Tigervision - Teldec) (7-008 - 3.60006 VG) (PAL)',
  b'Tp\xaf\xe5\x96\x1f\x9f\xdd/c\x88\xf5?h/2d\xa9\\\xe9': 'Miner 2049er - Starring Bounty Bob (1983) (Tigervision) (7-008) [fixed] ~',
  b'\x0eV\xb4\x8e\x88\xf6\x9d@^\xab\xf5D\xe5vc\xbd\x18\x0b;\x1e': 'Miner 2049er - Starring Bounty Bob (1983) (Tigervision) (7-008) ~',
  b'^\xdb\xf8\xa2O\xcb\xa9v9\x83\xbe\xfe \xe21\x1fa\xb9\x86\xd4': 'Miner 2049er Volume II (1983) (Tigervision - Teldec) (7-011 - 3.60015 VG) (PAL)',
  b'W_\xaa\xd9,\xb3\x89D\xb9\x88/\xfbi\x07>\n\xf9F\n\xba': 'Miner 2049er Volume II (1983) (Tigervision) (7-011) ~',
  b'4w9\x98\xd7t\x0e\x1e\x8c k;"\xa1\x9e(,\xa12\xe1': 'Mines of Minos (1982) (CommaVid, Irwin Gaines) (CM-005) ~',
  b'$\xb9F=>\x8d\xa1\x02\xa4\x15j\xc7\xe2\x8d\x9aZ\xfaE3\x95': 'Mines of Minos (Unknown) (PAL)',
  b'}q\xca>\x9d:\x81\xfbP\xd7o\x99\x83\xa1\xc0\xf6\x135\xe2\x11': 'Mines of Minos - Im Labyrinth des Roboters (1982) (CommaVid, Irwin Gaines - Ariola) (CM-005 - 712 005-720) (PAL)',
  b'\x9f\xee\xbc/A\xd0\x01>8ru\xbe-]R]\xd8\x1e\xd8\xdf': 'Miniature Golf (1979) (Atari, Tom Reuterdahl) (CX2626, CX2626P) (PAL)',
  b'\xbe$\xb4.7D\xa8\x1f\xb2\x17\xc8lN\xd5\xceQ\xbf\xf2\x8ee': 'Miniature Golf - Arcade Golf (1979) (Atari, Tom Reuterdahl - Sears) (CX2626 - 6-99829, 49-75116) ~',
  b"'s\x96\xe0\xd4\xfe\x82Q-\x99Lw\xe59\xbf\x061\x0b\xd8\xf1": 'Miniaturer Golf (AKA Miniature Golf) (32 in 1) (1988) (Atari, Tom Reuterdahl) (CX26163P) (PAL)',
  b"\x1e\x94C\xc7\xe6Q\x7fd\x9bG\xe0\x08\x1a\xaac^'\x1e\n\xef": 'Miss Pack Man (AKA Ms. Pac-Man) (Video Game Program) (PAL)',
  b'\xf7!\xd1\xf7P\xe1\x9b\x9e\x17\x88\xee\xd5\xe3\x87)#\xabF\xa9\x1d': "Miss Piggy's Wedding (06-24-1983) (Atari, Suki Lee) (CX26113) (Prototype) (8K) ~",
  b'`\xda\x8b\x01\x9f\x98ZZ\xead\xf5\x06\x91\xa5$$\xa6uU\x86': "Miss Piggy's Wedding (1983) (Atari, Suki Lee) (CX26113) (Prototype)",
  b'\xc6GP4\xad@\x18\x9f\xbc\x86\x8a\xab9\xdd\xc9Z\xe6.\xae\xaa': "Miss Piggy's Wedding (1983) (Atari, Suki Lee) (CX26113) (Prototype) [a]",
  b'@Zgc\x8f\x1f\xe8\x0b#4\xfe\xe1\xd3\xdf\xe8\x90e\x86\xdc\x1a': 'Missile Command (1981) (Atari, Rob Fulop - Sears) (CX2638 - 49-75166) (Prototype)',
  b'\x81\xff\xac\x1f\x10Q\xa4\xc9\x9e\x82\xe5:\xe6\x04\xf0\xb8\xec\x95\xad\x8e': 'Missile Command (1981) (Atari, Rob Fulop - Sears) (CX2638 - 49-75166) [no initials] ~',
  b"\xfa\xa0k\xb0d=\xbfUk\x13Y\x1c1\x91}'z\x83\x11\x0b": 'Missile Command (1981) (Atari, Rob Fulop - Sears) (CX2638 - 49-75166) ~',
  b'V\xe0c|)\xb9\xcf\xcb\xd9\xe7\xb0%\x19:\x80\xdd\xae\xb7\xe0\xb6': 'Missile Command (1981) (Atari, Rob Fulop) (CX2638) (PAL)',
  b'\xfe|\xa2\x1f\x96e\xa4\\hX\x1b\xda1\x01\xd8\xd7\xc2\x87\xd1\xed': 'Missile Command (1983) (CCE) (C-828)',
  b'\x84\x98\xca\xaa\x85\xd2\xfb\x85\xcf\xf4\xade\x91\xb3M\x9c\xcc\x0f\x7fA': 'Missile Command (Hack) (208 in 1) (Unknown) (PAL)',
  b'\x91h\x87\x16\xd3-\xbc\xd7\xc6\xb2\xc2$\xbdV\xa3D?S\\\xf5': 'Missile Command (Unknown) (PAL)',
  b'"Nz1\n\xfd\xb9\x1ci\x15t>r\xb7\xb5;8\xebWT': 'Missile Control (1983) (Video Gems) (VG-01) (PAL) ~',
  b'\xec\xa0vO&\x16\x84$c;b\n\x9f\xf9\xb2t\xb26\x13_': 'Missile War (AKA Astrowar) (1983) (Goliath - Hot Shot) (83-312) (PAL)',
  b'\xd8\x97\xb9\n\xe9-3\x86d\xe0,l\xa7\x0f\xf6\rY!\xb9\xa0': 'Mission 3,000 A.D. (208 in 1) (Unknown) (PAL)',
  b'e+\xb0\x92\xe9\xba)X\xe2B\xb1D\n\xcd\xd6\xc0\x8a\x88N\xc5': 'Mission 3,000 A.D. - Mission 3000 (1983) (Bit Corporation) (PG207)',
  b'\xda\xfaex\x0e\xbb[\xc6>H\xa6\xcf\xd0\xf7\x8e\xa3\xb1\x1az\x1b': 'Mission 3,000 A.D. - Mission 3000 (1983) (Bit Corporation) (PG207) (PAL) [demonstration cartridge]',
  b'\x99\x9d\xc3\x90\xa7\xa3\xf7\xbe|\x88\x02%\x06\xc7\x0b\xd4 \x8b&\xd8': 'Mission 3,000 A.D. - Mission 3000 (1983) (Bit Corporation) (PG207) (PAL) ~',
  b' \x9bTTk\xf9\xbf\xab\r\xaf:\xaa\xd4m\x8aS\x16z]\x0c': 'Mission Survive (1983) (Video Gems) (VG-04) (PAL) [a]',
  b'\x93R\x08!\xce@jz\xa6\xcc0G/v\xbc\xa5C\x80_\xd4': 'Mission Survive (1983) (Video Gems) (VG-04) (PAL) ~',
  b'!6m\x16\x08\xc0\x08\xa8\xfb~\x88Y\xca\xf3\xa3Za\x17g\x02': 'Misterious Thief, A (AKA A Mysterious Thief) (1983) (CCE) (C-839)',
  b'\xb2\x1d\x17\xae\x13\x1e\x93\xaa\x8c\xa3\xfdNL\xaf\x9fmZ\xc2\xf8n': 'Misterious Thief, A (AKA A Mysterious Thief) (1983) (CCE) (C-839) [a]',
  b'\x0bt\xa9\n"\xa7\xa1o\x9c!1\xfa\xbdv\xb7t-\xe0G>': 'Mogul Maniac (Joyboard) (1983) (Amiga - Video Soft) (3120) ~',
  b'M\xc2\xa1\xa7\x8b\x1c#|J{\x93\x19IJ\xd4\xd1c\x83\xb1\xe4': 'Mole Hunter (AKA Topy) (Suntek) (SS-023) (PAL)',
  b'\x81\xa4\xd5h \xb1\xe0\x010\xe3h\xa3S,@\x99)\xaf\xf5\xfb': "Monster Cise (Kid's Controller) (1984) (Atari - CCW) (CX26131) (Prototype) ~",
  b'}\xfe\xb1\xa8\xec\x86<\x1e\x0f)q\x13\xa1\xccA\x85\xc2\x15\xe8\x1c': "Montezuma's Revenge - Featuring Panama Joe (1984) (Parker Brothers - JWDA, Henry Will IV) (PB5760) ~",
  b'\xe0\x82\x92<\x89A\x8c\xe7\xc4\x9aw\x03i\xa7:9\xd2\xbb\x15\x83': 'Moon Patrol (05-16-1983) (Atari - GCC, Mark Ackerman, Noellie Alito) (CX2692) (Prototype)',
  b'\x14\x99\xf3\xb7\x87V\xe0r\x8b\x9c\xf6m\xf1G\xd4&:\x8ba\xfa': 'Moon Patrol (06-15-1983) (Atari - GCC, Mark Ackerman, Noellie Alito) (CX2692) (Prototype)',
  b'W\xe1\xf0\x98\x9f\x8a\x04Rg\xb4\xd1\xcb\x9eY\x9b,\x15V\x11\xcd': 'Moon Patrol (07-04-1983) (Atari - GCC, Mark Ackerman, Noellie Alito) (CX2692) (Prototype)',
  b'"\x86\x92\x94a\xdc\x07\xbb\xdcp\x92E\x01\x16c\xbae\xdb\x9f\x14': 'Moon Patrol (07-26-1983) (Atari - GCC, Mark Ackerman, Noellie Alito) (CX2692) (Prototype)',
  b'|\x0b"\xe8\xdcz\x0b\x85\xffng\xb7I6\x84\xdce-\x0e\'': 'Moon Patrol (07-31-1983) (Atari - GCC, Mark Ackerman, Noellie Alito) (CX2692) (Prototype)',
  b'\xdc\xe7x\xf3\x97\xa3%\x11?\x03W"\xb7v\x94\x92d]i\xeb': 'Moon Patrol (1983) (Atari - GCC, Mark Ackerman, Noellie Alito) (CX2692) ~',
  b'r>\xf7\xda!f\x0f\xe7]\x14\xd0\xf8N\xbaY\x01\xaa\xdb\xc4\xc2': 'Moon Patrol (1983) (Atari - GCC, Mark Ackerman, Noellie Alito) (CX2692, CX2692P) (PAL)',
  b'\xaby\x9f\xcfH\x94mA\xf0q\xfc~.\xd2Q.\xf2;\x07\xbe': 'Moon Patrol (1983) (CCE) (C-1006)',
  b'\xcc`T\xff\xef\x91\xab\x98\xfb\xef$\xb4P\xa4G\x0bY\x8e\xc1~': 'Moon Patrol (Canal 3 - Intellivision) (C 3004)',
  b'|\x14\xaeE\xed\x08\xc5\x9a\xf7\xc1\xc3\xeeF)\xd4\xed\x02\xf6\xa1\x87': 'Moon Patrol (Tron)',
  b'\x05\xab\x04\xdc0\xea\xe3\x1b\x98\xeb\xf6\xf4?\xecg\x93\xa5>\n#': 'Moonsweeper (1983) (Imagic, Bob Smith) (720114-1A, 03207, IZ-001-04) ~',
  b"\x17=t'\xd6\xab\x02L\xc9\xefu\xa0(\x1a\x85\xfdN\xe0\x13\xdd": 'Moonsweeper (1983) (Imagic, Bob Smith) (720114-2A, 13207, EIZ-001-04I) (PAL)',
  b'+\xb4\x8dR\xaaC\xb1\x0c\xdd)\xa1\xce\xb2\xbd\x0c\x83\xae\xfb\xf9j': 'Moonsweeper (1988) (Activision) (AIZ-001)',
  b'X[\x9f5)\x8b\r\xfbER^\xa2\xe6\x0c\x13~\xb3x\xd4\xc3': 'Morse Code Tutor (1979) (Atari, Brad Stewart) ~',
  b"\xc5\xdf'7f\x89\xb4\xa6\x05\x06\x95h!\x8ea\xc07\x952\xfc": 'Moto Laser (AKA Mega Force) (Star Game) (043)',
  b'\xec\xe9{\xdasO\xaf\xfc\xf8G\xa8\xbc\xdf\xa4tx\x9c7}\x8d': 'MotoRodeo (Motor Olympics, Motor Rodeo) (1991) (Atari - Axlon, Steve DeFrisco) (CX26171) (PAL)',
  b'\xde\xa1Pk\xa1\x07\xb9TL\xd9\xb1y\xf8;\xc6\x1c\xed\x91\x01\xac': 'MotoRodeo (Motor Olympics, Motor Rodeo) (1991) (Atari - Axlon, Steve DeFrisco) (CX26171) ~',
  b'\xd8\xa1\xeb\tHQ\xff\x06~\x07u\xfd9\xda\\\x0by\r\x82s': 'Motocross (AKA Motocross Racer) (Joystik)',
  b'\x0e\xc6\x7f\xfe`\xed\x91\xc6\x82\xe0\x14\x9a\x9eh\xe3\xa2~\x12\xde-': 'Motocross (Suntek) (SS-022) (PAL) ~',
  b'\xbc\xaf\xaa`3\x8c\xf5Z\xc7*\x94\r\x04\xa5\xadr\x00?|\xf0': 'Motocross - Motorcross - Motocross (1983) (Quelle - Otto Versand) (719.383 2 - 649635, 781393, 781784, 986404) (PAL)',
  b'\xc4\xd4\x95\xd4.\xa5\xbd5J\xf0N\x1f+h\xcc\xe0\xfbC\x17]': 'Motocross Racer (1983) (Xonox - K-Tel Software - Product Guild, Anthony R. Henderson) (99008, 6240) ~',
  b'\n\x84\xb0\xa6\xbd\x0ey\xf5\xfa\x0b\x1b\xb9\x11!`\xcbVJ\xb86': 'Mountain King (1983) (CBS Electronics - E.F. Dreyer - VSS, Ed Salvo) (4L 2738 0000) ~',
  b'H\xb5\xcb\xfd\x00\x16\x96\x17\x11wp6\x14\x8f\xa7\xd1\xb08\xfd\x18': 'Mouse Trap (1983) (CBS Electronics - JWDA, Sylvia Day, Todd Marshall, Robin McDaniel, Henry Will IV) (4L1818, 4L1819, 4L1820, 4L1821) (PAL)',
  b'\xcfcG\xde\xdc\xfe\xc2\x13\xc2\x8d\xd9!\x11\xecoA\xe7Kod': 'Mouse Trap (Mouse Attack) (1987) (Atari) (CX26146)',
  b'\x85\xc6\x93\xefu\xfbb\xf2o\x99w\xac\xf67\x90\xa0\x0c\xc0\xcb^': 'Mr. Do! (1983) (CBS Electronics, Ed English) (4L4478) (PAL)',
  b'\xe4\xc9\x12\x19\x97y\xbb\xa2_\x1b\x99P\x00\x7f\x14\xdc\xa3\xd1\x9c\x84': 'Mr. Do! (1983) (Coleco, Ed English) (2656) ~',
  b'-\x1c\x99\xa0\x15\x9f\x9b\xf5\xfer}4\xf4k\x07\xd9;\xc84\x1c': "Mr. Do!'s Castle (1984) (Parker Brothers) (PB5820) ~",
  b'\xe3\xb1g\x95\x1a\xe5\x08Bb\xae\xa3%7\x11\x083Y\xe2D\x1b': 'Mr. Postman (Digitel)',
  b'JHT\x80\x12\xbc\x97zJQ\xe4\xb1\x86_\x8c>\x80)\xd8\xac': 'Mr. Postman (Unknown)',
  b'\xf9\xe1I\x1dL\xfe\xa9x"\x94\x93IU\x9aA\xd8z#]\'': 'Mr. Postman - Der Postmann (1983) (Bit Corporation) (PG209) (PAL) ~',
  b'0F"\xbe\x85P\x1f\xd7\xf5\xc7\xeaQ\xb9e\x16\xd6\x85\xb8\xed\xaa': 'Mr. Postman - O Carteiro (1983) (CCE) (C-801)',
  b'\xcc\xd41:\x86C\x00\x85\x81\xd3\x03\x18]`\x1a\xd2\xfd\xad=Q': 'Ms. Pac-Man (1982) (Atari - GCC, Mark Ackerman, Glenn Parker) (CX2675) (Prototype)',
  b'b\xb93\xcd\xd8\x84K\xb1\x81l\xe5x\x89 9T\xfex&\x03': 'Ms. Pac-Man (1983) (Atari - GCC, Mark Ackerman, Glenn Parker) (CX2675) ~',
  b'\xfa\xb9\xf8*\xc3\xf09\xf7\x87\tZ\x91Z\xc7\x91J\xf0L\xe7\xd4': 'Ms. Pac-Man (1983) (Atari - GCC, Mark Ackerman, Glenn Parker) (CX2675, CX2675P) (PAL)',
  b'c\xb7\x9b\xd27\xa4\xf4\x883 \xc6re\x18\xf6\x14\x04\x91\xcd\xbd': 'Ms. Pac-Man (CCE)',
  b'\xbc\x1d\xd0\x1c\x06t\x02\xff\xc7\x02X\xce\xf9\x1c\xeb\xe3\x93\x15\xed\xc7': 'Music Demo (JWDA, Todd Marshall) ~',
  b'Zd\x1c\xaa*\xb3\xc7\xc0\xcd]\xeb\x02z\xcb\xc5\x8e\xfc\xcf\x8dj': 'Music Machine, The (Paddle) (1983) (Sparrow - HomeComputer Software, Dan Schafer, Glenn Stohel, Jon Tedesco) (GCG 1001T) ~',
  b'\xb2\xdf#\xb1\xbfm\xf9\xd2S\xad\x07\x05Y-?\xce5*\x83{': 'My Golf (1990) (HES - Imagineering, David Lubar) (535) (PAL) [fixed] ~',
  b'%\x0f\xbdc\xa8\x0f-\xdc\x03Ks\xde|\xe6\nZ\xd4\x9a\t\xc7': 'My Golf (1990) (HES - Imagineering, David Lubar) (535) (PAL) ~',
  b",,\xeb\x9f'\x9c\x8cp\xdbC\xe6Q+\xc7\x8cmcB\xc7%": 'My Golf (CCE) (PAL)',
  b'\x19\xc2\x88\r\xba\x08\xb5l\xe1\x04\x94\x13\x1dF\x8c9\x92\x1a\r4': 'Mysterious Thief, A (1983) (ZiMAG - Emag - Vidco) (GN-070) (Prototype) [a]',
  b'\r\x1c\x96>\xa0@\xc6\x0b)j&\xc4o\xaa\xadG(\x90I\x9f': 'Mysterious Thief, A (1983) (ZiMAG - Emag - Vidco) (GN-070) (Prototype) ~',
  b'\xbd<G|\x1e\x96,)W\xda\xad\x0c\xf8X"\x07\x10\x84\x1e\xb7': 'NFL Football (AKA Football) (32 in 1) (1988) (Atari, Bob Whitehead) (CX26163P) (PAL)',
  b'\x85\xb4\x8f_\xf8w\xe6\x89(\xab\xeb\xffm\xeb\xa1\xfa\x97# \xae': 'Name This Game (1983) (Digitel)',
  b'+J\x055\xca\x83\xb9c\x90n\xb0\xa5\xd6\x0c\xe0\xe2\x1f\x07\x90]': 'Name This Game (Guardians of Treasure) (1983) (U.S. Games Corporation - JWDA, Roger Booth, Sylvia Day, Ron Dubren, Todd Marshall, Robin McDaniel, Wes Trager, Henry Will IV) (VC1007) ~',
  b'\xa1P\xefV\xf8\x02B\xb0\xfa\x10\x879g\xe4\x9a\xfdy\xb6\xd6\xa9': 'Name This Game (Hack) (208 in 1) (Unknown) (PAL)',
  b"o\x8c\xbf\xcd\xb8'\x04UY\xad#\x01\xe0\x07\x1aaV\x7fRK": 'Name This Game (Unknown) (PAL)',
  b'\xb37N\x03\xea\xc5\xb6\x9eP\xe0\xc9WLVx!\xe5\x9d\xa3.': 'Netmaker (AKA Amidar) (Rainbow Vision - Suntek) (SS-006) (PAL)',
  b"7'q\xae\xb4\xe2\xfb,\xd1\xde\xadT\x97\xe3\x82\x1eB6\xd5\xfc": 'Night Driver (Paddle) (1980) (Atari, Rob Fulop - Sears) (CX2633 - 49-75119) ~',
  b'\xb0\xf0NE\x92\xfb\x9cU\xfa\xaai\xa1\xa4e\x04\x9c*\x9c|\xac': 'Night Driver (Paddle) (1980) (Atari, Rob Fulop) (CX2633, CX2633P) (PAL)',
  b'(\x1f\xf7\xe5\\\'ee"\xb1D\xb8L\xba\x08\xeb\x14\x8e/\n': 'Night Stalker (AKA Dark Cavern) (1988) (Telegames) (PAL)',
  b'*\x1e\xe3XL=\xf11\xa9}b\x91\xc9=\xe4k1`\xafT': "Nightmare (1983) (Sancho - Tang's Electronic Co.) (TEC004) (PAL) ~",
  b'Fj\x03>\xfdv:\xb7\xb1\xf6\xe1+\xe0B\xbc\x92\x92C\x83z': 'Nightmare (CCE)',
  b"\xe2\xe8u\x0b\x88V\xddD\xd9\x14\xc4:}'q\x88\xcc\x14\x8e\\": 'No Escape! (Escape from Argos) (1983) (Imagic, Michael Greene) (720055-1A, IA3312) ~',
  b'\xfb\x9f\xc3\xe2\xb2\xe3\x91\x95+\xfb\t\x0e\xc6\xe1\x89\xdc\xc4\n\xa0\xfa': 'No Escape! (Escape from Argos) (1983) (Imagic, Michael Greene) (720055-2A, IA3312P) (PAL)',
  b'\xd3\xc0\xf9s\xf2\xb3\x1c\x14\xdc:\xd2\xd9\x1f\xe8\xbb\xbbB\xbc\xe2i': 'Nuts (1983) (TechnoVision - Video Technology) (TVS1001) (PAL) ~',
  b'\xecy\x94?\x9b\xb4!t\xd9\xad\xb1S5\xee\x08\xaf_Xu\x9b': 'Nuts (Unknown)',
  b'\x9bd!2o\xd8CBw\xec\x88\xe0\xce\xc3\x07\xbep\x0b\xe9\x9d': 'Nuts (Unknown) (PAL)',
  b'\xf5\xf8a\x905\xc7>\xf3Zo\xa1\x97\xb3\x9b\xc5\x04\xe1\xa0\x9b\x88': 'Obelix (1984) (Atari, Andrew Fuchs, Jeffrey Gusman, Dave Jolly, Suki Lee) (CX26117) (PAL)',
  b'\x91U\xf7\xfaWH\n\x12\xa0<j\x84!<\xc5\xdc{\xe79\xb5': 'Obelix (1984) (Atari, Andrew Fuchs, Jeffrey Gusman, Dave Jolly, Suki Lee) (CX26117) ~',
  b'\x1f\x8d\x06\xb9\x9d\xb9K\n\xa8\xca2\x0c|\xb2\x12c\x9a\xc9Y\x1f': 'Ocean City Defender (AKA Atlantis) (Zellers)',
  b'!\xe3\x19\xd2\x85m\xbc \x7f\xd4\x02\x9abn\x1b`\xaf\xb4F\x91': 'Octopus (AKA Name This Game) (1983) (Carrere Video - JWDA, Roger Booth, Sylvia Day, Todd Marshall, Robin McDaniel, Wes Trager, Henry Will IV - Teldec - Prism) (USC1007) (PAL)',
  b'z\xd7J|61\x8f\x13\x04\xf5\xdcED\x01\xcf%\x7f\xa6\rz': 'Off Your Rocker (Joyboard) (1983) (Amiga - Video Soft) (3130) (Prototype) ~',
  b'\x9e\xbc\x01\xbd\x86\xfa\xce\xef\x87C%\x97\xf1\xfc_\xb6\xce\x08\x8e\xf5': "Off the Wall (Bizarre Breakout, Peasant King, Zip 'n' Zap) (1989) (Atari - Axlon, John Vifian) (CX26168) (PAL)",
  b'=\xcf\xe93\x99\x04AHV\x15\x86\x05b\x88\xc6\xf8\xe5\xc9n+': "Off the Wall (Bizarre Breakout, Peasant King, Zip 'n' Zap) (1989) (Atari - Axlon, John Vifian) (CX26168) ~",
  b'\xeagL\xf2\xc9\r@{\x8f\x8b\x96\xea\xc6\x92i\x0b`+s\xf9': 'Oink! (1983) (Activision, Mike Lorenzen) (AX-023) ~',
  b"\x03\xebJ;=\xb0L\x17\x82\xe1\xa9\x1d'\xeaQQc\xd2X\xfb": 'Oink! (CCE)',
  b'\xf8\xc9\xe7\xd6[\x99$\xdao\xac\x0c\xe7\xdc\x9e\xec\x8c\x00@\x05}': 'Oink! (Unknown) (PAL)',
  b'\x96\xf7\x1b\x08rZ\xb7\x8c\xac=\x0e\xbb\x9d\xcf\xfc\xf1\xd4\x89_\xba': 'Oink! - Das Schweinchen und der Wolf (1983) (Activision, Mike Lorenzen - Ariola) (EAX-023 - 711 023-720) (PAL)',
  b'\xdc\xaa\xb2Y\xe7a|z\xc7\xd3I\x894Q\x89j\x9c\xa0\xe2\x92': 'Omega Race (Booster Grip) (1983) (CBS Electronics) (4L 2737 0000) ~',
  b'\xccMf\xda\x01\x82\x81\x9a\x80g\xdb\xc6\xb7?\xf1\xbc3\xfc\x16[': 'Open Sesame (AKA I Want My Mommy) (1983) (Goliath) (5) (PAL)',
  b'?\xcf\xf0V#+\xb7\xdc\xa1>\x84\xeb\xe9\xbb-\xc6\xbb\\\xd5\xfd': 'Open Sesame (AKA I Want My Mommy) (4 Game in One) (1983) (Bit Corporation) (PGP204) (PAL)',
  b'\xcc\x00\xc5\x13\x8f\xd1\xa3\xb7#\xc5\xf1\xac\xc9\x0f\xf8\xe6\xe8yB\xc1': 'Open, Sesame! - Abre-te, Sesamo! (AKA I Want My Mommy) (1983) (CCE) (C-804)',
  b'\xcd\x96\x8c*\x98?\x9a\xdfM\x8d\x8dV\x829#\xb3\x1c3\x98\x0f': 'Open, Sesame! - Sesam, Oeffne Dich (AKA I Want My Mommy) (1982) (Bit Corporation) (PG204) (PAL)',
  b'R\x02\x822\xf4\xf3\x08\x00\xa2\x13\x1dS\xfc\xa4\x9b\x05\xf6\x14\x88h': "Oscar's Trash Race (Kid's Controller) (03-30-1983) (Atari - CCW, Christopher H. Omarzu, Preston Stuart, Bruce Williams) (CX26101) (Prototype)",
  b'\x99>\x99\x8c_\x85\xd1S\xc9\xa1F"@\xe4{\x06\xc2\x97\xca\x85': "Oscar's Trash Race (Kid's Controller) (10-20-1982) (Atari - CCW, Christopher H. Omarzu, Preston Stuart, Bruce Williams) (CX26101) (Prototype)",
  b'\xdey\xef2Mc\xb2\xbc\xad\xfc\x84_6\xb6SW\x93\xa3\xb2`': "Oscar's Trash Race (Kid's Controller) (12-03-1982) (Atari - CCW, Christopher H. Omarzu, Preston Stuart, Bruce Williams) (CX26101) (Prototype)",
  b'\x98\x00\x7f&5k@2\xa2\xaeN\x9f\xdd\xeaZ8\xa9\x88\xeb\x13': "Oscar's Trash Race (Kid's Controller) (1984) (Atari - CCW, Christopher H. Omarzu, Preston Stuart, Bruce Williams) (CX26101) (PAL)",
  b'y\x05p\x9f\xcc\x85\xcb\xcf\xc2\x8c\xa2\xedT?\xfaszT\x83\xae': "Oscar's Trash Race (Kid's Controller) (1984) (Atari - CCW, Christopher H. Omarzu, Preston Stuart, Bruce Williams) (CX26101) ~",
  b'[;\xbe\xfe\x03yQ0s\x17UZy\x98\x1cz|\x8f\xb2@': 'Othello (1981) (Atari, Ed Logg, Carol Shaw - Sears) (CX2639 - 49-75162) [no grid markers] ~',
  b'\xcb\xec\xf1\xa3-\x93f\xa3\xddJ\xd6C\x91l\xd5\x9c\xdc\x82\n\x8b': 'Othello (1981) (Atari, Ed Logg, Carol Shaw - Sears) (CX2639 - 49-75162) ~',
  b'z%:\xc3\xa4v\x9c\r\xbd?\xffF+\x00i5\x1f\x86+\xe9': 'Othello (1981) (Atari, Ed Logg, Carol Shaw) (CX2639, CX2639P) (PAL)',
  b'\xbf4\x0e\xf2|\xbcClo}o\xa9\x08\x97x0\xe8\xd1vu': 'Othello (1981) (Atari, Ed Logg, Carol Shaw) (CX2639, CX2639P) (PAL) [no grid markers]',
  b'4MiBr5\x13\xc3v\xa7\xa8Dw\x98\x04\xe1\x0f5{\x85': "Out of Control (1983) (Avalon Hill, Jean Baer, Bill 'Rebecca Ann' Heineman, Jim Jacob) (5005002) ~",
  b'J\x15\x14\xa7\xcfBy\xfd\xed\x1b\r\xb6\xf5\xb3\x18\x18\xb9\xff\x01\x1c': 'Outlaw (1978) (Atari, David Crane) (CX2605, CX2605P) (PAL)',
  b'\xb6\x10s\xbc0\xa9(]O\xa2\x1ef\xd0oRSa\xce\x05\xd9': 'Outlaw (32 in 1) (1988) (Atari, David Crane) (CX26163P) (PAL)',
  b'\xf8\xee\xaa\xf4cZ\xc3\x9bK\xdf}\xed\x13H\xbc\xe4c\x13\xef\x9f': 'Outlaw - Gunslinger (1978) (Atari, David Crane - Sears) (CX2605 - 6-99822, 49-75109) ~',
  b'\xca\xdf\x91\t\xe9u#\xd55%\xc2\xf9!\xcc\xc0\x83\xd8\xd8<\xf5': 'Pac Kong (AKA Inca Gold) (Funvision - Fund. International Co.) (PAL)',
  b'Y\xf8\xb3\xb9\xbeK\xf5#\x0c\xa2k\xb2\x1e&J\xbe+\x15G\x05': 'Pac Kong (AKA Inca Gold) (Unknown)',
  b'\xad7^0>\xf4\x98\xb7\x8e]\xd1\xa2\x9e\x9b\xfdw\xea\xea[\x1a': 'Pac Man (1983) (CCE) (C-812)',
  b'"\xb5\x90Z\xbcr\x98"z\xcc\x84\x15\x15G-\xd7FC8\xfc': 'Pac Man (1983) (CCE) (C-812) [a]',
  b'\x00\xfa\x1a\xfc\x97\xc5ZQ\xf5\xde\xe9\xf0V\x0e\x80Y\xd6\x81\x9d\x04': 'Pac Man (1983) (Digitel)',
  b'\x92<\x96\x9d1\xce\xf4P\x07Y2Co\x03\xf1@N\x1c\xab\x0e': 'Pac-Kong (AKA Inca Gold) (Rainbow Vision - Suntek) (SS-003) (PAL)',
  b'\t@\xfe\xa7\xf0L\xdbmK\x90\xc5\xad\x1a~4Nh\xf6\xdb\xb1': 'Pac-Man (1982) (Atari, Tod Frye - Sears) (CX2646 - 49-75185) ~',
  b'>{\xa14wU\xbd\xb3\x13\xe0jT\xaa\xec\xa9\xe6\x11\xea\xa2\xb8': 'Pac-Man (1982) (Atari, Tod Frye) (CX2646) (PAL)',
  b'z\xbc\xa7\xac\xe0O\xb6\x902L\xc3h\x02H\x94\x14\xb4\x9ap\x8e': 'Pac-Man (Hack) (Unknown) (PAL)',
  b',\x10\x8a\xd3\xfa\x80\xedC\x02B.\xdc\xd8\xf5E\xe0\x18\x88\x1f6': 'Panda (Quest) (AKA Panda Chase) (Suntek) (SS-035) (PAL)',
  b'\xf0{G\xcdo3b\xeb\x05\x1e\xab>\xcaVA\x83\xd0\xd4l\x0c': 'Panda Chase (Penda Chase) (1983) (Home Vision - Gem International Corporation - VDI) (VCS83105) (PAL) ~',
  b"o\xa7\x0c\xd7\x9c>\x18\xcdG\xd0'\xb24\xc4\xffK\x14`3'": 'Panda Chase (Unknown) (PAL)',
  b'8P6\xac\x98\xd4O/\xa3::\xb1\x1e5\xca\x9f\x01Q\xd6\xd2': 'Parachute (1983) (Home Vision - Gem International Corporation - VDI) (VCS83123) (PAL) ~',
  b'\xb5)\xc1fN\xd1\xab\xc8\xf5\xf9b\xa1\xfe\xd6\\\x0eD@!\x9c': "Peek-A-Boo (Dr. Salk) (Kid's Controller) (1984) (Atari, Dr. Lee Salk) (CX26135) (Prototype) ~",
  b'}\x15O\x99\x02t\x18\xa7q}\x10\x1c\xd7\xadi\xcb\xe1\xc4\x84\xb9': 'Pega Ladrão (AKA Keystone Kapers) (Dismac)',
  b'\x83"\x83S\x0f]\xee3/)\xcf\x8cHT\xddUO 0\xa0': "Pelé's Soccer - Pelé's Championship Soccer (AKA Championship Soccer) (1981) (Atari, Steve Wright) (CX2616)",
  b'\xc8\x87\xce\xed\xc2!o\x1b@=\xf8\xc4I\xc8\x9b\x083O\xf2\xf6': "Pelé's Soccer - Pelé's Championship Soccer (AKA Championship Soccer) (1981) (Atari, Steve Wright) (CX2616) (PAL)",
  b'F\x1c.\xa3\xe4\xd2O\x86\xec\x02!\\\x1fGC\xd2Pyl\x11': 'Pengo (1983) (Atari, Andrew Fuchs, Courtney Granner, Jeffrey Gusman, Mark R. Hahn) (CX2690) (Prototype)',
  b'^*\x8a\xe2TP\x91"{\x02NG\x8c\x13\x8b4\x12\x11)\x8b': 'Pengo (1984) (Atari, Andrew Fuchs, Courtney Granner, Jeffrey Gusman, Mark R. Hahn) (CX2690) (PAL)',
  b'\x89\xb9\x91\xa7\xa2Q\xf7\x8fB+\xcd\xf9\xcf}Du\xfd\xf3>\x97': 'Pengo (1984) (Atari, Andrew Fuchs, Courtney Granner, Jeffrey Gusman, Mark R. Hahn) (CX2690) ~',
  b'\xc51p5\xe7?`\xe9Y\xc1#\xd8\x96\x00\xc8\x1b|Ep\x1f': 'Pepsi Invaders - Coke Wins (Coca-Cola, Coke & Pepsi) (1983) (Atari, Richard Maurer, Christopher H. Omarzu - Coca Cola)',
  b'\x19\xc3\xad\x03Df\xc0C5\x01\xa4\x15\xa9\x96\xedqU\xd6\x06:': 'Pete Rose Baseball (1988) (Absolute Entertainment, Alex DeMeo) (AG-045-04, AK-045-04) ~',
  b'Y\xcc0\xa3v\x90\x14b\xe4;\xc7\x85\x86I\\r\xba\xb8\x1f\xce': 'Peter Penguin (AKA Frisco) (Pumuckl-Serie) (1983) (ITT Family Games) (554-37 338) (PAL)',
  b"\x02\xa5!5Eu\xd2'\x0fu%\x0ey\xbc\x18\xf6\x9ffl\x7f": 'Phantom Tank (AKA Tanks But No Tanks) (1983) (Goliath) (3) (PAL)',
  b'\x95\x9a\xcaKD&\x9b\x1eZ\xc5\x87\x91\xfc<|F\x1ajJ\x17': 'Phantom Tank - Phantom-Panzer (AKA Tanks But No Tanks) (1982) (Bit Corporation) (PG203) (PAL)',
  b'C,\x12d\xdc-$\xd7n%\xc8u\r\xfc8|\xcc\xb2\xbc/': 'Phantom Tank - Tanque Fantasma (AKA Tanks But No Tanks) (1983) (CCE) (C-808) (PAL)',
  b'9\xf9k\x96\xee\x99j\xb4\xaf\xd8,m\xea5\xc8\xd8e\x85\xb9\x0c': 'Phantom UFO (AKA Spider Fighter) (4 Game in One) (1983) (Bit Corporation) (PGP212) (PAL)',
  b'\x1a\x04\xff2]\xc3\xeb\xb28$(\xc6\xcdD\xda\xe7\xbcv\xffq': 'Phantompanzer II (AKA Thunderground) (1983) (Quelle) (343.273 9) (PAL)',
  b'\xac\x1c<t)\xd1\xd4E\x85\xb8\x89\x05\xbb\xc8\xa8\x1d<\xb2\xc7\xdf': 'Phanton Tank (AKA Tanks But No Tanks) (Digivision)',
  b'\xec\xd9\xe935\xefn\xda\x151\xb0\xb5\xf0:P\x15\x05Bw\xb0': "Pharaoh's Curse (1983) (TechnoVision - Video Technology) (TVS1003) (PAL) ~",
  b'\x97\tc\x97\x8d\x86\xf9\xb6\xcfm\xfe\n\xae\x0cMD\xa02s\x95': "Pharaoh's Curse (Unknown)",
  b']\xd8ST\x8bk\xf1\xab\xff\xaf(\xfa\x01p\x8f\xea \xd3F\xfd': "Pharaoh's Curse (Unknown) (PAL)",
  b"\xb2\x99\xdf'\x92\xc5\xcc\xa71\x18\x92]\xff\x85i[s\xa1b(": "Philly Flasher (AKA Beat 'Em & Eat 'Em) (Paddle) (1982) (PlayAround - J.H.M.) (201)",
  b'\x01\rQ\xe3\xf5"\xba`\xf0!\xd5h\x19C}|\x85\x89|\xdd': 'Phoenix (1983) (Atari - GCC, Mike Feinstein, John Mracek) (CX2673) ~',
  b'\x83z\x0f\x07\xc4\xec\xe7\xe6K\xbd\x08\x86\x05\x85\x9f{m\xad\xe5}': 'Phoenix (1983) (Atari - GCC, Mike Feinstein, John Mracek) (CX2673, CX2673P) (PAL)',
  b'\xd1K\x17\x84\x1f\xd2\xbc\xbfi\x14\n\x99x\xb2H \xd1\xe0W\x9a': 'Phoenix (1983) (CCE) (C-1012)',
  b'\x99\xd3\xfe\xd5\x80G\xaaa\xb4\x9e\xf0\xe0i\xf4\xfav\xbc\x1b\xe1\x8f': 'Phoenix (Unknown)',
  b' \x83\r\x17H\xad?\x88\x0e\xb6\xa4\xd6\xb3\x07\xdcs\xfe\xc7\xa0\x1c': 'Phoenix (Zirok)',
  b"\xa5\x91u7\xcf\x10\x93\xaa5\t\x03\xd8]\x9e'\x1e\x8a\x11\xd2\xcf": "Pick 'n' Pile (1990) (Salu - Ubi Soft, Dennis M. Kiss) (460673) (PAL) ~",
  b'\x97\x84S\x1d\x02 \x0b\xb7\x9aX\x98xw\xc6z\x0f\xc6X%.': 'Pick Up (1983) (20th Century Fox Video Games, Mark Klein) (11034) (Prototype) ~',
  b'H?\xc9\x07G\x1c\\5\x8f\xb3\xe6$\txa\xa2\xfc\x9c\x1eE': 'Picnic (Catch the Fly) (Paddle) (1983) (U.S. Games Corporation - Western Technologies, Tom Sloper) (VC2004) ~',
  b'p\xf4\xe4\xb3\x8d\x9dk!\x82CA8nq\xeb\x1cJ~0\xf4': 'Picnic (Paddle) (1983) (Carrere Video - Western Technologies, Tom Sloper - Teldec - Prism) (USC2004) (PAL)',
  b'WwA\x93\x08\x1a\xce\xa0\x10\xbd\x93Z\x04I\xbc\x8fS\x15q(': "Piece o' Cake (Bakery) (Paddle) (1983) (U.S. Games Corporation - Western Technologies) (VC2005) ~",
  b'\xae\xe0\x83$\xbc \x9f9\xc1FX\x98\x0f\xee\x064=\r\x8bc': 'Pigs in Space - Starring Miss Piggy (1983) (Atari, Bill Aspromonte, John Russell, Michael Sierchio, Robert Zdybel) (CX26114) (PAL)',
  b'!\xa6\x146]c\xff\x10\x1a\xf2\xf9\xcc\x9c\xb5\xf8\x19"\x9e\xc8\xdb': 'Pigs in Space - Starring Miss Piggy (1983) (Atari, Bill Aspromonte, John Russell, Michael Sierchio, Robert Zdybel) (CX26114) (Prototype)',
  b'\xd0\x8b0\xca.^5\x1c\xac;\xd3\xfbv\x0b\x87\xa1\xa3\n\xa3\x00': 'Pigs in Space - Starring Miss Piggy (1983) (Atari, Bill Aspromonte, John Russell, Michael Sierchio, Robert Zdybel) (CX26114) ~',
  b'\xcc\x14\xc6\x1bj\xb0\x87{\x1d\xd6\xa29\xca\xbf\xf9D\xf7\xf9wU': 'Pimball (AKA Video Pinball) (1983) (CCE) (C-856)',
  b'\x93f\x93\x95\x13\x83\xea\xaaq\x0f\x93\x8d\xeb\xd0\xf7p\x1a\x90\x9cm': 'Pitfall (1983) (CCE) (C-813)',
  b'\x86\x11\x96\xf2B[\x93t\xc1\xab\xa5p;\xc1[\x88$\x1e\x06k': 'Pitfall (AKA Pitfall!) (1984) (Supergame) (32)',
  b'\xa0\xa5H\xbcl\xc9\xe6\xaf\x14\x01\xc3}Cx5\x90\xb6p\xae\x1d': 'Pitfall (AKA Pitfall!) (Genus)',
  b'\xfdh\x86kY\xbc\x9f\xf2\xfd\xeb\xa0\xcb\x92\x01+\xfc\xc0D\xde2': 'Pitfall (AKA Pitfall!) (Star Game) (012)',
  b"\xd0\xec\x08\xb8\x8d\x03&'p\x1a\xd7#7RM\x91\xb2lek": 'Pitfall! - Abenteuer im Urwald (Jungle Runner) (1982) (Activision, David Crane - Ariola) (EAX-018, EAX-018-04B, EAX-018-04I - 711 018-725) (PAL)',
  b'\xfdyQo=\xb6\tx\xdc&Eb^\xe0(\xf5\xafv\xb67': "Pitfall! - Pitfall Harry's Jungle Adventure (Jungle Runner) (1981) (Activision, David Crane) (AX-018, AX-018-04) (Prototype)",
  b'\xc0\x84S\x9e\xf8\x88\xef\n\xc8\xbd\xe9\xe8\x1c\xd4\x97i{\xa2\xb4*': "Pitfall! - Pitfall Harry's Jungle Adventure (Jungle Runner) (1982) (Activision, David Crane) (AX-018, AX-018-04) (Prototype)",
  b'\x8dRT\x80D]H\xccHF\r\xc6f\xeb\xadx\xc8\xfb{s': "Pitfall! - Pitfall Harry's Jungle Adventure (Jungle Runner) (1982) (Activision, David Crane) (AX-018, AX-018-04) ~",
  b'\xf1\xb5\xfd\x9e\xef\xd0N\xec-\xf9\xac\xa6\xd7\xcaQO\x94\x95\xbaN': "Pitfall! - Pitfall Harry's Jungle Adventure (Unknown) (PAL)",
  b'\xdd\xe9\x13\xe8\x15\xfc\xe6o\x06\x88=\x9e\xb2\xbbuBn<\xc4H': 'Pizza Chef (1983) (CCE) (C-837)',
  b'\xf8\xdbHMEOQ\x8b!\xd9\xff/r\xed\x18\x874\xd5o\xf6': 'Pizza Chef - Pizza Time (1983) (ZiMAG - Emag - Vidco) (713-111 - GN-050) (Prototype) ~',
  b'q\xcc\x0b0U\xcbY\xcb7\xc9\x86^$\xec\xeb4\x99\x8fD\xb2': 'Planet Patrol (1982) (Play Video) (PAL)',
  b'\xcc\xfc\xbfR\x81ZD\x11X\x97r\x92\xb7\x19\xf7\xc5\xed\x80\xc5\x15': 'Planet Patrol (1982) (Spectravision - Spectravideo) (SA-202) ~',
  b'\x9b6\xe9\xaav(R8\xbav\xfal\xce\x8a\xac\xc5+^\xbc\xc9': 'Planet Patrol (1983) (CCE) (C-830)',
  b"\x8c\xc9i\x16\xd6+\xc5\xfdr\xd1'\xd9\x1b\x95;\xb6\xf3\xaa\x13Y": 'Planet Patrol (1983) (CCE) (C-830) [a]',
  b'!\xfdd\xe5p\xfcU^\x96`uO\x13"\xa1\xe8\xe1\xe0\x92\xca': 'Planet Patrol (Unknown) (PAL)',
  b'\x7f\x8d\x12\xb4\xb0(\xee\xba8u\x16\xad/\xa1K\xf3\x91\x14\xd5\x8f': 'Planet Patrol - Planeten Patrouilie (1982) (Spectravision - Spectravideo - Quelle) (SA-202 - 412.851 8) (PAL)',
  b"\x9f+\xdcS\xef\x1b\x92B'\x00Y\xf9\x90D\x8a;\x8d\xdd\x18&": 'Planet Patrol - Planeten Patrouilie (1982) (Spectravision - Spectravideo - Quelle) (SA-202 - 412.851 8) (PAL) [different spaceship]',
  b'\xdc\xca0\xe4\xaeX\xc8Z\x07\x0f\x0cl\xfa\xa4\xd2{\xe2\x97\ra': 'Planet of the Apes (1983) (20th Century Fox Video Games, John W.S. Marvin) (Prototype) ~',
  b'\x103\x98\xdd5\xeb\xd3\x94P\xc5\xca\xc7`\xfa3*\xac?\x94X': 'Plaque Attack (1983) (Activision, Steve Cartwright) (AX-027) ~',
  b'\xdb\xcd\xad\x8d@\n\x14\xf0^f\xa9\xd7\\\xd3AO\xc8\x0c\xc3W': 'Plaque Attack (1983) (Dynacom)',
  b'YH\xe4\x17\x1e\xea\xba6\x08\x1d\xdc\xc0H\xedtP\x88\x9b\x99\x1b': 'Plaque Attack (CCE)',
  b'P\x95q\x8cz\xba\xc1\x02\x1fQ\x91/\xfe\xf2H\x86WhP@': 'Plaque Attack (Digivision)',
  b'\xd8&\xa1^.5p,\r\xc3\xdf\xfe\xa0o\x1b\xeeF\x05\xcd\xbf': 'Plaque Attack (Unknown) (PAL)',
  b'\x16T\x02JA}\x1e\x82\x86\xd6\xe8\xaa]O\x85\xc1\xe6\xa5Gy': 'Plaque Attack - Schutzt Eure Zaehne (1983) (Activision, Steve Cartwright - Ariola) (EAX-027 - 711 027-722) (PAL)',
  b'\x02\x10O\xfa\xfbm7\xdc`\x83;\xa4/\xf4\x8dx|\xeeU\x0f': 'PlayAround Demo (1982) (PlayAround - J.H.M.)',
  b'c\xc1!F\xc1\x83\xbc\xcb\xf0\\\x00D\xa9a\xdc@y\x0e2\x12': 'Pleiades (1983) (UA Limited) (Prototype) ~',
  b'\xaa\x9f\xcbR\xcd\xa7Ni\xbf+\x99\x0e%lT\xc4&\xc5\xe2\x16': 'Plug Attack (AKA Plaque Attack) (Funvision - Fund. International Co.)',
  b'\xf1\x7f\x83\xaaAA\xda\xa4\xbc\x0c\x01PF\x9ay\x82\x87\x1c\x0e\xbb': "Polaris (02-17-1983) (Tigervision, Robert H. O'Neil) (7-007) (Prototype) (4K)",
  b'$\x10\x93\x1a\x8a\x18\xb9\x15\x99;i\x82\xfb\xab\xab\x0fCyg\xa4': "Polaris (1983) (Tigervision, Robert H. O'Neil - Teldec) (7-007 - 3.60005 VG) (PAL)",
  b"\x87'c\x0c\xa8@\xddvC\xbf\xc5v*oYq\x94\xec\xdd\xa3": "Polaris (1983) (Tigervision, Robert H. O'Neil) (7-007) (Prototype)",
  b'\xb7j\xb6\x91\x18\xb5y\xca\n\xcb\xbb\x8e\xbe\x80\x03\xee\xd5\xcb\xcbJ': "Polaris (1983) (Tigervision, Robert H. O'Neil) (7-007) ~",
  b'\x9a\xfd\xc3\xb6uj\xaf`\xe4\x94\x9fA@7#\x9czO\x7f\xd1': 'Pole Position (1983) (CCE) (C-1002)',
  b'Fe\xc4\\\x83\xbbd(\x9b5x,;Y\xaf\xed^\xc2\x15\x0c': 'Pole Position (RealSports Driving) (05-15-1983) (Atari - GCC, John Allred, Douglas B. Macrae, Betty Ryan Tylko) (CX2694) (Prototype)',
  b'\x9d3M\xa0sR\xa99\x9c\xbb\xd9\xb4\x1ci##-\x0c\xdc\xd3': 'Pole Position (RealSports Driving) (1983) (Atari - GCC, John Allred, Douglas B. Macrae, Betty Ryan Tylko) (CX2694) [a]',
  b'\xb7\x12/G\x8a4<\xff\xac\x17\xb7e\xe9d(\x93X~\x99\xa1': 'Pole Position (RealSports Driving) (1983) (Atari - GCC, John Allred, Douglas B. Macrae, Betty Ryan Tylko) (CX2694) ~',
  b'\xbf\x89\x87j\xba\xc4\xfa0\x0e\xd6\x86*\n\xe8zl_\xe8L\x13': 'Pole Position (RealSports Driving) (1983) (Atari - GCC, John Allred, Douglas B. Macrae, Betty Ryan Tylko) (CX2694, CX2694P) (PAL)',
  b'\xe9\x81\xb0z;\xbc\xe3\xea\x18|\x7f\xd2\x1b\xf4\xb8\xbe\xa0c\x07\x85': 'Pole Position (Tron)',
  b'\xc0\xaf\x01\x88\x02\x8c\xd8\x99\xc4\x9b\xa1\x8fR\xbd\x16x\xe5s\xbf\xf2': 'Polo (1978) (Atari, Carol Shaw - Ralph Lauren) (Prototype) ~',
  b'\x95M)\x80\xea\x8f\x8d\x9av\x92\x16\x12\xc3x\x88\x9f$\xc3V9': 'Pompeii (Labyrinth, Lavarinth) (1983) (Apollo) (AP-2011) (Prototype) ~',
  b'\x98\x85\xca\x8f;V\x12\xcb\xbcwI\xa3!\x94B[\xa4{|\xcf': 'Pooyan (1983) (Gakken) (001) (PAL)',
  b'\xb7\xa0\x02\x02\\$\xab.\xc4\xa0?b!-\xb7\xb9l\x0e_\xfd': 'Pooyan (1983) (Konami) (RC 100-X 02) ~',
  b'v\x00\x9f\xdf\xc6\x81\xa0L\xb9\x8fk\xf0!\xcd\xfe\xe2?s\x19\xae': 'Pooyan (Hack) (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\xff\x12\x12|\xdf\xef\\\xe3?oU\x93\xc3\xdf\x82\x01\xeb\xec\xd2\xe4': 'Pooyan (Unknown) (PAL)',
  b'\xfb\xfc\xa8"\x96+b<:\xc5\xa4X\\\x8d\x04\x04\xb6\x1d\x8c\xcf': 'Popeye (1983) (Parker Brothers, Joe Gaucher) (931519) (PAL)',
  b'\x17r\xa2-\xf3\xe9\xa1\xf3\x84#\x87\xacc\xee\xdd\xff\x7f\x04\xb0\x1c': 'Popeye (1983) (Parker Brothers, Joe Gaucher) (PB5370) ~',
  b'p\xaf\xc2\xcc\x87\x0b\xe5F\xdc\x97o\xa0\xc6\x81\x1f~\x01\xeb\xc4q': "Porky's (1983) (20th Century Fox - Lazer Micro Systems - Dunhill Electronics, B. Winston Hendrickson, Randall Hyde, Mark V. Rhoads, John Simonds) (11013) ~",
  b'\xa0\x9d\x91\xf1b\xd6e\xce\x86\x94\xd7_0\xf2]\xe6\xb9\xe6\n\xbe': 'Power Lords (1983) (Probe 2000 - North American Philips Consumer Electronics Corporation) (Prototype) ~',
  b'\x1e)\xc8Ph~\xb5\xcd\x84\xcb\xab6f\x08\xbd\x9a\x80\x16Ro': 'Power Play Arcade Video Game Album IV - Atlantis, Cosmic Ark, Dragonfire (1984) (Amiga) (1125) (Prototype)',
  b'N{Ck:l\xea\xe7\xf9h\xf3SAG\xb3\x15\xa8\xd3\xcd\x9d': "Power Play Arcade Video Game Album V - Mogul Maniac, Surf's Up, Off Your Rocker, S.A.C. Alert (1984) (Amiga) (1130) (Prototype)",
  b'\xaa\xbb\xbcU\x83\xe2\x87\x8b\xf1\x11w\x97\x8b\x07\xc226\x8f;2': 'Power Play Arcade Video Game Album, The - Ghost Attack, Genesis, Havoc (1983) (Amiga) (1110) (Prototype)',
  b'\x81\x16\xd0\xab5\x85\xca\xf1\xda\x04)\x07\xc2\xa5U\x11-\x03?I': 'Pressure Cooker (1983) (Activision, Garry Kitchen - Ariola) (EAZ-032 - 771 032-712) (PAL)',
  b'\x8b\x00\x13s\xbeHP`\xf8\x81\x82\xe9\xa7\xaf\xcfU\xb4\xd0zW': 'Pressure Cooker (1983) (Activision, Garry Kitchen) (AZ-032) ~',
  b'\xbbN\xbc\xb7\xa2R\x8c[\xeb\xf7\xa9W\x0c\xf3\xc2\x07\x050|C': 'Pressure Cooker (CCE)',
  b'\x1e\xa6\xbe\xa9\x07\xa6\xb5`|v\xf2"s\x0f\x81*\x99\xcd\x10\x15': 'Private Eye (1984) (Activision, Bob Whitehead) (AG-034-04) ~',
  b'$\xab\x03\xa5\x1b\xdd\x11:\x98E\x13$v\x04[sS"$\xdd': 'Private Eye (1984) (Activision, Bob Whitehead) (EAZ-034-04, EAZ-034-04I) (PAL)',
  b'\xc4\xde\x1a\xf8F\x8a\x81\x10\x89OK\x08\x0cf\xdc\xb4R\xebM\xeb': 'Private Eye (CCE)',
  b'\xfe\xb6\xbd7\xe5\xd7"\xbd\x08\x043Xyr\xb9\x80\xaf\xff_\xa5': 'Pumuckl I (AKA Panda Chase) (1983) (ITT Family Games) (PAL)',
  b'\xfb\x0e\xee*\\\x0bt\xee\xaa\xcb\xf4\xc7\x14\xb2\x08\x81\xc2\xf4\xd7G': "Pygmy (AKA Lock 'n' Chase) (Double-Game Package) (1983) (Otto Versand) (311377) (PAL)",
  b'\x86;\xe0Q\xc1\xc5_a\x15U\xd2(\xb6\xe3_\xd2E\t\xb6\xa0': 'Pyramid War (AKA Chopper Command) (Rainbow Vision - Suntek) (SS-004) (PAL)',
  b'\x96,\x9d\x8f\xf3\xa9\xa0\xe6`\x7f\x15\x9a\x92on\xf6\x81%\xa2\x8c': 'Pyramid War (AKA Chopper Command) (Rainbow Vision - Suntek) (SS-004) (PAL) [a2]',
  b'\xe5\xd0\xf6\x15\xfe\xd4n\xfc\xa3\xcc\x93\x1a\x1e_~\xbc\xces\xd5\x86': 'Q-bert (1983) (Parker Brothers - Western Technologies, Dave Hampton, Tom Sloper) (931517) (PAL)',
  b'\xf3\xef\x97\x87\xb4(z2\xe4\xd9\xac{\x9c3X\xed\xc1c\x15\xb2': 'Q-bert (1987) (Atari) (CX26150)',
  b'\x8b\xf3n\xc0\x12\xa4o\x8f{\x17\xccA\xea\n\xfd\xf2\xfe\x00\xd1\x91': 'Q-bert (1987) (Atari) (CX26150P) (PAL)',
  b'\x91\x96oO\xfd\x00\x04\xb1+\xa5\x0b\x99\x0c\x93\xb2[z\xe9TW': 'Q-bert (208 in 1) (Unknown) (PAL)',
  b'j,\xfaH\xe5\x13\xfb\xd4 \xdc\xd6@\x87 \xec\xd0^\x95\xc4\xf4': "Q-bert's Qubes (1984) (Parker Brothers - JWDA, Todd Marshall) (PB5550) (Prototype)",
  b'\xa6\x1b\xe3p$7\xb5\xd1n\x19\xc0\xd2\xcd\x9295\x15\xd4/#': "Q-bert's Qubes (1984) (Parker Brothers - JWDA, Todd Marshall) (PB5550) ~",
  b'\x16\xacW<\x82Z"\tY0\xf4\x9a)\xe0\x15z~K\xc9\n': 'Quadrun (12-06-1982) (Atari, Frank Hausman, Mimi Nyden, Steve Woita) (CX2686) (Prototype)',
  b'\x1ecJ\x873\xcb\xc5\x04b\xd3cV+\x80\x013C\xd2\xfa\xc3': 'Quadrun (1983) (Atari, Frank Hausman, Mimi Nyden, Steve Woita) (CX2686) (Prototype)',
  b'\xb8\xd6\xf5\x08\xed\xbfq>R\xf0\xcb\xf25\xd5\xe1z\xdd/\xbf.': 'Quadrun (1983) (Atari, Frank Hausman, Mimi Nyden, Steve Woita) (CX2686) ~',
  b'l\x0b\x91\x85#\x14\xb2\x93\xf6u\xc8\x89\xe4\xcfT~\xb9ItR': 'Quest for Quintana Roo (1988) (Telegames - VSS) (6057 A227) (PAL)',
  b'\xd8<t\r)h4>d\x01\x82\x8db\xf5\x8b\xe6\xae\xa8\xe8X': 'Quest for Quintana Roo (1988) (Telegames) (6057 A227)',
  b'3\xa4\x7fya\x0cE%\x80,\x98\x81\xf6z\xd1\xf3\xf8\xc1\xb5]': 'Quick Step! (Hop To It, Kwibble) (1983) (Imagic, Dave Johnson) (720119-1A, 03211) ~',
  b'\x80\x87^\xa3\x8cX\xc6Nx\x7f\xbf\x1cK\x99H\xadf%\xf2\xc9': 'Quick Step! (Hop To It, Kwibble) (1983) (Imagic, Dave Johnson) (720119-2A, 13211, EIX-004-04I) (PAL)',
  b'{\xf9E\xeaf~h>\xc2JN\xd7y\xe8\x8b\xbeU\xdcK&': 'Rabbit Transit (08-29-1983) (Atari) (Prototype)',
  b'\x1d\xf8>\x0e\x89\xa9u\x11\xee\x91\xa5\xa9\xf3v\xdf\xfc\xfb\xad\x06\xa9': 'Racer (1982) (Atari - Roklan, Joe Gaucher) (Prototype) ~',
  b'e\xbd\x03k\xff\xd9\n\xe3K\xaa8\x17\xfe\xe6\xeap\xc7dx\xac': 'Racing Car (1983) (Home Vision, R.J.P.G. - Gem International Corporation - VDI) (VCS83124) (PAL) ~',
  b'T\xaa&BT\xd8\xddn#\x7f?b\xb6\xe9\xc9\xa2\x1f\xcbN\xb1': 'Racing Car (Unknown)',
  b'F\x95V\xff\x9e\xfd$\xa6f\x8f\xc0d\x1f\xec<+\xadXU\xae': 'Racquetball (1982) (Apollo - Games by Apollo, Ed Salvo, Byron Parks) (AP-2003) (PAL)',
  b'\xc3\xc8W\x88\x12D%\xc5\xce\xd7\x81.H\x12\xe3\xe3K^\xe3\xd6': 'Racquetball (1982) (Apollo - Games by Apollo, Ed Salvo, Byron Parks) (AP-2003) [a]',
  b'J\xf6\x00\x81R\xf1\xd3\x86&\xd8@\x16\xa7\xefu4\x06\xb4\x8bF': 'Racquetball (1982) (Apollo - Games by Apollo, Ed Salvo, Byron Parks) (AP-2003) ~',
  b'\x02\x0c\x1e\xc7SX\xb8m\xea\x92\xe6\x1e~\xd7^4=h\x18\x8a': 'Racquetball (Unknown)',
  b'R\xa93\x1b\x04\xe4~\xe8\xd9\xcc\xc4e9U\xa1\xb61\xdcJ\x15': 'Racquetball (Unknown) (PAL)',
  b'\xf0\r\x10\x8e\x7f\x1cw\x16\xe8\xeb\xee\x92\x01Y\x9f\x82\x10;R\xff': 'Rad Action Pak - Kung-Fu Master, Freeway, Frostbite (1990) (HES - Activision) (559) (PAL)',
  b'\xffw\xba,\xeeb\x9d\xe5&\xfc\x9a\x0c\xc6\xd5\xcb\xcb^\xed\xbf4': 'Radar (AKA Exocet) (1983) (CCE) (C-867)',
  b'_\x1f+[@{\x06$\xb5\x94\t\xe0 `\xa3\xa9\xe8\xee\xd8\xfc': 'Radar (AKA Exocet) (Zellers)',
  b'\xcb6\x82\x8a*#\xea\xf8\xd76\xafG\x90\xff\xd7\xe9H\xeb\xbf\x1e': 'Radar Lock (Dog Fight) (1989) (Atari, Douglas Neubauer) (CX26176) (PAL)',
  b'3\xf0\x16\xc9A\xfa\xb0\x1e\x1e-\r{\xa7\x93\x0e;\xcd\x8f\xea\xa3': 'Radar Lock (Dog Fight) (1989) (Atari, Douglas Neubauer) (CX26176) ~',
  b'\xa7\x9fn\x0fO\xd7hx\xe5\xc3\xbakR\xd1~\x88\xac\xdb\xe9\xf6': 'Raft Rider (1983) (U.S. Games Corporation - Western Technologies, Dave Hampton) (VC2006) ~',
  b'\x11<\x17Af\xed\xc2\xe9B\xf5iu\xa1\x18~\xe7\xe4\x17\x91\x0f': 'Raft Rider (208 in 1) (Unknown) (PAL)',
  b'\x0f?\xe0\x17(\xda3\xf1w\xc4G\xe3\xb5F.\xb5\xc7-\xe3\xa8': 'Raft Rider (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\xe42\xcdl\xf1\x06\xechjf\x1e\x808\x0f\xfc\x8b\xf18\x08q': 'Raiders of the Lost Ark (06-14-82) (Atari, Jerome Domurat, Howard Scott Warshaw) (CX2659) (Prototype)',
  b'f\x804v\xa7\nRwA\xa3V7\xb9\x9b\xae;} \xd0\x0b': 'Raiders of the Lost Ark (1982) (Atari, Jerome Domurat, Howard Scott Warshaw) (CX2659) (PAL)',
  b'z\xe7\x07\x83\x96\x97\t1\x8eV\xf1\x89\xcf\x03\xda\x922\nj\xba': 'Raiders of the Lost Ark (1982) (Atari, Jerome Domurat, Howard Scott Warshaw) (CX2659) ~',
  b'U\xdb\xa8\xcff\xd8\xe6?\x95\xf8\xbeh\x98\x1f\xeb\xa82\xbd\xc0\xeb': 'Raketen-Angriff (AKA Missile Control) (Videospielkassette - Ariola) (PGP236) (PAL)',
  b'U\xbd\xd4\\}\xbb\x95nT\x9d\x19+\xd4\xdf8;\xda\xde\xf4\x9b': 'Ram It (1983) (Telesys, Jim Rupp) (1004) (PAL)',
  b'\xfd\ni\xc0n\xb3\xf7\xc92\x89Q\xc8\x90dO\x93\xc4\xba\xd6\xad': 'Ram It (1983) (Telesys, Jim Rupp) (1004) ~',
  b"\x82g\x1f(F\x84\xe5p\xbd\x8c\xe1\x8e\xf7'\x1d\xf7|GVR": 'Ram It (Unknown) (PAL)',
  b'{\xb7\xdf%X)\xd5\xfb\xbe\xe0\xd9D\x91^P\xf8\x9a^pu': 'Rampage! (1989) (Activision - Bobco, Robert C. Polaro) (AK-049-04) ~',
  b'\xc4/m\xa3\xaa\x0c\x8b]\xb6\x8d\xd2Z\xc5\xd6\xaaV\xd5\x9b\xafH': 'Rampage! (1989) (Activision - Bobco, Robert C. Polaro) (EAK-049-04B) (PAL)',
  b'\x14\xae\x88\xf2\x82\x93\xaaNc%\xb4\x14\xf6\xb6\xd3\xd9\xc9l\xf0b': 'Reactor (1983) (Parker Brothers, Charlie Heath) (931506) (PAL)',
  b'Z\xdf\x9bS\x03!G#\x80\xeb\xce\xb2S\x9d\xe2\xff\xbb\x03\x10\xbc': 'Reactor (1983) (Parker Brothers, Charlie Heath) (PB5330) ~',
  b':C\xe09i\xb0\x15Y\xd4$\xa5\xd4\xa7\xf7\x8d\xc9g\x99\xc07': 'RealSports Baseball (07-09-1982) (Atari, Eric Manghise, Mimi Nyden, Joseph Tung) (CX2640) (Prototype)',
  b'\xe9\xb0kV}\xdf\x80#\x06\xd5\xc8;\x89\xd3\xda\x80.\xf2\xe1o': 'RealSports Baseball (1982) (Atari, Eric Manghise, Mimi Nyden, Joseph Tung) (CX2640) (Prototype)',
  b'\xac\xe9{\x89\xb8\xb6\xab\x94t4\xdb\xfd&9Q\xc6\xc0\xb3I\xac': 'RealSports Baseball (1982) (Atari, Eric Manghise, Mimi Nyden, Joseph Tung) (CX2640) ~',
  b'\xbc.k\xda\xa9P\xbc\x06\xbe\x04\x08\x99\xdf\xeb\x9a\xd0\x93\x8fN\x98': 'RealSports Basketball (1983) (Atari - Roklan, Joe Gaucher) (CX2679) (Prototype) (PAL) ~',
  b'"\xde\xdb\xfc\xe6\xcc\x90U\xa6\xc4\xca\xec\x01<\xa8\x02\x00\xe5\x19q': 'RealSports Boxing (1987) (Atari - Imagineering, Alex DeMeo) (CX26135) ~',
  b'k\x94\rT2\x84\xc6\x83z1\xea\xa3(\xcaE\x8c\xc7|\x1as': 'RealSports Boxing (1987) (Atari - Imagineering, Alex DeMeo) (CX26135, CX26135P) (PAL)',
  b'\x00(\xd8\xcfF\xc4 _f\\\xa0Z\xd1\x15\xc1\xa5J\x85\xaa\x9f': 'RealSports Football (Football II) (1982) (Atari, Alan J. Murphy, Robert Zdybel) (CX2668) (Prototype)',
  b' \r\x04\xc1\xe7\xf4\x1aZ70(~\xd0\xc3\xf9)6(\xf1\x95': 'RealSports Football (Football II) (1982) (Atari, Alan J. Murphy, Robert Zdybel) (CX2668) ~',
  b'\xd7\xf2\xe5=\xeb?\xc4\xd2\xf8\xd8\x19\xcc\xf0\xfb\xfd\x7fQ\x0f\xdb\xee': 'RealSports Soccer - Football - RealSports Soccer (1983) (Atari, Jerome Domurat, Michael Sierchio) (CX2667) (Prototype)',
  b'\x9c\x88\xe0\xdb\xb2\x87\xd2|\x0e,2\xf6\xe8]\x82\x913\xbb\xe8\xa3': 'RealSports Soccer - Football - RealSports Soccer (1983) (Atari, Jerome Domurat, Michael Sierchio) (CX2667) [no opening tune] ~',
  b"\xe3\xd9d\xd9\x18\xb7\xf2\xc4 wj\xcd3p\xec\x1e\xe6'D\xea": 'RealSports Soccer - Football - RealSports Soccer (1983) (Atari, Jerome Domurat, Michael Sierchio) (CX2667) ~',
  b'\xf7\xd3O\x02x\xa7P\rC% g\xad\xaeR\x9au\x1a\xed\x9d': 'RealSports Soccer - Football - RealSports Soccer (1983) (Atari, Jerome Domurat, Michael Sierchio) (CX2667, CX2667P) (PAL)',
  b'\xdd\xb4\x14w\x03v3+\xb9k\xdd\x8bp\xe5\xbah\xc3\x12\x8a\xa7': 'RealSports Tennis (1982) (Atari - GCC) (CX2680) (Prototype)',
  b'p,\x1c}\x98]\r"\xf95&[\xd2\x84\xd1\xedP\xdf%\'': 'RealSports Tennis (1983) (Atari - GCC) (CX2680) ~',
  b'\x04gU\x93x\xcf\xba\x94\x9b\xe5y\xca\xda\xf2\xde\xadn%\xa5\\': 'RealSports Tennis (1983) (Atari - GCC) (CX2680, CX2680P) (PAL)',
  b' %\xe1\xd8hY_\xad6\xe5\xd9\xe78O\xfd$\xc2\x06 \x8d': 'RealSports Volleyball (1982) (Atari, Jim Huether, Alan J. Murphy, Robert C. Polaro) (CX2666) ~',
  b'TAr\x14\xea\x1b\x88\\\xff\xf7q7\x83s\xb9\xe7\xf9,\x81\xe2': 'RealSports Volleyball (1982) (Atari, Jim Huether, Alan J. Murphy, Robert C. Polaro) (CX2666, CX2666P) (PAL)',
  b'\xdd\x92\xf6xs\xd9(\xe9\x8c\xc4\xa2\xb2y\xb9\r}\xb8\xd5\x0f]': 'Red Sea Crossing (1983) (Inspirational Video Concepts, Steve Shustack) (321430) ~',
  b'v\xd2\x11v\x98\xbc\x90\xcdRw\xfd\xc0X\xa1\xab\xc2\xb3{\xd7g': 'Red Vs. Blue - Football (RealSports Football Beta) (1981) (Atari, Tod Frye) (Prototype)',
  b'\x94\xe9H\x10\xbflr\xee\xe4\x91W\xf9!\x8c<\x17\x0be\xc86': 'Rescue Terra I (1982) (VentureVision, Dan Oliver) (VV2001) ~',
  b"\xc2\x9a\x06\xc2\x95\x0cy\xed\xe3I\x05\xde]\xea'\xb1.\xf1\xf8\xbf": 'Resgate Espacial (AKA Moonsweeper) (CCE)',
  b'\xf8\xa9\xddF\xf9\xba\xd22\xf7M\x1e\xe2g\x1c\xcb&\xea\x1b0)': 'Revenge of the Beefsteak Tomatoes (Revenge of the Cherry Tomatoes) (1983) (20th Century Fox Video Games, John Russell) (11016) ~',
  b'0s\xc7J7V5\x94d\x82\xf6W\xd3g\x19\xdf\x11dDM': 'Reversi (AKA Othello) (32 in 1) (1988) (Atari, Ed Logg, Carol Shaw) (CX26163P) (PAL)',
  b'\xac\xb2C\x0bNlr\xce\x13\xf3!\xd9\xd3\xa3\x89\x86\xdcGh\xef': 'Riddle of the Sphinx (1982) (Imagic, Rob Fulop, Bob Smith) (720106-1A, IA3600) ~',
  b'\xc2\xbe\xa6x\xda\xa26\xd3\xf0\xe7\x9bQv`|\xc0\n\x83\xc6U': 'Riddle of the Sphinx (1982) (Imagic, Rob Fulop, Bob Smith) (720106-2A, IA3600P, EIX-009-04I) (PAL)',
  b'g\x15I=\xceT\xb2#bt\x12)\x07\x88\x15\xb36\t\x88\xae': 'River Patrol (1984) (Tigervision) (7-004) ~',
  b'@2\x97\x80@/\x82G\xf2\x94\xfe\x88O\xfcV\xcc=\xa0\xc6-': 'River Raid (1982) (Activision, Carol Shaw) (AX-020, AX-020-04) ~',
  b'\xde\x04\x1e.\xb7\xf4\x8f\xa3*G\xc8(\x87\xfa\xae\x81\xfby\x91"': 'River Raid (1982) (Activision, Carol Shaw) (EAX-020) (SECAM)',
  b'\xbf\xa0i\xc5\xadE\x99\x0cp\xab\xf8\x1d\xfbC\xe6\x92\xd1AKY': 'River Raid (1983) (CCE) (C-811)',
  b'\xc1\t\x9c\xa7F_\xf6\x17\xc6\xaa\x0e|\xe4\xb0\xf5\x05\x8b\xca\xf1\xfb': 'River Raid (1983) (CCE) (C-811) [a]',
  b'\rs\xd3\x08\xc8"~\xc4nC\xad[\x07\x8aZ\'\xab\xce-f': 'River Raid (1983) (Digitel)',
  b'2Z#t\x80\x0b,\xb7\x8a\xb7\xff\x9e@\x17u\x98e\x10\x9d}': 'River Raid (1984) (Galaga Games)',
  b'\x8b\xa7\x17z\xcd\xe6\xab\x99\xc0\xc6.#\x17\x00\xe1\xa7I^\xaa\xc2': 'River Raid (1984) (Galaga Games) (PAL)',
  b"\x98\xb9F\xbfp(\x0fq\x91\x01'\xd28d>\xe8\xc5g\xf2\x04": 'River Raid (1984) (Supergame) (71)',
  b"I\x1e\x99J\x89W\x0b\xe6A\xf5\xbd\x80\xa6\xe5x'=\t\x03g": 'River Raid (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\r\xdf\x0e\xd5\xcc\x81}\x89\x9c:\x80\xf1\xf7\xeaB\x84\x94\xc1\xe9\x96': 'River Raid (Canal 3 - Intellivision)',
  b'\x1bi\x83\xd1\x9748pU\x97\xfb\x10h\xaar\x86\x7f\xaaF\x93': 'River Raid (Digivision)',
  b'#\xe0\xa6h\xff\xda\xc02\x9b\xb6\x08\x00\xd9Y:\n5b\x190': 'River Raid (Genus)',
  b'\x96\xae9\xe1\x03\xfdFE{\r\xa50t\xbd\xf1o\xee7\x04\xfc': 'River Raid (Hack) (208 in 1) (Unknown) (PAL)',
  b'wj0g\x88\xe1\xae\xb5IdT\xf4\x93\xdb\x15\xc2\x01\x8f\xeb\xa9': 'River Raid (Hack) (208 in 1) (Unknown) (PAL) [a]',
  b'\xd35[n\xbc\x08\x85IL\xd6\xe6\xa1\xd9\xb6{\xe9\xea\xf3O&': 'River Raid (Hack) (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b',\xc6\xc9\x84B\x06\x9b\xf3+\x00\xcc\x98\xb2\xea\x15jN\xdf\xec\xb4': 'River Raid (Polyvox)',
  b'\xb8\xf1\x94\xcf\xa9%bJ\xb7\xfc\xaf\xfd\x98\xbb}rM\xca\x1d\xcf': 'River Raid (Robby) (18)',
  b'\x1d\xc4P\xbaB9:S\xf7\xd2\xee0\xe3Y\xd2\tc\xba\x12\xca': 'River Raid (Star Game) (003)',
  b'\xe6@\x029M\xe3!\xad\x03\xab\xc4nf\xf8\x97%\x97\x8f\xf7/': 'River Raid (Unknown) (PAL)',
  b'\xc5$5\xa4*\xb0>\xe2\xbf\x94t\xfa\xce$Qg\x07\xd2\x91S': 'River Raid - Jagdflieger (1982) (Activision, Carol Shaw - Ariola) (EAX-020, EAX-020-04B, EAX-020-04I - 711 020-720) (PAL)',
  b'\xa0\x8c>\xae3h3L\x93z^\x032\x97\x82\xe9_{W\xc7': 'River Raid II (1988) (Activision - Imagineering, Dan Kitchen, David Lubar) (AK-048-04) ~',
  b";\x83\x86\x81\xfc\x05'ap\xc6\x0e8\xfe\xc1\xde\xc6)\xcb\x06\x87": 'River Raid II (1988) (Activision - Imagineering, Dan Kitchen, David Lubar) (EAK-048-04, EAK-048-04B) (PAL)',
  b"\xbaL\x81\x0fC>Z'.\x87\xc1Q\xb3\x1a\x80\x1c\xdd\xcb\xf7\xc1": 'River Raid II (AKA River Raid) (1985) (Digitel)',
  b':\xd6\xf4e\xa1\x92\xa7\xe6\x93u\x9c\x8d\xa37\x9b-\xefS\x03$': 'River Raid II (AKA River Raid) (Digimax)',
  b'\n\xa8\x0cJ\x8f\xf6\x8b\x0c-i\xb2\xdc\x85\xa0\xf16"\xd3 \x9f': 'River Raid II (AKA River Raid) (Supergame)',
  b'\xcb\x19U\xc5U,\xe7e\xb2Q\xb4\xc3\xff\xbd\xec\x18\xd4\xc8\x1e\xbb': 'River Raid III (AKA River Raid) (1985) (Digitel)',
  b"N0\xa9\x99k\x9b\xf7X\x0f\x86\xbe'A \x98\x83gj\xca\xaf": 'River Raid III (AKA River Raid) (Supergame)',
  b'\xeei@\xc0\x18\xb3Hw\xe1rWb4\xd3\xd7\xb4\x08\xc0 :': 'Road Runner (06-25-1984) (Atari - Bobco, Robert C. Polaro) (CX2663) (Prototype)',
  b'\xee\xa4x\xef\x7f\xf8\xf5y\xddM\x90w\xd3%\xb9\xb6\xcb\x86Tp': 'Road Runner (1989) (Atari - Bobco, Robert C. Polaro) (CX2663) (PAL)',
  b'\x8b\xe5\xf9\xc2\xa1\x1fx\xacSnY\x8e>;}7\x13\x01T\xec': 'Road Runner (1989) (Atari - Bobco, Robert C. Polaro) (CX2663) ~',
  b'\xe3\xbb\x8d\xcb\x02\xe1\xfc\xa0w\x1emP\t\xd2~:b\xfa\x02V': 'Road Runner (CCE)',
  b'\x98\x07\x02u\x08\xbb6\x93~\x14\xcd\xce\x107j\xb0\x87\x01|\xc0': 'Robin Hood (1983) (Xonox - K-Tel Software - Computer Magic) (99005, 6220, 6250) (PAL)',
  b'\x7f\x9c#!\xc9\xf2,\xf2\xcd\xbc\xf1\xb3\xf0\xe5c\xa1\xc5?h\xca': 'Robin Hood (1983) (Xonox - K-Tel Software - Computer Magic) (99005, 6220, 6250) ~',
  b's\x8ediN\x8cnA"\xbf\x02L.\tGO!\xeb\xc4\xb0': 'Robin Hood (AKA Save Our Ship) (1983) (Quelle) (684.733 9) (PAL)',
  b'!\xa3\xeeW\xcbb/A\x0f\xfdQ\x98j\xb8\n\xca\xdb\x8dD\xb7': 'Robot Tank (Robotank) (1983) (Activision, Alan Miller) (AZ-028, AG-028-04) ~',
  b'A3A\xc9G\x15q\x97\xadL\x9d\xd3\xce\xf7\xd0\xda{.\x93a': 'Robot Tank - Rebellion der Roboter (Robotank) (1983) (Activision, Alan Miller - Ariola) (EAZ-028 - 711 028-725) (PAL)',
  b'\xec\x03\x0f\xfa\x9f\xf5t\xb2\xad\x10\xe0\xce\xcf\xf9\xaf+\x1f\x0c\xa2T': "Roc 'n Rope (1984) (CBS Electronics, Ed English) (4L1751) (PAL)",
  b'\n\xbf\n)-J$\xdfZ^\xbe\x19\xa9r\x9f:\x8f\x88<\x8b': "Roc 'n Rope (1984) (Coleco, Ed English) (2667) ~",
  b'\xb2\x7fQ\x90R\xbc\xf8\xfaw+\x07i#Z\xcdT\xdbo\xca\x1a': 'Rocky & Bullwinkle (04-20-1983) (M Network, Steve Crandall, Patricia Lewis Du Long) (MT4646) (Prototype) ~',
  b'\xa0"\x82\xe5\xb3\xde\xeb\x80\xa1l}\xde\x997]\xa8\xc3\x9b\xac,': 'Rodeo Champ (AKA Stampede) (4 Game in One) (1983) (Bit Corporation) (PGP218) (PAL)',
  b'\xfd$<H\x0ev\x9b \xb7\xbf>t\xbc\xd8nJ\xc9\x9d\xab\x19': 'Room of Doom (1982) (CommaVid, Irwin Gaines) (CM-004) ~',
  b'\xb0\x7f\xa1 \x13P\x82\xb8\xd2>{\xbc\xab\xf1\xdan\xe8\xeb-\xf1': 'Room of Doom (208 in 1) (Unknown) (PAL)',
  b'K\xe1@\xb4\xe53\x87\xd2\x80\xe5\xbf3j\x19\xd3t\xff1Ie': 'Room of Doom - Raum ohne Ausweg (1982) (CommaVid, Irwin Gaines - Ariola) (CM-004 - 712 004-720) (PAL)',
  b'\xad#\x10\xffeH\xd2\x97.!\x04\xf1\tW\x92\xb1\nX\xf6\x06': "Rubik's Cube (AKA Atari Video Cube) (1984) (Atari - GCC) (CX2698)",
  b'\xcb;\xb2\x8b\xc4\xf0\x83\x13\xea\x8e\xd9\xff9\\\xa3Bq\xd3\xc4\x99': "Rubik's Cube 3-D (1982) (Atari, Peter C. Niday) (Prototype) ~",
  b'\x11=9\xdb]"\x0b\xc9\xdad\xfa2\xb0\xb66\xaeA$G\'': 'Rush Hour (1983) (Commavid, Ben Burch) (CM-010) (Prototype) [a1]',
  b'\xf2rI\x11P<\x02/\xc4\x15;^F\x1a\x1d}\xba[d\x02': 'Rush Hour (1983) (Commavid, Ben Burch) (CM-010) (Prototype) [a2]',
  b'\xc1\x1d\xf5\xdd\xa5+ft\x85\xb7\xcdDR*\xc7]^\x0f\xe2N': 'Rush Hour (1983) (Commavid, Ben Burch) (CM-010) (Prototype) ~',
  b"'\xe9\x9eG\xfeN\x1b^\xfaI;T\xb0\xe5=*)\x97\xb4\xa1": 'S.A.C. Alert (Joyboard) (1983) (Amiga - Video Soft) (3135) (Prototype) (PAL)',
  b'{\x82\x9e\xdf\xfe\xe3\x03\xc6\xb7\x96\xc7\x15@\xb7cl\xa5h[\xcd': 'S.A.C. Alert (Joyboard) (1983) (Amiga - Video Soft) (3135) (Prototype) ~',
  b'-T\x19\xbf\xed,\xca\xbb\t\xfa%"\xed\xb3\xa8J\x8c\xea\xba\xe2': 'Saboteur (Sabotage) (05-20-1983) (Atari, Jerome Domurat, Howard Scott Warshaw) (CX26119) (Prototype)',
  b'\xc2\rB\xee\x01\xbc\x1a\xef$\xc1K\x84\xf9\xb0\x0e3\xb7\xc1\xba\xe3': 'Saboteur (Sabotage) (06-09-1983) (Atari, Jerome Domurat, Howard Scott Warshaw) (CX26119) (Prototype)',
  b'3\x94\xea\x0e\x9b\xd7\x9c\xb6\xeaO\x06B\xbd\x07%\x0e\xcbXq`': 'Saboteur (Sabotage) (06-15-1983) (Atari, Jerome Domurat, Howard Scott Warshaw) (CX26119) (Prototype)',
  b'\x85u*\xc6\xebpE\xa9\x084%\xcd\x16f\t\x88*\x1c,X': 'Saboteur (Sabotage) (09-02-1983) (Atari, Jerome Domurat, Howard Scott Warshaw) (CX26119) (Prototype)',
  b'\x90\xcd\x98|\xca\xff\xb4(\xe5?\xfd\x98V=\xbb\xe3\xba\xbd.s': 'Saboteur (Sabotage) (12-20-1983) (Atari, Jerome Domurat, Howard Scott Warshaw) (CX26119) (Prototype) ~',
  b'\x1d\xde\xdc\xf4\xff\xe4\xfc\x96\xa2\x87\x0e\x11\xd9\xcc\xa5\xfcs\xe9?Y': 'Save Mary! (Saving Mary) (04-03-1989) (Atari - Axlon, Tod Frye) (CX26178) (Prototype)',
  b'D5{\x82\x8b2\x14D\x1dp-I\x12\xa1\x94\x99Q\xd8\x04r': 'Save Mary! (Saving Mary) (10-24-1991) (Atari - Axlon, Tod Frye) (CX26178) (Prototype) (PAL)',
  b"\xec\xd8\xefI\xae#\xdd\xd3\xe1\x0e\xc6\x089\xb9\\\x8ewd\xea'": 'Save Mary! (Saving Mary) (11-21-1989) (Atari - Axlon, Tod Frye) (CX26178) (Prototype) ~',
  b'\x8f\x068O\xed`|Rg\xd6\xf2\xc1Zo\x19\xcb\xe5\xc5!\xd2': 'Save Our Ship (1983) (TechnoVision - Video Technology) (TVS1002) (PAL) ~',
  b'\x8e\x9bAc\xce\xe3d;\xf8\xfe\x0cZD9m\xdc\xf3\xc4\x01}': 'Save Our Ship (Hack) (Unknown)',
  b'\xb9\xdd\xb4\x1a[D\xafg\xa6\xcb8\xa7m\x12\x90\\\x80|\x0f\x93': 'Save Our Ship (Unknown)',
  b'\x9flr\x95\xc3\xa8.\x0b"\x03\xd8\x87\x15\xa6\xd9\xac\x17{;\xa9': 'Save the Whales (1983) (20th Century Fox Video Games, Steve Beck) (11035) (Prototype) ~',
  b')\xb6BW\xd4y1%\xdf\xa4\xc6\xcf\x05\xc2\xf7j\x11#[H': 'Schiessbude (AKA Carnival) (1983) (Quelle) (701.134 9) (PAL)',
  b'E\x11\x9e(@\xebW<\xa5\xb0X\xb6\x90\xd3\x98\x05$\xb2\x18\xc5': 'Schnapp die Apfeldiebe (AKA Plaque Attack) (1983) (Quelle) (429.663 8) (PAL)',
  b'\xe8\xf5\xae\x86\x1c\xa1\xf4\x10\xc6\xa9\xaf\x11j\x96\xede\xd9\xa3\xab\xb2': 'Scuba Diver (AKA Skindiver) (1983) (Panda) (104)',
  b'\xcf\xccO\x11\xac\xe2\xa7@\x8b\x81"\xdb\xf8\n\xb8\xeckV\xafv': 'Sea Battle (High Seas) (1983) (Intellivision Productions - M Network - APh Technological Consulting, Bruce Pedersen, Larry Zwick) (MT5860) ~',
  b'N\xc6\x98+-\xa2[)\x84\x04(\xfd\x99:9\x1ec\xf570': 'Sea Hawk (AKA Seahawk) (1983) (Panda) (108)',
  b'\xdcr$\t\xe83Ik|\xc8\x15\x9d\xae5\xcd\xd4|]k\\': 'Sea Hawk (AKA Seahawk) (CCE)',
  b'\x87\xcb.\xb0\x1f\x8e\x82\xfd\xdb:f\xe4h\xfa\xb6\x8a\xe5Wx\xda': 'Sea Hawk (AKA Seahawk) (Zellers)',
  b'\xe5U\x8a\xe3\n\xcc\x1f\xa5\xb4\xff\xe7\x82\xaeH\x06"Xj2\xca': 'Sea Hunt (AKA Skindiver) (1987) (Froggo) (FG1009)',
  b'v%$\xb1\xcaP\xee\x93qd\xa6\xf5,\xea\xfb\x8d?\xc6$\xc8': 'Sea Hunt (AKA Skindiver) (CCE)',
  b'\xc2\xf4?5\xd9W\x97\xc0jX\x16T@bV\xc7.B+\xa4': 'Sea Monster (1983) (Goliath) (8) (PAL)',
  b'\xee\xbb\xf7\x05,\xdaU\x1eV\x12\x84P7\xcf\x0f!\xa4\x0f"e': 'Sea Monster (Unknown) (PAL)',
  b'\x0f\xac\x0c\x86\x11\xfb\xe2\x80;*\x95\xd02\x82`\xd1\xc0:L\x8b': 'Sea Monster - O Monstro Marinho (1983) (CCE) (C-805)',
  b'_\xaf.7\xd8n\xbd\x84\xff\x06\x9d\xf2\x0b\xe0\xe6%\x8e\xdf\x0f\x0c': 'Sea Monster - See-Monster (1982) (Bit Corporation) (PG201) (PAL)',
  b'\xb4\xea\xd7ef\x96\xa9\x83\xde\xc3\x19,q\xccMc\xe9\xa6x%': "Seahawk (1982) (Sancho - Tang's Electronic Co.) (TEC002) (PAL) ~",
  b'[Z\xe5cc\x0b\xeb\x11\xf5\x1e\xe5\xa2D\xf12H\x1b\xf8Es': 'Seamonster - Monstre des Mers (1982) (Puzzy - Bit Corporation) (PG201)',
  b's$\xa1\xeb\xc6\x95\xa4w\xc8\x88G\x18\xff\xca\xd2w2\xa9\x8a\xb0': 'Seaquest (1983) (Activision, Steve Cartwright) (AX-022) ~',
  b'\xe0\x93\xedt\xed\xb7\xe4\x0b\x00\x9c\xab\xa3\xef1<\xa6dlUt': 'Seaquest (1983) (CCE) (C-815)',
  b'+CI@\xd2\x8c\xf8\xfe\x00\x1e(|D~A\xe2MC>\xe4': 'Seaquest (1983) (CCE) (C-815) [a]',
  b'\xb3\xb3\xac\x08\xeb\xfb\xf8\xca\xb3"\xeeT\x8d\x9e\x1a\xec\x0f\x1c\x1f\x7f': 'Seaquest (1983) (Dynacom)',
  b'\xed\r\xd4$\xcd\x9d\x00\x08\xf4V\xe2\xa3Sg\x1b\xb8\xfb\xb9\xc6\x1b': 'Seaquest (Digivision)',
  b'\xc2n\x1d\xac\xf3\xdd\xe2<E\xa0\x8b\xfe\x10\xbf\x11\xc15\xfeEp': 'Seaquest (Dinatronic)',
  b'\x1cC\x13\xa4\x9a\x1e\x0bh\x98\x88/\xc4/\x8e\xdf\xbeA\xfe\x8e|': 'Seaquest (Unknown) (PAL)',
  b'*\xbc\xd5su\r\xd2(\xecw\x14\xa7\x0c\xe5\x00\x86k\xee\xcce': 'Seaquest - Rettung aus der Tiefe (1983) (Activision, Steve Cartwright - Ariola) (EAX-022, EAX-022-04I - 711 022-720) (PAL)',
  b'\xb7C\xd4\xd1\x1a\x01\xd0%<\x14\xb1#J"\x05\x95**\x0e\xcf': 'Seawolf 3 (Submarine Commander Beta) (03-23-1981) (Sears Tele-Games, Marilyn Churchill, Matthew L. Hubbard) (CX2647 - 49-75142) (Prototype) (PAL)',
  b'\x19\x14\xf5z\xb0\xa6\xf2!\xf2\xad4K$J<\xdd{\x1d\x99\x1a': 'Secret Agent (Paddle) (1983) (Data Age) (Prototype) ~',
  b'\xaf\x11\xf1fm4Rg\x19j\x1c5"7\'\xe2\xef\x93H:': 'Secret Quest (1989) (Atari - Axlon, Steve DeFrisco) (CX26170) ~',
  b"\x849\xf2\xf8)\x15*K\xac'sa\x02\xae\x13\x1e\xac7m\x0e": 'Secret Quest (1989) (Atari - Axlon, Steve DeFrisco) (CX26170, CX26170P) (PAL)',
  b'\x16\x98{\x16\x07q\xe6\x8fiFN\xa9T\x0f\x12\x13\xc96\x07A': 'See Saw (AKA Circus Atari) (Double-Game Package) (1983) (Otto Versand) (649635) (PAL)',
  b'\xfc\xf5\xf8\xa7\xd6\xe5\x9a3\x9c \x02\xe3\xd4\x08M\x87\xde\xb6p\xfe': 'Sentinel (Light Gun) (1991) (Atari - Imagineering, David Lubar) (CX26183) ~',
  b'n\x91u\x97V\xc3O@\xa2\xc2i6\xdfl\x0c\xa1\xa3\x85\x0e\x80': 'Shark Attack (AKA Lochjaw) (1982) (Apollo - Games by Apollo, Steve Stringfellow) (AP-2005)',
  b']f|\xbb\xee\xbe\xc5\xf1\x8aQ>\x04W\x04\xe5\xcc~\xcc\x03\x01': 'Shark Attack (AKA Lochjaw) (1982) (Apollo - Games by Apollo, Steve Stringfellow) (AP-2005) (PAL)',
  b'\xcf\xb4\xb4\x1e1\x8c|\xd0\x07\x0eu\xe4\x12\xf6|\x97>\x12M\x8e': "Shootin' Gallery (1983) (Imagic, Dennis Koble) (720021-1A, IA3410) ~",
  b'\xf2\n\xb5z\xd7\xca`\xc5f7d\t\xc8\xf9L\x9f\x1b@\x03\xbd': 'Shooting Arcade (Light Gun) (01-16-1990) (Atari - Axlon, Tod Frye - Heuristica, Agustin Ortiz) (CX26169) (Prototype) (PAL)',
  b'nm\xaa4\x87\x8d>3\x1cc\x03Y\xc7\x12ZO\xfb\xa1\xb2-': 'Shooting Arcade (Light Gun) (09-19-1989) (Atari - Axlon, Tod Frye - Heuristica, Agustin Ortiz) (CX26169) (Prototype) ~',
  b'\x08\x95!\x92\xeak\xf0\xef\x9475 \xa7\xe8U\xf5\x8b\xaeay': 'Shuttle Orbiter (1983) (Avalon Hill, Jean Baer, Bill Hood) (5004002) ~',
  b'\x9c\x01\xb7W\x7f\xf3\xd4\xe7K\x91p]\xb8\xd8\xc9z\xda3\x10\x87': 'Sinistar (01-04-1984) (Atari, Lou Harp) (CX26122) (Prototype)',
  b'\xcc\xd5\x0f\x0eM\xdf\xae!B\xe4\xd0k\xeep\xe6\x08O\xbcL\xde': 'Sinistar (01-23-1984) (Atari, Lou Harp) (CX26122) (Prototype)',
  b'$/\xc2=\xef\x80\xda\x96\xda"\xc2\xc7#\x8dHcT\x89\xab\xb0': 'Sinistar (02-13-1984) (Atari, Lou Harp) (CX26122) (Prototype) ~',
  b'\x9d4,?\xb5D\x9f\x88\x7f\xc4\xe4R\xe3\x1f\xfc\xb0\xc9\x9e\x83\xcf': 'Sinistar (1984) (Atari, Lou Harp) (CX26122) (Prototype)',
  b"\xcf\xac;\xbeE\xaf_ '\xab\xe3,:Y\xd2\xfa\xd2\x9e\xf9\xb4": 'Sir Lancelot (1983) (Xonox - K-Tel Software - Product Guild, Anthony R. Henderson) (99006, 6220) (PAL)',
  b'\xfb@\x08\xb1<\xb9\x95|\xe5\xe2\xce\x15U\xc7\xae\xca\xc9\xe7s\xcc': 'Sir Lancelot (1983) (Xonox - K-Tel Software - Product Guild, Anthony R. Henderson) (99006, 6220) ~',
  b'\xa2o\xe0\xb5\xa4?\xe8\x11j\xb0\xaefV\xd6\xb1\x16D\xd8q\xec': "Skate Boardin' (1987) (Absolute Entertainment, David Crane) (AG-042-02, AG-042-04) ~",
  b'$\xa0\xeb#\xde\xdc\x8e\xccN;l\xefd-7\x04\x82O\xd1\xec': "Skate Boardin' (1987) (Absolute Entertainment, David Crane) (EAZ-042-04B, EAZ-042-04I) (PAL)",
  b'\x9cW\x14\xf5\x80\x9f\xae+$l\x98\x9f\xc5\xd8\xfc\xae\xca\x9e\xcet': 'Skeet Shoot (1981) (Apollo - Games by Apollo, Ed Salvo) (AP-1001) (PAL)',
  b"^\xa6\xd2\xeb'\xc7n\x85\xf4w\xbaly\x9d\xeb|An\xbb\xc3": 'Skeet Shoot (1981) (Apollo - Games by Apollo, Ed Salvo) (AP-1001) ~',
  b'\x93o\xce\x1f\xf6\x81\x13\xfa\xbf?\x97\x15\x92\xe9\x8a\xcd\xdc\xab\x97\xb7': 'Ski Hunt (Skiing Hunt) (1983) (Home Vision, R.J.P.G. - Gem International Corporation - VDI) (VCS83106) (PAL) ~',
  b'y\x8fY\x19\x91\x18gL\xec!U\xc1\x83-ys4\xb1i\x04': 'Ski Run (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'n\xa1\x0bD\t"[-\xe0\xf10\x83~}\x0e;\x16J\x16R': 'Ski Run (Video Game Cartridge - Ariola) (TP-607) (PAL)',
  b'j\x08\xa5\xc9\xd9\x83\xff)KW\xab\xb8\xee%$g"\xc8\x8aY': 'Skiing (208 in 1) (Unknown) (PAL)',
  b'\xc8p\xe4V&\xc2\\\xa0/\xa9Y\x94\xc6\xbb\x87\\\x01qz\xc2': 'Skiing (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'_\x92x\x99\xd1\x08o\x0br\x10\x85\xb0J\x13_T\xd5\xc2}\x9b': 'Skiing (32 in 1) (1988) (Atari) (CX26163P) (PAL)',
  b'%\xb8\xeei\xc7>m\x03\x0f\xe1\x19(\x95\xe7\xc54\x11\xd8Tu': 'Skiing (Dactari - Milmar)',
  b'e\x81\x84o\x98;P\xcf\xfbu\xd1\xc1\xb9\x02#\x8b\xa7\xddN\x92': 'Skiing - Le Ski (1980) (Activision, Bob Whitehead) (AG-005, CAG-005, AG-005-04) ~',
  b'C\xac\xd1\xff\x15\xa6\x96\xbb\x00M]\xbd]5\xa9\x05\xc5u\xd4\xe1': 'Skiing - Ski Weltcup (1980) (Activision, Bob Whitehead - Ariola) (EAG-005, PAG-005, EAG-005-04B - 711 005-715) (PAL)',
  b'\xa8k\xf5v,Gf\xc4\xde\x00\xb4\x9b)\xd0\xdf\x9a~\xe4\x17\xe0': "Skindiver (1983) (Sancho - Tang's Electronic Co.) (TEC003) (PAL) ~",
  b'\x12\xac\xae\x81\t{u-$F\xc9Ei\x96\xbd\xcc\xc3\xd6%\xd5': 'Sky Alien (Sky Aliem) (1983) (Home Vision - Gem International Corporation - VDI) (VCS83112) (PAL) ~',
  b'\x08V\xdd\x93\r\xb3(C\xb2_3E\x9a\xbe~|H\x95dn': 'Sky Diver (1979) (Atari, Jim Huether) (CX2629, CX2629P) (PAL)',
  b'\xaa1\x9aL9\xb7\x85\xb2\xe2V\xed\x1b\xb3mN\xb4\xc8\xf7\xc5]': 'Sky Diver (32 in 1) (1988) (Atari, Jim Huether) (CX26163P) (PAL)',
  b'\x9d\x97\x97\xf8\xe73\x81\xcb\xdd.uQt\x19e\x87\x89\x1b\x96(': 'Sky Diver (Hack) (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\x81\xa4\xe5\xa35\xba\x90\x18O\xa8\x84\x819\x14\xe7\x97\xd0\xa6\xa8\xd5': 'Sky Diver (Hack) (Unknown) (PAL)',
  b'M\xde\x18\xd4\xab\xc19V/\xddz\x9d/\xd4\x9a\x1f\x00\xa9\xe6J': 'Sky Diver - Dare Diver (1979) (Atari, Jim Huether - Sears) (CX2629 - 6-99843, 49-75118) ~',
  b'\x10_r-\xcf\x9a\x89\xb6\x83\xc1\r\xdd\x7fhLYf\xc8\xe1\xdb': 'Sky Jinks (1982) (Activision, Bob Whitehead) (AG-019) ~',
  b'\x8e\x16\xd6\rp\x0f\xc7\xa2\xa0!\xae+\x0e&C\x8ea\xc3\xdd\x0c': 'Sky Jinks (CCE)',
  b'\xc34/\xcdN\xbe\\w\xb7[{\x94o\xcd\x07\x17\xd8?I\xf7': 'Sky Jinks (Hack) (Unknown) (PAL)',
  b'\xdc\xcd\xe9\x89\x9b\xdc\x06>w\xa9\x91\x05\x13\x1e9\x14\x18\x1dr\xad': 'Sky Jinks - Wettflug gegen die Zeit (1982) (Activision, Bob Whitehead - Ariola) (EAG-019, EAG-019-04I - 711 019-715) (PAL)',
  b'\xfc_\x1e0\xdb;$i\xc9p\x1d\xad\xfa\x95\xf3&\x8f\xd1\xe4\xcb': 'Sky Patrol (Aerial Ace) (1982) (Imagic, Brad Stewart) (720106-1A, IA3409) (Prototype) ~',
  b"\xce\x8f\x0e\xcf'\x91\x94%\xc4\xc3\xfahlEi\xfd\xda\xa9\xce&": 'Sky Skipper (1983) (Parker Brothers) (931510) (PAL)',
  b'\xef\n~\xcf\xe8\xf3\xb5\xd1\xe6zseR\xa0\xcd\xc4r\x80;\xe9': 'Sky Skipper (1983) (Parker Brothers) (PB5350) ~',
  b'\x8e\x11\xf9\x80\xfc\n\xb13\xe8\to\xd1\xd4\xebs\xda\x15/\x82\xe5': 'Slot Machine (1979) (Atari, David Crane) (CX2653P) (PAL)',
  b'/z\xf6q&[M\x84\x06\xb9\x8ci\x15C\x06\xb5\x8e+\xd4\x88': 'Slot Machine (32 in 1) (1988) (Atari, David Crane) (CX26163P) (PAL)',
  b'r9\xd1\xc6O=\xfc*\x16\x13\xbe2\\\xce\x13\x80=\xd2\xba\xa5': 'Slot Machine - Slots (1979) (Atari, David Crane - Sears) (CX2653 - 6-99823, 49-75111) ~',
  b'\x8a\xf0It\xa9\x97\xa7\xebr\xc4\x89\xe6\x9f\x811\xd6]\xe0;\x1d': 'Slot Racers (1978) (Atari, Warren Robinett) (CX2606, CX2606P) (PAL)',
  b'\xdb\x1e\xcc\xfe\x16\x01\xf9\xc2\xa5\xcdz\x1d\xf8\x14xr\x00\x8b\x8d\x1d': 'Slot Racers (32 in 1) (1988) (Atari, Warren Robinett) (CX26163P) (PAL)',
  b'\xa2\xb10\x17\xd7Y4at\xe3\xd8\xddS\xb64r"\xd3\xb8]': 'Slot Racers - Maze (1978) (Atari, Warren Robinett - Sears) (CX2606 - 6-99825, 49-75112) ~',
  b'\xd5\xf7\xda\x87\x1d\x8e*\xf6\xf4t\xf0W\x00\xbdb\x82k\x01\xd8\xcd': 'Smash Hit Pak - Frogger, Boxing, Seaquest, Skiing, Stampede (HES) (498) (PAL)',
  b'S\x0cx\x83\xfe\xd4\xc5\xb9\xd7\x8e5\xd4\x87p\xb5n2\x89\x99\xa3': "Smurf - Rescue in Gargamel's Castle (Smurf, Smurf Action) (1983) (Coleco - JWDA, Todd Marshall, Robin McDaniel, Henry Will IV) (2465) ~",
  b'qj\xc3\x0b\xf2\xfc\x99\x88Q\xd3\xad\x1di\xbc\xd8\xc1\xca+:[': 'Smurf - Schtroumpfs - Pitufo (1983) (CBS Electronics - JWDA, Todd Marshall, Robin McDaniel, Henry Will IV) (4L1767, 4L1768, 4L1769, 4L1770) (PAL)',
  b'\xc0\xae9e\xfc\xfa\xb0)Ow\n\xf0\xafW\xd7\xd1\xad\xc1wP': 'Smurfs Save the Day (Kid Vid Voice Module) (1983) (Coleco) (2511) ~',
  b'0a\xb7\xaaM\x01\x0fg\x05{\x1c\xc6\x06\xa1\x8eo\x1a\xe7\xe1\xee': 'Snail Against Squirrel (Unknown) (PAL)',
  b'\xe7\xbfE\x0c\xf3\xa3\xf4\r\xe9\xd2M\x89\x96\x8aK\xc8\x9bS\xcb\x18': 'Snail Against Squirrel - Schnecke gegen Eichhoernchen (1983) (Bit Corporation) (PG208) (PAL) ~',
  b'\x84></\xc7\x1a\xf2\xdb?*\xe9\x8e\xb3P\xfd\xe2c4\xcf\xd1': "Sneak 'n Peek (Hide 'n Seek) (1982) (U.S. Games Corporation - Vidtec - JWDA, Garry Kitchen, Paul Willson) (VC1002) ~",
  b'@\xfdu7\x974\xb7U\x18\xd8u\xa7\x021\xfad\xdbJ\xfb[': "Sneak 'n Peek (Unknown) (PAL)",
  b'\x85\xd7 r\xbf\xe2!\x9f\xef\x91+~,\xf3m\x85\x1e^\xfe\x86': 'Snoopy (AKA Snoopy and the Red Baron) (1983) (Century) (PAL)',
  b'\x8b\x01n#\xfa\xae\xbbm\tj\xbdzd6\xb6\xd5\xce\x9f\xb0\xbc': 'Snoopy (AKA Snoopy and the Red Baron) (Digivision)',
  b'\xa4K\x8d\x17\xb6JS\xd6\n\xac\xa2\x13$\xb5@\xd3\xd1\x1d(3': "Snoopy and the Red Baron (05-27-1983) (Atari, Sam Comstock, Richard Dobbis, Nick 'Sandy Maiwald' Turner) (CX26111) (Prototype)",
  b'\r\xab\xf5O\xf8UA\x8f\xbc\xe2\xa7\x0b\xad\xf7\x16\x1d\x98^O\xe5': "Snoopy and the Red Baron (1983) (Atari, Sam Comstock, Richard Dobbis, Nick 'Sandy Maiwald' Turner) (CX26111) (PAL)",
  b"\x97+\xc0\xa7~v\xf3\xe4\xe1'\x0e\xc1\xc2\xfc9^\x98&\xbc\x07": "Snoopy and the Red Baron (1983) (Atari, Sam Comstock, Richard Dobbis, Nick 'Sandy Maiwald' Turner) (CX26111) ~",
  b'hz\x0c\x9b\x90\xbdG\xd9\xe0V*\xc9V6\xaf\tpQ\x98\xa4': 'Snoopy and the Red Baron (CCE)',
  b'ac\xd3f2\xa5\xe3\x13\xbfD9\xed-;\x9b\xf9\xb3r\xda#': 'Snoopy and the Red Baron (Canal 3 - Intellivision) (C 3007)',
  b',5\xfdT\xcd\x18\xe1\xa8\xb4\xeb&\x12hu\xcb\x0c\xf8\xa5\x8b\x19': 'Snow White (02-09-1983) (Atari, Greg Easter, Mimi Nyden) (CX26107) (Prototype) ~',
  b'\\\x96\x8cm\xc0\xdbed\xf4\xea\x83\xa5C\xc0\xbd|>\xfd\x102': 'Snow White (11-09-1982) (Atari, Greg Easter, Mimi Nyden) (CX26107) (Prototype)',
  b'\x9drP\x02\xe9K\x04\xe2\x9d\x8c\xbc\xe3\xc7\x1d;\xb2\xa8CR\xfa': 'Soccer (AKA International Soccer) (1988) (Telegames) (5687 A279)',
  b'\t\xeat\xf1M\xb8\xd2\x1e\xa7\x85\xd0\xc8 \x9e\xd6p\xe4\xce\x88\xbe': 'Solar Fox (1983) (CBS Electronics, Bob Curtiss) (4L 2487 5000) ~',
  b'\x83\xc4[`\xbaM\x17\\\xadmeY\xec\xfcb\xf9/\xdd<"': 'Solar Fox (1983) (CBS Electronics, Bob Curtiss) (4L1852, 4L1853, 4L1854, 4L1855) (PAL)',
  b'\xece\xef\x9eG#\x9a}\x15\xdb\x9b\xca~b[\x16n\x8a\xc2B': 'Solar Storm (Paddle) (1983) (Imagic, Wilfredo Aguilar, Michael Becker, Dennis Koble) (720113-1A, 03206) ~',
  b'\xf0\xa9\x82*\xb9\xc1\t\xf6\xe4\xf3\xe7W\x19\x95\x94\xdc.FU\xf0': 'Solar Storm (Paddle) (1983) (Imagic, Wilfredo Aguilar, Michael Becker, Dennis Koble) (720113-2A, 13206) (PAL)',
  b'&\n\xac7\xff=aR \x9c$\xef\xe8&i\xfe\x87\x1a\x13B': 'Solaris (The Last Starfighter, Star Raiders II, Universe) (1986) (Atari, Douglas Neubauer, Mimi Nyden) (CX26136) (PAL)',
  b'3\xb1o\xbc\x95\xc2\xcd\xc5-\x84\xd9\x8c\xa4q\xf1\r\xae?\x9d\xbf': 'Solaris (The Last Starfighter, Star Raiders II, Universe) (1986) (Atari, Douglas Neubauer, Mimi Nyden) (CX26136) ~',
  b'p\xe9\x127\x90`\xd84\xaa\x9f\xb2\xba\xa2\xe6\xa48\xf3\xb5\xd3\xb6': 'Sorcerer (1983) (Mythicon, Bill Bryner, Bruce de Graaf) (MA1001) ~',
  b'3\xf2\xc1EO(\x97M\x0e\x1fg\xef,\xe2\x16\xc9\xcf\x82\x02\xc8': "Sorcerer's Apprentice (1983) (Atari, Jerome Domurat, Peter C. Niday, Robert Vieira) (CX26109) (PAL)",
  b's3Ua\xc6Y\x84\xd7"\xd7{BB!\xda}0+$\x1d': "Sorcerer's Apprentice (1983) (Atari, Jerome Domurat, Peter C. Niday, Robert Vieira) (CX26109) (Prototype)",
  b']\xf0\x04\x89\x0c\xe6\xd8\xaf\x19\x86\xd3\xc7\xc2\x93Oz\xa8QY\xb9': "Sorcerer's Apprentice (1983) (Atari, Jerome Domurat, Peter C. Niday, Robert Vieira) (CX26109) (Prototype) [a]",
  b'\xae0\t\xe9!\xf22T\xbbq\xf6|\x8c\xb2\xd7\xd6\xde(E\xa5': "Sorcerer's Apprentice (1983) (Atari, Jerome Domurat, Peter C. Niday, Robert Vieira) (CX26109) ~",
  b'\xfb\x87\x0e\xc3\xd5\x14h\xfaL\xf4\x0e\x0e\xfa\xe9a~`\xc1\xc9\x1c': 'Space Adventure (AKA Flash Gordon) (Zellers)',
  b'V\x05ca;\xc3\t\xa52\xd6\x11\xf1\x1a\x1c\xf2\xb9\xaf\x1e/\x16': 'Space Attack (1982) (M Network - APh Technological Consulting, Hal Finney, Bruce Pedersen - INTV) (MT5659) ~',
  b'a\x05\xb1\x85\x9f\xf6\xb5I*\xa2\xf2\x14\x81"J\xc5\xe5\x9c\xd2\xee': 'Space Attack (1988) (Telegames) (PAL)',
  b'&\xc6\xc4~\x9beN\x81\xf4xu\xc5\xfc\xb4\xe6!!%\xf3)': 'Space Canyon (AKA Space Cavern) (1983) (Panda) (100)',
  b'\xb7W\xb8\x83\xee\x11@T\xc6P\x02\x7f;\x9a\x8f\x15T\x8c\xbf2': 'Space Cavern (1982) (Apollo - Games by Apollo, Dan Oliver) (AP-2002) ~',
  b'\xa4ja\x1f\x1bD\\\x94D\xaa\xfa\x9c$\xc1\xec)\x1c\x9e\xd1\xa7': "Space Cavern - Les guerriers de l'espace (1982) (Apollo - Games by Apollo, Dan Oliver - RCA Video Jeux) (AP-2002) (PAL)",
  b"\x07\xb1#W\x90N\xc3\xa4\x84m'\xb3y\x05\x81\xf6\xeb]\xa2\xb7": 'Space Eagle (AKA Exocet) (1983) (Goliath - Hot Shot) (83-213) (PAL)',
  b'1\xd9f\x8f\xe5\x81,=.\x07i\x87\xca2z\xc6\xb2\xe2\x80\xbf': 'Space Invaders (1980) (Atari, Richard Maurer - Sears) (CX2632 - 49-75153) ~',
  b'\x8f\xed1I\x0b\x10q\x94\xcf\xbf<\xd9?qK\x01\x86\x06\xeb\x9e': 'Space Invaders (1980) (Atari, Richard Maurer) (CX2632, CX2632P) (PAL)',
  b'\x82y\x8c\xf1\x99\xa8\xf4s\n\x93\x05\xfa\xb2a\xd9)\xa5x\r\x9b': 'Space Invaders (1980) (Atari, Richard Maurer) (CX2632, CX2632P) (PAL) [different speed and colors]',
  b'C\xa8U\xd5\xafn5\xaaw\xa5+\x03\xa4\xec\xd3%\xd7\xdb\xbb\x9b': 'Space Invaders (1980) (Atari, Richard Maurer) (CX2632, CX2632P) (PAL) [fixed]',
  b'q\xe8\xe75A\xb0\x8e\xa6o\xb4exo}\xb6r\xebd\xb2\xc9': 'Space Invaders (1983) (CCE) (C-820)',
  b'[\xdd\x8a\xf5@ \xfaC\x06WP\xbdB9\xa4\x97i]@;': 'Space Jockey (1982) (U.S. Games Corporation - Vidtec - JWDA, Garry Kitchen) (VC1001) ~',
  b'C\x1c\xa5\x9a4\xce~\x97}i\xf4&<\xf2\x85yQ\x7fE\xad': 'Space Jockey (1983) (Carrere Video - JWDA, Garry Kitchen - Teldec - Prism) (USC1001) (PAL)',
  b'\xca\xef"\xbc\x0e\xa0p@\xbb\x16\xe9\xc3\x00\xcat\xc2\x87\xf9f`': 'Space Jockey (208 in 1) (Unknown) (PAL)',
  b'\xdc\xdf/\xe2\xe3c&q\x81\xfb7h\x8c\xd7\x8f\xa8\xbc\xb9\xb0\xd3': 'Space Jockey (Unknown) (PAL)',
  b'\xc2EQ\xfc\xc2\xcf\xf7i\xc1"\xfe\x0f@\x89\xe8\xbbJw\xea\xa2': 'Space Raid (AKA Challenge of.... Nexar, The) (King Tripod Enterprise Co.) (SS - 007) (PAL)',
  b'"\xc0l\xc0j\x92\x90R\xe8\xc3\xb4\xecDA\x13\xf58[\xc3]': 'Space Raid (AKA MegaMania) (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\xbc\x9bA\x8f\x90\x05\x05\x96^\x0bnS\xa6\xc40\xf1\x00\xaeV\xfa': 'Space Raid (AKA MegaMania) (Rainbow Vision - Suntek) (SS-007) (PAL)',
  b'\xfa\xda\xe2\xe9\x87B\x9f\xaa\xa8U.\xa7@\x9a\xc7\x80\xb8n\xfe\x01': 'Space Raid (AKA MegaMania) (Rainbow Vision - Suntek) (SS-007) (PAL) [a]',
  b'=jC;\xc9=\xd0\x05\xa5x\xcd;tX\x96\xa7g\x08T-': 'Space Robot (Dimax - Sinmax) (SM8001)',
  b'S\xc8\x97\xf7$\xfe\xba\x0e{FF=\xaf\x85\x8b\xc9\xca\xa4\xdb\xa4': 'Space Robot (Dimax - Sinmax) (SM8001) (PAL)',
  b'\\\xf17<\x89\xdc\x15\xe6\x1e\xe1\xc3\xe3\xf5B5E\xf1\xbf\x80\x07': 'Space Robot (Unknown)',
  b'SL\n\t\xae`\xc0\xd9\xa2\xa2\xea\x9e*`/\x07\xb6\x95\xf4\xc5': 'Space Robot (Unknown) (PAL)',
  b'\x8c\x90^\xc6\x9e\x881\xc1\x8a\xdd\xc5\xf7*G2<\xd2L\x17\xbd': "Space Shuttle - A Journey Into Space (1983) (Activision - Woodside Design Associates, Steve 'Jessica Stevens' Kitchen) (AZ-033, AZ-033-04) [FE bankswitching] ~",
  b'\xbc\xecZf\xf8\xdf\xf1\xa7Qv\x96&\xb0\xfc\xe3\x05\xfa\xb4L\xa2': "Space Shuttle - A Journey Into Space (1983) (Activision - Woodside Design Associates, Steve 'Jessica Stevens' Kitchen) (AZ-033, AZ-033-04) ~",
  b'.\x15\x10\xea\xa9\xb5\x8d~\x0f;X\xb2bg\xc8\xcc\xd1aN<': "Space Shuttle - A Journey Into Space (1983) (Activision - Woodside Design Associates, Steve 'Jessica Stevens' Kitchen) (EAZ-033) (SECAM)",
  b'>\x9c\xd0l\xbaMy\xa9\xbc\x04\xfa\xf7\x99w\xb0\x1dIG\xf95': "Space Shuttle - A Journey Into Space - Eine Reise ins All (1983) (Activision - Woodside Design Associates, Steve 'Jessica Stevens' Kitchen - Ariola) (EAZ-033 - 711 033-725) (PAL)",
  b'\x9a,\x8a\xb7\xa3\xd0\xa4\xdf\xa1\x8f\xc1O1\xa4\x08,9X\xdbq': 'Space Tunnel (AKA Cosmic Corridor) (1983) (Goliath) (7) (PAL)',
  b'\xa4\x9b\xf7N\x00hg\xb4\xbd\xd6\xc0\xe1\xa7\xafv\xcd\xad\x8bv\xee': "Space Tunnel - Le Tunnel de L'Estace (AKA Cosmic Corridor) (1982) (Puzzy - Bit Corporation) (PG202)",
  b'\x9d\x97\xe4\x8b\xaf\x0f\x16t\xab\x87/\x17\xcc\xb3\xa1XZ*\xc1w': 'Space Tunnel - O Túnel Espacial (AKA Cosmic Corridor) (1983) (CCE) (C-807)',
  b'R\xc4%\xfc"\x02\x08\xd0\x8c\xb0\xb1\xf2\x90\x9c\xf7\x9f\xb5\xb3\xf3"': 'Space Tunnel - Weltraum-Tunnel (AKA Cosmic Corridor) (1982) (Bit Corporation) (PG202)',
  b'\xc6`$~\xe1(\xbe\xbe\x02\xae\x0b\x0b\xa6G\x8eH\xf5rF\x9f': 'Space Tunnel - Weltraum-Tunnel (AKA Cosmic Corridor) (1982) (Bit Corporation) (PG202) (PAL)',
  b'\x9a2\x10\xdf5\x86o\x81\xeb,\xbe\x8aMg}p\xe7U\x92J': 'Space War (1978) (Atari, Ian Shepard) (CX2604, CX2604P) (PAL)',
  b'\xf8W:\x97\x82\xad\xc2\xf1p\xeb\x8a\xd6\xa5\x0c6\xbb\xa7yZ[': 'Space War (208 in 1) (Unknown) (PAL)',
  b'\xc1\x93\r\x13\x0b\xdff\x8f<\x0bO\x02j\x98\x03(\xee\xb0\x04\x82': 'Space War (AKA Condor Attack) (1982) (Funvision - Fund. International Co.) (PAL)',
  b'#Q\x0b\xa6\x17C\x10\x97f\x8e\xaf\x10J\xa1\xe3b3\x170\x93': 'Space War - Space Combat (1978) (Atari, Ian Shepard - Sears) (CX2604 - 6-99812, 49-75106) ~',
  b'\x9d\xb1^\x10\xce\xa7\x99\xb6\x9e\n}\x89\t$\xc79\x1f\xd2\xcaE': 'Space War - Space Star (32 in 1) (1988) (Atari, Ian Shepard) (CX26163P) (PAL)',
  b'\x98;\x1a\xff\x97\xab\x12C\xe2\x83\xbab\xd3\xa6\xa7Z\xd1\x86\xd2%': 'SpaceMaster X-7 (1983) (20th Century Fox Video Games - Sirius Software, David Lubar) (11022) ~',
  b'\xa2x\xf1B\xb4\xfc\xae"~\xf9\x14\x7f\x12\xfc\xb8\x94\x82\x9d\x1a9': 'SpaceMaster X-7 (208 in 1) (Unknown) (PAL)',
  b'\x90\xadIc@\xff\x00\xb2EA(r\xe0bY#I\x84pz': 'Spacechase (1982) (Apollo - Games by Apollo, Ed Salvo, Byron Parks) (AP-2001) (PAL)',
  b'\xb3V)N5\x82{\xf8\x1a\xdd\x95\xfe\xe5E;\x0c\xa0\xf4\x97\xad': 'Spacechase (1982) (Apollo - Games by Apollo, Ed Salvo, Byron Parks) (AP-2001) ~',
  b'\x1e[`I\xceuD\x197\xb6\xaf\x87\x8c"\xdb\x1ch\x9d\xe6x': 'Spectracube Invasion (AKA Immies & Aggies) (Suntek) (SS-025) (PAL)',
  b'\x06\x82\n\xd3\xc9W\x918G\xf9\x84\x9d\x92\x0b\xc8r_S_\x11': 'Spider Fighter (1983) (Activision, Larry Miller) (AX-021) ~',
  b'\x80\xec\x8ec\x0c\x118@\x16\xf3R9\xb3\x8fj\x0f\xcf\x971|': 'Spider Fighter (1983) (CCE) (C-853)',
  b'\x13u\n\x9cxQ\xec\xe8$\x02\x11\x88W\xa3\xf5\xd8 f\n\xf6': 'Spider Fighter (1983) (Dynacom)',
  b'2\x855/\x8fsB\xd9\xf5\xef\x11\xd8kl[a\xc2\x87\x03B': 'Spider Fighter (Canal 3 - Intellivision)',
  b'3_\xda\x18_[\xbd~\x06\r\xa7\xb1IJ\xaf\xbe\x05D\x91\x14': 'Spider Fighter (Digivision)',
  b')\x08\x00X\x95\x02Zx0.\xa2\x84\x04\x93\xb8\xe4\x80\x9e\xa2\x08': 'Spider Fighter (Unknown) (PAL)',
  b'\xb8\xb8\xea\x8b\xa9\xe1\xa3\xb8\x03\x1eo\xf2\xd9\xbf\xcf\xde\xe9\xb5#n': 'Spider Fighter - Monster greifen an (1983) (Activision, Larry Miller - Ariola) (EAX-021, EAX-021-04I - 711 021-720) (PAL)',
  b'\x1a\x14\xa5\x08\xf0V\\\xdc\xf6\xd9\x941\xd4\xe4\x9b\x8e\x9e\n\x1c\xaa': 'Spider Kong (AKA Inca Gold) (1983) (Goliath) (6) (PAL) [a]',
  b'\\6\xd0\xbb\xe3\xfc\xb8\x96\x87{\x93\xcb\x0e\xe6\x95[x\xcc\x97\xff': 'Spider Kong (AKA Inca Gold) (208 in 1) (Unknown) (PAL)',
  b']o\x91\x8b\xbaK\xd0F\xe8[p}\xa3\xb7\xd6C\xcc.\x1f\x1f': 'Spider Maze (AKA Inca Gold) (1982) (K-Tel Vision)',
  b'\xed\x12\xb5\x0b\xedZ\xd0\x137-\x88\x10\r!8\xf1\x9e\xea\xf7L': 'Spider Maze (AKA Inca Gold) (1982) (K-Tel Vision) (PAL)',
  b'`\xaf#\xa8`\xb3>\x1a\x85\x08\x1b\x8d\xe7y\xd2\xdd\xfe6\xb1\x9a': 'Spider Monster - Inca Gold (1982) (Funvision - Fund. International Co.) (PAL) ~',
  b'dE\xfb/k\xbc\xc8Dh\xafx@Upb\x8d\xde\x976\\': 'Spider-Man (1982) (Parker Brothers, David Lamkins, Laura Nikolich) (931503) (PAL)',
  b'\x91,_Uq\xacY\xa6x-\xa4\x12\x18<\xddbw4X\x16': 'Spider-Man (1982) (Parker Brothers, David Lamkins, Laura Nikolich) (PB5900) ~',
  b'\x90A\x18\xb0\xc1\xbeHG\x82\xec*`\xa2D6\x05\x96\x08\xb3m': 'Spiderdroid (AKA Amidar) (1987) (Froggo) (FG1002)',
  b'\xfd\x0b$Hi\xf9R\x1a\x8f\xf6!\xd5\x9e"\x0f3\xf0Re6': 'Spiderman (AKA Spider-Man) (1983) (Quelle) (495.663 7) (PAL)',
  b';\x98h\xcf\x92Z^^\xe5\xdd\xee_%\x82\xd5\xe1\x89\x9d\xad\xe2': "Spike's Peak (1983) (Xonox - K-Tel Software - Beck-Tech, Steve Beck) (6210, 7210, 06003. 99001) (PAL)",
  b" RA\xa1'x\x82\x99\x81\xe9(\x1d\x9co\xa17\xf1\x1e\x13v": "Spike's Peak (1983) (Xonox - K-Tel Software - Beck-Tech, Steve Beck) (6210, 7210, 06003. 99001) ~",
  b'\xa9y\xbb\x0b\xc9M\xcbt\xd6|\xa1Y\x1eu\xd4#\x18vgg': 'Spinning Fireball (Unknown) (PAL)',
  b'\xe3\x0ea\x95d(\x84\x95K\xc9\x8d\xf2\xe8,>\x97-\xb4\x9aW': 'Spinning Fireball - Fire Spinner - Fireball Spinner (1983) (ZiMAG - Emag - Vidco) (GN-080) (Prototype) [a]',
  b'+\x18J\xcd\xf9\x082\x8a\x1c\x13L\x91\x8d\x16\xa6\x8c\xc2\xe5\x7f\x94': 'Spinning Fireball - Fire Spinner - Fireball Spinner (1983) (ZiMAG - Emag - Vidco) (GN-080) (Prototype) ~',
  b'\x16]\xe0\xeb\xcab\x8e\xb1\xe9\xf5d9\x0c\x9e\xed\xfe(\x9cz\x1d': 'Spitfire Attack (Flight Commander) (1983) (Milton Bradley Company) (4363) ~',
  b')\x88B\x1b]a!\xbc\xe7\x14i\xb8\xe4\xf3{\x0cK\x97\tt': 'Sports Action Pak - Enduro, Ice Hockey, Fishing Derby, Dragster (1988) (HES - Activision) (PAL)',
  b'\xca\r\xcb\xd9.\x9d\xbf\xeb\xc5\xbc!\x1dz\xce\x0e\\|#\x831': 'Springer (1983) (Tigervision - Software Electronics Corporation - Teldec) (7-006 - 3.60008 VG) (PAL)',
  b'm\xa0\xaa\x8a\xa4\x0c\xd9\xc7\x8d\xc0\x14\xde\xb9\x07E)h\x8d\x91\xd0': 'Springer (1983) (Tigervision - Software Electronics Corporation) (7-006) ~',
  b'\xc0\xe2\x9b\x86\xfc\x1c\xc4\x1a\x1c\x8a\xfa7W,<V\x98\xaep\xb2': 'Sprint Master (Sprint 88, Sprint 2600) (1988) (Atari - Bobco, Robert C. Polaro) (CX26155) ~',
  b'\xdc\xaa\xc1\xfa\xc0\xfe\x15\xaa\xa4U</\xa0+\x92:m\xfbh\x80': 'Sprint Master (Sprint 88, Sprint 2600) (1988) (Atari - Bobco, Robert C. Polaro) (CX26155P) (PAL)',
  b'\x1d\n\xcf\x06M\x06\xa0&\xa0K`((]\xb7\x8c\x83N\x98T': 'Spy Hunter (Dual Control Module) (1984) (Sega, Jeff Lorenz) (011-01, 011-02) ~',
  b'\x85\xbf\xd2\xaa\xad\xa3\xee\xea\x1e\xb0\x9c\xb8\xd7\xa29D\xc1\xear%': 'Spy Vs. Spy (AKA Chopper Command) (4 Game in One) (1983) (Bit Corporation) (PGP213) (PAL)',
  b'\x031H\xfa\xeb\xc9}N\xd3\xa8l\x97\xfe\x0c\xde\xe2\x1b\xd2a\xf7': 'Squeeze Box (1983) (U.S. Games Corporation - JWDA, Todd Marshall, Robin McDaniel, Henry Will IV) (VC2002) ~',
  b'\x87\xab\xe4\xab\x10\xc9\xa7\xca\\U\xd7\xa3J1\x81\xbb\xde\x0b\n\x90': 'Squirrel - O Esquilo (AKA Snail Against Squirrel) (1983) (CCE) (C-809)',
  b'\x11\xa9\xddDx\x7f\x01\x1e\xc5@\x15\x92H7|\xb2\x7f\xb8\xf7\xbb': "Squoosh (Vat's Incredible!, The Grape Escape) (1983) (Apollo) (AP-2012) (Prototype) [a]",
  b"t\x05\xd4\xb4'\xcb'\x80b\xfaD\xdb\x91\x8c\np b\xe5\\": "Squoosh (Vat's Incredible!, The Grape Escape) (1983) (Apollo) (AP-2012) (Prototype) ~",
  b'F\xaa\xbd\xe3\x07J\xcd\xed\x88\x90\xa2\xef\xa5Xmk\x8b\xd7k]': 'Sssnake (1982) (Data Age) (DA1003) ~',
  b'VU|\x04\xe0\xad\xfd\x00\xcbG\x84/D\xb1AY\xad\x93\xc6%': 'Sssnake (1983) (Gameworld) (133-003) (PAL)',
  b"'q\x84\xc4\xe6\x1c\xed\x1490I\xa2\x1a0N\x94\x1d\x05\x99?": 'Stampede (1981) (Activision, Bob Whitehead) (AG-011) ~',
  b'\x9b\x8d\xff,\xc1\xf7!"\x85\xbe\xbf^R\x9e\xc3\xc7\xec\x97\x10\x17': 'Stampede (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\xe0\x13\xde\xc08\xbd\xb0\xe2\xb5|G0\xe9d\xe5I\x99\xd6BZ': 'Stampede (32 in 1) (1988) (Atari) (CX26163P) (PAL)',
  b'$\xbd\xc9!\x95\x9e]>\x9d\xb8\xe1O\x17\xa4{\xf0\x05\xf6\xb2\xba': 'Stampede (Canal 3 - Intellivision)',
  b'\xd6e\xb6\xbc1\xc3p\xb7\xeb\xb1[\xcaD\xf3Qh&-\xe8\x1e': 'Stampede (Hack) (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'M\x7f\x86\xacJ\x82\xf6mfW\xbf\xb5\r.\x07\xb3I\x81$)': 'Stampede (Unknown) (PAL)',
  b'\x0fP\xa4\xbb\n\xaaE\x08\xd7eK\xaczF\x0cT\xce|Z\xfe': 'Stampede - Lasso-Helden (1981) (Activision, Bob Whitehead - Ariola) (EAG-011, PAG-011 - 711 011-715) (PAL)',
  b'\x1b\x95\xe0t7\xdd\xc1R=~\xc2\x1cF\x02s\xe9\x1d\xbf6\xc7': 'Star Fox (1983) (Mythicon, Bill Bryner, Bruce de Graaf) (MA1003) ~',
  b'\xb8\xf4\x12\x8f\x8eD\xfb\x93\xd2\xb1\x115\xa4n\xabB\xcde\xb9\xf0': 'Star Raiders (1981) (Atari, Carla Meninsky - Sears) (CX2660 - 49-75187) (Prototype)',
  b'\xe1\x0c\xce\x1aC\x8c\x82\xbdI\x9e\x1e\xb3\x1a?\x07\xd7%A\x98\xf5': 'Star Raiders (Video Touch Pad) (1982) (Atari, Carla Meninsky - Sears) (CX2660 - 49-75187) ~',
  b'|*-\xdb\xde\xf69\xed)\x85\xcef\xaeq{"\x85\xa9J\xe0': 'Star Raiders (Video Touch Pad) (1982) (Atari, Carla Meninsky) (CX2660) (PAL)',
  b'\x87\x8ex\xedF\xe2\x9cD\x94\x9d\t\x04\xa2\x19\x88&\xe4\x12\xed\x81': 'Star Ship - Outer Space (Star Trek, Space, Space Mission) (1977) (Atari, Bob Whitehead - Sears) (CX2603 - 99803, 49-75601) ~',
  b'\xde\x05\xd1\xca\x8a\xd1\xe7\xa8]\xf3\xfa\xf2[\x1a\xa9\x0b\x15\x9a\xfd\xed': 'Star Strike (1983) (M Network, David Akers, Patricia Lewis Du Long - INTV) (MT4313) ~',
  b"\xa0\xe2\x94\x05\xa9's\xbf;\xaa(Ex\x8a\x8a\xdd\x8f;\xc0\xb1": 'Star Strike (1988) (Telegames) (PAL)',
  b'fzR\x8e\x8c\xf3\xfd\x7fS<\xf6y\x93\xf4\x8b\xcc<\x10\x0e\r': 'Star Trek - Strategic Operations Simulator (1983) (Sega, Jeff Lorenz - Teldec) (004-01) (PAL)',
  b'a\xa3\xeb\xbf\xfa\x0b\xfbv\x12\x95\xc6n\x18\x9bb\x91_H\x18\xd9': 'Star Trek - Strategic Operations Simulator (1983) (Sega, Jeff Lorenz) (004-01) ~',
  b'\xcc\xc5\xb8)\xc4\xaaq\xac\xb7\x97nt\x1f\xdb\xf5\x9c\x8e\xf9\xebU': 'Star Voyager (1982) (Imagic, Bob Smith) (720000-201, 720102-1B, IA3201) ~',
  b'B,eD\x99\x8a\xc6\xc5#\x90\x83\xa8\\\xe9OP\xf9\xa1*\x00': 'Star Voyager (1982) (Imagic, Bob Smith) (720102-2B, IA3201P, EIX-011-04I) (PAL)',
  b'\xef\xfd\x88dX^\xfc~\xfa\xbf\x11\x96,\x15\xd3\xa6\xb0bki': 'Star Voyager (1983) (CCE) (C-818)',
  b'A{\xb8\x9e\x11\x17A3!\xff\xbaH\xa1~\x00[&\x87h\x0b': 'Star Voyager (1983) (CCE) (C-818) [a]',
  b'\xacQ\xaa\x01\x92\xccA\xbc\x082\x97v\xc1\xb0*\xe3\x88\xa9\xc5X': 'Star Wars - Jedi Arena (Paddle) (1983) (Parker Brothers, Rex Bradford) (931507) (PAL)',
  b'6\xb9\xed\xc7\x15\x03\x11 ?7\\\x1b\xe1\r\x05\x10\xef\xdedv': 'Star Wars - Jedi Arena (Paddle) (1983) (Parker Brothers, Rex Bradford) (PB5000) ~',
  b'\x19D}~\xab\xde\xcd\x99\x84\xbe\x1a*\xd5\x00\xba4%\x1cE\x8a': 'Star Wars - Return of the Jedi - Death Star Battle (Revenge of the Jedi - Game II) (1983) (Parker Brothers - JWDA, Todd Marshall, Robin McDaniel, Ray Miller) (931513) (PAL)',
  b'*\xd9\xdbKZ\xec-\xa3n\xcc1xY\x9b\x02a\x9c<F.': 'Star Wars - Return of the Jedi - Death Star Battle (Revenge of the Jedi - Game II) (1983) (Parker Brothers - JWDA, Todd Marshall, Robin McDaniel, Ray Miller) (PB5060) ~',
  b'\xc9\xd2\x01\x93[\xbecsy2A\xba\x9c\x03\xcc\x02\xf1\xdf1\xc9': 'Star Wars - Return of the Jedi - Ewok Adventure (Revenge of the Jedi - Game I) (1983) (Parker Brothers, Larry Gelberg) (PB5065) (Prototype) (PAL)',
  b'\xb7Y\xea\xbf\r\xcb\x11,\x94\xb9\xfdfE\x1a\x88!0fx`': 'Star Wars - Return of the Jedi - Ewok Adventure (Revenge of the Jedi - Game I) (1983) (Parker Brothers, Larry Gelberg) (PB5065) (Prototype) ~',
  b'\x96W\xf1\xe4E\xd6\xe4\x06\xa4\x81\x15\x84cR\xf4\xf9\xbc\x81^\xc8': 'Star Wars - The Arcade Game (01-03-1984) (Parker Brothers, Wilfredo Aguilar, Michael Becker, Neil McKenzie, Bob Smith, Brad Stewart) (PB5540) (Prototype)',
  b'\xe3\xc6\xf3\xddC\x90\xfb`\xd8\x8a\x05\xef\x08IGWM1;\xa5': 'Star Wars - The Arcade Game (04-05-1984) (Parker Brothers, Wilfredo Aguilar, Michael Becker, Neil McKenzie, Bob Smith, Brad Stewart) (PB5540) (Prototype) (8K)',
  b'\xe5\x12/u\x12\x90\x85\xd8\x03\x99rD$mE\xeb\x00~\xde\x9a': 'Star Wars - The Arcade Game (12-05-1983) (Parker Brothers, Wilfredo Aguilar, Michael Becker, Neil McKenzie, Bob Smith, Brad Stewart) (PB5540) (Prototype)',
  b'\xdc\x07A\xb2\x8c\xdf)U\x8b\x0c^\xe3v\xb4\xdc\x90\x98K\x92\xd7': 'Star Wars - The Arcade Game (12-15-1983) (Parker Brothers, Wilfredo Aguilar, Michael Becker, Neil McKenzie, Bob Smith, Brad Stewart) (PB5540) (Prototype)',
  b'X\xba\xf6s\x9ea(\x04\x8c\xb6\xd5\x99q\xe97\x00\xa9\xdeo/': 'Star Wars - The Arcade Game (12-23-1983) (Parker Brothers, Wilfredo Aguilar, Michael Becker, Neil McKenzie, Bob Smith, Brad Stewart) (PB5540) (Prototype)',
  b'\x91N\xb4R\x9f_\xa0DQk\xa2\xcfZ`n\x84|\xbf\r\x15': 'Star Wars - The Arcade Game (1984) (Parker Brothers, Wilfredo Aguilar, Michael Becker, Neil McKenzie, Bob Smith, Brad Stewart) (PB5540) (PAL)',
  b'\x88#\xfe=\x8e:\xea\xdcka\xcaQ\x91N;\x15\xaa\x13\x80\x1c': 'Star Wars - The Arcade Game (1984) (Parker Brothers, Wilfredo Aguilar, Michael Becker, Neil McKenzie, Bob Smith, Brad Stewart) (PB5540) ~',
  b'rv\xae\x9e\x8a\x99a\x83x\x86\x8d\x8d\xe2\xcf}\xe6\xael\x1d\xc0': 'Star Wars - The Empire Strikes Back (1982) (Parker Brothers, Rex Bradford, Sam Kjellman) (931501) (PAL)',
  b'\xad[,\x9d\xf5X\xab#\xad)T\xfeI\xed[7\xa0`\t\xbf': 'Star Wars - The Empire Strikes Back (1982) (Parker Brothers, Rex Bradford, Sam Kjellman) (PB5050) ~',
  b"\x81Hv\xed'\x01\x14\x91#c\xe4q\x8a\x84\x12=\xee!;o": 'StarMaster (1982) (Activision, Alan Miller) (AX-016) ~',
  b'\xbc.\xa6\xe6c\xd6y\xad.\\\xf8\xef\xbf\x85\xa5X\x08\xe7\xd5d': 'StarMaster (Unknown) (PAL)',
  b'\x1b\x15\xda\x84\xa0\xe5\n\x91\xed\xed\xdeo\xf2\xefG\x81\xd0\xee\xf4\x81': 'StarMaster - Kommando Galaxis (1982) (Activision, Alan Miller - Ariola) (EAX-016, PAX-016 - 711 016-725) (PAL)',
  b'I\xd9\xc8\x11\x99.o\x90\n>\xf7\x95\xa7\x92\xe4\x8d\xd4Q\x87\xae': 'StarMaster - Kommando Galaxis (1982) (Activision, Alan Miller - Ariola) (EAX-016, PAX-016 - 711 016-725) (PAL) [fixed]',
  b"\x07\x92\x00F\xedUQ\x97\x97'A\x17\xf2\xf8\xc8\xc7\xe50\xf6#": 'Stargate (1984) (Atari, Bill Aspromonte, Andrew Fuchs) (CX26120) (PAL)',
  b'O\x87\xbe\x0e\xf1j\x1d\x03\x89"m\x1f\xbd\xa9\xb4\xc1k\x06\xe1>': 'Stargate (1984) (Atari, Bill Aspromonte, Andrew Fuchs) (CX26120) ~',
  b'3Y\xaazj_\xa2[\xea\xa3\xaeXh\xd0\x03MR\xde\x98\x82': 'Stargunner (1983) (Telesys, Alex Leavens) (1005) ~',
  b'\xbd\x7f\x00\x05\xfa\x01\x8f\x13\xed~\x94,\x83\xc1u\x1f\xb7F\xa3\x17': 'Steeplechase (1983) (Video Gems) (VG-03) (PAL) ~',
  b'\xc0\x9c\xc1\x7fSc\x0b\x07\xf5\xabg#\x16\xf5\x8b-\xfe\xdfA\x97': 'Steeplechase (Paddle) (04-15-1980) (Sears Tele-Games, Jim Huether) (CX2614 - 49-75126) (Prototype)',
  b'\xe5n\xf1\xc01=m\x04\xe2TF\xc4\xe3O\x9b\xb7\xed\xa8\xef\xac': 'Steeplechase (Paddle) (1980) (Sears Tele-Games, Jim Huether) (CX2614 - 49-75126) ~',
  b'}\x97\xd0\x14\xc2*.\xd3\xa5\xbcK1\x0fZ{\xe1\xb1\xd3R\x0f': 'Stellar Track (Stella Trek) (1981) (Sears Tele-Games, Robert Zdybel) (CX2619 - 49-75159) ~',
  b'E\xd3~7\xde\xa7c\x95g7\x11\x07p\xe3\xd3=\xe9\xc8s\xa5': 'Stone Age (1983) (CCE) (C-840) ~',
  b'c\x7f7)\xbd9\x94#\x9d\x1e\xf5\xe8h\xe8\xfc?W\xf3\xb2k': 'Stopp die Gangster (AKA Gangster Alley) (1983) (Quelle) (377.943 6) (PAL)',
  b'\xaf|\xc0e\xa4\x0b!Gp\xbfY|}\xa8]\xe9\xd3~\xb3A': 'Strahlen der Teufelsvoegel (AKA Atlantis) (1983) (Quelle) (463.860 7) (PAL)',
  b'\x96\xdb\xd3\xa9\xcb\x98\x1cBT\x0e\xac\x99\xb5\x86\x91c\xa6\xfb\x81\xc4': 'Strategy X (1983) (Gakken) (010) (PAL)',
  b'\x9dm\xec\xdan\x8a\xb2c\xf78\x0f\xf6b\xc8\x14\xb8\xcb\x8c\xaf4': 'Strategy X (1983) (Konami) (RC 101-X 02) ~',
  b'\x8d\xf4\x06`A\x91\xaa\xbeC\xd6#\xd4\x82\xff_8\xe6gF4': 'Strawberry Shortcake - Musical Match-Ups (1983) (Parker Brothers, Dawn Stockbridge) (PB5910) (PAL)',
  b"|\xa8\xf9\xcdt\xcf\xa5\x05\xc4\x93u\x7f\xf3{\xf1'\xffF{\xb4": 'Strawberry Shortcake - Musical Match-Ups (1983) (Parker Brothers, Dawn Stockbridge) (PB5910) ~',
  b'\x10\xf6kQn\xae[\xb9a\x93\xfd\xeaF\xaa\xfb6!:\xb3\xbe': 'Street Racer (Wheels) (Paddle) (1977) (Atari, Larry Kaplan) (CX2612, CX2612P) (PAL)',
  b'\xbf\xfb=A\x91l\x839\x86$\x15\x1e\xb0\n\xa2\xa3\xac\xd2:\xb8': 'Street Racer - Speedway II (Wheels) (Paddle) (1977) (Atari, Larry Kaplan - Sears) (CX2612 - 99804, 49-75103) ~',
  b',\xfe(\x0f\xdb\xb6\xb5\xc8\xcd\xa8\xa4b\r\xf1*QT\xe1#\xbe': 'Stronghold (1983) (CommaVid, Joseph Biel) (CM-009) ~',
  b'.\x19\xd7\xe1l\xf1v\x82\xb0C\xba\xaa0\xe3E\xe6\xfaE@\xe5': 'Stunt Cycle (Paddle) (07-21-1980) (Atari, Robert C. Polaro) (CX26157) (Prototype) ~',
  b':\xec~\xa8\xafr\xbb\xe1\x05\xb9\xd2\x90:\x92\xf5\xad+7\xbd\xdb': 'Stunt Man (AKA Nightmare) (1983) (Panda) (105)',
  b'\xcc\xd7_\x01A\xb9\x17en\xf2\xb8l\x06\x8f\xba28\xd1\x8a\x0c': 'Sub-Scan (Subterfuge) (1983) (Sega) (002-01) ~',
  b'5\xcc\xe0>\xec&`Y8\x81\xbbwZ\xc9\x07\xae\xdf\xf0s\x93': 'Sub-Scan (Unknown) (PAL)',
  b'\xb2+\xa7\xcb\xde`\xa2\x1e\xcb\xbe9S\xccJ\\\x0b\xf0\x07\xcc&': 'Submarine Commander (Seawolf 3) (1982) (Sears Tele-Games, Marilyn Churchill, Matthew L. Hubbard) (CX2647 - 49-75142) ~',
  b"*\xbck\xbc\xab'\x98_\x19\xe4)\x15S\x0f\xd5V\xb6\xb1\xae#": 'Subterranea (Tarantula) (1983) (Imagic, Mark Klein) (720112-1A, 03213) ~',
  b'G\xa8F\xbc\xe5?s\x9e\xc2\xcfWt\xfe\t\xd7\x0euQ\x11\xaf': 'Subterranea (Tarantula) (1983) (Imagic, Mark Klein) (EIZ-003-04I) (PAL)',
  b";SE\xfa\x19ND\x9b'\xd0\x14\xcc\x994\xb6bI=\xc9\xba": 'Summer Games (1987) (Epyx, Steven A. Baker, Tod Frye, Peter Engelbrite) (80561-00250) (PAL)',
  b'e\xf4\xa7\x08\xe6\xafV_\x1fu\xd0\xfb\xdc\x89B\xcb\x14\x9c\xf2\x99': 'Summer Games (1987) (Epyx, Steven A. Baker, Tod Frye, Peter Engelbrite) (80561-00250) ~',
  b'\x0f\xd7\xb4m~\x1c^\xee\xa6\x81\x08\xd0\xbb\x92\xec8r\xab=V': 'Super Action Pak - Pitfall, Barnstorming, Grand Prix, Laser Blast (1988) (HES - Activision) (223) (PAL)',
  b'\xb0f\xa6\x0e\xa1\xdf\x1d\xb0\xa5Rq\xc7`\x8b\x0e\x19\xe4\xd1\x8a\x1e': 'Super Baseball (1988) (Atari, Mimi Nyden, Joseph Tung) (CX26152)',
  b'L\x83s\x12X\x12n\xccpD\x15]\x17\x99V\x01\x10\x9aoi': 'Super Baseball (1988) (Atari, Mimi Nyden, Joseph Tung) (CX26152) (PAL)',
  b'\x8c\x878\xfa\xbd\xc3\xb4 F\xe3\xc9\xc2\x19\xf3\x8d\xd5\xb7\x05\x19R': 'Super Baseball (AKA RealSports Baseball) (CCE)',
  b'\xbd\xc2\xb5Qpr\xa1_eh<f.\xa54K\xcb\xa8\xa0\x92': 'Super Box (AKA RealSports Boxing) (CCE)',
  b'\xe3\x80\xe2C\xc6q\xe9T\xe8j\xa1\xa3\xa0\xbf\xeb6\xd5\xe0\xc3\xe2': "Super Breakout (Paddle) (1982 - 1981) (Atari, Carol Shaw, Nick 'Sandy Maiwald' Turner - Sears) (CX2608 - 49-75165) [a]",
  b'\xac*\xad!\x96\xc1U\xc1\xd8}oB\xfa\x88\x89\x18%\xf4\xfd\xe6': "Super Breakout (Paddle) (1982 - 1981) (Atari, Carol Shaw, Nick 'Sandy Maiwald' Turner - Sears) (CX2608 - 49-75165) ~",
  b':\x9e)iP\xaa\x8e\x93\xe7R\x18\xb9q\xb8aE\x16)/\x86': "Super Breakout (Paddle) (1982 - 1981) (Atari, Carol Shaw, Nick 'Sandy Maiwald' Turner) (CX2608) (PAL)",
  b'[w\x99\xe1\xc9\x8c\xa8ie\xfa\x8bs\xb8u\x00B\xed\xb7\tR': 'Super Challenge Baseball (208 in 1) (Unknown) (PAL)',
  b'h\xb3\x1e\x9c\xf6&\xf0)\x1d\xbc\xf08D\x7f\xbf\x87\x8b\xb7\xb3~': 'Super Challenge Baseball (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\xdf\xceMd6\xf9\x1d\x8d8_\x8b\x01\xf0\xd8\xe3H\x84\x00@{': 'Super Challenge Baseball - Baseball (Big League Baseball) (1982) (M Network - APh Technological Consulting, David Rolfe - INTV) (MT5665) ~',
  b'\\\x138\xecv\x82\x8c\xfaJ\x85\xb5\xbd\x8d\xb1\xc0\x0c\x80\x95\xc30': 'Super Challenge Football - Football (Pro Football) (1982) (M Network - APh Technological Consulting, Ken Smith - INTV) (MT5658) ~',
  b'>\xa9g\x937@\xda6\xf4|\x04\xb2iX}\xa14\xa9<4': 'Super Cobra (1983) (Parker Brothers, Mike Brodie - Roklan, Paul Crowley) (931505) (PAL)',
  b'\xba\xc0\xa0%e\t\xf8\xfd\x1f\xee\xa9=t\xbaL}\x82\xc1\xed\xc6': 'Super Cobra (1983) (Parker Brothers, Mike Brodie - Roklan, Paul Crowley) (PB5320) ~',
  b'tS\xf7D\xe2P\xbd\xa7`\x1e\x1e(\xb3\x0b}\xbc\xba\x8b]\xf1': 'Super Ferrari (AKA Enduro) (Rainbow Vision - Suntek) (SS-011) (PAL)',
  b'\xea\xcakGO\xd5R\xabJ\xafuRf\x18\x82\x81e\xa9\x194': 'Super Football (1988) (Atari, Douglas Neubauer) (CX26154) ~',
  b'\xfa\x8d\xb9T\xef\xf40.\xc5\x18\xaa\xf7\xa4w\xa0s\xc1\x96z\x08': 'Super Football (1988) (Atari, Douglas Neubauer) (CX26154, CX26154P) (PAL)',
  b'\xc8\rIz\x99\xb2\xfe\x1e<\xe6\xeac_q\xa23~g\xc3\xd5': 'Super Futebol (AKA RealSports Football) (CCE)',
  b'HM\x97\xe7\x8f\xf82\xed\xcaS\x82\xf3\x1e\x08D\x97\xe5\x91\xccr': 'Super Futebol (AKA RealSports Soccer) (CCE)',
  b'\xc0x\xeb$\xcc\x8c\x9aP@\x89\x14~\x80XjN\xdb0>\x0e': 'Super Hit Pak - River Raid, Sky Jinks, Grand Prix, Fishing Derby, Checkers (HES - Activision) (PAL)',
  b'P\x16M\xdd\xba\xe3\x17+\xfa:n\xd2\xaa\xf3\xd58~\xd71M': 'Super Kung-Fu (1983) (Xonox - K-Tel Software - VSS, Robert Weatherby) (6230, 6250) (PAL)',
  b'\xb7\xdf#)j\xf4\x95\xbbE<\xef\xb8\xe9.\xd8NK\xc4\\\x14': 'Super Soccer (AKA RealSports Soccer) (Digivision)',
  b'%A\xd8\x7f\xa9\x02\xf0N\xf3s\x825a\xf4\x97?\xfa\xa3I@': 'Super Tenis (AKA RealSports Tennis) (VGS)',
  b'\xba\xb9)~\xe2\xba\xf7p\xf4d\x9b\x85U\xe5.\xdfW\x9da\x91': 'Super Tennis (AKA RealSports Tennis) (1983) (CCE) (C-1005)',
  b'\xed\x906X\x00\xdb\x94g\xfd\xa6db\xb9\xac\x83\xeb\x04\xa4\x0e\xab': 'Super Tennis (AKA RealSports Tennis) (Tron)',
  b'\xb7\x9cOQ\xd4\x80\t\xe4.\x8d\xec&\xa9\x13\x05\xac\xfe\x7fC\xa8': 'Super Voleyball (AKA RealSports Volleyball) (CCE)',
  b'9T\xfegs\x14\x02\xadr\x81]\x81\x9b\xb1\xfc\x1f<{\xbf3': 'Super-Cowboy beim Rodeo (AKA Stampede) (1983) (Quelle) (874.254 6) (PAL)',
  b'\x82\xa7\x04E3\xf9Uj\x9c\xd0e\x88\x1c\xcb\x9e\xd3\x1f\x14\xdf\xd0': 'Super-Ferrari (AKA Enduro) (1983) (Quelle) (402.272 9) (PAL)',
  b'"\xa6\x128\xc2\x18\xa8?\xdf\x0c\xceY\x86$\x8d\x97W~\xb4g': 'Super-Ferrari (AKA Enduro) (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\xfb\x84\x86e\xf7\xfe\x80)\xf8\x0b\x1c\n\xaeYr\xf2b\x85+\xf7': 'Super-Ferrari (AKA Enduro) (Unknown)',
  b'Ai\x1d\x8fV\xbf*\xdb<\x0f\x12~r\x95c\xc9\xfb\xa5\xd9\x9d': 'Supercharger BIOS (1982) (Arcadia Corporation) (PAL)',
  b'\xcd\x9d\x03\nY\xf5\xbd\xcc\x88\xf1\xf8\nl\xc9\xfc,\xd92\xf7\xdb': 'Supercharger BIOS (1982) (Arcadia Corporation) ~',
  b'\xf0\x0b\xcb\xda=p\xc0\xc90&#O\xc7u\xda\xf8\x1d\xb8V\x9a': 'Superman (1979) (Atari, John Dunn - Sears) (CX2631 - 49-75152) [fixed] ~',
  b"\xb9\xde\xe0'\xc8\xd7\xdd*F\xbe\x11\x1a\xb0\xb86<\x1b\xec\xc0\x81": 'Superman (1979) (Atari, John Dunn - Sears) (CX2631 - 49-75152) ~',
  b'\x84\x96\x85\x17\xda\x87\x9f0~\x1fm-\xe3F++\x0b\xdb\x95\xf2': 'Superman (1979) (Atari, John Dunn) (CX2631, CX2631P) (PAL)',
  b'|K<\xceP\xed\x1b_\x18\x17\xc8\xfb\xf9\xeb\xc4./\x88U\xe8': 'Superman (1983) (CCE) (C-857)',
  b'\xa9tg=\xcc\xfbxm\xa6\x15\r\xf8\xd4XJU\x95\x08\xa6\xde': 'Superman (Unknown) (PAL)',
  b'i\xb7\xff\x92A\xa1\xb5\x06\xadS\x96\xcd8E1\xaf\xf7\x08\xe7[': 'Superman (Unknown) (PAL) [a]',
  b'\xbbV\xb4\x1ci\xad?\xfb\x17\xf0\xd7\xc2sa\x98*2n\x04i': "Surf's Up (Joyboard) (1983) (Amiga - Video Soft) (3125) (Prototype) (4K)",
  b'\x9c\x11\xb5\xb0\xf7[\x11\xee\xa1\x16\xcc\x80\xf6\rI\xe2\x02\x89c{': "Surf's Up (Joyboard) (1983) (Amiga - Video Soft) (3125) (Prototype) [a1]",
  b"\xf9\x03\xa3\xe5\xedECV\xda}'\xd3\xc6\xbd\x1e\x15\xf1\xd67\xa6": "Surf's Up (Joyboard) (1983) (Amiga - Video Soft) (3125) (Prototype) [a2]",
  b'\xcf\x84\xe2\x1a\xdaUs\rh\x9c\xfa\xc7\xd2n"\x951r"\xbc': "Surf's Up (Joyboard) (1983) (Amiga - Video Soft) (3125) (Prototype) ~",
  b'\xe7T\xc8\x98\\\xa7\xf5x\x0c#\xa8Ve`\x99\xb7\x10\xe8\x99\x19': "Surfer's Paradise - But Danger Below! (1983) (Video Gems) (VG-02) (PAL) ~",
  b'\x94MN*>\xdd1 \xb8YZ|j\xad\x85|\xff\x04E4': 'Surround (1977) (Blockade) (Atari, Alan Miller) (CX2641, CX2641P) (PAL)',
  b'b\xcc\x0b\xc42\xc8\x1e\x9f\xcb\xc8\xa1\x9b\xbc\xb6\x85Y\xb5H\xff\xdf': 'Surround (32 in 1) (1988) (Atari, Alan Miller) (CX26163P) (PAL)',
  b'\xb7\x98\x83s\xb8\x19\x92\xd0\x80VV\r\x15\xd3\xe3-\x9d8\x88\xbc': 'Surround - Chase (Blockade) (1977) (Atari, Alan Miller - Sears) (CX2641 - 99807, 49-75105) ~',
  b'\xbc\xdecXP\x92\xf8NC}\xd5\x7f\x83\xbdg\xbdf\xb1VF': 'Survival Run (1983) (Data Age) (Prototype) ~',
  b'l\x99;Lp\xcf\xed9\x0f\x1fCo\xdb\xaa\x1f\x81I[\xe1\x8e': 'Survival Run (Cosmic Commander) (1983) (Milton Bradley Company - Renaissance Technology, Ty Roberts) (4362) ~',
  b'\xd8\x9f\xf9\xb3\xf5H\xfaIl!\\\xdeL\x18\xbc/<\xce\xf3\xef': 'SwordQuest - EarthWorld (Adventure I, SwordQuest I - EarthWorld) (1982) (Atari, Dan Hitchens, Mimi Nyden) (CX2656) (PAL)',
  b'=\xebe\n\xe2k\x86\xe2P\xae\xa8\xf7\xcam\x06t\xe6I\x8e\xbb': 'SwordQuest - EarthWorld (Adventure I, SwordQuest I - EarthWorld) (1982) (Atari, Dan Hitchens, Mimi Nyden) (CX2656) ~',
  b"\xa6\xfax\xed\xffk{a\xc7\xb5'\xfe{ix\xeb7\x10\x81\xed": 'SwordQuest - FireWorld (Adventure II, SwordQuest II - FireWorld) (1983) (Atari, Tod Frye, Mimi Nyden) (CX2657) (PAL)',
  b'\\<\xf9v\xed\xbe\xa5\xde\xd6f4\xa2\x84x\x7f\x96V\x16\xd9~': 'SwordQuest - FireWorld (Adventure II, SwordQuest II - FireWorld) (1983) (Atari, Tod Frye, Mimi Nyden) (CX2657) ~',
  b'V\x9f\xcbg\xca\x16t\xb4\x8e/:.z\xf7\x07z7D\x02\xde': 'SwordQuest - WaterWorld (1983) (Atari, Tod Frye) (CX2671) ~',
  b'\rYT["\xe1P\x19\xa3=\xe1i\x99\xa5}\xae\x1f\x99\x82\x83': 'Swordfight (Sword, Swordfighting) (1983) (Intellivision Productions - M Network, Patricia Lewis Du Long, Stephen Tatsumi) ~',
  b"\x0e\xc5\x8a:Z'\xd1\xb8*_\x9a\xab\xab\x02\xf9\xa88{iV": 'TRON - Deadly Discs (TRON Joystick) (1983) (M Network - APh Technological Consulting, Jeff Ronne, Brett Stutz - INTV) (MT5662) ~',
  b'+\n\xa1\xc0\x98\xdc\x1fx\xc0@R\x80y\xd2\x13\xfa\x7fo0\xea': 'Tac-Scan (Canal 3 - Intellivision)',
  b'U\xe9\x8f\xe1K\x07F\x074x\x1aj\xa2\xf4\xf1dh0\xc0\xaf': 'Tac-Scan (Paddle) (1983) (Sega, Jeff Lorenz) (001-01) ~',
  b'CC\x8f]\x7f\x03\x8a\x9b\xf1N{c"\x8d.iH\t\xe0\xfe': 'Tac-Scan (Unknown) (PAL)',
  b"K\xfaD'\xf8#C>b\x8e\x7f\xd0B\x00\xf9\x9a\xb4&#6": 'Tank Brigade (AKA Tanks But No Tanks) (1983) (Panda) (101)',
  b'e\xa0\xd7\x89,y\x9e`\xa2\x01\xb1\xb1\xb5\xca\x15j(\xaf\x0e\xe9': 'Tank City (AKA Thunderground) (Funvision - Fund. International Co.)',
  b'\x13\xa9\xd8l\xbd\xe3*\x14x\xef\x0c~\xf4\x12B{\x13\xbdb"': 'Tanks But No Tanks (1983) (ZiMAG - Emag - Vidco) (707-111 - GN-030) ~',
  b'\xac\x10\x19;\xd7\x86\xca<*sX\t\xc9]\x13\x08\x98\xe6\xc7\x84': 'Tanks But No Tanks (Unknown) (PAL)',
  b'><Fl\xd1*\x86\xae\xb0<\x19\x9d\x831\x02\xa0r\xf3?\x96': 'Tanks War (AKA Tanks But No Tanks) (1983) (Home Vision - Gem International Corporation - VDI) (VCS83135) (PAL)',
  b'\xbfIQ\x97\xfdM\x9c\x84\xe05\xb6\xbe9zX#\x8b\xff\xe5\x1b': 'Tapeworm (1982) (Spectravision - Spectravideo) (SA-204) (PAL)',
  b'\xee\x8b\xc1q\ng\xc3>\x9f\x95\xbb\x05\xcc=\x8f\x84\x10\x93\xfd\xe2': 'Tapeworm (1982) (Spectravision - Spectravideo) (SA-204) ~',
  b'c>\xf65\x96V\xab\xe9b{\x9d[\x97\xec\xcb\x98X\xa4\xdc\x8c': 'Tapeworm (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\xe9\x86\xe1\x81\x8et{\xeb\x9b3\xceM\xff\x1c\xdckU\xbd\xb6 ': 'Tapper (1984) (Sega - Beck-Tech, Steve Beck) (010-01) ~',
  b'6\x13\x95\x02\xd3\xb4?\xc3Xt\xaf\x8b?\x06\xc6\x9d]\xc4\xfa\x97': 'Targ (1983) (CBS Electronics - VSS) (80110) (Prototype)',
  b'\xce\xef\xab\xdf\xd9}-x<\xb3\xd1F\x04\xf9\xc0Y\xa9^\xeaR': 'Target Practice (Carnival) (AKA Carnival) (1983) (CCE) (C-833)',
  b'/a\x14\xe0"\xbfW\xffXE\rv^`\xd2\xf3\xe3\x91\x042': 'Target Practice (Carnival) (AKA Carnival) (1983) (CCE) (C-833) [a]',
  b'\xba\xe77\x00\xbae2\xe9\xe6A[dq\xd1\x15\xbd\xb7\x80Td': 'Task Force (AKA Gangster Alley) (1987) (Froggo) (FG1003)',
  b'z\xafk\xe6\x10\xban\xa1 [\xdd^\xd6\x088\xcc\xb8(\rW': 'Tax Avoiders (1986) (American Videogame - Dunhill Electronics, Darrell Wagner, Todd Clark Holm, John Simonds) ~',
  b"'%\xeey\xa5WBC0.z=2\xdeC6\x96\xf5\x97\xe5": 'Taz (Tazz) (06-15-1983) (Atari, Jerome Domurat, Steve Woita) (CX2699) (Prototype) (PAL)',
  b'\xff\xa7\x9f\x03\xbb\x93\xd5\x1c\xe6\x80\x1dgZb!\xdc\xd0\xdfMa': 'Taz (Tazz) (07-13-1983) (Atari, Jerome Domurat, Steve Woita) (CX2699) (Prototype)',
  b'g\x0b=\xed\xe2p,\x96IF6\xb6\xd7\xf4\xae\x81\x12P\x1c\xb4': 'Taz (Tazz) (07-15-1983) (Atari, Jerome Domurat, Steve Woita) (CX2699) (Prototype)',
  b'\x0b\xcf\xd8\xaal3\xd6\x01\xc2$\x08P\xf1\x18\x84\x07\xe2\xf4s1': 'Taz (Tazz) (1983) (Atari, Jerome Domurat, Steve Woita) (CX2699) (Prototype)',
  b"\xc8^\xf5\x97\xd5n\xe0\x8d)\x15\x97\xec\xf6\x80\x84'\xd90Y\xc9": 'Taz (Tazz) (1983) (Atari, Jerome Domurat, Steve Woita) (CX2699) (Prototype) [a]',
  b'\xfaJ\xeeyHp6ej\xab\xb42\xd7\xc6\xe1>\xc2\x1e:<': 'Taz (Tazz) (1984) (Atari, Jerome Domurat, Steve Woita) (CX2699) ~',
  b'E\xb1\x8e\xfdWa\xc6\x13\x89PQ\x1fX\x0e\xbe\x10GA\x9eK': 'Teddy Apple (AKA I Want My Mommy) (1983) (Home Vision - Gem International Corporation - VDI) (VCS83137) (PAL)',
  b'~\xfc\x0e\xbe3M\xde\x84\xe2_\xa0 \xec\xdeO\xdd\xcb\xea\x9e\x8f': 'Telepathy (Mindlink Controller) (1983) (Atari, Dan Oliver) (Prototype) ~',
  b'\xd1D$<\n\xfcT\x1d\xf0@\x06T\xae7\xf3\x8f\x0b\xdbYg': 'Teller-Jonglieren! (AKA Dishaster) (1983) (Quelle) (685.996 1) (PAL)',
  b'\xbfMW\x0c\x1cs\x8aMm\x00#~%\xc6.\x9c2%\xf9\x8f': 'Tempest (01-05-1984) (Atari, Michael Kosaka, Carla Meninsky) (CX2687) (Prototype) ~',
  b'\\\x8a\xb1\x91\xca7\x8d\x9e\xd7w\xe9\xa5}\xa5J\x08\xca\xae>\xa5': 'Tennis (1981) (Activision, Alan Miller - Ariola) (EAG-007, EAG-007-04I, PAG-007 - 711 007-720) (PAL)',
  b'\xf6\xcd2I\x03\xf67\x19\xb8\xdb\xbc\xad=>*@XcO\xd5': 'Tennis (1983) (CCE) (C-858) (4K)',
  b'$p\x15&\x1d;f\xd3\x1a\x0be&\xe2rq\xab\xc8\x11(Q': 'Tennis (32 in 1) (1988) (Atari) (CX26163P) (PAL)',
  b'\x06.\xd7\xdbb)\xca\x94.\x83\xe6\x89\xe47\x9e\xf1\xd1\x02F\x9d': 'Tennis (Canal 3 - Intellivision) (4K)',
  b'\xed\xc2\x06\xb9\xcb\xa5H_\xa8\x99\x8c\x87\xac\x05\xe0\x94D\xf4\xd3\xe0': 'Tennis (Dactari - Milmar)',
  b'\xf6\x15r\xb5t^\xee\xfeK\xa9Io\x0e]\n\x1b\xfeEe\xa1': 'Tennis (Hack) (208 in 1) (Unknown) (PAL)',
  b'V\xac[Zc\xc8+\xfc\x11\xa7|\x90d\xf0o\x8d\x9a\xa8yL': 'Tennis (Pet Boat) (PAL)',
  b'x=\x81\xf4\x10v\x98\x8d\xac\xd8\xde\x9a\xb4w\x9bU\xa1\xce\x94\x97': 'Tennis (Star Game) (007)',
  b"|\xb7iI\xb0\x85\x00\x8e\x9e\x86'\x9b\xff\x98\x03\xc0LU\xd09": 'Tennis (Tennis Game) (AKA Tennis) (1983) (Home Vision - Gem International Corporation - VDI) (VCS83107) (PAL) (4K)',
  b'W}\x9b6\xf7\xd5\xe4\xfe\x8f\x85j:\xca\xed\xb2\xdb\xe3\xcb\x04d': 'Tennis (Unknown)',
  b'\xbf\xdc \x04\xbes_%U\xfe\x93u\x08\x86\x84\xe3Q\x86\x9cY': 'Tennis (Zirok)',
  b'=0\xe7\xedaqh\xd8R\x92=\xef \x00\xc9\xc0\xa8\xb7(\xc6': 'Tennis - Le Tennis (1981) (Activision, Alan Miller) (AG-007, CAG-007) ~',
  b'\x1f\x16`":\xbe\xee1\x13\xc3\xf7[\xf4\xd9\xcb\xb3\xb1\xcfI\xe4': "Texas Chainsaw Massacre, The (1982) (Wizard Video Games, Bob Davis, Robert H. O'Neil) (Prototype) ~",
  b'qq"\xa4\x18K\xc8\xdbA\xe6Z\xb7\xc3i\xc4\x0b!\xc0H\xd9': 'Texas Chainsaw Massacre, The (1983) (Wizard Video Games - VSS, Ed Salvo) (008) ~',
  b'1\x8c@=5\x82\xed\n\x90\xb5Q\xbc\x18\x16\x99\x8c\xf9\x9ef\xe7': 'Threshold (1982) (Tigervision, Warren Schwader - Teldec) (7-003) (PAL)',
  b'\x9aR\xfa\x88\xbdtU\x04O\x00T\x8e\x96\x15E!1\xd1\xc7\x11': 'Threshold (1982) (Tigervision, Warren Schwader) (7-003) ~',
  b'/\x12\xc7mt# U\xb7gk\x1b\xe4.\x90\xcf\xcb\x1eMK': 'Thunderground (Canal 3 - Intellivision)',
  b'<\xc8\xbc\xc0\xffQd043\xf4i\xaaM\xa2\xeb%m\x1a\xd0': 'Thunderground (Underground) (1983) (Sega, Jeff Lorenz) (003-01) ~',
  b"4\xc3\x16\x86\x9e\x87I\xe34\xeai\x89'H\xb2\x81\xce6\xf7\xfc": 'Thunderground (Unknown) (PAL)',
  b'S\xeep\xd4\xb3^\xe3\xdf?\xfb\x95\xfa6\x0b\xdd\xb4\xf2\xf5j\xb2': 'Thwocker (04-09-1984) (Activision, Charlie Heath) (Prototype) ~',
  b'B\x9c<!\xbf\x89\xdc\x08\x89-\x16\xc4 \xb7+\x0c\x88\xc9\xe0<': 'Time Machine (AKA Asteroid Fire) (1983) (Goliath - Hot Shot) (83-112) (PAL)',
  b'8sXQId\xd0\xb6\xb5_\x941WjY\xb5Xi\xf7\xab': 'Time Pilot (1983) (Coleco - Woodside Design Associates, Harley H. Puthuff Jr.) (2663) ~',
  b'z\x05\xa7}d\x8e5lq\xadm<2\x9a\xabH\xd7\xd8\x8en': 'Time Race (AKA Space Jockey) (1983) (Goliath - Hot Shot) (83-212) (PAL)',
  b'\xde\xe8\xaf\xa6$\xc7\x16\xb7\x14_;\x927\xa2\x11i\xd8\x9b\x1a\x1a': 'Time Race (AKA Space Jockey) (Rainbow Vision - Suntek - Sunteck Corporation) (SS-001) (PAL)',
  b'\x90\x04\x1bv\xb8s\xd2\x14\'jV|\xca:\xf5;\xfa2"\xac': 'Time Race (AKA Time Warp) (Funvision - Fund. International Co.) (PAL)',
  b'YM\xb9\x13\xe2u\t\xfe$L\\\x90\x0b]\xbfY\x91\xd1\x9a>': 'Time Warp (1982) (Funvision - Fund. International Co.) (PAL) ~',
  b'W\xf4\xe0\xea\x1c\x81\xea&3@\xbefkqS\x16\xad\xff\xf4\x91': 'Time Warp (1983) (CCE) (C-845)',
  b'\x124\t\x88\x85*\xd7%\xc2\xac\xeeWR\xa9{\xee\xe8\xe6 \x81': 'Time Warp (Unknown)',
  b'.\x8eb\xa5\xdcE\xc7\x80auv\xad\x8a\xca\xf5\xcb\xf8\xd3m-': 'Time Warp (Unknown) (PAL)',
  b'rX\xe8\xdaQw\xf3u\x83\xef\xca\x1b\xf4>Y\x87\xc9\x84]L': 'Time Warp (Zellers)',
  b'\x97\x9d\x9b\x0b\x0f2\xb4\x0c\n\x05h\xbee\xa0\xbc^\xf3l\xa6\xd0': 'Title Match Pro Wrestling (1987) (Absolute Entertainment, Alex DeMeo) (AG-041-04) ~',
  b'\xb5\xe9\xe2\x07\xd0$V\xe7\xec\xcf:\xe9\xb7\xdf`w{\xb3\xfbg': 'Title Match Pro Wrestling - Pro Wrestling (1987) (Absolute Entertainment, Alex DeMeo) (EAZ-041-04I) (PAL)',
  b'\xf56\xde\xa3\x9fd\xd1\xfdr\xe4\xee\n\x989N\xda9\xdd\xc3\xda': 'Tom Boy (AKA Pitfall!) (Rainbow Vision - Suntek) (SS-005) (PAL)',
  b'9j\x96\xc9\x1c\t\x92\xf8ri3v`\x99\xa3\xf2\xc0f\x9b1': 'Tom Boy (AKA Pitfall!) (Unknown)',
  b"o,\xce\x077\xf6\x03\x10\x89(\x01\xc4&s\xee'B\x0f\x96\xdc": "Tom's Eierjagd (AKA Play Farm) (1983) (Quelle) (731.503 9) (PAL)",
  b'H\x9c\x9bW%5r\x1a\x05\x16\xa2\xb7Y\xe0\xb9\xc7\xf7\xa5\xb3\xcc': 'Tomarc the Barbarian (Thundarr the Barbarian) (1983) (99007, 6240) (Xonox - K-Tel Software - Product Guild, Anthony R. Henderson) ~',
  b"['B(\x1f\xea\x96\xabj:/0\xe6v5+\xcfBC\x90": 'Tomcat - The F-14 Fighter Simulator (1988) (Absolute Entertainment, Dan Kitchen) (AK-046-04) ~',
  b'\xd8*\xc7#}\xf5L\xc8h\x8e0t\xb5\x843\xa7\xddk}\x11': 'Tooth Protectors (1983) (Camelot - DSD, Michael Doherty, Clyde Hager - Johnson & Johnson) ~',
  b'\xe8S\x95N\xae}\x0bx9x\x7f\x05i\x88\x8as\x9eV\xcf\xaf': 'Top Gun - Air Patrol (AKA Air Raiders) (1983) (Quelle - Otto Versand) (626.502 9 - 746381) (PAL)',
  b'Y\xb5\xf6\x1a\x8a`\t\xa2\xac\x99\x03\xfb=r\xee\x81\x7f\xd8I\xbb': 'Topy (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co)',
  b'\xb3D\xb3\xe0BDz\xfb\xb3\xe4\x02\x92\xdcL\xa0c\xd5\xd1\x11\r': 'Towering Inferno (1982) (U.S. Games Corporation - Western Technologies, Jeff Corsiglia, Paul Allen Newell, Tom Sloper) (VC1009) ~',
  b'\xda\xb9\x12M\xe3\xc1\xd7v\x0c\x97\x96P\xd4\xd8;\x1c\x0el\x14-': 'Towering Inferno (208 in 1) (Unknown) (PAL)',
  b'\x00ZjS\xf5\xa8V\xf0\xbd\xbc\xa5\x19\xaf\x1e\xf26\xaa\xa1IM': 'Track and Field (Los Angeles 1984 Games) (Track & Field Controller) (1984) (Atari - GCC, Jaques Hugon, Seth Lipkin) (CX26125, CX26127) ~',
  b'\x9a\x99\x17\xd8#b\xc7{M9oV\x96b\x19\xfc$\x8e\xdfG': 'Treasure Below (1983) (Video Gems) (VG-05) (PAL) ~',
  b'F\x1d\x91\xde\xe9Y\x95\x8f\x91\x07\xf1\xd7.qj\x17\xaf<\x1c\x92': 'Treasure Hunting (AKA Pitfall!) (Funvision - Fund. International Co.)',
  b"\xd1C~)\x1f\xbc\x19'\xfc\xce\x14\xab\xc2\x1aX\xa4#\xe1Sh": 'Treasure Island (AKA Treasure Discovery) (Suntek) (SS-026) (PAL)',
  b'\x86\xc5c\xdb\x11\xdb\x9a\xfb\xff\xbds\xc5^\x9f\xae\x9b/i\xbeO': 'Trick Shot (1982) (Imagic, Dennis Koble) (720000-100, 720100-1B, IA3000, IA3000C) ~',
  b'y\x9e$\xa0\xcfL\x06\x95\xe3N\x94\x84\x7f"7\x06\xa4\x0f\xcc\xac': 'Trick Shot (1982) (Imagic, Dennis Koble) (720100-2B, IA3000P) (PAL)',
  b'\x7f\xfe\xc0\xee\xb2`\x85x#\x8e\xf9\x8e\xa2\x87\xd7\xb4\x87\xe9\xd4\xff': 'Tuby Bird (AKA Dolphin) (208 in 1) (Unknown) (PAL)',
  b'\xaf .\xe5\x15\xb4\x11\x13f\x0b\xdfE\x00\x80!\x08\xb1/\x85\xa3': 'Tunnel Runner (Black Box) (1983) (CBS Electronics, Richard K. Balaska Jr., Andy Frank, Stuart Ross) (4L 2520 5000) (Prototype)',
  b'\xfc\x1a\x0bXvZ}\xcb\xd8\xe35b\xe1\x07M\xdd\x9e\n\xc6$': 'Tunnel Runner (Black Box) (1983) (CBS Electronics, Richard K. Balaska Jr., Andy Frank, Stuart Ross) (4L 2520 5000) ~',
  b'\x8a\xb3[\xc1r\x14\xdf;+\xaej2\xb9n_\x1dJ(\xca\x17': 'Turbo (1982) (Coleco - Product Guild - GMA, Michael Green, Anthony R. Henderson, Gary Littleton) (2455) (Prototype) ~',
  b'\x11b\xfeF\x97\x7f\x01\xb4\xd2^\xfa\xb8\x13\xe0\xd0^\xc9\n\xea\xdc': 'Turmoil (1982) (20th Century Fox Video Games - Sirius Software, Mark Turmell) (11007) ~',
  b'\xb6\x8aE^~\xc4\xab\xfa\xf1\xbe\xb9\x05l\xaf\xd5*\xc5\x9eT\xc0': 'Turmoil (Unknown) (PAL)',
  b"\xd8\r\xa5@\xbb\x1f\xbb\xe5'\x1e\x07B\xcf/j\xa6\rq\xbd\x97": 'Turmoil (Zellers)',
  b'\xd2\x05\xe4\xc7<d3^\x18x\xb7\x8f\x97\x0c_\xb5 `\xc7\xdb': 'Tutankham (1983) (Parker Brothers, Dave Engman, Dawn Stockbridge) (931509) (PAL)',
  b'\xa4\xd6\xba\xc8T\xa7\r,U\x94i2\xf1Q\x1c\xc6-\xb7\xd4\xaa': 'Tutankham (1983) (Parker Brothers, Dave Engman, Dawn Stockbridge) (PB5340) ~',
  b'LA7\x9f\r\xd9\x88\x03\x84\xfc\xbbF\xba\xd9\xfb\xaa\xf1\t\xa4w': 'UFI und sein gefaehrlicher Einsatz (AKA Go Go Home) (1983) (Quelle) (732.174 8) (PAL)',
  b'\x04\xe12\xf9\x01\xf8,\x89\xeb\x1a\x1f*K\xa7\x9e\x97\xe8\xb9\xd7\x86': 'UFO (AKA Space Jockey) (32 in 1) (1988) (Atari) (CX26163P) (PAL)',
  b'\xd6\x8d\x9f\xfc\xb6\xdb\x02\xea\x0e\x92\x99\x19/T\x8e=e\x94\xc0\xf3': "UFO Patrol (AKA X'Mission) (Suntek) (SS-031) (PAL)",
  b'\xd2\xbc\xc7g\xb1p\xb0\xfc\n*\t\x03\xf43z\x82\x16#\xab\xb4': 'Ungeheuer der Tiefe (AKA Skindiver) (1983) (Quelle) (719.013 5) (PAL)',
  b"P\xfb\x0e'\xee\x99J\x9eQ\xb6?\x84CJ%\xa0\x03n\xeab": 'Universal Chaos (AKA Targ) (1988) (Telegames - VSS) (7062 A305) (PAL)',
  b'(a\x06\xfb\x9750\xbc>*\xf12@\xf2\x8cK\xcb7\xe6B': 'Universal Chaos (AKA Targ) (1988) (Telegames - VSS) (7062 A305) ~',
  b'=\xa6\xa2\xcci\x99E\xf7\x08\xda\xc4\xe8\x80\xffn\x08\\c[\xbd': 'Unknown 20th Century Fox Game (1983) (20th Century Fox Video Games) (Prototype) ~',
  b'\x8f\x0c+\xe5\x12\xd29\xf2^\xe4q\xd3>\xcb\xe2\xe5\x89i\x99\xbd': 'Unknown Activision Game (10-22-1982) (Activision) (Prototype)',
  b'\xd9<\x86\x97\r=\xf2\x93&\xac\xc3\\\x0c\xe6g\x9b\x83\xc2\xaf\xf2': 'Unknown Activision Game (10-29-1982) (Activision) (Prototype) ~',
  b'Q\x12\xed\xc1\x88\x85\xb2\xfb\xdeJ\xdc!\x11\x17l\x87\xb4\x1f\x03R': 'Unknown Datatech Game (2600 Screen Search Console) (Jone Yuan Telephonic Enterprise Co) ~',
  b'k\xdeg\x1aP3\n\xf1T\xed\x15\xe7?\xdb\xa3\xfaU\xf2=\x87': "Up 'n Down (1984) (Sega - Beck-Tech, Steve Beck, Phat Ho) (009-01) ~",
  b"\xe4\x1c7\xd2\xf5\xcc\xa3\xc7\x88\xf9\xd4z\x9f\xa6B\xe0\x1d5\xa8'": 'Vanguard (1982) (Atari - GCC, Dave Payne) (CX2669) (Prototype)',
  b'\x01G]\x03|\xb7\xa2\xa8\x92\xbe\t\xd6p\x83\x10/\xa9\x15\x92\x16': 'Vanguard (1983) (Atari - GCC, Dave Payne) (CX2669) ~',
  b'\x1aD\x91\x8f\xa7d\xda\x8d\xde\x97\xa7T:\x18\x1c\xcfgV\x86\x8b': 'Vanguard (1983) (Atari - GCC, Dave Payne) (CX2669, CX2669P) (PAL)',
  b'^\x01\xdb\xa1\xf4#\xbb;2\xb5\x1b\x15\xfc1\x82}%\xd4\xf8\xdf': 'Vanguard (CCE)',
  b'D\x92\x0b\xd0Y\x98s\xc7z\xb1\x7f6\xffI\xe2\x15\xcb\xf9)\x17': 'Vanguard (VGS)',
  b'\x81\xf4\x85\xb5EDP\x80\xe0\t\xad\xf0\xf2\xe9=\xd4\xf1k\x92\xce': 'Venetian Blinds Demo (1982) (Activision, David Crane) ~',
  b'q\xfc\x0c\xf5\xca\xad\x1c\xea\xc6\xc3\x92<\xfe\xb4P\xb2\x8a)\xa8(': 'Venture (1982) (Coleco, Joseph Biel) (2457) (Prototype) (2K)',
  b'\x03\x05\xdf\xc9\x9b\xf1\x92\xe54R\xa1\xe0@\x8c\xcc\x14\x89@\xaf\xcd': 'Venture (1982) (Coleco, Joseph Biel) (2457) ~',
  b'wR6\x03u\xac\x7f5\xc6I\xeb\x81BH\x97\xdbVan=': 'Venture (1983) (CBS Electronics, Joseph Biel) (4L1802, 4L1803, 4L1804, 4L2278) (PAL)',
  b'c\xaf\xb1\xb7Q\x14\xd7\x08\r\x978\xea\x1d\x90(*\xf5\xd1\xfa\xde': 'Venture (1987) (Atari) (CX26145)',
  b'\xdc4{\xd6\x05\xd8\xd0\xa0\xca:G\xe7\x07\xa0\xd5\x8a\xde\x82/\xd3': 'Viagem Espacial (AKA Star Voyager) (Dismac)',
  b'$\xc1\xc6\xe1\xa4\x95\x13{\xecS\x81\x8c,\xd8\xbd\x83\xc83m!': 'Video Checkers - Atari Video Checkers (1980) (Atari, Carol Shaw) (CX2636, CX2636P) (PAL)',
  b'\xba\xba\xe8\x8a\x83+v\xd8\xc5\xafn\xa6;\x8f\x10\xa0\xda[\xb9\x92': 'Video Checkers - Checkers - Atari Video Checkers (1980) (Atari, Carol Shaw - Sears) (CX2636 - 49-75156) ~',
  b'\x04>\xf5#\xe4\xfc\xb9\xfc/\xc2\xfd\xa2\x1f\x15g\x1b\xf8b\x0f\xc3': 'Video Chess (Computer Chess) (1979) (Atari, Larry Wagner, Bob Whitehead - Sears) (CX2645 - 49-75181) ~',
  b'\xc7\xc7\x92\xc9\x1f\x9f\xa7\xc3\xe2\x91\xd0\xe0\xf0\xddq\xa7\x18\x13i\xa5': 'Video Chess (Computer Chess) (1979) (Atari, Larry Wagner, Bob Whitehead) (CX2645, CX2645P) (PAL)',
  b'\x9d\xd3\x87\xb0\xea+\x96\xde\x89i\x9a\xe6oT\xad\xf9\xd6sg\xb8': 'Video Chess (Unknown) (PAL)',
  b'\x15T\xb1F\xd0v\xb6Gv\xbfI\x13l\xea\x01\xf6\x0e\xeb\xa4\xc1': 'Video Jogger (Foot Craz) (1983) (Exus Corporation) ~',
  b'\x08\xc7\xc4\xab[%\xc0\x08\xcc\x8a\x87\xfb\x14\x06\xb7\xd1-\xf5a$': 'Video Life (1984) (CommaVid, John Bronstein) (CM-002) [higher sounds] ~',
  b';\x18\xdbs\x937G\x85\x1e\xba\x9a\x0f\xfa<\x12\xb9\xf6\x02\xa9\\': 'Video Life (1984) (CommaVid, John Bronstein) (CM-002) ~',
  b'\xa9\x1c\xf3\xf9w\xd9\xae\x0cP^\xf4\xe7\xb3S\xcc\xdc\xb0\xbd\xe5n': 'Video Olympics (Paddle) (1977) (Atari, Joe Decuir) (CX2621, CX2621P) (PAL)',
  b'\x1f\xfe\x89\xd7\x9dU\xad\xab\xc0\x91k\x95\xcc7\xe1\x86\x19\xefx0': 'Video Olympics - Pong Sports (Paddle) (1977) (Atari, Joe Decuir - Sears) (CX2621 - 99806, 6-99806, 49-75104) ~',
  b'2Mg\xf2\xb5e\x17\xc2)>\x93Q\xc7\x13\xb9}<\x98+\x0b': 'Video Pinball (1981) (Atari, Bob Smith) (CX2648) (PAL)',
  b'!)\xa7/0\xdc\t\x98q\x12\xbe\xce+\x1c\r3H\x18p1': 'Video Pinball (Unknown) (PAL)',
  b',\x16\xc1\xa67L\x8e"\']\x15-\x93\xdd1\xff\xba&\'\x1f': 'Video Pinball - Arcade Pinball (1981) (Atari, Bob Smith - Sears) (CX2648 - 49-75161) ~',
  b'\xfc\xdeu\x9b\xf4\xa2\xffc`%\x9c#\xaa\xefG9\xa4\x9d\xdf4': 'Video Reflex (Foot Craz) (1983) (Exus Corporation) [no roman numbers] ~',
  b'\xde\xc2\xa3\xe9\xb3f\xdc\xe2\xb6=\xc1\xc16b\xd3\xf2$ \xa2.': 'Video Reflex (Foot Craz) (1983) (Exus Corporation) ~',
  b'\xf4\xf9A\xb7y\xbc\xb7\x90-\xf0\xeb|\xf6\xb9\x85\xf5nu\x11\x83': 'Vogel Flieh (AKA Dolphin) (1983) (Quelle) (465.302 8) (PAL)',
  b'\x17\x18\x1c\x8f\xa8\xa2\x8d\x84\xf7\x83\xc1\xa3\x8e\x1e=)g\x89\xbf\xc5': 'Volleyball (AKA RealSports Volleyball) (1983) (Digitel)',
  b'\xb6\x1a\xf8\x07\xd5\x7f%\xae\x10\xae\x13U\xa2\x87\xdbj\x8f\x05\x18\xa0': 'Volleyball (AKA RealSports Volleyball) (Dactari - Milmar)',
  b'\xfd\xbft\xd7&\xb4\x85z\x87\x99DSA\x08\x8f*J\xd82\x0c': 'Volleyball (AKA RealSports Volleyball) (Double-Game Package) (1983) (Otto Versand) (781698) (PAL)',
  b'\xd0ztj\xde\x99\n\x990\xb3\xf5E\\\x1f;\x91\xa4\xf8\x91\x87': 'Volleyball (AKA RealSports Volleyball) (Robby)',
  b'\xce\xaahh\xad%\xadL\x06\xb8\x9e\xf6\xfaNT\x11\xad\x1f\xcck': 'Vom Himmel durch die Hoelle (AKA Parachute) (1983) (Quelle) (719.941 7) (PAL)',
  b'.\xbd\x0fC\xeev\x83?uu\x9a\xc1\xbb\xb4Z\x8e\x0c;\x86\xe9': 'Vulture Attack (AKA Condor Attack) (1982) (K-Tel Vision)',
  b'\x05\xde\xde\x1d\xeb\xc0O\xb1\xc8\x1a\x8d\x94\x0e2\xac\x08\xe8Z\xf2\x9b': 'Vulture Attack (AKA Condor Attack) (1982) (K-Tel Vision) (PAL)',
  b'?*\xd3fn\xb7\x13\xb2\x04\x84\xe3g}\x8e\x1c\xac\x8f\xbd2>': 'Vídeo Cube (AKA Atari Video Cube) (CCE)',
  b's\x07"\x95r\x14S\x06]b\xd9\x13cC\xb8\x13\x10\xa4\xd2%': 'Wabbit (1982) (Apollo, Ban Tran) (AP-2010) ~',
  b'K\x80\x8c\x97_\x8a\x83E\xc7-\x0fb\x81\xea[\xa99c\xd7\xfb': 'Wachroboter jagt Jupy (AKA Keystone Kapers) (1983) (Quelle) (715.853 5) (PAL)',
  b'\r\x98i\x813\x06e\x884\x9a\x80\x84\xf9\xbd\xd1;\x80#.\xd3': 'Walker (Suntek) (SS-032) (PAL) [a]',
  b'%\xe2\x047u\xa0\x87b\xe3\xfc\x98\xec\xb9o\xbfF\xb2\x0e\x86\x0e': 'Walker (Suntek) (SS-032) (PAL) ~',
  b'H+\xd3I"+\x8cp.\x12\\\'\xfdQns\xaf\x13\x96{': 'Wall Ball (1983) (Avalon Hill, Duncan Scott) (5003002) ~',
  b'm\xe4+\xbcGf\xb2c\x01\xe2\x91\xba\x00\xf7\xf7\xa9\xact\x869': 'Wall Break (Unknown)',
  b'\x00\x9aB\xd7\x12b\xd3&z\xe3\x15\xa3\x92Q\x96d\xcc/$\xb4': 'Wall Defender (AKA Wall Break) (HES) (PAL)',
  b'\xbcB\x7f\xf6M\xd0\xb6\xe8\xc5\xf1/\xb8\xb1\x84\x1d>\xdd\xa0)_': 'Wall-Defender (AKA Wall Break) (1983) (Bomb - Onbase) (CA285)',
  b'n\xf3\xc2\xe8\xbb\r6\x1dZ\x99\xaf\xd9_N)\xa6\x9ba\xbcJ': 'Wall-Defender (AKA Wall Break) (1983) (Bomb - Onbase) (CA285) (PAL)',
  b'\\\xc1"\xf1\x0e\xb1\xa8\x14\xd0G@KF+2_Q\xf3],': 'War 2000 (AKA Astrowar) (1983) (Home Vision - Gem International Corporation - VDI) (VCS83102) (PAL)',
  b'-uc\xd37\xcb\xc0\xcd\xf4\xfc\x14\xf6\x98S\xabgWiw\x88': 'Warlords (Kings in the Corner) (Paddle) (1981) (Atari, Carla Meninsky - Sears) (CX2610 - 49-75127) ~',
  b'\xc3\xfb\xa4\xbe\xbc-\xda\x846\x12d3\xe1\xf4\xc1^Q\xef\x1b\xd4': 'Warlords (Kings in the Corner) (Paddle) (1981) (Atari, Carla Meninsky) (CX2610) (PAL)',
  b'#*7\x05\x19\xa7\xfc\xce\x12\x1e\x15\xf8P\xd0\xd3g\x19\t\xf8\xb8': 'Warplock (Paddle) (1982) (Data Age) (DA1002) ~',
  b'<\xfd\xd9)R\xdcJ\x01\x9a|\x9b\x05\x0fnL\xc6ej\x01I': 'Warplock (Paddle) (1983) (Gameworld) (133-002) (PAL)',
  b'\xc5\x13p=c\x8c\x01\xb0\xc2i"\xd5\xe3\xe7\xbf\xb6^\xa5\x97\xda': 'Weltraumtunnel (AKA Innerspace) (1983) (Quelle) (292.651 7) (PAL)',
  b'G"\x15\xfc\xb4l\xec\x90Uv\xd59\xef\xc8\x044\x88\xef\xc4\xed': "Westward Ho (AKA Custer's Revenge) (1982) (PlayAround - J.H.M.) (206) (PAL)",
  b'c\xf4wj\xa4\xc3]\x12@\x01\x91\x8bs<\xdbMF\xdf\xbe\x9b': 'Wilma Wanderer (AKA Lilly Adventure) (1983) (ITT Family Games) (PAL)',
  b'\x1c\xe2Bj\x1aq\xeb\xac\x81p\x9c\x88\xeb0\xe4a\xb2\x91X\xe2': 'Wing War (Flap!) (1983) (Imagic, Michael Greene) (EIZ-002-04I) (PAL) ~',
  b'A\x9e}\xd2L\x81\n\xfb\x8b\x8eU^\xd8H\x98S\xb0\xbf\x05\xd8': 'Wings (06-03-1983) (CBS Electronics, Stuart Ross) (Prototype) ~',
  b'>c\xfe\xca1\xfb\x98\xccF\xfd\xfe\x90\xeci\xe4\x93\x7f\x9d\xef\xa6': 'Wings (10-10-1983) (CBS Electronics, Stuart Ross) (Prototype) (PAL)',
  b'\xb5\xd3\xd0\xc0\x84\x16\x97\xb9@\xfe\x80\xea\xe2I\xa8\x84\xd6\xa8\x89\xc8': 'Winter Games (1987) (Epyx, Steven A. Baker, Tod Frye, Peter Engelbrite) (80561-00251) (PAL)',
  b"hP\xd3)\xe8\xab@;\xda\xe3\x88PfZ.\xff\x91'\x8e\x92": 'Winter Games (1987) (Epyx, Steven A. Baker, Tod Frye, Peter Engelbrite) (80561-00251) ~',
  b'\x07\xd9T\xeb\xe8\xa69$\x9b!\x85\xd2\xb4[\xc9!l\xae\x14\xe6': 'Winterjagd (AKA Ski Hunt) (1983) (Quelle) (343.073 3) (PAL)',
  b'\xbb\xc4\x84@:#\xdb\xb4\xea\xcb\x85\xfe[h\\\x93\x9c\x93\xcd\x08': 'Wizard (1980) (Atari, Chris Crawford) (Prototype) [a]',
  b'$\x99\xce|wSk) u9h\x13_\x91\x80\xbd\x9a\xfen': 'Wizard (1980) (Atari, Chris Crawford) (Prototype) ~',
  b'\x08\x8d\xc3?\xadO\xd9\r\xa7\x1e!\xb6\xf1\x9cV_\xf6-\xef\x1f': 'Wizard of Wor (1982) (CBS Electronics - Roklan, Joe Hellesen, Joe Wagner) (4L1720, 4L1721, 4L1722, 4L2276) (PAL)',
  b'2n^c\xa5N\xc6\xa0#\x1f\xd3\x8eb\xe3R\x00MG\x19\xfe': 'Wizard of Wor (1982) (CBS Electronics - Roklan, Joe Hellesen, Joe Wagner) (M8774, M8794) ~',
  b'%7\x888\x0fowW\xcd\xa3\x9drV\xc7\x96\xa9X\x00\xc9\t': 'Word Zapper (208 in 1) (Unknown) (PAL)',
  b'\x98 \xce\xbf\xbd\xb8\xb7\xff\x88p\xb0p\x088\x15\x16\xfe\xb2%7': 'Word Zapper (Unknown)',
  b'\xe3^w:\x89Z\xb7\xd1!a\x1f\xcdP~W\\[\xabD\x13': 'Word Zapper (Unknown) (PAL)',
  b'\x80lZ\x8a{\x04*\x1a:\xda\x1bo)E\x1a4F\xf9=\xa3': 'Word Zapper (Word Grabber) (1982) (U.S. Games Corporation - Vidtec - JWDA, Todd Marshall, Henry Will IV) (VC1003) ~',
  b'\x84\xf4\xb6\x0f5\x1f#\xe6da:\xc3\xd1U\xbb5K\x80\x0c\xd5': "Words-Attack (1983) (Sancho - Tang's Electronic Co.) (Prototype) (PAL) ~",
  b'\x0e\x98\x1dk\xaa~>!\x86\x08\xebF\xadQ\xfb0g[\xab\xac': 'World End (1983) (Home Vision - Gem International Corporation - VDI) (VCS83109) (PAL) ~',
  b'6\xa1\xe7>\xb5\xaa\\<\xd0\xb0\x1a\xf5\x11}\x19\xb8\xc3`q\xe4': 'Worm War I (1982) (20th Century Fox Video Games - Sirius Software, David Lubar) (11001) ~',
  b'<;Ih\xdd:\xb0\xe4\xcf\xec\xc5\xac\x0c\x85\x05g\xcf?dC': 'Worm War I (1983) (CCE) (C-843)',
  b'Nr\xaap2\xe3\x82\xa7\x92\xb0\xf4\x83\xcd\xe9\x98o\x12Y\x80\x18': 'Worm War I (208 in 1) (Unknown) (PAL)',
  b'\xe4( \xd4fw\xa5P&rj\x0e\x17W\x0c\xf6\xdc\x8b\x04\x8b': 'Worm War I (Unknown) (PAL)',
  b'\xbcH0\x97\xf9O+\xf9\xc4\xd36k;\xd2R\x00\xf2\xe2\xf2*': 'Wuestenschlacht (AKA Chopper Command) (1983) (Quelle) (262.794 1) (PAL)',
  b'\x99g\xa7n\xfbh\x01\x7fy1\x88\xf6\x91\x15\x9f\x04\xe6\xbbDG': "X'Mission (Unknown) (PAL)",
  b'\xa5\xc2\xa3\x97\x8e\xbd\x8ej\x1ez\xee\xe7(\x19\x04\x985\x16\xf6\x9e': 'X-Man (1983) (Universal Gamex Corporation, Miguel Castillo, H.K. Poon) (GX-001) (PAL)',
  b'\x1c]\x15\x1e\x86\xc0\xa0\xbb\xdf;3\xef\x158\x88\xc6\xbex\xc3k': 'X-Man (1983) (Universal Gamex Corporation, Miguel Castillo, H.K. Poon) (GX-001) ~',
  b"u\x03\x91'\xcb|S\xfa\t\xd1\xc2#r\x9aLZR\x1a\xed\xef": 'Xenophobe (1991) (Atari) (CX26172) (PAL)',
  b'\x16\x0bn6Cz\xd6\xac\xbc&\x86\xfb\xde\x10\x02\xe2\xfa\x88\xc5\xfb': 'Xenophobe (1991) (Atari) (CX26172) ~',
  b'S\xf9A\xca\xbff\x1a\x9c\xe7c\x14+\x07\x05q\xf6\xfdL\x88 ': 'Xevious (01-18-1984) (Atari, Tod Frye) (CX2695) (Prototype) ~',
  b'\xfa\xc2\x9f!\xb7\x11\xd3\x1e?\xfc!\xdf\xdc\xaf\xef\x05\xaa\xfa\xbf\x02': 'Xevious (05-25-1983) (Atari, Tod Frye) (CX2695) (Prototype)',
  b's\x13;\x81\x19n\\\xbc\x1c\xec\x99\xee\xfc\x12#\xdd\xb8\xf4\xca\x83': 'Xevious (08-02-1983) (Atari, Tod Frye) (CX2695) (Prototype)',
  b'\xc4\xf9V\xccjx\xf3\xd7\xb5\xc2\xfb\x16\xdeX\xed|\xdbl\x82\xa8': 'Xevious (CCE)',
  b'_\xeb\xb9\xf5j\x98\x9e\xe6\xa6r\xd4\x98FGD\xc6"\x14xi': "Yars' Revenge (Canal 3 - Intellivision)",
  b'\x88FG\xeb\xc1]\xf7\xd98\x8b,\xba\x03\x18\xc2@\xe5\x93\xdb\xc0': "Yars' Revenge (Time Freeze) (09-01-81) (Atari, Howard Scott Warshaw - Sears) (CX2655 - 49-75167) (Prototype)",
  b'\xe2\xcd\x89\x96\xc1\xcf\x92\x9e)\x13\x06\x90\x02M\x1e\xc2=;\x0b\xde': "Yars' Revenge (Time Freeze) (1982) (Atari, Howard Scott Warshaw - Sears) (CX2655 - 49-75167) ~",
  b'\xaa\xeb\x19\xd1\xe5\xb3\x16FY\xb3\xf1\x8c\xbf2\xf3\xdb\x92NP\x0c': "Yars' Revenge (Time Freeze) (1982) (Atari, Howard Scott Warshaw) (CX2655, CX2655P) (PAL)",
  b'\xa8\x0e\xa7\x086\x15\xfcu`\xbe\xec\xfc\x0e\x99\xfax\xfa~\xee\xf7': 'Year 1999, The (AKA Condor Attack) (Rainbow Vision - Suntek) (SS-008) (PAL)',
  b'U.+\xd3v\x01\xdaIa\xb04cS\xc9<\xce">B\x9e': 'Z-Tack (AKA Base Attack) (1983) (Bomb - Onbase) (CA283)',
  b'N,\xcb\n\xfe#jv\x19:\xd5\x00\xba\xb5z\xbb\x82c*Z': 'Z-Tack (AKA Base Attack) (1983) (Bomb - Onbase) (CA283) (PAL)',
  b'\xfe^\x7f\x02k1T\xa2"$O\xe5\x95\xcf\xeb\xbf\x8a\x08\x0f\xcb': 'Zaxxon (1983) (CBS Electronics) (4L1784, 4L1786, 4L1787, 4L2277) (PAL)',
  b'X\xc2\xf6\xab\xc5Y\x9c\xd3\\\x0er/$\xbc\xc1(\xac\x8f\x9a0': 'Zaxxon (1983) (Coleco) (2454) ~',
  b"\xf1M'\rB\x9f\xc3\xec\xf5\xdd\x04\xd7\x1e\x87\xe7\x9d\xafR\xfd_": 'Zoo Fun (AKA Panda Chase) (Suntek) (SS-027) (PAL)',
  b'\x7f\x11\\\xb4\x1f\xc7\x08\x89\xb93\xfd,\x80\x80D\xa2\xb66ra': 'Zoo Keeper Sounds (1984) (Atari, Christopher H. Omarzu, Robert Vieira) (CX26121) (Prototype) ~',
},
}  # end sha1_signatures
